@@ -9,8 +9,10 @@
 #include <algorithm>  // For std::stable_sort, std::min
 #include <utility>
 
-#include "js/ForOfIterator.h"  // For JS::ForOfIterator
 #include "jsapi.h"             // For most JSAPI
+#include "js/ForOfIterator.h"  // For JS::ForOfIterator
+#include "js/PropertyAndElement.h"  // JS_Enumerate, JS_GetProperty, JS_GetPropertyById
+#include "mozilla/AnimatedPropertyID.h"
 #include "mozilla/ComputedStyle.h"
 #include "mozilla/ErrorResult.h"
 #include "mozilla/RangedArray.h"
@@ -22,7 +24,6 @@
 #include "mozilla/TimingParams.h"
 #include "mozilla/dom/BaseKeyframeTypesBinding.h"  // For FastBaseKeyframe etc.
 #include "mozilla/dom/BindingCallContext.h"
-#include "mozilla/dom/Document.h"  // For Document::AreWebAnimationsImplicitKeyframesEnabled
 #include "mozilla/dom/Element.h"
 #include "mozilla/dom/KeyframeEffect.h"  // For PropertyValuesPair etc.
 #include "mozilla/dom/KeyframeEffectBinding.h"
@@ -34,6 +35,7 @@
 #include "nsContentUtils.h"  // For GetContextForContent
 #include "nsIScriptError.h"
 #include "nsPresContextInlines.h"
+#include "nsString.h"
 #include "nsTArray.h"
 
 using mozilla::dom::Nullable;
@@ -59,7 +61,9 @@ enum class ListAllowance { eDisallow, eAllow };
  * mValues.
  */
 struct PropertyValuesPair {
-  nsCSSPropertyID mProperty;
+  PropertyValuesPair() : mProperty(eCSSProperty_UNKNOWN) {}
+
+  AnimatedPropertyID mProperty;
   nsTArray<nsCString> mValues;
 };
 
@@ -68,8 +72,8 @@ struct PropertyValuesPair {
  * BaseKeyframe or BasePropertyIndexedKeyframe object.
  */
 struct AdditionalProperty {
-  nsCSSPropertyID mProperty;
-  size_t mJsidIndex;  // Index into |ids| in GetPropertyValuesPairs.
+  AnimatedPropertyID mProperty;
+  size_t mJsidIndex = 0;  // Index into |ids| in GetPropertyValuesPairs.
 
   struct PropertyComparator {
     bool Equals(const AdditionalProperty& aLhs,
@@ -78,8 +82,24 @@ struct AdditionalProperty {
     }
     bool LessThan(const AdditionalProperty& aLhs,
                   const AdditionalProperty& aRhs) const {
-      return nsCSSProps::PropertyIDLNameSortPosition(aLhs.mProperty) <
-             nsCSSProps::PropertyIDLNameSortPosition(aRhs.mProperty);
+      bool customLhs =
+          aLhs.mProperty.mID == nsCSSPropertyID::eCSSPropertyExtra_variable;
+      bool customRhs =
+          aRhs.mProperty.mID == nsCSSPropertyID::eCSSPropertyExtra_variable;
+      if (!customLhs && !customRhs) {
+        // Compare by IDL names.
+        return nsCSSProps::PropertyIDLNameSortPosition(aLhs.mProperty.mID) <
+               nsCSSProps::PropertyIDLNameSortPosition(aRhs.mProperty.mID);
+      }
+      if (customLhs && customRhs) {
+        // Compare by custom property names.
+        return nsDependentAtomString(aLhs.mProperty.mCustomName) <
+               nsDependentAtomString(aRhs.mProperty.mCustomName);
+      }
+      // Custom properties should be ordered before normal CSS properties, as if
+      // the custom property name starts with `--`.
+      // <https://drafts.csswg.org/web-animations-1/#idl-attribute-name-to-animation-property-name>
+      return !customLhs && customRhs;
     }
   };
 };
@@ -92,11 +112,14 @@ struct AdditionalProperty {
  * to gather data for each individual segment.
  */
 struct KeyframeValueEntry {
-  nsCSSPropertyID mProperty;
+  KeyframeValueEntry()
+      : mProperty(eCSSProperty_UNKNOWN), mOffset(), mComposite() {}
+
+  AnimatedPropertyID mProperty;
   AnimationValue mValue;
 
   float mOffset;
-  Maybe<ComputedTimingFunction> mTimingFunction;
+  Maybe<StyleComputedTimingFunction> mTimingFunction;
   dom::CompositeOperation mComposite;
 
   struct PropertyOffsetComparator {
@@ -106,11 +129,28 @@ struct KeyframeValueEntry {
     }
     static bool LessThan(const KeyframeValueEntry& aLhs,
                          const KeyframeValueEntry& aRhs) {
-      // First, sort by property IDL name.
-      int32_t order = nsCSSProps::PropertyIDLNameSortPosition(aLhs.mProperty) -
-                      nsCSSProps::PropertyIDLNameSortPosition(aRhs.mProperty);
-      if (order != 0) {
-        return order < 0;
+      // First, sort by property name.
+      bool customLhs =
+          aLhs.mProperty.mID == nsCSSPropertyID::eCSSPropertyExtra_variable;
+      bool customRhs =
+          aRhs.mProperty.mID == nsCSSPropertyID::eCSSPropertyExtra_variable;
+      if (!customLhs && !customRhs) {
+        // Compare by IDL names.
+        int32_t order =
+            nsCSSProps::PropertyIDLNameSortPosition(aLhs.mProperty.mID) -
+            nsCSSProps::PropertyIDLNameSortPosition(aRhs.mProperty.mID);
+        if (order != 0) {
+          return order < 0;
+        }
+      } else if (customLhs && customRhs) {
+        // Compare by custom property names.
+        int order = Compare(nsDependentAtomString(aLhs.mProperty.mCustomName),
+                            nsDependentAtomString(aRhs.mProperty.mCustomName));
+        if (order != 0) {
+          return order < 0;
+        }
+      } else {
+        return !customLhs && customRhs;
       }
 
       // Then, by offset.
@@ -159,7 +199,7 @@ static bool AppendValueAsString(JSContext* aCx, nsTArray<nsCString>& aValues,
                                 JS::Handle<JS::Value> aValue);
 
 static Maybe<PropertyValuePair> MakePropertyValuePair(
-    nsCSSPropertyID aProperty, const nsACString& aStringValue,
+    const AnimatedPropertyID& aProperty, const nsACString& aStringValue,
     dom::Document* aDocument);
 
 static bool HasValidOffsets(const nsTArray<Keyframe>& aKeyframes);
@@ -181,9 +221,6 @@ static void GetKeyframeListFromPropertyIndexedKeyframe(
     JSContext* aCx, dom::Document* aDocument, JS::Handle<JS::Value> aValue,
     nsTArray<Keyframe>& aResult, ErrorResult& aRv);
 
-static bool HasImplicitKeyframeValues(const nsTArray<Keyframe>& aKeyframes,
-                                      dom::Document* aDocument);
-
 static void DistributeRange(const Range<Keyframe>& aRange);
 
 // ------------------------------------------------------------------
@@ -229,13 +266,6 @@ nsTArray<Keyframe> KeyframeUtils::GetKeyframesFromObject(
     return keyframes;
   }
 
-  if (!dom::Document::AreWebAnimationsImplicitKeyframesEnabled(aCx, nullptr) &&
-      HasImplicitKeyframeValues(keyframes, aDocument)) {
-    keyframes.Clear();
-    aRv.ThrowNotSupportedError(
-        "Animation to or from an underlying value is not yet supported");
-  }
-
   return keyframes;
 }
 
@@ -319,14 +349,14 @@ nsTArray<AnimationProperty> KeyframeUtils::GetAnimationPropertiesFromKeyframes(
 }
 
 /* static */
-bool KeyframeUtils::IsAnimatableProperty(nsCSSPropertyID aProperty) {
+bool KeyframeUtils::IsAnimatableProperty(const AnimatedPropertyID& aProperty) {
   // Regardless of the backend type, treat the 'display' property as not
   // animatable. (Servo will report it as being animatable, since it is
   // in fact animatable by SMIL.)
-  if (aProperty == eCSSProperty_display) {
+  if (aProperty.mID == eCSSProperty_display) {
     return false;
   }
-  return Servo_Property_IsAnimatable(aProperty);
+  return Servo_Property_IsAnimatable(&aProperty);
 }
 
 // ------------------------------------------------------------------
@@ -430,9 +460,7 @@ static bool ConvertKeyframeSequence(JSContext* aCx, dom::Document* aDocument,
       keyframe->mOffset.emplace(keyframeDict.mOffset.Value());
     }
 
-    if (StaticPrefs::dom_animations_api_compositing_enabled()) {
-      keyframe->mComposite = keyframeDict.mComposite;
-    }
+    keyframe->mComposite = keyframeDict.mComposite;
 
     // Look for additional property-values pairs on the object.
     nsTArray<PropertyValuesPair> propertyValuePairs;
@@ -471,7 +499,7 @@ static bool ConvertKeyframeSequence(JSContext* aCx, dom::Document* aDocument,
       // includes a chrome-only member that can be set to indicate that
       // ComputeValues should fail for shorthand property values on that
       // keyframe.
-      if (nsCSSProps::IsShorthand(pair.mProperty) &&
+      if (nsCSSProps::IsShorthand(pair.mProperty.mID) &&
           keyframeDict.mSimulateComputeValuesFailure) {
         MarkAsComputeValuesFailureKey(keyframe->mPropertyValues.LastElement());
       }
@@ -527,21 +555,29 @@ static bool GetPropertyValuesPairs(JSContext* aCx,
     // This means if the attribute is the string "cssOffset"/"cssFloat", we use
     // CSS "offset"/"float" property.
     // https://drafts.csswg.org/web-animations/#property-name-conversion
-    nsCSSPropertyID property = nsCSSPropertyID::eCSSProperty_UNKNOWN;
-    if (propName.EqualsLiteral("cssOffset")) {
-      property = nsCSSPropertyID::eCSSProperty_offset;
+    nsCSSPropertyID propertyID = nsCSSPropertyID::eCSSProperty_UNKNOWN;
+    if (nsCSSProps::IsCustomPropertyName(propName)) {
+      propertyID = eCSSPropertyExtra_variable;
+    } else if (propName.EqualsLiteral("cssOffset")) {
+      propertyID = nsCSSPropertyID::eCSSProperty_offset;
     } else if (propName.EqualsLiteral("cssFloat")) {
-      property = nsCSSPropertyID::eCSSProperty_float;
+      propertyID = nsCSSPropertyID::eCSSProperty_float;
     } else if (!propName.EqualsLiteral("offset") &&
                !propName.EqualsLiteral("float")) {
-      property = nsCSSProps::LookupPropertyByIDLName(
+      propertyID = nsCSSProps::LookupPropertyByIDLName(
           propName, CSSEnabledState::ForAllContent);
     }
 
+    // TODO(zrhoffman, bug 1811897) Add test coverage for removing the `--`
+    // prefix here.
+    AnimatedPropertyID property =
+        propertyID == eCSSPropertyExtra_variable
+            ? AnimatedPropertyID(
+                  NS_Atomize(Substring(propName, 2, propName.Length() - 2)))
+            : AnimatedPropertyID(propertyID);
+
     if (KeyframeUtils::IsAnimatableProperty(property)) {
-      AdditionalProperty* p = properties.AppendElement();
-      p->mProperty = property;
-      p->mJsidIndex = i;
+      properties.AppendElement(AdditionalProperty{std::move(property), i});
     }
   }
 
@@ -621,12 +657,11 @@ static bool AppendValueAsString(JSContext* aCx, nsTArray<nsCString>& aValues,
 }
 
 static void ReportInvalidPropertyValueToConsole(
-    nsCSSPropertyID aProperty, const nsACString& aInvalidPropertyValue,
-    dom::Document* aDoc) {
+    const AnimatedPropertyID& aProperty,
+    const nsACString& aInvalidPropertyValue, dom::Document* aDoc) {
   AutoTArray<nsString, 2> params;
   params.AppendElement(NS_ConvertUTF8toUTF16(aInvalidPropertyValue));
-  CopyASCIItoUTF16(nsCSSProps::GetStringValue(aProperty),
-                   *params.AppendElement());
+  aProperty.ToString(*params.AppendElement());
   nsContentUtils::ReportToConsole(nsIScriptError::warningFlag, "Animation"_ns,
                                   aDoc, nsContentUtils::eDOM_PROPERTIES,
                                   "InvalidKeyframePropertyValue", params);
@@ -643,15 +678,16 @@ static void ReportInvalidPropertyValueToConsole(
  *   an invalid property value.
  */
 static Maybe<PropertyValuePair> MakePropertyValuePair(
-    nsCSSPropertyID aProperty, const nsACString& aStringValue,
+    const AnimatedPropertyID& aProperty, const nsACString& aStringValue,
     dom::Document* aDocument) {
   MOZ_ASSERT(aDocument);
   Maybe<PropertyValuePair> result;
 
   ServoCSSParser::ParsingEnvironment env =
       ServoCSSParser::GetParsingEnvironment(aDocument);
-  RefPtr<RawServoDeclarationBlock> servoDeclarationBlock =
-      ServoCSSParser::ParseProperty(aProperty, aStringValue, env);
+  RefPtr<StyleLockedDeclarationBlock> servoDeclarationBlock =
+      ServoCSSParser::ParseProperty(aProperty, aStringValue, env,
+                                    StyleParsingMode::DEFAULT);
 
   if (servoDeclarationBlock) {
     result.emplace(aProperty, std::move(servoDeclarationBlock));
@@ -693,7 +729,7 @@ static bool HasValidOffsets(const nsTArray<Keyframe>& aKeyframes) {
  *              a shorthand property.
  */
 static void MarkAsComputeValuesFailureKey(PropertyValuePair& aPair) {
-  MOZ_ASSERT(nsCSSProps::IsShorthand(aPair.mProperty),
+  MOZ_ASSERT(nsCSSProps::IsShorthand(aPair.mProperty.mID),
              "Only shorthand property values can be marked as failure values");
 
   aPair.mSimulateComputeValuesFailure = true;
@@ -754,12 +790,6 @@ static AnimationProperty* HandleMissingInitialKeyframe(
   MOZ_ASSERT(aEntry.mOffset != 0.0f,
              "The offset of the entry should not be 0.0");
 
-  // If the preference for implicit keyframes is not enabled, don't fill in the
-  // missing keyframe.
-  if (!StaticPrefs::dom_animations_api_implicit_keyframes_enabled()) {
-    return nullptr;
-  }
-
   AnimationProperty* result = aResult.AppendElement();
   result->mProperty = aEntry.mProperty;
 
@@ -774,17 +804,6 @@ static void HandleMissingFinalKeyframe(
   MOZ_ASSERT(aEntry.mOffset != 1.0f,
              "The offset of the entry should not be 1.0");
 
-  // If the preference for implicit keyframes is not enabled, don't fill
-  // in the missing keyframe.
-  if (!StaticPrefs::dom_animations_api_implicit_keyframes_enabled()) {
-    // If we have already appended a new entry for the property so we have to
-    // remove it.
-    if (aCurrentAnimationProperty) {
-      aResult.RemoveLastElement();
-    }
-    return;
-  }
-
   // If |aCurrentAnimationProperty| is nullptr, that means this is the first
   // entry for the property, we have to append a new AnimationProperty for this
   // property.
@@ -840,7 +859,7 @@ static void BuildSegmentsFromValueEntries(
   // care to identify properties that lack a value at offset 0.0/1.0 and drops
   // those properties from |aResult|.
 
-  nsCSSPropertyID lastProperty = eCSSProperty_UNKNOWN;
+  AnimatedPropertyID lastProperty(eCSSProperty_UNKNOWN);
   AnimationProperty* animationProperty = nullptr;
 
   size_t i = 0, n = aEntries.Length();
@@ -861,9 +880,9 @@ static void BuildSegmentsFromValueEntries(
       break;
     }
 
-    MOZ_ASSERT(aEntries[i].mProperty != eCSSProperty_UNKNOWN &&
-                   aEntries[i + 1].mProperty != eCSSProperty_UNKNOWN,
-               "Each entry should specify a valid property");
+    MOZ_ASSERT(
+        aEntries[i].mProperty.IsValid() && aEntries[i + 1].mProperty.IsValid(),
+        "Each entry should specify a valid property");
 
     // No keyframe for this property at offset 0.
     if (aEntries[i].mProperty != lastProperty && aEntries[i].mOffset != 0.0f) {
@@ -1007,15 +1026,6 @@ static void GetKeyframeListFromPropertyIndexedKeyframe(
       continue;
     }
 
-    // If we only have one value, we should animate from the underlying value
-    // but not if the pref for supporting implicit keyframes is disabled.
-    if (!StaticPrefs::dom_animations_api_implicit_keyframes_enabled() &&
-        count == 1) {
-      aRv.ThrowNotSupportedError(
-          "Animation to or from an underlying value is not yet supported");
-      return;
-    }
-
     size_t n = pair.mValues.Length() - 1;
     size_t i = 0;
 
@@ -1099,7 +1109,7 @@ static void GetKeyframeListFromPropertyIndexedKeyframe(
   //
   // This corresponds to step 5, "Otherwise," branch, substeps 7-11 of
   // https://drafts.csswg.org/web-animations/#processing-a-keyframes-argument
-  FallibleTArray<Maybe<ComputedTimingFunction>> easings;
+  FallibleTArray<Maybe<StyleComputedTimingFunction>> easings;
   auto parseAndAppendEasing = [&](const nsACString& easingString,
                                   ErrorResult& aRv) {
     auto easing = TimingParams::ParseEasing(easingString, aRv);
@@ -1143,93 +1153,25 @@ static void GetKeyframeListFromPropertyIndexedKeyframe(
   //
   // This corresponds to step 5, "Otherwise," branch, substep 12 of
   // https://drafts.csswg.org/web-animations/#processing-a-keyframes-argument
-  if (StaticPrefs::dom_animations_api_compositing_enabled()) {
-    const FallibleTArray<dom::CompositeOperationOrAuto>* compositeOps = nullptr;
-    AutoTArray<dom::CompositeOperationOrAuto, 1> singleCompositeOp;
-    auto& composite = keyframeDict.mComposite;
-    if (composite.IsCompositeOperationOrAuto()) {
-      singleCompositeOp.AppendElement(
-          composite.GetAsCompositeOperationOrAuto());
-      const FallibleTArray<dom::CompositeOperationOrAuto>& asFallibleArray =
-          singleCompositeOp;
-      compositeOps = &asFallibleArray;
-    } else if (composite.IsCompositeOperationOrAutoSequence()) {
-      compositeOps = &composite.GetAsCompositeOperationOrAutoSequence();
-    }
-
-    // Fill in and repeat as needed.
-    if (compositeOps && !compositeOps->IsEmpty()) {
-      size_t length = compositeOps->Length();
-      for (size_t i = 0; i < aResult.Length(); i++) {
-        aResult[i].mComposite = compositeOps->ElementAt(i % length);
-      }
-    }
-  }
-}
-
-/**
- * Returns true if the supplied set of keyframes has keyframe values for
- * any property for which it does not also supply a value for the 0% and 100%
- * offsets. The check is not entirely accurate but should detect most common
- * cases.
- *
- * @param aKeyframes The set of keyframes to analyze.
- * @param aDocument The document to use when parsing keyframes so we can
- *   try to detect where we have an invalid value at 0%/100%.
- */
-static bool HasImplicitKeyframeValues(const nsTArray<Keyframe>& aKeyframes,
-                                      dom::Document* aDocument) {
-  // We are looking to see if that every property referenced in |aKeyframes|
-  // has a valid property at offset 0.0 and 1.0. The check as to whether a
-  // property is valid or not, however, is not precise. We only check if the
-  // property can be parsed, NOT whether it can also be converted to a
-  // StyleAnimationValue since doing that requires a target element bound to
-  // a document which we might not always have at the point where we want to
-  // perform this check.
-  //
-  // This is only a temporary measure until we ship implicit keyframes and
-  // remove the corresponding pref.
-  // So as long as this check catches most cases, and we don't do anything
-  // horrible in one of the cases we can't detect, it should be sufficient.
-
-  nsCSSPropertyIDSet properties;               // All properties encountered.
-  nsCSSPropertyIDSet propertiesWithFromValue;  // Those with a defined 0% value.
-  nsCSSPropertyIDSet propertiesWithToValue;  // Those with a defined 100% value.
-
-  auto addToPropertySets = [&](nsCSSPropertyID aProperty, double aOffset) {
-    properties.AddProperty(aProperty);
-    if (aOffset == 0.0) {
-      propertiesWithFromValue.AddProperty(aProperty);
-    } else if (aOffset == 1.0) {
-      propertiesWithToValue.AddProperty(aProperty);
-    }
-  };
-
-  for (size_t i = 0, len = aKeyframes.Length(); i < len; i++) {
-    const Keyframe& frame = aKeyframes[i];
-
-    // We won't have called DistributeKeyframes when this is called so
-    // we can't use frame.mComputedOffset. Instead we do a rough version
-    // of that algorithm that substitutes null offsets with 0.0 for the first
-    // frame, 1.0 for the last frame, and 0.5 for everything else.
-    double computedOffset = i == len - 1 ? 1.0 : i == 0 ? 0.0 : 0.5;
-    double offsetToUse = frame.mOffset ? frame.mOffset.value() : computedOffset;
-
-    for (const PropertyValuePair& pair : frame.mPropertyValues) {
-      if (nsCSSProps::IsShorthand(pair.mProperty)) {
-        MOZ_ASSERT(pair.mServoDeclarationBlock);
-        CSSPROPS_FOR_SHORTHAND_SUBPROPERTIES(prop, pair.mProperty,
-                                             CSSEnabledState::ForAllContent) {
-          addToPropertySets(*prop, offsetToUse);
-        }
-      } else {
-        addToPropertySets(pair.mProperty, offsetToUse);
-      }
-    }
+  const FallibleTArray<dom::CompositeOperationOrAuto>* compositeOps = nullptr;
+  AutoTArray<dom::CompositeOperationOrAuto, 1> singleCompositeOp;
+  auto& composite = keyframeDict.mComposite;
+  if (composite.IsCompositeOperationOrAuto()) {
+    singleCompositeOp.AppendElement(composite.GetAsCompositeOperationOrAuto());
+    const FallibleTArray<dom::CompositeOperationOrAuto>& asFallibleArray =
+        singleCompositeOp;
+    compositeOps = &asFallibleArray;
+  } else if (composite.IsCompositeOperationOrAutoSequence()) {
+    compositeOps = &composite.GetAsCompositeOperationOrAutoSequence();
   }
 
-  return !propertiesWithFromValue.Equals(properties) ||
-         !propertiesWithToValue.Equals(properties);
+  // Fill in and repeat as needed.
+  if (compositeOps && !compositeOps->IsEmpty()) {
+    size_t length = compositeOps->Length();
+    for (size_t i = 0; i < aResult.Length(); i++) {
+      aResult[i].mComposite = compositeOps->ElementAt(i % length);
+    }
+  }
 }
 
 /**
