@@ -92,8 +92,8 @@ TimingParams TimingParams::FromEffectTiming(
   if (aRv.Failed()) {
     return result;
   }
-  Maybe<ComputedTimingFunction> easing =
-      TimingParams::ParseEasing(aEffectTiming.mEasing, aRv);
+  Maybe<StyleComputedTimingFunction> easing =
+      ParseEasing(aEffectTiming.mEasing, aRv);
   if (aRv.Failed()) {
     return result;
   }
@@ -105,7 +105,7 @@ TimingParams TimingParams::FromEffectTiming(
   result.mIterationStart = aEffectTiming.mIterationStart;
   result.mDirection = aEffectTiming.mDirection;
   result.mFill = aEffectTiming.mFill;
-  result.mFunction = easing;
+  result.mFunction = std::move(easing);
 
   result.Update();
 
@@ -146,9 +146,9 @@ TimingParams TimingParams::MergeOptionalEffectTiming(
     }
   }
 
-  Maybe<ComputedTimingFunction> easing;
+  Maybe<StyleComputedTimingFunction> easing;
   if (aEffectTiming.mEasing.WasPassed()) {
-    easing = TimingParams::ParseEasing(aEffectTiming.mEasing.Value(), aRv);
+    easing = ParseEasing(aEffectTiming.mEasing.Value(), aRv);
     if (aRv.Failed()) {
       return result;
     }
@@ -189,19 +189,19 @@ TimingParams TimingParams::MergeOptionalEffectTiming(
 }
 
 /* static */
-Maybe<ComputedTimingFunction> TimingParams::ParseEasing(
+Maybe<StyleComputedTimingFunction> TimingParams::ParseEasing(
     const nsACString& aEasing, ErrorResult& aRv) {
-  nsTimingFunction timingFunction;
+  auto timingFunction = StyleComputedTimingFunction::LinearKeyword();
   if (!ServoCSSParser::ParseEasing(aEasing, timingFunction)) {
     aRv.ThrowTypeError<dom::MSG_INVALID_EASING_ERROR>(aEasing);
     return Nothing();
   }
 
-  if (timingFunction.IsLinear()) {
+  if (timingFunction.IsLinearKeyword()) {
     return Nothing();
   }
 
-  return Some(ComputedTimingFunction(timingFunction));
+  return Some(std::move(timingFunction));
 }
 
 bool TimingParams::operator==(const TimingParams& aOther) const {
@@ -214,4 +214,78 @@ bool TimingParams::operator==(const TimingParams& aOther) const {
          mFunction == aOther.mFunction;
 }
 
+// FIXME: This is a tentative way to normalize the timing which is defined in
+// [web-animations-2] [1]. I borrow this implementation and some concepts for
+// the edge cases from Chromium [2] so we can match the behavior with them. The
+// implementation here ignores the case of percentage of start delay, end delay,
+// and duration because Gecko doesn't support them. We may have to update the
+// calculation if the spec issue [3] gets any update.
+//
+// [1]
+// https://drafts.csswg.org/web-animations-2/#time-based-animation-to-a-proportional-animation
+// [2] https://chromium-review.googlesource.com/c/chromium/src/+/2992387
+// [3] https://github.com/w3c/csswg-drafts/issues/4862
+TimingParams TimingParams::Normalize(
+    const TimeDuration& aTimelineDuration) const {
+  MOZ_ASSERT(aTimelineDuration,
+             "the timeline duration of scroll-timeline is always non-zero now");
+
+  TimingParams normalizedTiming(*this);
+
+  // Handle iteration duration value of "auto" first.
+  // FIXME: Bug 1676794: Gecko doesn't support `animation-duration:auto` and we
+  // don't support JS-generated scroll animations, so we don't fall into this
+  // case for now. Need to check this again after we support ScrollTimeline
+  // interface.
+  if (!mDuration) {
+    // If the iteration duration is auto, then:
+    //   Set start delay and end delay to 0, as it is not possible to mix time
+    //   and proportions.
+    normalizedTiming.mDelay = TimeDuration();
+    normalizedTiming.mEndDelay = TimeDuration();
+    normalizedTiming.Update();
+    return normalizedTiming;
+  }
+
+  if (mEndTime.IsZero()) {
+    // mEndTime of zero causes division by zero so we handle it here.
+    //
+    // FIXME: The spec doesn't mention this case, so we might have to update
+    // this based on the spec issue,
+    // https://github.com/w3c/csswg-drafts/issues/7459.
+    normalizedTiming.mDelay = TimeDuration();
+    normalizedTiming.mEndDelay = TimeDuration();
+    normalizedTiming.mDuration = Some(TimeDuration());
+  } else if (mEndTime == TimeDuration::Forever()) {
+    // The iteration count or duration may be infinite; however, start and
+    // end delays are strictly finite. Thus, in the limit when end time
+    // approaches infinity:
+    //    start delay / end time = finite / infinite = 0
+    //    end delay / end time = finite / infinite = 0
+    //    iteration duration / end time = 1 / iteration count
+    // This condition can be reached by switching to a scroll timeline on
+    // an existing infinite duration animation.
+    //
+    // FIXME: The spec doesn't mention this case, so we might have to update
+    // this based on the spec issue,
+    // https://github.com/w3c/csswg-drafts/issues/7459.
+    normalizedTiming.mDelay = TimeDuration();
+    normalizedTiming.mEndDelay = TimeDuration();
+    normalizedTiming.mDuration =
+        Some(aTimelineDuration.MultDouble(1.0 / mIterations));
+  } else {
+    // Convert to percentages then multiply by the timeline duration.
+    const double endTimeInSec = mEndTime.ToSeconds();
+    normalizedTiming.mDelay =
+        aTimelineDuration.MultDouble(mDelay.ToSeconds() / endTimeInSec);
+    normalizedTiming.mEndDelay =
+        aTimelineDuration.MultDouble(mEndDelay.ToSeconds() / endTimeInSec);
+    normalizedTiming.mDuration = Some(StickyTimeDuration(
+        aTimelineDuration.MultDouble(mDuration->ToSeconds() / endTimeInSec)));
+  }
+
+  normalizedTiming.Update();
+  return normalizedTiming;
+}
+
 }  // namespace mozilla
