@@ -4,24 +4,45 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-#include "StructuredCloneHolder.h"
+#include "mozilla/dom/StructuredCloneHolder.h"
 
-#include "ImageContainer.h"
+#include <new>
+#include "ErrorList.h"
+#include "MainThreadUtils.h"
+#include "js/CallArgs.h"
+#include "js/Value.h"
+#include "js/WasmModule.h"
+#include "js/Wrapper.h"
+#include "jsapi.h"
+#include "mozilla/AlreadyAddRefed.h"
 #include "mozilla/AutoRestore.h"
+#include "mozilla/ErrorResult.h"
+#include "mozilla/OwningNonNull.h"
+#include "mozilla/RefPtr.h"
+#include "mozilla/ScopeExit.h"
+#include "mozilla/StaticPrefs_dom.h"
+#include "mozilla/dom/BindingDeclarations.h"
 #include "mozilla/dom/BindingUtils.h"
+#include "mozilla/dom/Blob.h"
 #include "mozilla/dom/BlobBinding.h"
+#include "mozilla/dom/BlobImpl.h"
 #include "mozilla/dom/BrowsingContext.h"
-#include "mozilla/dom/BrowsingContextBinding.h"
 #include "mozilla/dom/ClonedErrorHolder.h"
 #include "mozilla/dom/ClonedErrorHolderBinding.h"
-#include "mozilla/dom/StructuredCloneBlob.h"
-#include "mozilla/dom/DocGroup.h"
-#include "mozilla/dom/Directory.h"
 #include "mozilla/dom/DirectoryBinding.h"
+#include "mozilla/dom/DOMJSClass.h"
+#include "mozilla/dom/DOMTypes.h"
+#include "mozilla/dom/Directory.h"
+#include "mozilla/dom/DocGroup.h"
+#include "mozilla/dom/EncodedAudioChunk.h"
+#include "mozilla/dom/EncodedAudioChunkBinding.h"
+#include "mozilla/dom/EncodedVideoChunk.h"
+#include "mozilla/dom/EncodedVideoChunkBinding.h"
 #include "mozilla/dom/File.h"
 #include "mozilla/dom/FileList.h"
 #include "mozilla/dom/FileListBinding.h"
 #include "mozilla/dom/FormData.h"
+#include "mozilla/dom/FormDataBinding.h"
 #include "mozilla/dom/ImageBitmap.h"
 #include "mozilla/dom/ImageBitmapBinding.h"
 #include "mozilla/dom/JSExecutionManager.h"
@@ -29,23 +50,45 @@
 #include "mozilla/dom/MessagePortBinding.h"
 #include "mozilla/dom/OffscreenCanvas.h"
 #include "mozilla/dom/OffscreenCanvasBinding.h"
-#include "mozilla/dom/PMessagePort.h"
+#include "mozilla/dom/ReadableStream.h"
+#include "mozilla/dom/ReadableStreamBinding.h"
 #include "mozilla/dom/ScriptSettings.h"
+#include "mozilla/dom/StructuredCloneBlob.h"
+#include "mozilla/dom/StructuredCloneHolderBinding.h"
 #include "mozilla/dom/StructuredCloneTags.h"
 #include "mozilla/dom/ToJSValue.h"
+#include "mozilla/dom/TransformStream.h"
+#include "mozilla/dom/TransformStreamBinding.h"
+#include "mozilla/dom/VideoFrame.h"
+#include "mozilla/dom/AudioData.h"
+#include "mozilla/dom/VideoFrameBinding.h"
 #include "mozilla/dom/WebIDLSerializable.h"
+#include "mozilla/dom/WritableStream.h"
+#include "mozilla/dom/WritableStreamBinding.h"
+#include "mozilla/dom/WorkerCommon.h"
 #include "mozilla/dom/WorkerPrivate.h"
+#include "mozilla/fallible.h"
 #include "mozilla/gfx/2D.h"
-#include "mozilla/ipc/BackgroundChild.h"
-#include "mozilla/ipc/BackgroundUtils.h"
-#include "mozilla/ipc/PBackgroundSharedTypes.h"
-#include "MultipartBlobImpl.h"
-#include "nsQueryObject.h"
+#include "nsContentUtils.h"
+#include "nsDebug.h"
+#include "nsError.h"
+#include "nsID.h"
+#include "nsIEventTarget.h"
+#include "nsIFile.h"
+#include "nsIGlobalObject.h"
+#include "nsIInputStream.h"
+#include "nsIPrincipal.h"
+#include "nsISupports.h"
+#include "nsJSPrincipals.h"
+#include "nsPIDOMWindow.h"
+#include "nsString.h"
+#include "nsThreadUtils.h"
+#include "nsXPCOM.h"
+#include "xpcpublic.h"
 
 using namespace mozilla::ipc;
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 namespace {
 
@@ -73,14 +116,15 @@ bool StructuredCloneCallbacksWrite(JSContext* aCx,
 }
 
 bool StructuredCloneCallbacksReadTransfer(
-    JSContext* aCx, JSStructuredCloneReader* aReader, uint32_t aTag,
-    void* aContent, uint64_t aExtraData, void* aClosure,
-    JS::MutableHandleObject aReturnObject) {
+    JSContext* aCx, JSStructuredCloneReader* aReader,
+    const JS::CloneDataPolicy& aCloneDataPolicy, uint32_t aTag, void* aContent,
+    uint64_t aExtraData, void* aClosure,
+    JS::MutableHandle<JSObject*> aReturnObject) {
   StructuredCloneHolderBase* holder =
       static_cast<StructuredCloneHolderBase*>(aClosure);
   MOZ_ASSERT(holder);
-  return holder->CustomReadTransferHandler(aCx, aReader, aTag, aContent,
-                                           aExtraData, aReturnObject);
+  return holder->CustomReadTransferHandler(aCx, aReader, aCloneDataPolicy, aTag,
+                                           aContent, aExtraData, aReturnObject);
 }
 
 bool StructuredCloneCallbacksWriteTransfer(
@@ -133,7 +177,7 @@ bool StructuredCloneCallbacksSharedArrayBuffer(JSContext* cx, bool aReceiving,
 
     nsPIDOMWindowInner* innerWindow = nullptr;
     if (global) {
-      innerWindow = global->AsInnerWindow();
+      innerWindow = global->GetAsInnerWindow();
     }
 
     DocGroup* docGroup = nullptr;
@@ -174,7 +218,10 @@ void AssertTagValues() {
                     SCTAG_DOM_DOMMATRIX == 0xffff8013 &&
                     SCTAG_DOM_URLSEARCHPARAMS == 0xffff8014 &&
                     SCTAG_DOM_DOMMATRIXREADONLY == 0xffff8015 &&
-                    SCTAG_DOM_STRUCTUREDCLONETESTER == 0xffff8018,
+                    SCTAG_DOM_STRUCTUREDCLONETESTER == 0xffff8018 &&
+                    SCTAG_DOM_FILESYSTEMHANDLE == 0xffff8019 &&
+                    SCTAG_DOM_FILESYSTEMFILEHANDLE == 0xffff801a &&
+                    SCTAG_DOM_FILESYSTEMDIRECTORYHANDLE == 0xffff801b,
                 "Something has changed the sctag values. This is wrong!");
 }
 
@@ -262,9 +309,9 @@ bool StructuredCloneHolderBase::Read(
 }
 
 bool StructuredCloneHolderBase::CustomReadTransferHandler(
-    JSContext* aCx, JSStructuredCloneReader* aReader, uint32_t aTag,
-    void* aContent, uint64_t aExtraData,
-    JS::MutableHandleObject aReturnObject) {
+    JSContext* aCx, JSStructuredCloneReader* aReader,
+    const JS::CloneDataPolicy& aCloneDataPolicy, uint32_t aTag, void* aContent,
+    uint64_t aExtraData, JS::MutableHandle<JSObject*> aReturnObject) {
   MOZ_CRASH("Nothing to read.");
   return false;
 }
@@ -300,7 +347,7 @@ StructuredCloneHolder::StructuredCloneHolder(
       mGlobal(nullptr)
 #ifdef DEBUG
       ,
-      mCreationEventTarget(GetCurrentThreadEventTarget())
+      mCreationEventTarget(GetCurrentSerialEventTarget())
 #endif
 {
 }
@@ -343,6 +390,7 @@ void StructuredCloneHolder::Read(nsIGlobalObject* aGlobal, JSContext* aCx,
   mGlobal = aGlobal;
 
   if (!StructuredCloneHolderBase::Read(aCx, aValue, aCloneDataPolicy)) {
+    mTransferredPorts.Clear();
     JS_ClearPendingException(aCx);
     aRv.ThrowDataCloneError(mErrorMessage);
     return;
@@ -354,6 +402,9 @@ void StructuredCloneHolder::Read(nsIGlobalObject* aGlobal, JSContext* aCx,
     mWasmModuleArray.Clear();
     mClonedSurfaces.Clear();
     mInputStreamArray.Clear();
+    mVideoFrames.Clear();
+    mEncodedAudioChunks.Clear();
+    mEncodedVideoChunks.Clear();
     Clear();
   }
 }
@@ -384,9 +435,35 @@ void StructuredCloneHolder::ReadFromBuffer(
   }
 }
 
+static bool CheckExposedGlobals(JSContext* aCx, nsIGlobalObject* aGlobal,
+                                uint16_t aExposedGlobals) {
+  JS::Rooted<JSObject*> global(aCx, aGlobal->GetGlobalJSObject());
+
+  // Sandboxes aren't really DOM globals (though they do set the
+  // JSCLASS_DOM_GLOBAL flag), and so we can't simply do the exposure check.
+  // Some sandboxes do have a DOM global as their prototype, so using the
+  // prototype to check for exposure will at least make it work for those
+  // specific cases.
+  {
+    JSObject* proto = xpc::SandboxPrototypeOrNull(aCx, global);
+    if (proto) {
+      global = proto;
+    }
+  }
+
+  if (!IsGlobalInExposureSet(aCx, global, aExposedGlobals)) {
+    ErrorResult error;
+    error.ThrowDataCloneError("Interface is not exposed.");
+    MOZ_ALWAYS_TRUE(error.MaybeSetPendingException(aCx));
+    return false;
+  }
+  return true;
+}
+
 /* static */
 JSObject* StructuredCloneHolder::ReadFullySerializableObjects(
-    JSContext* aCx, JSStructuredCloneReader* aReader, uint32_t aTag) {
+    JSContext* aCx, JSStructuredCloneReader* aReader, uint32_t aTag,
+    bool aIsForIndexedDB) {
   AssertTagValues();
 
   nsIGlobalObject* global = xpc::CurrentNativeGlobal(aCx);
@@ -394,10 +471,29 @@ JSObject* StructuredCloneHolder::ReadFullySerializableObjects(
     return nullptr;
   }
 
-  WebIDLDeserializer deserializer =
+  Maybe<std::pair<uint16_t, WebIDLDeserializer>> deserializer =
       LookupDeserializer(StructuredCloneTags(aTag));
-  if (deserializer) {
-    return deserializer(aCx, global, aReader);
+  if (deserializer.isSome()) {
+    uint16_t exposedGlobals;
+    WebIDLDeserializer deserialize;
+    std::tie(exposedGlobals, deserialize) = deserializer.ref();
+
+    // https://html.spec.whatwg.org/multipage/structured-data.html#structureddeserialize
+    //
+    // 22. Otherwise:
+    //
+    //   1. Let interfaceName be serialized.[[Type]].
+    //   2. If the interface identified by interfaceName is not exposed in
+    //      targetRealm, then throw a "DataCloneError" DOMException.
+    //
+    // The special-casing for IndexedDB is because it uses a sandbox to
+    // deserialize, which means we don't actually have access to exposure
+    // information.
+    if (!aIsForIndexedDB && !CheckExposedGlobals(aCx, global, exposedGlobals)) {
+      return nullptr;
+    }
+
+    return deserialize(aCx, global, aReader);
   }
 
   if (aTag == SCTAG_DOM_NULL_PRINCIPAL || aTag == SCTAG_DOM_SYSTEM_PRINCIPAL ||
@@ -408,7 +504,7 @@ JSObject* StructuredCloneHolder::ReadFullySerializableObjects(
       return nullptr;
     }
 
-    JS::RootedValue result(aCx);
+    JS::Rooted<JS::Value> result(aCx);
     {
       // nsJSPrincipals::ReadKnownPrincipalType addrefs for us, but because of
       // the casting between JSPrincipals* and nsIPrincipal* we can't use
@@ -460,13 +556,17 @@ bool StructuredCloneHolder::WriteFullySerializableObjects(
   }
 
   // Don't know what this is
-  xpc::Throw(aCx, NS_ERROR_DOM_DATA_CLONE_ERR);
+  ErrorResult rv;
+  const char* className = JS::GetClass(obj)->name;
+  rv.ThrowDataCloneError(nsDependentCString(className) +
+                         " object could not be cloned."_ns);
+  MOZ_ALWAYS_TRUE(rv.MaybeSetPendingException(aCx));
   return false;
 }
 
-/* static */
-bool StructuredCloneHolder::ReadString(JSStructuredCloneReader* aReader,
-                                       nsString& aString) {
+template <typename char_type>
+static bool ReadTString(JSStructuredCloneReader* aReader,
+                        nsTString<char_type>& aString) {
   uint32_t length, zero;
   if (!JS_ReadUint32Pair(aReader, &length, &zero)) {
     return false;
@@ -475,17 +575,42 @@ bool StructuredCloneHolder::ReadString(JSStructuredCloneReader* aReader,
   if (NS_WARN_IF(!aString.SetLength(length, fallible))) {
     return false;
   }
-  size_t charSize = sizeof(nsString::char_type);
+  size_t charSize = sizeof(char_type);
   return JS_ReadBytes(aReader, (void*)aString.BeginWriting(),
                       length * charSize);
 }
 
+template <typename char_type>
+static bool WriteTString(JSStructuredCloneWriter* aWriter,
+                         const nsTSubstring<char_type>& aString) {
+  size_t charSize = sizeof(char_type);
+  return JS_WriteUint32Pair(aWriter, aString.Length(), 0) &&
+         JS_WriteBytes(aWriter, aString.BeginReading(),
+                       aString.Length() * charSize);
+}
+
+/* static */
+bool StructuredCloneHolder::ReadString(JSStructuredCloneReader* aReader,
+                                       nsString& aString) {
+  return ReadTString(aReader, aString);
+}
+
 /* static */
 bool StructuredCloneHolder::WriteString(JSStructuredCloneWriter* aWriter,
-                                        const nsString& aString) {
-  size_t charSize = sizeof(nsString::char_type);
-  return JS_WriteUint32Pair(aWriter, aString.Length(), 0) &&
-         JS_WriteBytes(aWriter, aString.get(), aString.Length() * charSize);
+                                        const nsAString& aString) {
+  return WriteTString(aWriter, aString);
+}
+
+/* static */
+bool StructuredCloneHolder::ReadCString(JSStructuredCloneReader* aReader,
+                                        nsCString& aString) {
+  return ReadTString(aReader, aString);
+}
+
+/* static */
+bool StructuredCloneHolder::WriteCString(JSStructuredCloneWriter* aWriter,
+                                         const nsACString& aString) {
+  return WriteTString(aWriter, aString);
 }
 
 namespace {
@@ -620,12 +745,10 @@ JSObject* ReadFileList(JSContext* aCx, JSStructuredCloneReader* aReader,
 
     uint32_t zero, index;
     // |index| is the index of the first blobImpl.
-    if (!JS_ReadUint32Pair(aReader, &zero, &index)) {
+    if (!JS_ReadUint32Pair(aReader, &zero, &index) || zero != 0) {
       return nullptr;
     }
 
-    MOZ_ASSERT(zero == 0);
-
     // |aCount| is the number of BlobImpls to use from the |index|.
     for (uint32_t i = 0; i < aCount; ++i) {
       uint32_t pos = index + i;
@@ -741,7 +864,9 @@ JSObject* ReadFormData(JSContext* aCx, JSStructuredCloneReader* aReader,
         formData->Append(name, directory);
 
       } else {
-        MOZ_ASSERT(tag == 0);
+        if (NS_WARN_IF(tag != 0)) {
+          return nullptr;
+        }
 
         nsAutoString value;
         if (NS_WARN_IF(!value.SetLength(indexOrLengthOfString, fallible))) {
@@ -793,52 +918,36 @@ bool WriteFormData(JSStructuredCloneWriter* aWriter, FormData* aFormData,
     return false;
   }
 
-  class MOZ_STACK_CLASS Closure final {
-    JSStructuredCloneWriter* mWriter;
-    StructuredCloneHolder* mHolder;
+  auto write = [aWriter, aHolder](
+                   const nsString& aName,
+                   const OwningBlobOrDirectoryOrUSVString& aValue) {
+    if (!StructuredCloneHolder::WriteString(aWriter, aName)) {
+      return false;
+    }
 
-   public:
-    Closure(JSStructuredCloneWriter* aWriter, StructuredCloneHolder* aHolder)
-        : mWriter(aWriter), mHolder(aHolder) {}
-
-    static bool Write(const nsString& aName,
-                      const OwningBlobOrDirectoryOrUSVString& aValue,
-                      void* aClosure) {
-      Closure* closure = static_cast<Closure*>(aClosure);
-      if (!StructuredCloneHolder::WriteString(closure->mWriter, aName)) {
+    if (aValue.IsBlob()) {
+      if (!JS_WriteUint32Pair(aWriter, SCTAG_DOM_BLOB,
+                              aHolder->BlobImpls().Length())) {
         return false;
       }
 
-      if (aValue.IsBlob()) {
-        if (!JS_WriteUint32Pair(closure->mWriter, SCTAG_DOM_BLOB,
-                                closure->mHolder->BlobImpls().Length())) {
-          return false;
-        }
-
-        RefPtr<BlobImpl> blobImpl = aValue.GetAsBlob()->Impl();
-
-        closure->mHolder->BlobImpls().AppendElement(blobImpl);
-        return true;
-      }
-
-      if (aValue.IsDirectory()) {
-        Directory* directory = aValue.GetAsDirectory();
-        return WriteDirectory(closure->mWriter, directory);
-      }
-
-      size_t charSize = sizeof(nsString::char_type);
-      if (!JS_WriteUint32Pair(closure->mWriter, 0,
-                              aValue.GetAsUSVString().Length()) ||
-          !JS_WriteBytes(closure->mWriter, aValue.GetAsUSVString().get(),
-                         aValue.GetAsUSVString().Length() * charSize)) {
-        return false;
-      }
+      RefPtr<BlobImpl> blobImpl = aValue.GetAsBlob()->Impl();
 
+      aHolder->BlobImpls().AppendElement(blobImpl);
       return true;
     }
+
+    if (aValue.IsDirectory()) {
+      Directory* directory = aValue.GetAsDirectory();
+      return WriteDirectory(aWriter, directory);
+    }
+
+    const size_t charSize = sizeof(nsString::char_type);
+    return JS_WriteUint32Pair(aWriter, 0, aValue.GetAsUSVString().Length()) &&
+           JS_WriteBytes(aWriter, aValue.GetAsUSVString().get(),
+                         aValue.GetAsUSVString().Length() * charSize);
   };
-  Closure closure(aWriter, aHolder);
-  return aFormData->ForEach(Closure::Write, &closure);
+  return aFormData->ForEach(write);
 }
 
 JSObject* ReadWasmModule(JSContext* aCx, uint32_t aIndex,
@@ -884,7 +993,7 @@ JSObject* ReadInputStream(JSContext* aCx, uint32_t aIndex,
   }
 #endif
   MOZ_ASSERT(aIndex < aHolder->InputStreams().Length());
-  JS::RootedValue result(aCx);
+  JS::Rooted<JS::Value> result(aCx);
   {
     nsCOMPtr<nsIInputStream> inputStream = aHolder->InputStreams()[aIndex];
 
@@ -917,6 +1026,12 @@ bool WriteInputStream(JSStructuredCloneWriter* aWriter,
 
 }  // anonymous namespace
 
+static const uint16_t sWindowOrWorker =
+    GlobalNames::DedicatedWorkerGlobalScope |
+    GlobalNames::ServiceWorkerGlobalScope |
+    GlobalNames::SharedWorkerGlobalScope | GlobalNames::Window |
+    GlobalNames::WorkerDebuggerGlobalScope;
+
 JSObject* StructuredCloneHolder::CustomReadHandler(
     JSContext* aCx, JSStructuredCloneReader* aReader,
     const JS::CloneDataPolicy& aCloneDataPolicy, uint32_t aTag,
@@ -924,26 +1039,41 @@ JSObject* StructuredCloneHolder::CustomReadHandler(
   MOZ_ASSERT(mSupportsCloning);
 
   if (aTag == SCTAG_DOM_BLOB) {
+    if (!CheckExposedGlobals(aCx, mGlobal, sWindowOrWorker)) {
+      return nullptr;
+    }
     return ReadBlob(aCx, aIndex, this);
   }
 
   if (aTag == SCTAG_DOM_DIRECTORY) {
+    if (!CheckExposedGlobals(aCx, mGlobal, sWindowOrWorker)) {
+      return nullptr;
+    }
     return ReadDirectory(aCx, aReader, aIndex, this);
   }
 
   if (aTag == SCTAG_DOM_FILELIST) {
+    if (!CheckExposedGlobals(aCx, mGlobal, sWindowOrWorker)) {
+      return nullptr;
+    }
     return ReadFileList(aCx, aReader, aIndex, this);
   }
 
   if (aTag == SCTAG_DOM_FORMDATA) {
+    if (!CheckExposedGlobals(aCx, mGlobal, sWindowOrWorker)) {
+      return nullptr;
+    }
     return ReadFormData(aCx, aReader, aIndex, this);
   }
 
   if (aTag == SCTAG_DOM_IMAGEBITMAP &&
       CloneScope() == StructuredCloneScope::SameProcess) {
+    if (!CheckExposedGlobals(aCx, mGlobal, sWindowOrWorker)) {
+      return nullptr;
+    }
     // Get the current global object.
     // This can be null.
-    JS::RootedObject result(aCx);
+    JS::Rooted<JSObject*> result(aCx);
     {
       // aIndex is the index of the cloned image.
       result = ImageBitmap::ReadStructuredClone(aCx, aReader, mGlobal,
@@ -967,14 +1097,64 @@ JSObject* StructuredCloneHolder::CustomReadHandler(
   }
 
   if (aTag == SCTAG_DOM_BROWSING_CONTEXT) {
+    if (!CheckExposedGlobals(aCx, mGlobal, GlobalNames::Window)) {
+      return nullptr;
+    }
     return BrowsingContext::ReadStructuredClone(aCx, aReader, this);
   }
 
   if (aTag == SCTAG_DOM_CLONED_ERROR_OBJECT) {
+    if (!CheckExposedGlobals(aCx, mGlobal, sWindowOrWorker)) {
+      return nullptr;
+    }
     return ClonedErrorHolder::ReadStructuredClone(aCx, aReader, this);
   }
 
-  return ReadFullySerializableObjects(aCx, aReader, aTag);
+  if (StaticPrefs::dom_media_webcodecs_enabled() &&
+      aTag == SCTAG_DOM_VIDEOFRAME &&
+      CloneScope() == StructuredCloneScope::SameProcess &&
+      aCloneDataPolicy.areIntraClusterClonableSharedObjectsAllowed()) {
+    JS::Rooted<JSObject*> global(aCx, mGlobal->GetGlobalJSObject());
+    if (VideoFrame_Binding::ConstructorEnabled(aCx, global)) {
+      return VideoFrame::ReadStructuredClone(aCx, mGlobal, aReader,
+                                             VideoFrames()[aIndex]);
+    }
+  }
+
+  if (StaticPrefs::dom_media_webcodecs_enabled() &&
+      aTag == SCTAG_DOM_ENCODEDVIDEOCHUNK &&
+      CloneScope() == StructuredCloneScope::SameProcess &&
+      aCloneDataPolicy.areIntraClusterClonableSharedObjectsAllowed()) {
+    JS::Rooted<JSObject*> global(aCx, mGlobal->GetGlobalJSObject());
+    if (EncodedVideoChunk_Binding::ConstructorEnabled(aCx, global)) {
+      return EncodedVideoChunk::ReadStructuredClone(
+          aCx, mGlobal, aReader, EncodedVideoChunks()[aIndex]);
+    }
+  }
+
+  if (StaticPrefs::dom_media_webcodecs_enabled() &&
+      aTag == SCTAG_DOM_AUDIODATA &&
+      CloneScope() == StructuredCloneScope::SameProcess &&
+      aCloneDataPolicy.areIntraClusterClonableSharedObjectsAllowed()) {
+    JS::Rooted<JSObject*> global(aCx, mGlobal->GetGlobalJSObject());
+    if (AudioData_Binding::ConstructorEnabled(aCx, global)) {
+      return AudioData::ReadStructuredClone(aCx, mGlobal, aReader,
+                                            AudioData()[aIndex]);
+    }
+  }
+
+  if (StaticPrefs::dom_media_webcodecs_enabled() &&
+      aTag == SCTAG_DOM_ENCODEDAUDIOCHUNK &&
+      CloneScope() == StructuredCloneScope::SameProcess &&
+      aCloneDataPolicy.areIntraClusterClonableSharedObjectsAllowed()) {
+    JS::Rooted<JSObject*> global(aCx, mGlobal->GetGlobalJSObject());
+    if (EncodedAudioChunk_Binding::ConstructorEnabled(aCx, global)) {
+      return EncodedAudioChunk::ReadStructuredClone(
+          aCx, mGlobal, aReader, EncodedAudioChunks()[aIndex]);
+    }
+  }
+
+  return ReadFullySerializableObjects(aCx, aReader, aTag, false);
 }
 
 bool StructuredCloneHolder::CustomWriteHandler(
@@ -1022,15 +1202,13 @@ bool StructuredCloneHolder::CustomWriteHandler(
   {
     ImageBitmap* imageBitmap = nullptr;
     if (NS_SUCCEEDED(UNWRAP_OBJECT(ImageBitmap, &obj, imageBitmap))) {
-      if (imageBitmap->IsWriteOnly()) {
-        return false;
-      }
-
       SameProcessScopeRequired(aSameProcessScopeRequired);
 
       if (CloneScope() == StructuredCloneScope::SameProcess) {
-        return ImageBitmap::WriteStructuredClone(aWriter, GetSurfaces(),
-                                                 imageBitmap);
+        ErrorResult rv;
+        ImageBitmap::WriteStructuredClone(aWriter, GetSurfaces(), imageBitmap,
+                                          rv);
+        return !rv.MaybeSetPendingException(aCx);
       }
       return false;
     }
@@ -1072,6 +1250,52 @@ bool StructuredCloneHolder::CustomWriteHandler(
     return false;
   }
 
+  // See if this is a VideoFrame object.
+  if (StaticPrefs::dom_media_webcodecs_enabled()) {
+    VideoFrame* videoFrame = nullptr;
+    if (NS_SUCCEEDED(UNWRAP_OBJECT(VideoFrame, &obj, videoFrame))) {
+      SameProcessScopeRequired(aSameProcessScopeRequired);
+      return CloneScope() == StructuredCloneScope::SameProcess
+                 ? videoFrame->WriteStructuredClone(aWriter, this)
+                 : false;
+    }
+  }
+
+  // See if this is a EncodedVideoChunk object.
+  if (StaticPrefs::dom_media_webcodecs_enabled()) {
+    EncodedVideoChunk* encodedVideoChunk = nullptr;
+    if (NS_SUCCEEDED(
+            UNWRAP_OBJECT(EncodedVideoChunk, &obj, encodedVideoChunk))) {
+      SameProcessScopeRequired(aSameProcessScopeRequired);
+      return CloneScope() == StructuredCloneScope::SameProcess
+                 ? encodedVideoChunk->WriteStructuredClone(aWriter, this)
+                 : false;
+    }
+  }
+
+  // See if this is an AudioData object.
+  if (StaticPrefs::dom_media_webcodecs_enabled()) {
+    mozilla::dom::AudioData* audioData = nullptr;
+    if (NS_SUCCEEDED(UNWRAP_OBJECT(AudioData, &obj, audioData))) {
+      SameProcessScopeRequired(aSameProcessScopeRequired);
+      return CloneScope() == StructuredCloneScope::SameProcess
+                 ? audioData->WriteStructuredClone(aWriter, this)
+                 : false;
+    }
+  }
+
+  // See if this is a EncodedAudioChunk object.
+  if (StaticPrefs::dom_media_webcodecs_enabled()) {
+    EncodedAudioChunk* encodedAudioChunk = nullptr;
+    if (NS_SUCCEEDED(
+            UNWRAP_OBJECT(EncodedAudioChunk, &obj, encodedAudioChunk))) {
+      SameProcessScopeRequired(aSameProcessScopeRequired);
+      return CloneScope() == StructuredCloneScope::SameProcess
+                 ? encodedAudioChunk->WriteStructuredClone(aWriter, this)
+                 : false;
+    }
+  }
+
   {
     // We only care about streams, so ReflectorToISupportsStatic is fine.
     nsCOMPtr<nsISupports> base = xpc::ReflectorToISupportsStatic(aObj);
@@ -1084,28 +1308,46 @@ bool StructuredCloneHolder::CustomWriteHandler(
   return WriteFullySerializableObjects(aCx, aWriter, aObj);
 }
 
-bool StructuredCloneHolder::CustomReadTransferHandler(
-    JSContext* aCx, JSStructuredCloneReader* aReader, uint32_t aTag,
-    void* aContent, uint64_t aExtraData,
-    JS::MutableHandleObject aReturnObject) {
+already_AddRefed<MessagePort> StructuredCloneHolder::ReceiveMessagePort(
+    uint64_t aIndex) {
+  if (NS_WARN_IF(aIndex >= mPortIdentifiers.Length())) {
+    return nullptr;
+  }
+  UniqueMessagePortId portId(mPortIdentifiers[aIndex]);
+
+  ErrorResult rv;
+  RefPtr<MessagePort> port = MessagePort::Create(mGlobal, portId, rv);
+  if (NS_WARN_IF(rv.Failed())) {
+    rv.SuppressException();
+    return nullptr;
+  }
+
+  return port.forget();
+}
+
+// TODO: Convert this to MOZ_CAN_RUN_SCRIPT (bug 1415230)
+MOZ_CAN_RUN_SCRIPT_BOUNDARY bool
+StructuredCloneHolder::CustomReadTransferHandler(
+    JSContext* aCx, JSStructuredCloneReader* aReader,
+    const JS::CloneDataPolicy& aCloneDataPolicy, uint32_t aTag, void* aContent,
+    uint64_t aExtraData, JS::MutableHandle<JSObject*> aReturnObject) {
   MOZ_ASSERT(mSupportsTransferring);
 
   if (aTag == SCTAG_DOM_MAP_MESSAGEPORT) {
+    if (!CheckExposedGlobals(
+            aCx, mGlobal,
+            sWindowOrWorker | GlobalNames::AudioWorkletGlobalScope)) {
+      return false;
+    }
 #ifdef FUZZING
     if (aExtraData >= mPortIdentifiers.Length()) {
       return false;
     }
 #endif
-    MOZ_ASSERT(aExtraData < mPortIdentifiers.Length());
-    UniqueMessagePortId portIdentifier(mPortIdentifiers[aExtraData]);
-
-    ErrorResult rv;
-    RefPtr<MessagePort> port = MessagePort::Create(mGlobal, portIdentifier, rv);
-    if (NS_WARN_IF(rv.Failed())) {
-      rv.SuppressException();
+    RefPtr<MessagePort> port = ReceiveMessagePort(aExtraData);
+    if (!port) {
       return false;
     }
-
     mTransferredPorts.AppendElement(port);
 
     JS::Rooted<JS::Value> value(aCx);
@@ -1120,12 +1362,14 @@ bool StructuredCloneHolder::CustomReadTransferHandler(
 
   if (aTag == SCTAG_DOM_CANVAS &&
       CloneScope() == StructuredCloneScope::SameProcess) {
+    if (!CheckExposedGlobals(aCx, mGlobal, sWindowOrWorker)) {
+      return false;
+    }
     MOZ_ASSERT(aContent);
     OffscreenCanvasCloneData* data =
         static_cast<OffscreenCanvasCloneData*>(aContent);
     RefPtr<OffscreenCanvas> canvas =
         OffscreenCanvas::CreateFromCloneData(mGlobal, data);
-    delete data;
 
     JS::Rooted<JS::Value> value(aCx);
     if (!GetOrCreateDOMReflector(aCx, canvas, &value)) {
@@ -1133,17 +1377,20 @@ bool StructuredCloneHolder::CustomReadTransferHandler(
       return false;
     }
 
+    delete data;
     aReturnObject.set(&value.toObject());
     return true;
   }
 
   if (aTag == SCTAG_DOM_IMAGEBITMAP &&
       CloneScope() == StructuredCloneScope::SameProcess) {
+    if (!CheckExposedGlobals(aCx, mGlobal, sWindowOrWorker)) {
+      return false;
+    }
     MOZ_ASSERT(aContent);
     ImageBitmapCloneData* data = static_cast<ImageBitmapCloneData*>(aContent);
     RefPtr<ImageBitmap> bitmap =
         ImageBitmap::CreateFromCloneData(mGlobal, data);
-    delete data;
 
     JS::Rooted<JS::Value> value(aCx);
     if (!GetOrCreateDOMReflector(aCx, bitmap, &value)) {
@@ -1151,6 +1398,116 @@ bool StructuredCloneHolder::CustomReadTransferHandler(
       return false;
     }
 
+    delete data;
+    aReturnObject.set(&value.toObject());
+    return true;
+  }
+
+  if (aTag == SCTAG_DOM_READABLESTREAM) {
+#ifdef FUZZING
+    if (aExtraData >= mPortIdentifiers.Length()) {
+      return false;
+    }
+#endif
+    RefPtr<MessagePort> port = ReceiveMessagePort(aExtraData);
+    if (!port) {
+      return false;
+    }
+    nsCOMPtr<nsIGlobalObject> global = mGlobal;
+    return ReadableStream::ReceiveTransfer(aCx, global, *port, aReturnObject);
+  }
+
+  if (aTag == SCTAG_DOM_WRITABLESTREAM) {
+#ifdef FUZZING
+    if (aExtraData >= mPortIdentifiers.Length()) {
+      return false;
+    }
+#endif
+    RefPtr<MessagePort> port = ReceiveMessagePort(aExtraData);
+    if (!port) {
+      return false;
+    }
+    nsCOMPtr<nsIGlobalObject> global = mGlobal;
+    return WritableStream::ReceiveTransfer(aCx, global, *port, aReturnObject);
+  }
+
+  if (aTag == SCTAG_DOM_TRANSFORMSTREAM) {
+#ifdef FUZZING
+    if (aExtraData + 1 >= mPortIdentifiers.Length()) {
+      return false;
+    }
+#endif
+    RefPtr<MessagePort> port1 = ReceiveMessagePort(aExtraData);
+    RefPtr<MessagePort> port2 = ReceiveMessagePort(aExtraData + 1);
+    if (!port1 || !port2) {
+      return false;
+    }
+    nsCOMPtr<nsIGlobalObject> global = mGlobal;
+    return TransformStream::ReceiveTransfer(aCx, global, *port1, *port2,
+                                            aReturnObject);
+  }
+
+  if (StaticPrefs::dom_media_webcodecs_enabled() &&
+      aTag == SCTAG_DOM_VIDEOFRAME &&
+      CloneScope() == StructuredCloneScope::SameProcess &&
+      aCloneDataPolicy.areIntraClusterClonableSharedObjectsAllowed()) {
+    MOZ_ASSERT(aContent);
+
+    JS::Rooted<JSObject*> globalObj(aCx, mGlobal->GetGlobalJSObject());
+    // aContent will be released in CustomFreeTransferHandler.
+    if (!VideoFrame_Binding::ConstructorEnabled(aCx, globalObj)) {
+      return false;
+    }
+
+    VideoFrame::TransferredData* data =
+        static_cast<VideoFrame::TransferredData*>(aContent);
+    nsCOMPtr<nsIGlobalObject> global = mGlobal;
+    RefPtr<VideoFrame> frame = VideoFrame::FromTransferred(global.get(), data);
+    // aContent will be released in CustomFreeTransferHandler if frame is null.
+    if (!frame) {
+      return false;
+    }
+
+    JS::Rooted<JS::Value> value(aCx);
+    if (!GetOrCreateDOMReflector(aCx, frame, &value)) {
+      JS_ClearPendingException(aCx);
+      return false;
+    }
+    delete data;
+    aContent = nullptr;
+    aReturnObject.set(&value.toObject());
+    return true;
+  }
+
+  if (StaticPrefs::dom_media_webcodecs_enabled() &&
+      aTag == SCTAG_DOM_AUDIODATA &&
+      CloneScope() == StructuredCloneScope::SameProcess &&
+      aCloneDataPolicy.areIntraClusterClonableSharedObjectsAllowed()) {
+    MOZ_ASSERT(aContent);
+
+    JS::Rooted<JSObject*> globalObj(aCx, mGlobal->GetGlobalJSObject());
+    // aContent will be released in CustomFreeTransferHandler.
+    if (!AudioData_Binding::ConstructorEnabled(aCx, globalObj)) {
+      return false;
+    }
+
+    AudioData::TransferredData* data =
+        static_cast<AudioData::TransferredData*>(aContent);
+    nsCOMPtr<nsIGlobalObject> global = mGlobal;
+    RefPtr<mozilla::dom::AudioData> audioData =
+        AudioData::FromTransferred(global.get(), data);
+    // aContent will be released in CustomFreeTransferHandler if frame is null.
+    if (!audioData) {
+      return false;
+    }
+
+    JS::Rooted<JS::Value> value(aCx);
+    if (!GetOrCreateDOMReflector(aCx, audioData, &value)) {
+      JS_ClearPendingException(aCx);
+      return false;
+    }
+    delete data;
+    aContent = nullptr;
     aReturnObject.set(&value.toObject());
     return true;
   }
@@ -1158,7 +1515,9 @@ bool StructuredCloneHolder::CustomReadTransferHandler(
   return false;
 }
 
-bool StructuredCloneHolder::CustomWriteTransferHandler(
+// TODO: Convert this to MOZ_CAN_RUN_SCRIPT (bug 1415230)
+MOZ_CAN_RUN_SCRIPT_BOUNDARY bool
+StructuredCloneHolder::CustomWriteTransferHandler(
     JSContext* aCx, JS::Handle<JSObject*> aObj, uint32_t* aTag,
     JS::TransferableOwnership* aOwnership, void** aContent,
     uint64_t* aExtraData) {
@@ -1184,8 +1543,8 @@ bool StructuredCloneHolder::CustomWriteTransferHandler(
       mPortIdentifiers.AppendElement(identifier.release());
 
       *aTag = SCTAG_DOM_MAP_MESSAGEPORT;
-      *aOwnership = JS::SCTAG_TMO_CUSTOM;
       *aContent = nullptr;
+      *aOwnership = JS::SCTAG_TMO_CUSTOM;
 
       return true;
     }
@@ -1196,16 +1555,17 @@ bool StructuredCloneHolder::CustomWriteTransferHandler(
       if (NS_SUCCEEDED(rv)) {
         MOZ_ASSERT(canvas);
 
-        if (canvas->IsNeutered()) {
+        UniquePtr<OffscreenCanvasCloneData> clonedCanvas =
+            canvas->ToCloneData(aCx);
+        if (!clonedCanvas) {
           return false;
         }
 
         *aExtraData = 0;
         *aTag = SCTAG_DOM_CANVAS;
-        *aOwnership = JS::SCTAG_TMO_CUSTOM;
-        *aContent = canvas->ToCloneData();
+        *aContent = clonedCanvas.release();
         MOZ_ASSERT(*aContent);
-        canvas->SetNeutered();
+        *aOwnership = JS::SCTAG_TMO_CUSTOM;
 
         return true;
       }
@@ -1218,7 +1578,6 @@ bool StructuredCloneHolder::CustomWriteTransferHandler(
 
         *aExtraData = 0;
         *aTag = SCTAG_DOM_IMAGEBITMAP;
-        *aOwnership = JS::SCTAG_TMO_CUSTOM;
 
         UniquePtr<ImageBitmapCloneData> clonedBitmap = bitmap->ToCloneData();
         if (!clonedBitmap) {
@@ -1227,10 +1586,115 @@ bool StructuredCloneHolder::CustomWriteTransferHandler(
 
         *aContent = clonedBitmap.release();
         MOZ_ASSERT(*aContent);
+        *aOwnership = JS::SCTAG_TMO_CUSTOM;
+
         bitmap->Close();
 
         return true;
       }
+
+      if (StaticPrefs::dom_media_webcodecs_enabled()) {
+        VideoFrame* videoFrame = nullptr;
+        rv = UNWRAP_OBJECT(VideoFrame, &obj, videoFrame);
+        if (NS_SUCCEEDED(rv)) {
+          MOZ_ASSERT(videoFrame);
+
+          *aExtraData = 0;
+          *aTag = SCTAG_DOM_VIDEOFRAME;
+          *aContent = nullptr;
+
+          UniquePtr<VideoFrame::TransferredData> data = videoFrame->Transfer();
+          if (!data) {
+            return false;
+          }
+          *aContent = data.release();
+          MOZ_ASSERT(*aContent);
+          *aOwnership = JS::SCTAG_TMO_CUSTOM;
+          return true;
+        }
+      }
+      if (StaticPrefs::dom_media_webcodecs_enabled()) {
+        mozilla::dom::AudioData* audioData = nullptr;
+        rv = UNWRAP_OBJECT(AudioData, &obj, audioData);
+        if (NS_SUCCEEDED(rv)) {
+          MOZ_ASSERT(audioData);
+
+          *aExtraData = 0;
+          *aTag = SCTAG_DOM_AUDIODATA;
+          *aContent = nullptr;
+
+          UniquePtr<AudioData::TransferredData> data = audioData->Transfer();
+          if (!data) {
+            return false;
+          }
+          *aContent = data.release();
+          MOZ_ASSERT(*aContent);
+          *aOwnership = JS::SCTAG_TMO_CUSTOM;
+          return true;
+        }
+      }
+    }
+
+    {
+      RefPtr<ReadableStream> stream;
+      rv = UNWRAP_OBJECT(ReadableStream, &obj, stream);
+      if (NS_SUCCEEDED(rv)) {
+        MOZ_ASSERT(stream);
+
+        *aTag = SCTAG_DOM_READABLESTREAM;
+        *aContent = nullptr;
+
+        UniqueMessagePortId id;
+        if (!stream->Transfer(aCx, id)) {
+          return false;
+        }
+        *aExtraData = mPortIdentifiers.Length();
+        mPortIdentifiers.AppendElement(id.release());
+        *aOwnership = JS::SCTAG_TMO_CUSTOM;
+        return true;
+      }
+    }
+
+    {
+      RefPtr<WritableStream> stream;
+      rv = UNWRAP_OBJECT(WritableStream, &obj, stream);
+      if (NS_SUCCEEDED(rv)) {
+        MOZ_ASSERT(stream);
+
+        *aTag = SCTAG_DOM_WRITABLESTREAM;
+        *aContent = nullptr;
+
+        UniqueMessagePortId id;
+        if (!stream->Transfer(aCx, id)) {
+          return false;
+        }
+        *aExtraData = mPortIdentifiers.Length();
+        mPortIdentifiers.AppendElement(id.release());
+        *aOwnership = JS::SCTAG_TMO_CUSTOM;
+        return true;
+      }
+    }
+
+    {
+      RefPtr<TransformStream> stream;
+      rv = UNWRAP_OBJECT(TransformStream, &obj, stream);
+      if (NS_SUCCEEDED(rv)) {
+        MOZ_ASSERT(stream);
+
+        *aTag = SCTAG_DOM_TRANSFORMSTREAM;
+        *aContent = nullptr;
+
+        UniqueMessagePortId id1;
+        UniqueMessagePortId id2;
+        if (!stream->Transfer(aCx, id1, id2)) {
+          return false;
+        }
+        *aExtraData = mPortIdentifiers.Length();
+        mPortIdentifiers.AppendElement(id1.release());
+        mPortIdentifiers.AppendElement(id2.release());
+        *aOwnership = JS::SCTAG_TMO_CUSTOM;
+        return true;
+      }
     }
   }
 
@@ -1270,6 +1734,52 @@ void StructuredCloneHolder::CustomFreeTransferHandler(
     delete data;
     return;
   }
+
+  if (aTag == SCTAG_DOM_READABLESTREAM || aTag == SCTAG_DOM_WRITABLESTREAM) {
+    MOZ_ASSERT(!aContent);
+#ifdef FUZZING
+    if (aExtraData >= mPortIdentifiers.Length()) {
+      return;
+    }
+#endif
+    MOZ_ASSERT(aExtraData < mPortIdentifiers.Length());
+    MessagePort::ForceClose(mPortIdentifiers[aExtraData]);
+    return;
+  }
+
+  if (aTag == SCTAG_DOM_TRANSFORMSTREAM) {
+    MOZ_ASSERT(!aContent);
+#ifdef FUZZING
+    if (aExtraData + 1 >= mPortIdentifiers.Length()) {
+      return;
+    }
+#endif
+    MOZ_ASSERT(aExtraData + 1 < mPortIdentifiers.Length());
+    MessagePort::ForceClose(mPortIdentifiers[aExtraData]);
+    MessagePort::ForceClose(mPortIdentifiers[aExtraData + 1]);
+    return;
+  }
+
+  if (StaticPrefs::dom_media_webcodecs_enabled() &&
+      aTag == SCTAG_DOM_VIDEOFRAME &&
+      CloneScope() == StructuredCloneScope::SameProcess) {
+    if (aContent) {
+      VideoFrame::TransferredData* data =
+          static_cast<VideoFrame::TransferredData*>(aContent);
+      delete data;
+    }
+    return;
+  }
+  if (StaticPrefs::dom_media_webcodecs_enabled() &&
+      aTag == SCTAG_DOM_AUDIODATA &&
+      CloneScope() == StructuredCloneScope::SameProcess) {
+    if (aContent) {
+      AudioData::TransferredData* data =
+          static_cast<AudioData::TransferredData*>(aContent);
+      delete data;
+    }
+    return;
+  }
 }
 
 bool StructuredCloneHolder::CustomCanTransferHandler(
@@ -1311,6 +1821,57 @@ bool StructuredCloneHolder::CustomCanTransferHandler(
     }
   }
 
+  {
+    ReadableStream* stream = nullptr;
+    nsresult rv = UNWRAP_OBJECT(ReadableStream, &obj, stream);
+    if (NS_SUCCEEDED(rv)) {
+      // https://streams.spec.whatwg.org/#ref-for-transfer-steps
+      // Step 1: If ! IsReadableStreamLocked(value) is true, throw a
+      // "DataCloneError" DOMException.
+      return !stream->Locked();
+    }
+  }
+
+  {
+    WritableStream* stream = nullptr;
+    nsresult rv = UNWRAP_OBJECT(WritableStream, &obj, stream);
+    if (NS_SUCCEEDED(rv)) {
+      // https://streams.spec.whatwg.org/#ref-for-transfer-steps①
+      // Step 1: If ! IsWritableStreamLocked(value) is true, throw a
+      // "DataCloneError" DOMException.
+      return !stream->Locked();
+    }
+  }
+
+  {
+    TransformStream* stream = nullptr;
+    nsresult rv = UNWRAP_OBJECT(TransformStream, &obj, stream);
+    if (NS_SUCCEEDED(rv)) {
+      // https://streams.spec.whatwg.org/#ref-for-transfer-steps②
+      // Step 3 + 4: If ! Is{Readable,Writable}StreamLocked(value) is true,
+      // throw a "DataCloneError" DOMException.
+      return !stream->Readable()->Locked() && !stream->Writable()->Locked();
+    }
+  }
+
+  if (StaticPrefs::dom_media_webcodecs_enabled()) {
+    VideoFrame* videoframe = nullptr;
+    nsresult rv = UNWRAP_OBJECT(VideoFrame, &obj, videoframe);
+    if (NS_SUCCEEDED(rv)) {
+      SameProcessScopeRequired(aSameProcessScopeRequired);
+      return CloneScope() == StructuredCloneScope::SameProcess;
+    }
+  }
+
+  if (StaticPrefs::dom_media_webcodecs_enabled()) {
+    mozilla::dom::AudioData* audioData = nullptr;
+    nsresult rv = UNWRAP_OBJECT(AudioData, &obj, audioData);
+    if (NS_SUCCEEDED(rv)) {
+      SameProcessScopeRequired(aSameProcessScopeRequired);
+      return CloneScope() == StructuredCloneScope::SameProcess;
+    }
+  }
+
   return false;
 }
 
@@ -1337,5 +1898,4 @@ void StructuredCloneHolder::SameProcessScopeRequired(
   }
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
