@@ -515,7 +515,7 @@ class nsAttrValue {
   // aType can be ePercent or eDoubleValue.
   void SetDoubleValueAndType(double aValue, ValueType aType,
                              const nsAString* aStringValue);
-  void SetColorValue(nscolor aColor, const nsAString& aString);
+  bool SetColorValue(nscolor aColor, const nsAString& aString);
   void SetMiscAtomOrString(const nsAString* aValue);
   void ResetMiscAtomOrString();
   void SetSVGType(ValueType aType, const void* aValue,
