@@ -5,6 +5,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "nsCopySupport.h"
+#include "nsGlobalWindowInner.h"
 #include "nsIDocumentEncoder.h"
 #include "nsISupports.h"
 #include "nsIContent.h"
@@ -22,9 +23,10 @@
 #include "nsServiceManagerUtils.h"
 #include "mozilla/ScopeExit.h"
 #include "mozilla/dom/DataTransfer.h"
+#include "mozilla/dom/BrowsingContext.h"
 
 #include "nsIDocShell.h"
-#include "nsIContentViewerEdit.h"
+#include "nsIDocumentViewerEdit.h"
 #include "nsISelectionController.h"
 
 #include "nsPIDOMWindow.h"
@@ -32,16 +34,14 @@
 #include "nsHTMLDocument.h"
 #include "nsGkAtoms.h"
 #include "nsIFrame.h"
-#include "nsIURI.h"
-#include "nsGenericHTMLElement.h"
 
 // image copy stuff
 #include "nsIImageLoadingContent.h"
 #include "nsIInterfaceRequestorUtils.h"
 #include "nsContentUtils.h"
-#include "nsContentCID.h"
 
 #ifdef XP_WIN
+#  include "mozilla/StaticPrefs_clipboard.h"
 #  include "nsCExternalHandlerService.h"
 #  include "nsEscape.h"
 #  include "nsIMIMEInfo.h"
@@ -81,7 +81,7 @@ static nsresult AppendDOMNode(nsITransferable* aTransferable,
 // copy image as file promise onto the transferable
 static nsresult AppendImagePromise(nsITransferable* aTransferable,
                                    imgIRequest* aImgRequest,
-                                   nsIImageLoadingContent* aImageElement);
+                                   nsINode* aImageNode);
 #endif
 
 static nsresult EncodeForTextUnicode(nsIDocumentEncoder& aEncoder,
@@ -95,13 +95,14 @@ static nsresult EncodeForTextUnicode(nsIDocumentEncoder& aEncoder,
   // html content with pre-wrap style : text/plain. Otherwise text/html. see
   // nsHTMLCopyEncoder::SetSelection
   nsAutoString mimeType;
-  mimeType.AssignLiteral(kUnicodeMime);
+  mimeType.AssignLiteral("text/unicode");
 
   // Do the first and potentially trial encoding as preformatted and raw.
   uint32_t flags = aAdditionalEncoderFlags |
                    nsIDocumentEncoder::OutputPreformatted |
                    nsIDocumentEncoder::OutputRaw |
-                   nsIDocumentEncoder::OutputForPlainTextClipboardCopy;
+                   nsIDocumentEncoder::OutputForPlainTextClipboardCopy |
+                   nsIDocumentEncoder::OutputPersistNBSP;
 
   nsresult rv = aEncoder.Init(&aDocument, mimeType, flags);
   NS_ENSURE_SUCCESS(rv, rv);
@@ -172,7 +173,7 @@ static nsresult EncodeAsTextHTMLWithContext(
 }
 
 struct EncodedDocumentWithContext {
-  // When determening `mSerializationForTextUnicode`, `text/unicode` is passed
+  // When determining `mSerializationForTextUnicode`, `text/unicode` is passed
   // as mime type to the encoder. It uses this as a switch to decide whether to
   // encode the document as `text/html` or `text/plain`. It  is `true` iff
   // `text/html` was used.
@@ -244,6 +245,7 @@ static nsresult CreateTransferable(
   NS_ENSURE_TRUE(aTransferable, NS_ERROR_NULL_POINTER);
 
   aTransferable->Init(aDocument.GetLoadContext());
+  aTransferable->SetDataPrincipal(aDocument.NodePrincipal());
   if (aEncodedDocumentWithContext.mUnicodeEncodingIsTextHTML) {
     // Set up a format converter so that clipboard flavor queries work.
     // This converter isn't really used for conversions.
@@ -275,14 +277,13 @@ static nsresult CreateTransferable(
 
     if (!aEncodedDocumentWithContext.mSerializationForTextUnicode.IsEmpty()) {
       // unicode text
-      // Add the unicode DataFlavor to the transferable
+      // Add the plain text DataFlavor to the transferable
       // If we didn't have this, then nsDataObj::GetData matches
-      // text/unicode against the kURLMime flavour which is not desirable
+      // text/plain against the kURLMime flavour which is not desirable
       // (eg. when pasting into Notepad)
-      rv =
-          AppendString(aTransferable,
-                       aEncodedDocumentWithContext.mSerializationForTextUnicode,
-                       kUnicodeMime);
+      rv = AppendString(
+          aTransferable,
+          aEncodedDocumentWithContext.mSerializationForTextUnicode, kTextMime);
       NS_ENSURE_SUCCESS(rv, rv);
     }
 
@@ -309,10 +310,9 @@ static nsresult CreateTransferable(
   } else {
     if (!aEncodedDocumentWithContext.mSerializationForTextUnicode.IsEmpty()) {
       // Add the unicode DataFlavor to the transferable
-      rv =
-          AppendString(aTransferable,
-                       aEncodedDocumentWithContext.mSerializationForTextUnicode,
-                       kUnicodeMime);
+      rv = AppendString(
+          aTransferable,
+          aEncodedDocumentWithContext.mSerializationForTextUnicode, kTextMime);
       NS_ENSURE_SUCCESS(rv, rv);
     }
   }
@@ -332,7 +332,8 @@ static nsresult PutToClipboard(
   rv = CreateTransferable(aEncodedDocumentWithContext, aDocument, transferable);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  rv = clipboard->SetData(transferable, nullptr, aClipboardID);
+  rv = clipboard->SetData(transferable, nullptr, aClipboardID,
+                          aDocument.GetWindowContext());
   NS_ENSURE_SUCCESS(rv, rv);
 
   return rv;
@@ -418,7 +419,8 @@ nsresult nsCopySupport::GetTransferableForNode(
   if (NS_WARN_IF(result.Failed())) {
     return result.StealNSResult();
   }
-  selection->AddRangeAndSelectFramesAndNotifyListeners(*range, aDoc, result);
+  selection->AddRangeAndSelectFramesAndNotifyListenersInternal(*range, aDoc,
+                                                               result);
   if (NS_WARN_IF(result.Failed())) {
     return result.StealNSResult();
   }
@@ -454,17 +456,21 @@ nsresult nsCopySupport::GetContents(const nsACString& aMimeType,
   return docEncoder->EncodeToString(outdata);
 }
 
-nsresult nsCopySupport::ImageCopy(nsIImageLoadingContent* aImageElement,
-                                  nsILoadContext* aLoadContext,
-                                  int32_t aCopyFlags) {
+nsresult nsCopySupport::ImageCopy(
+    nsIImageLoadingContent* aImageElement, nsILoadContext* aLoadContext,
+    int32_t aCopyFlags, mozilla::dom::WindowContext* aSettingWindowContext) {
   nsresult rv;
 
+  nsCOMPtr<nsINode> imageNode = do_QueryInterface(aImageElement, &rv);
+  NS_ENSURE_SUCCESS(rv, rv);
+
   // create a transferable for putting data on the Clipboard
   nsCOMPtr<nsITransferable> trans(do_CreateInstance(kCTransferableCID, &rv));
   NS_ENSURE_SUCCESS(rv, rv);
   trans->Init(aLoadContext);
+  trans->SetDataPrincipal(imageNode->NodePrincipal());
 
-  if (aCopyFlags & nsIContentViewerEdit::COPY_IMAGE_TEXT) {
+  if (aCopyFlags & nsIDocumentViewerEdit::COPY_IMAGE_TEXT) {
     // get the location from the element
     nsCOMPtr<nsIURI> uri;
     rv = aImageElement->GetCurrentURI(getter_AddRefs(uri));
@@ -476,11 +482,11 @@ nsresult nsCopySupport::ImageCopy(nsIImageLoadingContent* aImageElement,
     NS_ENSURE_SUCCESS(rv, rv);
 
     // append the string to the transferable
-    rv = AppendString(trans, NS_ConvertUTF8toUTF16(location), kUnicodeMime);
+    rv = AppendString(trans, NS_ConvertUTF8toUTF16(location), kTextMime);
     NS_ENSURE_SUCCESS(rv, rv);
   }
 
-  if (aCopyFlags & nsIContentViewerEdit::COPY_IMAGE_HTML) {
+  if (aCopyFlags & nsIDocumentViewerEdit::COPY_IMAGE_HTML) {
     // append HTML data to the transferable
     nsCOMPtr<nsINode> node(do_QueryInterface(aImageElement, &rv));
     NS_ENSURE_SUCCESS(rv, rv);
@@ -489,16 +495,25 @@ nsresult nsCopySupport::ImageCopy(nsIImageLoadingContent* aImageElement,
     NS_ENSURE_SUCCESS(rv, rv);
   }
 
-  if (aCopyFlags & nsIContentViewerEdit::COPY_IMAGE_DATA) {
+  if (aCopyFlags & nsIDocumentViewerEdit::COPY_IMAGE_DATA) {
     // get the image data and its request from the element
     nsCOMPtr<imgIRequest> imgRequest;
     nsCOMPtr<imgIContainer> image = nsContentUtils::GetImageFromContent(
         aImageElement, getter_AddRefs(imgRequest));
     NS_ENSURE_TRUE(image, NS_ERROR_FAILURE);
 
+    if (imgRequest) {
+      // Remember the referrer used for this image request.
+      nsCOMPtr<nsIReferrerInfo> referrerInfo;
+      imgRequest->GetReferrerInfo(getter_AddRefs(referrerInfo));
+      trans->SetReferrerInfo(referrerInfo);
+    }
+
 #ifdef XP_WIN
-    rv = AppendImagePromise(trans, imgRequest, aImageElement);
-    NS_ENSURE_SUCCESS(rv, rv);
+    if (StaticPrefs::clipboard_imageAsFile_enabled()) {
+      rv = AppendImagePromise(trans, imgRequest, imageNode);
+      NS_ENSURE_SUCCESS(rv, rv);
+    }
 #endif
 
     // copy the image data onto the transferable
@@ -511,17 +526,15 @@ nsresult nsCopySupport::ImageCopy(nsIImageLoadingContent* aImageElement,
   NS_ENSURE_SUCCESS(rv, rv);
 
   // check whether the system supports the selection clipboard or not.
-  bool selectionSupported;
-  rv = clipboard->SupportsSelectionClipboard(&selectionSupported);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  // put the transferable on the clipboard
-  if (selectionSupported) {
-    rv = clipboard->SetData(trans, nullptr, nsIClipboard::kSelectionClipboard);
+  if (clipboard->IsClipboardTypeSupported(nsIClipboard::kSelectionClipboard)) {
+    // put the transferable on the clipboard
+    rv = clipboard->SetData(trans, nullptr, nsIClipboard::kSelectionClipboard,
+                            aSettingWindowContext);
     NS_ENSURE_SUCCESS(rv, rv);
   }
 
-  return clipboard->SetData(trans, nullptr, nsIClipboard::kGlobalClipboard);
+  return clipboard->SetData(trans, nullptr, nsIClipboard::kGlobalClipboard,
+                            aSettingWindowContext);
 }
 
 static nsresult AppendString(nsITransferable* aTransferable,
@@ -589,10 +602,10 @@ static nsresult AppendDOMNode(nsITransferable* aTransferable,
 #ifdef XP_WIN
 static nsresult AppendImagePromise(nsITransferable* aTransferable,
                                    imgIRequest* aImgRequest,
-                                   nsIImageLoadingContent* aImageElement) {
+                                   nsINode* aImageNode) {
   nsresult rv;
 
-  NS_ENSURE_TRUE(aImgRequest, NS_OK);
+  NS_ENSURE_TRUE(aImgRequest && aImageNode, NS_OK);
 
   bool isMultipart;
   rv = aImgRequest->GetMultipart(&isMultipart);
@@ -601,76 +614,43 @@ static nsresult AppendImagePromise(nsITransferable* aTransferable,
     return NS_OK;
   }
 
-  nsCOMPtr<nsINode> node = do_QueryInterface(aImageElement, &rv);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  // Fix the file extension in the URL if necessary
-  nsCOMPtr<nsIMIMEService> mimeService =
-      do_GetService(NS_MIMESERVICE_CONTRACTID);
-  NS_ENSURE_TRUE(mimeService, NS_OK);
+  nsCOMPtr<nsIMIMEService> mimeService = do_GetService("@mozilla.org/mime;1");
+  if (NS_WARN_IF(!mimeService)) {
+    return NS_ERROR_FAILURE;
+  }
 
   nsCOMPtr<nsIURI> imgUri;
   rv = aImgRequest->GetFinalURI(getter_AddRefs(imgUri));
   NS_ENSURE_SUCCESS(rv, rv);
 
-  nsCOMPtr<nsIURL> imgUrl = do_QueryInterface(imgUri);
-  NS_ENSURE_TRUE(imgUrl, NS_OK);
-
-  nsAutoCString extension;
-  rv = imgUrl->GetFileExtension(extension);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  nsCString mimeType;
-  rv = aImgRequest->GetMimeType(getter_Copies(mimeType));
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  nsCOMPtr<nsIMIMEInfo> mimeInfo;
-  mimeService->GetFromTypeAndExtension(mimeType, ""_ns,
-                                       getter_AddRefs(mimeInfo));
-  NS_ENSURE_TRUE(mimeInfo, NS_OK);
-
   nsAutoCString spec;
-  rv = imgUrl->GetSpec(spec);
+  rv = imgUri->GetSpec(spec);
   NS_ENSURE_SUCCESS(rv, rv);
 
   // pass out the image source string
   nsString imageSourceString;
   CopyUTF8toUTF16(spec, imageSourceString);
 
-  bool validExtension;
-  if (extension.IsEmpty() ||
-      NS_FAILED(mimeInfo->ExtensionExists(extension, &validExtension)) ||
-      !validExtension) {
-    // Fix the file extension in the URL
-    nsAutoCString primaryExtension;
-    mimeInfo->GetPrimaryExtension(primaryExtension);
-    if (!primaryExtension.IsEmpty()) {
-      rv = NS_MutateURI(imgUri)
-               .Apply(NS_MutatorMethod(&nsIURLMutator::SetFileExtension,
-                                       primaryExtension, nullptr))
-               .Finalize(imgUrl);
-      NS_ENSURE_SUCCESS(rv, rv);
-    }
-  }
+  nsCString mimeType;
+  rv = aImgRequest->GetMimeType(getter_Copies(mimeType));
+  NS_ENSURE_SUCCESS(rv, rv);
 
   nsAutoCString fileName;
-  imgUrl->GetFileName(fileName);
+  rv = aImgRequest->GetFileName(fileName);
+  NS_ENSURE_SUCCESS(rv, rv);
 
-  NS_UnescapeURL(fileName);
-
-  // make the filename safe for the filesystem
-  fileName.ReplaceChar(FILE_PATH_SEPARATOR FILE_ILLEGAL_CHARACTERS, '-');
-
-  nsString imageDestFileName;
-  CopyUTF8toUTF16(fileName, imageDestFileName);
+  nsAutoString validFileName = NS_ConvertUTF8toUTF16(fileName);
+  mimeService->ValidateFileNameForSaving(
+      validFileName, mimeType, nsIMIMEService::VALIDATE_DEFAULT, validFileName);
 
   rv = AppendString(aTransferable, imageSourceString, kFilePromiseURLMime);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  rv = AppendString(aTransferable, imageDestFileName, kFilePromiseDestFilename);
+  rv = AppendString(aTransferable, validFileName, kFilePromiseDestFilename);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  aTransferable->SetRequestingPrincipal(node->NodePrincipal());
+  aTransferable->SetCookieJarSettings(
+      aImageNode->OwnerDoc()->CookieJarSettings());
   aTransferable->SetContentPolicyType(nsIContentPolicy::TYPE_INTERNAL_IMAGE);
 
   // add the dataless file promise flavor
@@ -681,12 +661,12 @@ static nsresult AppendImagePromise(nsITransferable* aTransferable,
 already_AddRefed<Selection> nsCopySupport::GetSelectionForCopy(
     Document* aDocument) {
   PresShell* presShell = aDocument->GetPresShell();
-  if (!presShell) {
+  if (NS_WARN_IF(!presShell)) {
     return nullptr;
   }
 
   RefPtr<nsFrameSelection> frameSel = presShell->GetLastFocusedFrameSelection();
-  if (!frameSel) {
+  if (NS_WARN_IF(!frameSel)) {
     return nullptr;
   }
 
@@ -695,7 +675,9 @@ already_AddRefed<Selection> nsCopySupport::GetSelectionForCopy(
 }
 
 bool nsCopySupport::CanCopy(Document* aDocument) {
-  if (!aDocument) return false;
+  if (!aDocument) {
+    return false;
+  }
 
   RefPtr<Selection> sel = GetSelectionForCopy(aDocument);
   return sel && !sel->IsCollapsed();
@@ -713,6 +695,7 @@ static bool IsInsideRuby(nsINode* aNode) {
 static bool IsSelectionInsideRuby(Selection* aSelection) {
   uint32_t rangeCount = aSelection->RangeCount();
   for (auto i : IntegerRange(rangeCount)) {
+    MOZ_ASSERT(aSelection->RangeCount() == rangeCount);
     const nsRange* range = aSelection->GetRangeAt(i);
     if (!IsInsideRuby(range->GetClosestCommonInclusiveAncestor())) {
       return false;
@@ -734,6 +717,33 @@ static Element* GetElementOrNearestFlattenedTreeParentElement(nsINode* aNode) {
   return nullptr;
 }
 
+/**
+ * This class is used while processing clipboard paste event.
+ */
+class MOZ_RAII AutoHandlingPasteEvent final {
+ public:
+  explicit AutoHandlingPasteEvent(nsGlobalWindowInner* aWindow,
+                                  DataTransfer* aDataTransfer,
+                                  const EventMessage& aEventMessage,
+                                  const int32_t& aClipboardType) {
+    MOZ_ASSERT(aDataTransfer);
+    if (aWindow && aEventMessage == ePaste &&
+        aClipboardType == nsIClipboard::kGlobalClipboard) {
+      aWindow->SetCurrentPasteDataTransfer(aDataTransfer);
+      mInnerWindow = aWindow;
+    }
+  }
+
+  ~AutoHandlingPasteEvent() {
+    if (mInnerWindow) {
+      mInnerWindow->SetCurrentPasteDataTransfer(nullptr);
+    }
+  }
+
+ private:
+  RefPtr<nsGlobalWindowInner> mInnerWindow;
+};
+
 bool nsCopySupport::FireClipboardEvent(EventMessage aEventMessage,
                                        int32_t aClipboardType,
                                        PresShell* aPresShell,
@@ -810,8 +820,17 @@ bool nsCopySupport::FireClipboardEvent(EventMessage aEventMessage,
     nsEventStatus status = nsEventStatus_eIgnore;
     InternalClipboardEvent evt(true, originalEventMessage);
     evt.mClipboardData = clipboardData;
-    EventDispatcher::Dispatch(targetElement, presShell->GetPresContext(), &evt,
-                              nullptr, &status);
+
+    {
+      AutoHandlingPasteEvent autoHandlingPasteEvent(
+          nsGlobalWindowInner::Cast(doc->GetInnerWindow()), clipboardData,
+          aEventMessage, aClipboardType);
+
+      RefPtr<nsPresContext> presContext = presShell->GetPresContext();
+      EventDispatcher::Dispatch(targetElement, presContext, &evt, nullptr,
+                                &status);
+    }
+
     // If the event was cancelled, don't do the clipboard operation
     doDefault = (status != nsEventStatus_eConsumeNoDefault);
   }
@@ -912,7 +931,12 @@ bool nsCopySupport::FireClipboardEvent(EventMessage aEventMessage,
       NS_ENSURE_TRUE(transferable, false);
 
       // put the transferable on the clipboard
-      nsresult rv = clipboard->SetData(transferable, nullptr, aClipboardType);
+      WindowContext* settingWindowContext = nullptr;
+      if (aPresShell && aPresShell->GetDocument()) {
+        settingWindowContext = aPresShell->GetDocument()->GetWindowContext();
+      }
+      nsresult rv = clipboard->SetData(transferable, nullptr, aClipboardType,
+                                       settingWindowContext);
       if (NS_FAILED(rv)) {
         return false;
       }
@@ -922,7 +946,7 @@ bool nsCopySupport::FireClipboardEvent(EventMessage aEventMessage,
   // Now that we have copied, update the clipboard commands. This should have
   // the effect of updating the enabled state of the paste menu item.
   if (doDefault || count) {
-    piWindow->UpdateCommands(u"clipboard"_ns, nullptr, 0);
+    piWindow->UpdateCommands(u"clipboard"_ns);
   }
 
   if (aActionTaken) {
