@@ -9,69 +9,178 @@
 #include "gfxTextRun.h"  // for the gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_* values
 #include "nsHyphenationManager.h"
 #include "nsHyphenator.h"
+#include "mozilla/AutoRestore.h"
+#include "mozilla/ClearOnShutdown.h"
 #include "mozilla/gfx/2D.h"
-#include "mozilla/intl/LineBreaker.h"
+#include "mozilla/intl/LineBreaker.h"  // for LineBreaker::ComputeBreakPositions
+#include "mozilla/intl/Locale.h"
+#include "mozilla/intl/UnicodeProperties.h"
+#include "mozilla/ScopeExit.h"
+#include "mozilla/StaticPrefs_intl.h"
 
+using mozilla::AutoRestore;
 using mozilla::intl::LineBreaker;
+using mozilla::intl::LineBreakRule;
+using mozilla::intl::Locale;
+using mozilla::intl::LocaleParser;
+using mozilla::intl::UnicodeProperties;
+using mozilla::intl::WordBreakRule;
+
+// There is no break opportunity between any pair of characters that has line
+// break class of either AL (Alphabetic), IS (Infix Numeric Separator), NU
+// (Numeric), or QU (Quotation). See
+// https://www.unicode.org/Public/UCD/latest/ucd/LineBreak.txt for Unicode code
+// point and line break class mapping.
+static constexpr uint8_t kNonBreakableASCII[] = {
+    // clang-format off
+// 0x20-0x2f
+0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0, 1, 0,
+// 0x30-0x3f
+1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0,
+// 0x40-0x4f
+1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+// 0x50-0x5f
+1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1,
+// 0x60-0x6f
+1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+// 0x70-0x7f
+1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0,
+    // clang-format on
+};
+
+template <typename T>
+static constexpr bool IsNonBreakableChar(T aChar, bool aLegacyBehavior) {
+  if (aLegacyBehavior) {
+    // If not using ICU4X, line break rules aren't compatible with UAX#14. Use
+    // old way.
+    return (0x0030 <= aChar && aChar <= 0x0039) ||
+           (0x0041 <= aChar && aChar <= 0x005A) ||
+           (0x0061 <= aChar && aChar <= 0x007A) || (0x000a == aChar);
+  }
+  if (aChar < 0x20 || aChar > 0x7f) {
+    return false;
+  }
+  return !!kNonBreakableASCII[aChar - 0x20];
+}
 
 nsLineBreaker::nsLineBreaker()
     : mCurrentWordLanguage(nullptr),
       mCurrentWordContainsMixedLang(false),
-      mCurrentWordContainsComplexChar(false),
+      mScriptIsChineseOrJapanese(false),
       mAfterBreakableSpace(false),
       mBreakHere(false),
-      mWordBreak(LineBreaker::kWordBreak_Normal) {}
+      mWordBreak(WordBreakRule::Normal),
+      mLineBreak(LineBreakRule::Auto),
+      mWordContinuation(false),
+      mLegacyBehavior(!mozilla::StaticPrefs::intl_icu4x_segmenter_enabled()) {}
 
 nsLineBreaker::~nsLineBreaker() {
   NS_ASSERTION(mCurrentWord.Length() == 0,
                "Should have Reset() before destruction!");
 }
 
+/* static */
+bool nsLineBreaker::ShouldCapitalize(uint32_t aChar, bool& aCapitalizeNext) {
+  using mozilla::intl::GeneralCategory;
+  auto category = UnicodeProperties::CharType(aChar);
+  switch (category) {
+    case GeneralCategory::Uppercase_Letter:
+    case GeneralCategory::Lowercase_Letter:
+    case GeneralCategory::Titlecase_Letter:
+    case GeneralCategory::Modifier_Letter:
+    case GeneralCategory::Other_Letter:
+    case GeneralCategory::Decimal_Number:
+    case GeneralCategory::Letter_Number:
+    case GeneralCategory::Other_Number:
+      if (aCapitalizeNext) {
+        aCapitalizeNext = false;
+        return true;
+      }
+      break;
+    case GeneralCategory::Space_Separator:
+    case GeneralCategory::Line_Separator:
+    case GeneralCategory::Paragraph_Separator:
+    case GeneralCategory::Dash_Punctuation:
+    case GeneralCategory::Initial_Punctuation:
+      /* These punctuation categories are excluded, for examples like
+       *   "what colo[u]r" -> "What Colo[u]r?" (rather than "What Colo[U]R?")
+       * and
+       *   "snake_case" -> "Snake_case" (to match word selection behavior)
+      case GeneralCategory::Open_Punctuation:
+      case GeneralCategory::Close_Punctuation:
+      case GeneralCategory::Connector_Punctuation:
+       */
+      aCapitalizeNext = true;
+      break;
+    case GeneralCategory::Final_Punctuation:
+      /* Special-case: exclude Unicode single-close-quote/apostrophe,
+         for examples like "Lowe’s" etc. */
+      if (aChar != 0x2019) {
+        aCapitalizeNext = true;
+      }
+      break;
+    case GeneralCategory::Other_Punctuation:
+      /* Special-case: exclude ASCII apostrophe, for "Lowe's" etc.,
+         and MIDDLE DOT, for Catalan "l·l". */
+      if (aChar != '\'' && aChar != 0x00B7) {
+        aCapitalizeNext = true;
+      }
+      break;
+    default:
+      break;
+  }
+  return false;
+}
+
 static void SetupCapitalization(const char16_t* aWord, uint32_t aLength,
                                 bool* aCapitalization) {
-  // Capitalize the first alphanumeric character after a space or start
-  // of the word.
-  // The only space character a word can contain is NBSP.
+  // Capitalize the first alphanumeric character after a space or punctuation.
   bool capitalizeNextChar = true;
   for (uint32_t i = 0; i < aLength; ++i) {
     uint32_t ch = aWord[i];
-    if (capitalizeNextChar) {
-      if (NS_IS_HIGH_SURROGATE(ch) && i + 1 < aLength &&
-          NS_IS_LOW_SURROGATE(aWord[i + 1])) {
-        ch = SURROGATE_TO_UCS4(ch, aWord[i + 1]);
-      }
-      if (nsContentUtils::IsAlphanumeric(ch)) {
-        aCapitalization[i] = true;
-        capitalizeNextChar = false;
-      }
-      if (!IS_IN_BMP(ch)) {
-        ++i;
-      }
+    if (i + 1 < aLength && NS_IS_SURROGATE_PAIR(ch, aWord[i + 1])) {
+      ch = SURROGATE_TO_UCS4(ch, aWord[i + 1]);
     }
-    if (ch == 0xA0 /*NBSP*/) {
-      capitalizeNextChar = true;
+    aCapitalization[i] =
+        nsLineBreaker::ShouldCapitalize(ch, capitalizeNextChar);
+
+    if (!IS_IN_BMP(ch)) {
+      ++i;
     }
   }
 }
 
 nsresult nsLineBreaker::FlushCurrentWord() {
+  auto cleanup = mozilla::MakeScopeExit([&] {
+    mCurrentWord.Clear();
+    mTextItems.Clear();
+    mCurrentWordMightBeBreakable = false;
+    mCurrentWordContainsMixedLang = false;
+    mCurrentWordLanguage = nullptr;
+    mWordContinuation = false;
+  });
+
   uint32_t length = mCurrentWord.Length();
   AutoTArray<uint8_t, 4000> breakState;
-  if (!breakState.AppendElements(length)) return NS_ERROR_OUT_OF_MEMORY;
+  if (!breakState.AppendElements(length, mozilla::fallible)) {
+    return NS_ERROR_OUT_OF_MEMORY;
+  }
 
-  nsTArray<bool> capitalizationState;
-
-  if (!mCurrentWordContainsComplexChar) {
-    // For break-strict set everything internal to "break", otherwise
-    // to "no break"!
+  if (mLineBreak == LineBreakRule::Anywhere) {
     memset(breakState.Elements(),
-           mWordBreak == LineBreaker::kWordBreak_BreakAll
-               ? gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NORMAL
-               : gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NONE,
+           gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NORMAL,
+           length * sizeof(uint8_t));
+  } else if (!mCurrentWordMightBeBreakable &&
+             mWordBreak != WordBreakRule::BreakAll) {
+    // word-break: normal or keep-all has no break opportunity if the word
+    // is non-breakable. (See the comment of kNonBreakableASCII).
+    memset(breakState.Elements(),
+           gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NONE,
            length * sizeof(uint8_t));
   } else {
-    nsContentUtils::LineBreaker()->GetJISx4051Breaks(
-        mCurrentWord.Elements(), length, mWordBreak, breakState.Elements());
+    LineBreaker::ComputeBreakPositions(
+        mCurrentWord.Elements(), length, mWordBreak, mLineBreak,
+        mScriptIsChineseOrJapanese, breakState.Elements());
   }
 
   bool autoHyphenate = mCurrentWordLanguage && !mCurrentWordContainsMixedLang;
@@ -92,6 +201,7 @@ nsresult nsLineBreaker::FlushCurrentWord() {
     }
   }
 
+  nsTArray<bool> capitalizationState;
   uint32_t offset = 0;
   for (i = 0; i < mTextItems.Length(); ++i) {
     TextItem* ti = &mTextItems[i];
@@ -115,10 +225,11 @@ nsresult nsLineBreaker::FlushCurrentWord() {
       ti->mSink->SetBreaks(ti->mSinkOffset + skipSet, ti->mLength - skipSet,
                            breakState.Elements() + offset + skipSet);
 
-      if (ti->mFlags & BREAK_NEED_CAPITALIZATION) {
+      if (!mWordContinuation && (ti->mFlags & BREAK_NEED_CAPITALIZATION)) {
         if (capitalizationState.Length() == 0) {
-          if (!capitalizationState.AppendElements(length))
+          if (!capitalizationState.AppendElements(length, mozilla::fallible)) {
             return NS_ERROR_OUT_OF_MEMORY;
+          }
           memset(capitalizationState.Elements(), false, length * sizeof(bool));
           SetupCapitalization(mCurrentWord.Elements(), length,
                               capitalizationState.Elements());
@@ -131,11 +242,6 @@ nsresult nsLineBreaker::FlushCurrentWord() {
     offset += ti->mLength;
   }
 
-  mCurrentWord.Clear();
-  mTextItems.Clear();
-  mCurrentWordContainsComplexChar = false;
-  mCurrentWordContainsMixedLang = false;
-  mCurrentWordLanguage = nullptr;
   return NS_OK;
 }
 
@@ -158,10 +264,11 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
     NS_ASSERTION(!mAfterBreakableSpace && !mBreakHere,
                  "These should not be set");
 
-    while (offset < aLength && !IsSpace(aText[offset])) {
+    while (offset < aLength && !IsSegmentSpace(aText[offset])) {
       mCurrentWord.AppendElement(aText[offset]);
-      if (!mCurrentWordContainsComplexChar && IsComplexChar(aText[offset])) {
-        mCurrentWordContainsComplexChar = true;
+      if (!mCurrentWordMightBeBreakable &&
+          !IsNonBreakableChar<char16_t>(aText[offset], mLegacyBehavior)) {
+        mCurrentWordMightBeBreakable = true;
       }
       UpdateCurrentWordLanguage(aHyphenationLanguage);
       ++offset;
@@ -171,23 +278,30 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
       mTextItems.AppendElement(TextItem(aSink, 0, offset, aFlags));
     }
 
-    if (offset == aLength) return NS_OK;
+    if (offset == aLength) {
+      return NS_OK;
+    }
 
     // We encountered whitespace, so we're done with this word
     nsresult rv = FlushCurrentWord();
-    if (NS_FAILED(rv)) return rv;
+    if (NS_FAILED(rv)) {
+      return rv;
+    }
   }
 
   AutoTArray<uint8_t, 4000> breakState;
   if (aSink) {
-    if (!breakState.AppendElements(aLength)) return NS_ERROR_OUT_OF_MEMORY;
+    if (!breakState.AppendElements(aLength, mozilla::fallible)) {
+      return NS_ERROR_OUT_OF_MEMORY;
+    }
   }
 
   bool noCapitalizationNeeded = true;
   nsTArray<bool> capitalizationState;
   if (aSink && (aFlags & BREAK_NEED_CAPITALIZATION)) {
-    if (!capitalizationState.AppendElements(aLength))
+    if (!capitalizationState.AppendElements(aLength, mozilla::fallible)) {
       return NS_ERROR_OUT_OF_MEMORY;
+    }
     memset(capitalizationState.Elements(), false, aLength * sizeof(bool));
     noCapitalizationNeeded = false;
   }
@@ -204,11 +318,13 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
     offset = aLength;
     while (offset > start) {
       --offset;
-      if (IsSpace(aText[offset])) break;
+      if (IsSegmentSpace(aText[offset])) {
+        break;
+      }
     }
   }
   uint32_t wordStart = offset;
-  bool wordHasComplexChar = false;
+  bool wordMightBeBreakable = false;
 
   RefPtr<nsHyphenator> hyphenator;
   if ((aFlags & BREAK_USE_AUTO_HYPHENATION) &&
@@ -219,13 +335,14 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
 
   for (;;) {
     char16_t ch = aText[offset];
-    bool isSpace = IsSpace(ch);
+    bool isSpace = IsSegmentSpace(ch);
     bool isBreakableSpace = isSpace && !(aFlags & BREAK_SUPPRESS_INSIDE);
 
     if (aSink && !noBreaksNeeded) {
       breakState[offset] =
           mBreakHere || (mAfterBreakableSpace && !isBreakableSpace) ||
-                  (mWordBreak == LineBreaker::kWordBreak_BreakAll)
+                  mWordBreak == WordBreakRule::BreakAll ||
+                  mLineBreak == LineBreakRule::Anywhere
               ? gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NORMAL
               : gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NONE;
     }
@@ -235,47 +352,57 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
     if (isSpace || ch == '\n') {
       if (offset > wordStart && aSink) {
         if (!(aFlags & BREAK_SUPPRESS_INSIDE)) {
-          if (wordHasComplexChar) {
-            // Save current start-of-word state because GetJISx4051Breaks will
-            // set it to false
-            uint8_t currentStart = breakState[wordStart];
-            nsContentUtils::LineBreaker()->GetJISx4051Breaks(
-                aText + wordStart, offset - wordStart, mWordBreak,
-                breakState.Elements() + wordStart);
-            breakState[wordStart] = currentStart;
+          if (mLineBreak == LineBreakRule::Anywhere) {
+            memset(breakState.Elements() + wordStart,
+                   gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NORMAL,
+                   offset - wordStart);
+          } else if (wordMightBeBreakable) {
+            // Save current start-of-word state because ComputeBreakPositions()
+            // will set it to false.
+            AutoRestore<uint8_t> saveWordStartBreakState(breakState[wordStart]);
+            LineBreaker::ComputeBreakPositions(
+                aText + wordStart, offset - wordStart, mWordBreak, mLineBreak,
+                mScriptIsChineseOrJapanese, breakState.Elements() + wordStart);
           }
           if (hyphenator) {
             FindHyphenationPoints(hyphenator, aText + wordStart, aText + offset,
                                   breakState.Elements() + wordStart);
           }
         }
-        if (!noCapitalizationNeeded) {
+        if (!mWordContinuation && !noCapitalizationNeeded) {
           SetupCapitalization(aText + wordStart, offset - wordStart,
                               capitalizationState.Elements() + wordStart);
         }
       }
-      wordHasComplexChar = false;
-      ++offset;
-      if (offset >= aLength) break;
-      wordStart = offset;
-    } else {
-      if (!wordHasComplexChar && IsComplexChar(ch)) {
-        wordHasComplexChar = true;
-      }
+      wordMightBeBreakable = false;
+      mWordContinuation = false;
       ++offset;
       if (offset >= aLength) {
-        // Save this word
-        mCurrentWordContainsComplexChar = wordHasComplexChar;
-        uint32_t len = offset - wordStart;
-        char16_t* elems = mCurrentWord.AppendElements(len);
-        if (!elems) return NS_ERROR_OUT_OF_MEMORY;
-        memcpy(elems, aText + wordStart, sizeof(char16_t) * len);
-        mTextItems.AppendElement(TextItem(aSink, wordStart, len, aFlags));
-        // Ensure that the break-before for this word is written out
-        offset = wordStart + 1;
-        UpdateCurrentWordLanguage(aHyphenationLanguage);
         break;
       }
+      wordStart = offset;
+      continue;
+    }
+
+    if (!wordMightBeBreakable &&
+        !IsNonBreakableChar<char16_t>(ch, mLegacyBehavior)) {
+      wordMightBeBreakable = true;
+    }
+    ++offset;
+    if (offset >= aLength) {
+      // Save this word
+      mCurrentWordMightBeBreakable = wordMightBeBreakable;
+      uint32_t len = offset - wordStart;
+      char16_t* elems = mCurrentWord.AppendElements(len, mozilla::fallible);
+      if (!elems) {
+        return NS_ERROR_OUT_OF_MEMORY;
+      }
+      memcpy(elems, aText + wordStart, sizeof(char16_t) * len);
+      mTextItems.AppendElement(TextItem(aSink, wordStart, len, aFlags));
+      // Ensure that the break-before for this word is written out
+      offset = wordStart + 1;
+      UpdateCurrentWordLanguage(aHyphenationLanguage);
+      break;
     }
   }
 
@@ -327,11 +454,11 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
     NS_ASSERTION(!mAfterBreakableSpace && !mBreakHere,
                  "These should not be set");
 
-    while (offset < aLength && !IsSpace(aText[offset])) {
+    while (offset < aLength && !IsSegmentSpace(aText[offset])) {
       mCurrentWord.AppendElement(aText[offset]);
-      if (!mCurrentWordContainsComplexChar &&
-          IsComplexASCIIChar(aText[offset])) {
-        mCurrentWordContainsComplexChar = true;
+      if (!mCurrentWordMightBeBreakable &&
+          !IsNonBreakableChar<uint8_t>(aText[offset], mLegacyBehavior)) {
+        mCurrentWordMightBeBreakable = true;
       }
       ++offset;
     }
@@ -347,12 +474,16 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
 
     // We encountered whitespace, so we're done with this word
     nsresult rv = FlushCurrentWord();
-    if (NS_FAILED(rv)) return rv;
+    if (NS_FAILED(rv)) {
+      return rv;
+    }
   }
 
   AutoTArray<uint8_t, 4000> breakState;
   if (aSink) {
-    if (!breakState.AppendElements(aLength)) return NS_ERROR_OUT_OF_MEMORY;
+    if (!breakState.AppendElements(aLength, mozilla::fallible)) {
+      return NS_ERROR_OUT_OF_MEMORY;
+    }
   }
 
   uint32_t start = offset;
@@ -367,15 +498,17 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
     offset = aLength;
     while (offset > start) {
       --offset;
-      if (IsSpace(aText[offset])) break;
+      if (IsSegmentSpace(aText[offset])) {
+        break;
+      }
     }
   }
   uint32_t wordStart = offset;
-  bool wordHasComplexChar = false;
+  bool wordMightBeBreakable = false;
 
   for (;;) {
     uint8_t ch = aText[offset];
-    bool isSpace = IsSpace(ch);
+    bool isSpace = IsSegmentSpace(ch);
     bool isBreakableSpace = isSpace && !(aFlags & BREAK_SUPPRESS_INSIDE);
 
     if (aSink) {
@@ -383,7 +516,8 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
       // will be set by nsILineBreaker, we don't consider CJK at this point.
       breakState[offset] =
           mBreakHere || (mAfterBreakableSpace && !isBreakableSpace) ||
-                  (mWordBreak == LineBreaker::kWordBreak_BreakAll)
+                  mWordBreak == WordBreakRule::BreakAll ||
+                  mLineBreak == LineBreakRule::Anywhere
               ? gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NORMAL
               : gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NONE;
     }
@@ -391,42 +525,52 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
     mAfterBreakableSpace = isBreakableSpace;
 
     if (isSpace) {
-      if (offset > wordStart && wordHasComplexChar) {
-        if (aSink && !(aFlags & BREAK_SUPPRESS_INSIDE)) {
-          // Save current start-of-word state because GetJISx4051Breaks will
-          // set it to false
-          uint8_t currentStart = breakState[wordStart];
-          nsContentUtils::LineBreaker()->GetJISx4051Breaks(
-              aText + wordStart, offset - wordStart, mWordBreak,
-              breakState.Elements() + wordStart);
-          breakState[wordStart] = currentStart;
+      if (offset > wordStart && aSink && !(aFlags & BREAK_SUPPRESS_INSIDE)) {
+        if (mLineBreak == LineBreakRule::Anywhere) {
+          memset(breakState.Elements() + wordStart,
+                 gfxTextRun::CompressedGlyph::FLAG_BREAK_TYPE_NORMAL,
+                 offset - wordStart);
+        } else if (wordMightBeBreakable) {
+          // Save current start-of-word state because ComputeBreakPositions()
+          // will set it to false.
+          AutoRestore<uint8_t> saveWordStartBreakState(breakState[wordStart]);
+          LineBreaker::ComputeBreakPositions(
+              aText + wordStart, offset - wordStart, mWordBreak, mLineBreak,
+              mScriptIsChineseOrJapanese, breakState.Elements() + wordStart);
         }
-        wordHasComplexChar = false;
       }
 
-      ++offset;
-      if (offset >= aLength) break;
-      wordStart = offset;
-    } else {
-      if (!wordHasComplexChar && IsComplexASCIIChar(ch)) {
-        wordHasComplexChar = true;
-      }
+      wordMightBeBreakable = false;
+      mWordContinuation = false;
       ++offset;
       if (offset >= aLength) {
-        // Save this word
-        mCurrentWordContainsComplexChar = wordHasComplexChar;
-        uint32_t len = offset - wordStart;
-        char16_t* elems = mCurrentWord.AppendElements(len);
-        if (!elems) return NS_ERROR_OUT_OF_MEMORY;
-        uint32_t i;
-        for (i = wordStart; i < offset; ++i) {
-          elems[i - wordStart] = aText[i];
-        }
-        mTextItems.AppendElement(TextItem(aSink, wordStart, len, aFlags));
-        // Ensure that the break-before for this word is written out
-        offset = wordStart + 1;
         break;
       }
+      wordStart = offset;
+      continue;
+    }
+
+    if (!wordMightBeBreakable &&
+        !IsNonBreakableChar<uint8_t>(ch, mLegacyBehavior)) {
+      wordMightBeBreakable = true;
+    }
+    ++offset;
+    if (offset >= aLength) {
+      // Save this word
+      mCurrentWordMightBeBreakable = wordMightBeBreakable;
+      uint32_t len = offset - wordStart;
+      char16_t* elems = mCurrentWord.AppendElements(len, mozilla::fallible);
+      if (!elems) {
+        return NS_ERROR_OUT_OF_MEMORY;
+      }
+      uint32_t i;
+      for (i = wordStart; i < offset; ++i) {
+        elems[i - wordStart] = aText[i];
+      }
+      mTextItems.AppendElement(TextItem(aSink, wordStart, len, aFlags));
+      // Ensure that the break-before for this word is written out
+      offset = wordStart + 1;
+      break;
     }
   }
 
@@ -439,26 +583,66 @@ nsresult nsLineBreaker::AppendText(nsAtom* aHyphenationLanguage,
 void nsLineBreaker::UpdateCurrentWordLanguage(nsAtom* aHyphenationLanguage) {
   if (mCurrentWordLanguage && mCurrentWordLanguage != aHyphenationLanguage) {
     mCurrentWordContainsMixedLang = true;
-  } else {
-    mCurrentWordLanguage = aHyphenationLanguage;
+    mScriptIsChineseOrJapanese = false;
+    return;
   }
+
+  if (aHyphenationLanguage && !mCurrentWordLanguage) {
+    static mozilla::StaticRefPtr<nsAtom> sLastHyphenationLanguage;
+    static bool sLastScriptIsChineseOrJapanese = false;
+    static bool sInit = false;
+
+    if (!sInit) {
+      mozilla::ClearOnShutdown(&sLastHyphenationLanguage);
+      sInit = true;
+    }
+
+    if (sLastHyphenationLanguage == aHyphenationLanguage) {
+      MOZ_ASSERT(nsAtomString(sLastHyphenationLanguage)
+                     .Equals(nsAtomString(aHyphenationLanguage)));
+      mScriptIsChineseOrJapanese = sLastScriptIsChineseOrJapanese;
+    } else {
+      Locale loc;
+      auto result =
+          LocaleParser::TryParse(nsAtomCString(aHyphenationLanguage), loc);
+
+      if (result.isErr()) {
+        return;
+      }
+      if (loc.Script().Missing() && loc.AddLikelySubtags().isErr()) {
+        return;
+      }
+      mScriptIsChineseOrJapanese =
+          loc.Script().EqualTo("Hans") || loc.Script().EqualTo("Hant") ||
+          loc.Script().EqualTo("Jpan") || loc.Script().EqualTo("Hrkt");
+
+      sLastHyphenationLanguage = aHyphenationLanguage;
+      sLastScriptIsChineseOrJapanese = mScriptIsChineseOrJapanese;
+    }
+  }
+  mCurrentWordLanguage = aHyphenationLanguage;
 }
 
 nsresult nsLineBreaker::AppendInvisibleWhitespace(uint32_t aFlags) {
   nsresult rv = FlushCurrentWord();
-  if (NS_FAILED(rv)) return rv;
+  if (NS_FAILED(rv)) {
+    return rv;
+  }
 
   bool isBreakableSpace = !(aFlags & BREAK_SUPPRESS_INSIDE);
   if (mAfterBreakableSpace && !isBreakableSpace) {
     mBreakHere = true;
   }
   mAfterBreakableSpace = isBreakableSpace;
+  mWordContinuation = false;
   return NS_OK;
 }
 
 nsresult nsLineBreaker::Reset(bool* aTrailingBreak) {
   nsresult rv = FlushCurrentWord();
-  if (NS_FAILED(rv)) return rv;
+  if (NS_FAILED(rv)) {
+    return rv;
+  }
 
   *aTrailingBreak = mBreakHere || mAfterBreakableSpace;
   mBreakHere = false;
