@@ -26,10 +26,6 @@
 #include "WebGLValidateStrings.h"
 #include "WebGLVertexArray.h"
 
-#if defined(MOZ_WIDGET_COCOA)
-#  include "nsCocoaFeatures.h"
-#endif
-
 ////////////////////
 // Minimum value constants defined in GLES 2.0.25 $6.2 "State Tables":
 const uint32_t kMinMaxVertexAttribs = 8;            // Page 164
@@ -202,6 +198,8 @@ static webgl::Limits MakeLimits(const WebGLContext& webgl) {
   // GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS value is the accurate value.
   gl.GetUIntegerv(LOCAL_GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS,
                   &limits.maxTexUnits);
+  limits.maxTexUnits = std::min(
+      limits.maxTexUnits, uint32_t{UINT8_MAX});  // We want to use uint8_t.
 
   gl.GetUIntegerv(LOCAL_GL_MAX_TEXTURE_SIZE, &limits.maxTex2dSize);
   gl.GetUIntegerv(LOCAL_GL_MAX_CUBE_MAP_TEXTURE_SIZE, &limits.maxTexCubeSize);
@@ -486,16 +484,6 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
     gl->fEnable(LOCAL_GL_PROGRAM_POINT_SIZE);
   }
 
-#ifdef XP_MACOSX
-  if (gl->WorkAroundDriverBugs() && gl->Vendor() == gl::GLVendor::ATI &&
-      !nsCocoaFeatures::IsAtLeastVersion(10, 9)) {
-    // The Mac ATI driver, in all known OSX version up to and including
-    // 10.8, renders points sprites upside-down. (Apple bug 11778921)
-    gl->fPointParameterf(LOCAL_GL_POINT_SPRITE_COORD_ORIGIN,
-                         LOCAL_GL_LOWER_LEFT);
-  }
-#endif
-
   if (gl->IsSupported(gl::GLFeature::seamless_cube_map_opt_in)) {
     gl->fEnable(LOCAL_GL_TEXTURE_CUBE_MAP_SEAMLESS);
   }
@@ -571,10 +559,10 @@ bool WebGLContext::InitAndValidateGL(FailureReason* const out_failReason) {
   mMaybeNeedsLegacyVertexAttrib0Handling |= mNeedsLegacyVertexAttrib0Handling;
 
   if (const auto& env =
-          PR_GetEnv("MOZ_WEBGL_WORKAROUND_FIRST_AFFECTS_INSTANCE_ID")) {
+          gfxEnv::MOZ_WEBGL_WORKAROUND_FIRST_AFFECTS_INSTANCE_ID()) {
     const auto was = mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst;
     mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst =
-        (std::string{env} != "0");
+        (env.as_str != "0");
     printf_stderr(
         "mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst: %i -> %i\n",
         int(was), int(mBug_DrawArraysInstancedUserAttribFetchAffectedByFirst));
