@@ -6,15 +6,21 @@
 
 #include "Geolocation.h"
 
+#include "GeolocationIPCUtils.h"
+#include "GeolocationSystem.h"
 #include "mozilla/ClearOnShutdown.h"
 #include "mozilla/CycleCollectedJSContext.h"  // for nsAutoMicroTask
+#include "mozilla/dom/BrowserChild.h"
 #include "mozilla/dom/ContentChild.h"
 #include "mozilla/dom/PermissionMessageUtils.h"
 #include "mozilla/dom/GeolocationPositionError.h"
 #include "mozilla/dom/GeolocationPositionErrorBinding.h"
+#include "mozilla/glean/GleanMetrics.h"
+#include "mozilla/ipc/MessageChannel.h"
 #include "mozilla/Preferences.h"
 #include "mozilla/Services.h"
 #include "mozilla/StaticPrefs_geo.h"
+#include "mozilla/StaticPtr.h"
 #include "mozilla/Telemetry.h"
 #include "mozilla/UniquePtr.h"
 #include "mozilla/Unused.h"
@@ -23,10 +29,11 @@
 #include "nsComponentManagerUtils.h"
 #include "nsContentPermissionHelper.h"
 #include "nsContentUtils.h"
-#include "nsGlobalWindow.h"
+#include "nsGlobalWindowInner.h"
 #include "mozilla/dom/Document.h"
 #include "nsINamed.h"
 #include "nsIObserverService.h"
+#include "nsIPromptService.h"
 #include "nsIScriptError.h"
 #include "nsPIDOMWindow.h"
 #include "nsServiceManagerUtils.h"
@@ -43,13 +50,18 @@ class nsIPrincipal;
 #  include "GpsdLocationProvider.h"
 #endif
 
+#ifdef MOZ_ENABLE_DBUS
+#  include "mozilla/WidgetUtilsGtk.h"
+#  include "GeoclueLocationProvider.h"
+#  include "PortalLocationProvider.h"
+#endif
+
 #ifdef MOZ_WIDGET_COCOA
 #  include "CoreLocationLocationProvider.h"
 #endif
 
 #ifdef XP_WIN
 #  include "WindowsLocationProvider.h"
-#  include "mozilla/WindowsVersion.h"
 #endif
 
 // Some limit to the number of get or watch geolocation requests
@@ -63,11 +75,13 @@ class nsIPrincipal;
 using mozilla::Unused;  // <snicker>
 using namespace mozilla;
 using namespace mozilla::dom;
+using namespace mozilla::dom::geolocation;
 
-class nsGeolocationRequest final
-    : public ContentPermissionRequestBase,
-      public nsIGeolocationUpdate,
-      public SupportsWeakPtr<nsGeolocationRequest> {
+mozilla::LazyLogModule gGeolocationLog("Geolocation");
+
+class nsGeolocationRequest final : public ContentPermissionRequestBase,
+                                   public nsIGeolocationUpdate,
+                                   public SupportsWeakPtr {
  public:
   NS_DECL_ISUPPORTS_INHERITED
   NS_DECL_NSIGEOLOCATIONUPDATE
@@ -78,15 +92,14 @@ class nsGeolocationRequest final
   nsGeolocationRequest(Geolocation* aLocator, GeoPositionCallback aCallback,
                        GeoPositionErrorCallback aErrorCallback,
                        UniquePtr<PositionOptions>&& aOptions,
-                       uint8_t aProtocolType, nsIEventTarget* aMainThreadTarget,
+                       nsIEventTarget* aMainThreadSerialEventTarget,
                        bool aWatchPositionRequest = false,
                        int32_t aWatchId = 0);
 
   // nsIContentPermissionRequest
   MOZ_CAN_RUN_SCRIPT NS_IMETHOD Cancel(void) override;
-  MOZ_CAN_RUN_SCRIPT NS_IMETHOD Allow(JS::HandleValue choices) override;
-
-  MOZ_DECLARE_WEAKREFERENCE_TYPENAME(nsGeolocationRequest)
+  MOZ_CAN_RUN_SCRIPT NS_IMETHOD Allow(JS::Handle<JS::Value> choices) override;
+  NS_IMETHOD GetTypes(nsIArray** aTypes) override;
 
   void Shutdown();
 
@@ -108,6 +121,11 @@ class nsGeolocationRequest final
   bool IsWatch() { return mIsWatchPositionRequest; }
   int32_t WatchId() { return mWatchId; }
 
+  void SetPromptBehavior(
+      geolocation::SystemGeolocationPermissionBehavior aBehavior) {
+    mBehavior = aBehavior;
+  }
+
  private:
   virtual ~nsGeolocationRequest();
 
@@ -143,9 +161,10 @@ class nsGeolocationRequest final
 
   int32_t mWatchId;
   bool mShutdown;
-  nsCOMPtr<nsIContentPermissionRequester> mRequester;
-  uint8_t mProtocolType;
-  nsCOMPtr<nsIEventTarget> mMainThreadTarget;
+  bool mSeenAnySignal = false;
+  nsCOMPtr<nsIEventTarget> mMainThreadSerialEventTarget;
+
+  SystemGeolocationPermissionBehavior mBehavior;
 };
 
 static UniquePtr<PositionOptions> CreatePositionOptionsCopy(
@@ -195,13 +214,13 @@ static nsPIDOMWindowInner* ConvertWeakReferenceToWindow(
 nsGeolocationRequest::nsGeolocationRequest(
     Geolocation* aLocator, GeoPositionCallback aCallback,
     GeoPositionErrorCallback aErrorCallback,
-    UniquePtr<PositionOptions>&& aOptions, uint8_t aProtocolType,
-    nsIEventTarget* aMainThreadTarget, bool aWatchPositionRequest,
+    UniquePtr<PositionOptions>&& aOptions,
+    nsIEventTarget* aMainThreadSerialEventTarget, bool aWatchPositionRequest,
     int32_t aWatchId)
     : ContentPermissionRequestBase(
           aLocator->GetPrincipal(),
-          ConvertWeakReferenceToWindow(aLocator->GetOwner()),
-          NS_LITERAL_CSTRING("geo"), NS_LITERAL_CSTRING("geolocation")),
+          ConvertWeakReferenceToWindow(aLocator->GetOwner()), "geo"_ns,
+          "geolocation"_ns),
       mIsWatchPositionRequest(aWatchPositionRequest),
       mCallback(std::move(aCallback)),
       mErrorCallback(std::move(aErrorCallback)),
@@ -209,12 +228,8 @@ nsGeolocationRequest::nsGeolocationRequest(
       mLocator(aLocator),
       mWatchId(aWatchId),
       mShutdown(false),
-      mProtocolType(aProtocolType),
-      mMainThreadTarget(aMainThreadTarget) {
-  if (nsCOMPtr<nsPIDOMWindowInner> win =
-          do_QueryReferent(mLocator->GetOwner())) {
-  }
-}
+      mMainThreadSerialEventTarget(aMainThreadSerialEventTarget),
+      mBehavior(SystemGeolocationPermissionBehavior::NoPrompt) {}
 
 nsGeolocationRequest::~nsGeolocationRequest() { StopTimeoutTimer(); }
 
@@ -229,11 +244,15 @@ NS_IMPL_CYCLE_COLLECTION_WEAK_PTR_INHERITED(nsGeolocationRequest,
                                             mCallback, mErrorCallback, mLocator)
 
 void nsGeolocationRequest::Notify() {
+  MOZ_LOG(gGeolocationLog, LogLevel::Debug, ("nsGeolocationRequest::Notify"));
   SetTimeoutTimer();
   NotifyErrorAndShutdown(GeolocationPositionError_Binding::TIMEOUT);
 }
 
 void nsGeolocationRequest::NotifyErrorAndShutdown(uint16_t aErrorCode) {
+  MOZ_LOG(gGeolocationLog, LogLevel::Debug,
+          ("nsGeolocationRequest::NotifyErrorAndShutdown with error code: %u",
+           aErrorCode));
   MOZ_ASSERT(!mShutdown, "timeout after shutdown");
   if (!mIsWatchPositionRequest) {
     Shutdown();
@@ -245,14 +264,6 @@ void nsGeolocationRequest::NotifyErrorAndShutdown(uint16_t aErrorCode) {
 
 NS_IMETHODIMP
 nsGeolocationRequest::Cancel() {
-  if (mRequester) {
-    // Record the number of denied requests for regular web content.
-    // This method is only called when the user explicitly denies the request,
-    // and is not called when the page is simply unloaded, or similar.
-    Telemetry::Accumulate(Telemetry::GEOLOCATION_REQUEST_GRANTED,
-                          mProtocolType);
-  }
-
   if (mLocator->ClearPendingRequest(this)) {
     return NS_OK;
   }
@@ -261,46 +272,177 @@ nsGeolocationRequest::Cancel() {
   return NS_OK;
 }
 
-NS_IMETHODIMP
-nsGeolocationRequest::Allow(JS::HandleValue aChoices) {
-  MOZ_ASSERT(aChoices.isUndefined());
+/**
+ * When the promise for the cancel dialog is resolved or rejected, we should
+ * stop waiting for permission.  If it was granted then the
+ * SystemGeolocationPermissionRequest should already be resolved, so we do
+ * nothing.  Otherwise, we were either cancelled or got an error, so we cancel
+ * the SystemGeolocationPermissionRequest.
+ */
+class CancelSystemGeolocationPermissionRequest : public PromiseNativeHandler {
+ public:
+  NS_DECL_ISUPPORTS
 
-  if (mRequester) {
-    // Record the number of granted requests for regular web content.
-    Telemetry::Accumulate(Telemetry::GEOLOCATION_REQUEST_GRANTED,
-                          mProtocolType + 10);
+  explicit CancelSystemGeolocationPermissionRequest(
+      SystemGeolocationPermissionRequest* aRequest)
+      : mRequest(aRequest) {}
 
-    // Record whether a location callback is fulfilled while the owner window
-    // is not visible.
-    bool isVisible = false;
-    nsCOMPtr<nsPIDOMWindowInner> window = mLocator->GetParentObject();
-
-    if (window) {
-      nsCOMPtr<Document> doc = window->GetDoc();
-      isVisible = doc && !doc->Hidden();
-    }
-
-    if (IsWatch()) {
-      mozilla::Telemetry::Accumulate(
-          mozilla::Telemetry::GEOLOCATION_WATCHPOSITION_VISIBLE, isVisible);
-    } else {
-      mozilla::Telemetry::Accumulate(
-          mozilla::Telemetry::GEOLOCATION_GETCURRENTPOSITION_VISIBLE,
-          isVisible);
-    }
+  MOZ_CAN_RUN_SCRIPT_BOUNDARY
+  void ResolvedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue,
+                        ErrorResult& aRv) override {
+    mRequest->Stop();
   }
 
+  MOZ_CAN_RUN_SCRIPT_BOUNDARY
+  void RejectedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue,
+                        ErrorResult& aRv) override {
+    mRequest->Stop();
+  }
+
+ private:
+  ~CancelSystemGeolocationPermissionRequest() = default;
+  RefPtr<SystemGeolocationPermissionRequest> mRequest;
+};
+
+NS_IMPL_ISUPPORTS0(CancelSystemGeolocationPermissionRequest)
+
+/* static */
+void Geolocation::ReallowWithSystemPermissionOrCancel(
+    BrowsingContext* aBrowsingContext,
+    geolocation::ParentRequestResolver&& aResolver) {
+  // Make sure we don't return without responding to the geolocation request.
+  auto denyPermissionOnError =
+      MakeScopeExit([&aResolver]() MOZ_CAN_RUN_SCRIPT_BOUNDARY_LAMBDA {
+        aResolver(GeolocationPermissionStatus::Error);
+      });
+
+  NS_ENSURE_TRUE_VOID(aBrowsingContext);
+
+  nsCOMPtr<nsIStringBundle> bundle;
+  nsCOMPtr<nsIStringBundleService> sbs =
+      do_GetService(NS_STRINGBUNDLE_CONTRACTID);
+  NS_ENSURE_TRUE_VOID(sbs);
+
+  sbs->CreateBundle("chrome://browser/locale/browser.properties",
+                    getter_AddRefs(bundle));
+  NS_ENSURE_TRUE_VOID(bundle);
+
+  nsAutoString title;
+  nsresult rv =
+      bundle->GetStringFromName("geolocation.systemSettingsTitle", title);
+  NS_ENSURE_SUCCESS_VOID(rv);
+
+  nsAutoString brandName;
+  rv = nsContentUtils::GetLocalizedString(nsContentUtils::eBRAND_PROPERTIES,
+                                          "brandShortName", brandName);
+  NS_ENSURE_SUCCESS_VOID(rv);
+  AutoTArray<nsString, 1> formatParams;
+  formatParams.AppendElement(brandName);
+  nsAutoString message;
+  rv = bundle->FormatStringFromName("geolocation.systemSettingsMessage",
+                                    formatParams, message);
+  NS_ENSURE_SUCCESS_VOID(rv);
+
+  // We MUST do this because aResolver is moved below.
+  denyPermissionOnError.release();
+
+  RefPtr<SystemGeolocationPermissionRequest> permissionRequest =
+      geolocation::RequestLocationPermissionFromUser(aBrowsingContext,
+                                                     std::move(aResolver));
+  NS_ENSURE_TRUE_VOID(permissionRequest);
+
+  auto cancelRequestOnError = MakeScopeExit([&]() {
+    // Stop waiting for the system permission and just leave it up to the user.
+    permissionRequest->Stop();
+  });
+
+  nsCOMPtr<nsIPromptService> promptSvc =
+      do_GetService("@mozilla.org/prompter;1", &rv);
+  NS_ENSURE_SUCCESS_VOID(rv);
+
+  // The dialog should include a cancel button if Gecko is prompting the user
+  // for system permission.  It should have no buttons if the OS will be
+  // doing the prompting.
+  bool geckoWillPrompt =
+      GetLocationOSPermission() ==
+      geolocation::SystemGeolocationPermissionBehavior::GeckoWillPromptUser;
+  // This combination of flags removes the default yes and no buttons and adds a
+  // spinner to the title.
+  const auto kSpinnerNoButtonFlags = nsIPromptService::BUTTON_TITLE_IS_STRING *
+                                         nsIPromptService::BUTTON_POS_0 +
+                                     nsIPromptService::BUTTON_TITLE_IS_STRING *
+                                         nsIPromptService::BUTTON_POS_1 +
+                                     nsIPromptService::SHOW_SPINNER;
+  // This combination of flags indicates there is only one button labeled
+  // "Cancel".
+  const auto kCancelButtonFlags =
+      nsIPromptService::BUTTON_TITLE_CANCEL * nsIPromptService::BUTTON_POS_0;
+  RefPtr<mozilla::dom::Promise> tabBlockingDialogPromise;
+  rv = promptSvc->AsyncConfirmEx(
+      aBrowsingContext, nsIPromptService::MODAL_TYPE_TAB, title.get(),
+      message.get(),
+      geckoWillPrompt ? kCancelButtonFlags : kSpinnerNoButtonFlags, nullptr,
+      nullptr, nullptr, nullptr, false, JS::UndefinedHandleValue,
+      getter_AddRefs(tabBlockingDialogPromise));
+  NS_ENSURE_SUCCESS_VOID(rv);
+  MOZ_ASSERT(tabBlockingDialogPromise);
+
+  // If the tab blocking dialog promise is resolved or rejected then the dialog
+  // is no longer visible so we should stop waiting for permission, whether it
+  // was granted or not.
+  tabBlockingDialogPromise->AppendNativeHandler(
+      new CancelSystemGeolocationPermissionRequest(permissionRequest));
+
+  cancelRequestOnError.release();
+}
+
+NS_IMETHODIMP
+nsGeolocationRequest::Allow(JS::Handle<JS::Value> aChoices) {
+  MOZ_ASSERT(aChoices.isUndefined());
+
   if (mLocator->ClearPendingRequest(this)) {
     return NS_OK;
   }
 
+  if (mBehavior != SystemGeolocationPermissionBehavior::NoPrompt) {
+    // Asynchronously present the system dialog or open system preferences
+    // (RequestGeolocationPermissionFromUser will know which to do), and wait
+    // for the permission to change or the request to be canceled.  If the
+    // permission is (maybe) granted then it will call Allow again.  It actually
+    // will also re-call Allow if the permission is denied, in order to get the
+    // "denied permission" behavior.
+    mBehavior = SystemGeolocationPermissionBehavior::NoPrompt;
+    RefPtr<BrowsingContext> browsingContext = mWindow->GetBrowsingContext();
+    if (ContentChild* cc = ContentChild::GetSingleton()) {
+      cc->SendRequestGeolocationPermissionFromUser(
+          browsingContext,
+          [self = RefPtr{this}](GeolocationPermissionStatus aResult)
+              MOZ_CAN_RUN_SCRIPT_BOUNDARY_LAMBDA {
+                self->Allow(JS::UndefinedHandleValue);
+              },
+          [self = RefPtr{this}](mozilla::ipc::ResponseRejectReason aReason)
+              MOZ_CAN_RUN_SCRIPT_BOUNDARY_LAMBDA {
+                self->Allow(JS::UndefinedHandleValue);
+              });
+      return NS_OK;
+    }
+
+    Geolocation::ReallowWithSystemPermissionOrCancel(
+        browsingContext,
+        [self = RefPtr{this}](GeolocationPermissionStatus aResult)
+            MOZ_CAN_RUN_SCRIPT_BOUNDARY_LAMBDA {
+              self->Allow(JS::UndefinedHandleValue);
+            });
+    return NS_OK;
+  }
+
   RefPtr<nsGeolocationService> gs =
       nsGeolocationService::GetGeolocationService();
 
   bool canUseCache = false;
   CachedPositionAndAccuracy lastPosition = gs->GetCachedPosition();
   if (lastPosition.position) {
-    DOMTimeStamp cachedPositionTime_ms;
+    EpochTimeStamp cachedPositionTime_ms;
     lastPosition.position->GetTimestamp(&cachedPositionTime_ms);
     // check to see if we can use a cached value
     // if the user has specified a maximumAge, return a cached value.
@@ -309,7 +451,7 @@ nsGeolocationRequest::Allow(JS::HandleValue aChoices) {
       bool isCachedWithinRequestedAccuracy =
           WantsHighAccuracy() <= lastPosition.isHighAccuracy;
       bool isCachedWithinRequestedTime =
-          DOMTimeStamp(PR_Now() / PR_USEC_PER_MSEC - maximumAge_ms) <=
+          EpochTimeStamp(PR_Now() / PR_USEC_PER_MSEC - maximumAge_ms) <=
           cachedPositionTime_ms;
       canUseCache =
           isCachedWithinRequestedAccuracy && isCachedWithinRequestedTime;
@@ -323,7 +465,7 @@ nsGeolocationRequest::Allow(JS::HandleValue aChoices) {
     // will now be owned by the RequestSendLocationEvent
     Update(lastPosition.position);
 
-    // After Update is called, getCurrentPosition finishes it's job.
+    // After Update is called, getCurrentPosition finishes its job.
     if (!mIsWatchPositionRequest) {
       return NS_OK;
     }
@@ -337,22 +479,26 @@ nsGeolocationRequest::Allow(JS::HandleValue aChoices) {
     }
   }
 
-  // Kick off the geo device, if it isn't already running
-  nsCOMPtr<nsIPrincipal> principal = GetPrincipal();
-  nsresult rv = gs->StartDevice(principal);
-
-  if (NS_FAILED(rv)) {
-    // Location provider error
-    NotifyError(GeolocationPositionError_Binding::POSITION_UNAVAILABLE);
-    return NS_OK;
-  }
-
-  if (mIsWatchPositionRequest || !canUseCache) {
+  // Non-cached location request
+  bool allowedRequest = mIsWatchPositionRequest || !canUseCache;
+  if (allowedRequest) {
     // let the locator know we're pending
     // we will now be owned by the locator
     mLocator->NotifyAllowedRequest(this);
   }
 
+  // Kick off the geo device, if it isn't already running
+  nsresult rv = gs->StartDevice();
+
+  if (NS_FAILED(rv)) {
+    if (allowedRequest) {
+      mLocator->RemoveRequest(this);
+    }
+    // Location provider error
+    NotifyError(GeolocationPositionError_Binding::POSITION_UNAVAILABLE);
+    return NS_OK;
+  }
+
   SetTimeoutTimer();
 
   return NS_OK;
@@ -384,11 +530,11 @@ void nsGeolocationRequest::SendLocation(nsIDOMGeoPosition* aPosition) {
   }
 
   if (mOptions && mOptions->mMaximumAge > 0) {
-    DOMTimeStamp positionTime_ms;
+    EpochTimeStamp positionTime_ms;
     aPosition->GetTimestamp(&positionTime_ms);
     const uint32_t maximumAge_ms = mOptions->mMaximumAge;
-    const bool isTooOld = DOMTimeStamp(PR_Now() / PR_USEC_PER_MSEC -
-                                       maximumAge_ms) > positionTime_ms;
+    const bool isTooOld = EpochTimeStamp(PR_Now() / PR_USEC_PER_MSEC -
+                                         maximumAge_ms) > positionTime_ms;
     if (isTooOld) {
       return;
     }
@@ -416,6 +562,10 @@ void nsGeolocationRequest::SendLocation(nsIDOMGeoPosition* aPosition) {
     Shutdown();
   }
 
+  nsCOMPtr<nsIObserverService> obs = services::GetObserverService();
+  obs->NotifyObservers(wrapped, "geolocation-position-events",
+                       u"location-updated");
+
   nsAutoMicroTask mt;
   if (mCallback.HasWebIDLCallback()) {
     RefPtr<PositionCallback> callback = mCallback.GetWebIDLCallback();
@@ -444,16 +594,46 @@ nsIPrincipal* nsGeolocationRequest::GetPrincipal() {
 
 NS_IMETHODIMP
 nsGeolocationRequest::Update(nsIDOMGeoPosition* aPosition) {
+  if (!mSeenAnySignal) {
+    mSeenAnySignal = true;
+    glean::geolocation::request_result
+        .EnumGet(glean::geolocation::RequestResultLabel::eSuccess)
+        .Add();
+  }
   nsCOMPtr<nsIRunnable> ev = new RequestSendLocationEvent(aPosition, this);
-  mMainThreadTarget->Dispatch(ev.forget());
+  mMainThreadSerialEventTarget->Dispatch(ev.forget());
   return NS_OK;
 }
 
+static glean::geolocation::RequestResultLabel MapErrorToLabel(
+    uint16_t aErrorCode) {
+  using Label = glean::geolocation::RequestResultLabel;
+  switch (aErrorCode) {
+    case GeolocationPositionError_Binding::PERMISSION_DENIED:
+      return Label::ePermissionDenied;
+    case GeolocationPositionError_Binding::POSITION_UNAVAILABLE:
+      return Label::ePositionUnavailable;
+    case GeolocationPositionError_Binding::TIMEOUT:
+      return Label::eTimeout;
+    default:
+      MOZ_CRASH("Unknown geolocation error label");
+      return Label::ePositionUnavailable;
+  }
+}
+
 NS_IMETHODIMP
 nsGeolocationRequest::NotifyError(uint16_t aErrorCode) {
+  MOZ_LOG(
+      gGeolocationLog, LogLevel::Debug,
+      ("nsGeolocationRequest::NotifyError with error code: %u", aErrorCode));
   MOZ_ASSERT(NS_IsMainThread());
+  if (!mSeenAnySignal) {
+    mSeenAnySignal = true;
+    glean::geolocation::request_result.EnumGet(MapErrorToLabel(aErrorCode))
+        .Add();
+  }
   RefPtr<GeolocationPositionError> positionError =
-      new GeolocationPositionError(mLocator, aErrorCode);
+      new GeolocationPositionError(mLocator, static_cast<int16_t>(aErrorCode));
   positionError->NotifyCallback(mErrorCallback);
   return NS_OK;
 }
@@ -475,6 +655,24 @@ void nsGeolocationRequest::Shutdown() {
   }
 }
 
+NS_IMETHODIMP
+nsGeolocationRequest::GetTypes(nsIArray** aTypes) {
+  AutoTArray<nsString, 2> options;
+
+  switch (mBehavior) {
+    case SystemGeolocationPermissionBehavior::SystemWillPromptUser:
+      options.AppendElement(u"sysdlg"_ns);
+      break;
+    case SystemGeolocationPermissionBehavior::GeckoWillPromptUser:
+      options.AppendElement(u"syssetting"_ns);
+      break;
+    default:
+      break;
+  }
+  return nsContentPermissionUtils::CreatePermissionArray(mType, options,
+                                                         aTypes);
+}
+
 ////////////////////////////////////////////////////
 // nsGeolocationRequest::TimerCallbackHolder
 ////////////////////////////////////////////////////
@@ -526,10 +724,41 @@ nsresult nsGeolocationService::Init() {
 #endif
 
 #ifdef MOZ_WIDGET_GTK
-#  ifdef MOZ_GPSD
-  if (Preferences::GetBool("geo.provider.use_gpsd", false)) {
-    mProvider = new GpsdLocationProvider();
+#  ifdef MOZ_ENABLE_DBUS
+  if (!mProvider && widget::ShouldUsePortal(widget::PortalKind::Location)) {
+    mProvider = new PortalLocationProvider();
+    MOZ_LOG(gGeolocationLog, LogLevel::Debug,
+            ("Selected PortalLocationProvider"));
+    glean::geolocation::linux_provider
+        .EnumGet(glean::geolocation::LinuxProviderLabel::ePortal)
+        .Set(true);
   }
+  // Geoclue includes GPS data so it has higher priority than raw GPSD
+  if (!mProvider && StaticPrefs::geo_provider_use_geoclue()) {
+    nsCOMPtr<nsIGeolocationProvider> gcProvider = new GeoclueLocationProvider();
+    MOZ_LOG(gGeolocationLog, LogLevel::Debug,
+            ("Checking GeoclueLocationProvider"));
+    // The Startup() method will only succeed if Geoclue is available on D-Bus
+    if (NS_SUCCEEDED(gcProvider->Startup())) {
+      gcProvider->Shutdown();
+      mProvider = std::move(gcProvider);
+      MOZ_LOG(gGeolocationLog, LogLevel::Debug,
+              ("Selected GeoclueLocationProvider"));
+      glean::geolocation::linux_provider
+          .EnumGet(glean::geolocation::LinuxProviderLabel::eGeoclue)
+          .Set(true);
+    }
+  }
+#    ifdef MOZ_GPSD
+  if (!mProvider && Preferences::GetBool("geo.provider.use_gpsd", false)) {
+    mProvider = new GpsdLocationProvider();
+    MOZ_LOG(gGeolocationLog, LogLevel::Debug,
+            ("Selected GpsdLocationProvider"));
+    glean::geolocation::linux_provider
+        .EnumGet(glean::geolocation::LinuxProviderLabel::eGpsd)
+        .Set(true);
+  }
+#    endif
 #  endif
 #endif
 
@@ -540,8 +769,7 @@ nsresult nsGeolocationService::Init() {
 #endif
 
 #ifdef XP_WIN
-  if (Preferences::GetBool("geo.provider.ms-windows-location", false) &&
-      IsWin8OrLater()) {
+  if (Preferences::GetBool("geo.provider.ms-windows-location", false)) {
     mProvider = new WindowsLocationProvider();
   }
 #endif
@@ -618,6 +846,9 @@ nsGeolocationService::Update(nsIDOMGeoPosition* aSomewhere) {
 
 NS_IMETHODIMP
 nsGeolocationService::NotifyError(uint16_t aErrorCode) {
+  MOZ_LOG(
+      gGeolocationLog, LogLevel::Debug,
+      ("nsGeolocationService::NotifyError with error code: %u", aErrorCode));
   // nsTArray doesn't have a constructors that takes a different-type TArray.
   nsTArray<RefPtr<Geolocation>> geolocators;
   geolocators.AppendElements(mGeolocators);
@@ -637,7 +868,7 @@ CachedPositionAndAccuracy nsGeolocationService::GetCachedPosition() {
   return mLastPosition;
 }
 
-nsresult nsGeolocationService::StartDevice(nsIPrincipal* aPrincipal) {
+nsresult nsGeolocationService::StartDevice() {
   if (!StaticPrefs::geo_enabled()) {
     return NS_ERROR_NOT_AVAILABLE;
   }
@@ -649,8 +880,7 @@ nsresult nsGeolocationService::StartDevice(nsIPrincipal* aPrincipal) {
 
   if (XRE_IsContentProcess()) {
     ContentChild* cpc = ContentChild::GetSingleton();
-    cpc->SendAddGeolocationListener(IPC::Principal(aPrincipal),
-                                    HighAccuracyRequested());
+    cpc->SendAddGeolocationListener(HighAccuracyRequested());
     return NS_OK;
   }
 
@@ -788,7 +1018,7 @@ NS_IMPL_CYCLE_COLLECTION_WRAPPERCACHE(Geolocation, mPendingCallbacks,
                                       mWatchingCallbacks, mPendingRequests)
 
 Geolocation::Geolocation()
-    : mProtocolType(ProtocolType::OTHER), mLastWatchId(0) {}
+    : mProtocolType(ProtocolType::OTHER), mLastWatchId(1) {}
 
 Geolocation::~Geolocation() {
   if (mService) {
@@ -899,6 +1129,18 @@ Geolocation::Update(nsIDOMGeoPosition* aSomewhere) {
     return NS_OK;
   }
 
+  // Don't update position if window is not fully active or the document is
+  // hidden. We keep the pending callaback and watchers waiting for the next
+  // update.
+  nsCOMPtr<nsPIDOMWindowInner> window = do_QueryReferent(this->GetOwner());
+  if (window) {
+    nsCOMPtr<Document> document = window->GetDoc();
+    bool isHidden = document && document->Hidden();
+    if (isHidden || !window->IsFullyActive()) {
+      return NS_OK;
+    }
+  }
+
   if (aSomewhere) {
     nsCOMPtr<nsIDOMGeoPositionCoords> coords;
     aSomewhere->GetCoords(getter_AddRefs(coords));
@@ -906,7 +1148,10 @@ Geolocation::Update(nsIDOMGeoPosition* aSomewhere) {
       double accuracy = -1;
       coords->GetAccuracy(&accuracy);
       mozilla::Telemetry::Accumulate(
-          mozilla::Telemetry::GEOLOCATION_ACCURACY_EXPONENTIAL, accuracy);
+          mozilla::Telemetry::GEOLOCATION_ACCURACY_EXPONENTIAL,
+          static_cast<uint32_t>(accuracy));
+      glean::geolocation::accuracy.AccumulateSingleSample(
+          static_cast<uint64_t>(accuracy));
     }
   }
 
@@ -925,6 +1170,8 @@ Geolocation::Update(nsIDOMGeoPosition* aSomewhere) {
 
 NS_IMETHODIMP
 Geolocation::NotifyError(uint16_t aErrorCode) {
+  MOZ_LOG(gGeolocationLog, LogLevel::Debug,
+          ("Geolocation::NotifyError with error code: %u", aErrorCode));
   if (!WindowOwnerStillExists()) {
     Shutdown();
     return NS_OK;
@@ -947,6 +1194,16 @@ Geolocation::NotifyError(uint16_t aErrorCode) {
   return NS_OK;
 }
 
+bool Geolocation::IsFullyActiveOrChrome() {
+  // For regular content window, only allow this proceed if the window is "fully
+  // active".
+  if (nsPIDOMWindowInner* window = this->GetParentObject()) {
+    return window->IsFullyActive();
+  }
+  // Calls coming from chrome code don't have window, so we can proceed.
+  return true;
+}
+
 bool Geolocation::IsAlreadyCleared(nsGeolocationRequest* aRequest) {
   for (uint32_t i = 0, length = mClearedWatchIDs.Length(); i < length; ++i) {
     if (mClearedWatchIDs[i] == aRequest->WatchId()) {
@@ -973,8 +1230,7 @@ bool Geolocation::ShouldBlockInsecureRequests() const {
   }
 
   if (!nsGlobalWindowInner::Cast(win)->IsSecureContext()) {
-    nsContentUtils::ReportToConsole(nsIScriptError::errorFlag,
-                                    NS_LITERAL_CSTRING("DOM"), doc,
+    nsContentUtils::ReportToConsole(nsIScriptError::errorFlag, "DOM"_ns, doc,
                                     nsContentUtils::eDOM_PROPERTIES,
                                     "GeolocationInsecureRequestIsForbidden");
     return true;
@@ -1006,33 +1262,28 @@ void Geolocation::GetCurrentPosition(PositionCallback& aCallback,
   }
 }
 
-static nsIEventTarget* MainThreadTarget(Geolocation* geo) {
-  nsCOMPtr<nsPIDOMWindowInner> window = do_QueryReferent(geo->GetOwner());
-  if (!window) {
-    return GetMainThreadEventTarget();
-  }
-  return nsGlobalWindowInner::Cast(window)->EventTargetFor(
-      mozilla::TaskCategory::Other);
-}
-
 nsresult Geolocation::GetCurrentPosition(GeoPositionCallback callback,
                                          GeoPositionErrorCallback errorCallback,
                                          UniquePtr<PositionOptions>&& options,
                                          CallerType aCallerType) {
+  if (!IsFullyActiveOrChrome()) {
+    RefPtr<GeolocationPositionError> positionError =
+        new GeolocationPositionError(
+            this, GeolocationPositionError_Binding::POSITION_UNAVAILABLE);
+    positionError->NotifyCallback(errorCallback);
+    return NS_OK;
+  }
+
   if (mPendingCallbacks.Length() > MAX_GEO_REQUESTS_PER_WINDOW) {
     return NS_ERROR_NOT_AVAILABLE;
   }
 
   // After this we hand over ownership of options to our nsGeolocationRequest.
 
-  // Count the number of requests per protocol/scheme.
-  Telemetry::Accumulate(Telemetry::GEOLOCATION_GETCURRENTPOSITION_SECURE_ORIGIN,
-                        static_cast<uint8_t>(mProtocolType));
-
-  nsIEventTarget* target = MainThreadTarget(this);
+  nsIEventTarget* target = GetMainThreadSerialEventTarget();
   RefPtr<nsGeolocationRequest> request = new nsGeolocationRequest(
       this, std::move(callback), std::move(errorCallback), std::move(options),
-      static_cast<uint8_t>(mProtocolType), target);
+      target);
 
   if (!StaticPrefs::geo_enabled() || ShouldBlockInsecureRequests() ||
       !request->CheckPermissionDelegate()) {
@@ -1046,10 +1297,7 @@ nsresult Geolocation::GetCurrentPosition(GeoPositionCallback callback,
   }
 
   if (mOwner) {
-    if (!RegisterRequestWithPrompt(request)) {
-      return NS_ERROR_NOT_AVAILABLE;
-    }
-
+    RequestIfPermitted(request);
     return NS_OK;
   }
 
@@ -1083,29 +1331,32 @@ int32_t Geolocation::WatchPosition(
                        std::move(aOptions), CallerType::System, IgnoreErrors());
 }
 
-// On errors we return -1 because that's not a valid watch id and will
+// On errors we return 0 because that's not a valid watch id and will
 // get ignored in clearWatch.
 int32_t Geolocation::WatchPosition(GeoPositionCallback aCallback,
                                    GeoPositionErrorCallback aErrorCallback,
                                    UniquePtr<PositionOptions>&& aOptions,
                                    CallerType aCallerType, ErrorResult& aRv) {
-  if (mWatchingCallbacks.Length() > MAX_GEO_REQUESTS_PER_WINDOW) {
-    aRv.Throw(NS_ERROR_NOT_AVAILABLE);
-    return -1;
+  if (!IsFullyActiveOrChrome()) {
+    RefPtr<GeolocationPositionError> positionError =
+        new GeolocationPositionError(
+            this, GeolocationPositionError_Binding::POSITION_UNAVAILABLE);
+    positionError->NotifyCallback(aErrorCallback);
+    return 0;
   }
 
-  // Count the number of requests per protocol/scheme.
-  Telemetry::Accumulate(Telemetry::GEOLOCATION_WATCHPOSITION_SECURE_ORIGIN,
-                        static_cast<uint8_t>(mProtocolType));
+  if (mWatchingCallbacks.Length() > MAX_GEO_REQUESTS_PER_WINDOW) {
+    aRv.Throw(NS_ERROR_NOT_AVAILABLE);
+    return 0;
+  }
 
   // The watch ID:
   int32_t watchId = mLastWatchId++;
 
-  nsIEventTarget* target = MainThreadTarget(this);
+  nsIEventTarget* target = GetMainThreadSerialEventTarget();
   RefPtr<nsGeolocationRequest> request = new nsGeolocationRequest(
       this, std::move(aCallback), std::move(aErrorCallback),
-      std::move(aOptions), static_cast<uint8_t>(mProtocolType), target, true,
-      watchId);
+      std::move(aOptions), target, true, watchId);
 
   if (!StaticPrefs::geo_enabled() || ShouldBlockInsecureRequests() ||
       !request->CheckPermissionDelegate()) {
@@ -1116,21 +1367,17 @@ int32_t Geolocation::WatchPosition(GeoPositionCallback aCallback,
 
   if (!mOwner && aCallerType != CallerType::System) {
     aRv.Throw(NS_ERROR_FAILURE);
-    return -1;
+    return 0;
   }
 
   if (mOwner) {
-    if (!RegisterRequestWithPrompt(request)) {
-      aRv.Throw(NS_ERROR_NOT_AVAILABLE);
-      return -1;
-    }
-
+    RequestIfPermitted(request);
     return watchId;
   }
 
   if (aCallerType != CallerType::System) {
     aRv.Throw(NS_ERROR_FAILURE);
-    return -1;
+    return 0;
   }
 
   request->Allow(JS::UndefinedHandleValue);
@@ -1138,7 +1385,7 @@ int32_t Geolocation::WatchPosition(GeoPositionCallback aCallback,
 }
 
 void Geolocation::ClearWatch(int32_t aWatchId) {
-  if (aWatchId < 0) {
+  if (aWatchId < 1) {
     return;
   }
 
@@ -1195,8 +1442,9 @@ void Geolocation::NotifyAllowedRequest(nsGeolocationRequest* aRequest) {
   }
 }
 
-bool Geolocation::RegisterRequestWithPrompt(nsGeolocationRequest* request) {
-  nsIEventTarget* target = MainThreadTarget(this);
+/* static */ bool Geolocation::RegisterRequestWithPrompt(
+    nsGeolocationRequest* request) {
+  nsIEventTarget* target = GetMainThreadSerialEventTarget();
   ContentPermissionRequestBase::PromptResult pr = request->CheckPromptPrefs();
   if (pr == ContentPermissionRequestBase::PromptResult::Granted) {
     request->RequestDelayedTask(target,
@@ -1214,6 +1462,55 @@ bool Geolocation::RegisterRequestWithPrompt(nsGeolocationRequest* request) {
   return true;
 }
 
+/* static */ geolocation::SystemGeolocationPermissionBehavior
+Geolocation::GetLocationOSPermission() {
+  auto permission = geolocation::GetGeolocationPermissionBehavior();
+
+  if (!StaticPrefs::geo_prompt_open_system_prefs() &&
+      permission == geolocation::SystemGeolocationPermissionBehavior::
+                        GeckoWillPromptUser) {
+    return geolocation::SystemGeolocationPermissionBehavior::NoPrompt;
+  }
+  return permission;
+}
+
+void Geolocation::RequestIfPermitted(nsGeolocationRequest* request) {
+  auto getPermission = [request = RefPtr{request}](auto aPermission) {
+    switch (aPermission) {
+      case geolocation::SystemGeolocationPermissionBehavior::
+          SystemWillPromptUser:
+      case geolocation::SystemGeolocationPermissionBehavior::
+          GeckoWillPromptUser:
+        request->SetPromptBehavior(aPermission);
+        break;
+      case geolocation::SystemGeolocationPermissionBehavior::NoPrompt:
+        // Either location access is already permitted by OS or the system
+        // permission UX is not available for this platform.  Do nothing.
+        break;
+      default:
+        MOZ_ASSERT_UNREACHABLE(
+            "unexpected GeolocationPermissionBehavior value");
+        break;
+    }
+    RegisterRequestWithPrompt(request);
+  };
+
+  if (auto* contentChild = ContentChild::GetSingleton()) {
+    contentChild->SendGetSystemGeolocationPermissionBehavior(
+        std::move(getPermission),
+        [request =
+             RefPtr{request}](mozilla::ipc::ResponseRejectReason aReason) {
+          NS_WARNING("Error sending GetSystemGeolocationPermissionBehavior");
+          // We still need to run the location request, even if we don't
+          // have permission.
+          RegisterRequestWithPrompt(request);
+        });
+  } else {
+    MOZ_ASSERT(XRE_IsParentProcess());
+    getPermission(GetLocationOSPermission());
+  }
+}
+
 JSObject* Geolocation::WrapObject(JSContext* aCtx,
                                   JS::Handle<JSObject*> aGivenProto) {
   return Geolocation_Binding::Wrap(aCtx, this, aGivenProto);
