@@ -12,17 +12,11 @@
 #include "mozilla/UniquePtr.h"
 #include "mozilla/dom/BrowsingContext.h"
 #include "mozilla/dom/PopupBlocker.h"
-#include "mozilla/dom/RadioGroupManager.h"
-#include "nsCOMPtr.h"
-#include "nsIForm.h"
+#include "mozilla/dom/RadioGroupContainer.h"
 #include "nsIFormControl.h"
 #include "nsGenericHTMLElement.h"
-#include "nsIRadioGroupContainer.h"
-#include "nsIWeakReferenceUtils.h"
 #include "nsThreadUtils.h"
 #include "nsInterfaceHashtable.h"
-#include "nsRefPtrHashtable.h"
-#include "nsTHashMap.h"
 #include "js/friend/DOMProxy.h"  // JS::ExpandoAndGeneration
 
 class nsIMutableArray;
@@ -32,14 +26,13 @@ namespace mozilla {
 class EventChainPostVisitor;
 class EventChainPreVisitor;
 namespace dom {
+class DialogFormSubmission;
 class HTMLFormControlsCollection;
+class HTMLFormSubmission;
 class HTMLImageElement;
 class FormData;
 
-class HTMLFormElement final : public nsGenericHTMLElement,
-                              public nsIForm,
-                              public nsIRadioGroupContainer,
-                              RadioGroupManager {
+class HTMLFormElement final : public nsGenericHTMLElement {
   friend class HTMLFormControlsCollection;
 
  public:
@@ -53,51 +46,33 @@ class HTMLFormElement final : public nsGenericHTMLElement,
   // nsISupports
   NS_DECL_ISUPPORTS_INHERITED
 
-  // nsIForm
-  NS_IMETHOD_(nsIFormControl*) GetElementAt(int32_t aIndex) const override;
-  NS_IMETHOD_(uint32_t) GetElementCount() const override;
-  NS_IMETHOD_(int32_t) IndexOfControl(nsIFormControl* aControl) override;
-  NS_IMETHOD_(nsIFormControl*) GetDefaultSubmitElement() const override;
-
-  // nsIRadioGroupContainer
-  void SetCurrentRadioButton(const nsAString& aName,
-                             HTMLInputElement* aRadio) override;
-  HTMLInputElement* GetCurrentRadioButton(const nsAString& aName) override;
-  NS_IMETHOD GetNextRadioButton(const nsAString& aName, const bool aPrevious,
-                                HTMLInputElement* aFocusedRadio,
-                                HTMLInputElement** aRadioOut) override;
-  NS_IMETHOD WalkRadioGroup(const nsAString& aName,
-                            nsIRadioVisitor* aVisitor) override;
-  void AddToRadioGroup(const nsAString& aName,
-                       HTMLInputElement* aRadio) override;
-  void RemoveFromRadioGroup(const nsAString& aName,
-                            HTMLInputElement* aRadio) override;
-  virtual uint32_t GetRequiredRadioCount(const nsAString& aName) const override;
-  virtual void RadioRequiredWillChange(const nsAString& aName,
-                                       bool aRequiredAdded) override;
-  virtual bool GetValueMissingState(const nsAString& aName) const override;
-  virtual void SetValueMissingState(const nsAString& aName,
-                                    bool aValue) override;
-
-  virtual EventStates IntrinsicState() const override;
+  int32_t IndexOfContent(nsIContent* aContent);
+  nsGenericHTMLFormElement* GetDefaultSubmitElement() const;
+  bool IsDefaultSubmitElement(nsGenericHTMLFormElement* aElement) const {
+    return aElement == mDefaultSubmitElement;
+  }
 
   // EventTarget
-  virtual void AsyncEventRunning(AsyncEventDispatcher* aEvent) override;
+  void AsyncEventRunning(AsyncEventDispatcher* aEvent) override;
+
+  /** Whether we already dispatched a DOMFormHasPassword event or not */
+  bool mHasPendingPasswordEvent = false;
+  /** Whether we already dispatched a DOMFormHasPossibleUsername event or not */
+  bool mHasPendingPossibleUsernameEvent = false;
 
   // nsIContent
-  virtual bool ParseAttribute(int32_t aNamespaceID, nsAtom* aAttribute,
-                              const nsAString& aValue,
-                              nsIPrincipal* aMaybeScriptedPrincipal,
-                              nsAttrValue& aResult) override;
+  bool ParseAttribute(int32_t aNamespaceID, nsAtom* aAttribute,
+                      const nsAString& aValue,
+                      nsIPrincipal* aMaybeScriptedPrincipal,
+                      nsAttrValue& aResult) override;
   void GetEventTargetParent(EventChainPreVisitor& aVisitor) override;
   void WillHandleEvent(EventChainPostVisitor& aVisitor) override;
-  virtual nsresult PostHandleEvent(EventChainPostVisitor& aVisitor) override;
+  nsresult PostHandleEvent(EventChainPostVisitor& aVisitor) override;
 
-  virtual nsresult BindToTree(BindContext&, nsINode& aParent) override;
-  virtual void UnbindFromTree(bool aNullParent = true) override;
-  virtual nsresult BeforeSetAttr(int32_t aNamespaceID, nsAtom* aName,
-                                 const nsAttrValueOrString* aValue,
-                                 bool aNotify) override;
+  nsresult BindToTree(BindContext&, nsINode& aParent) override;
+  void UnbindFromTree(UnbindContext&) override;
+  void BeforeSetAttr(int32_t aNamespaceID, nsAtom* aName,
+                     const nsAttrValue* aValue, bool aNotify) override;
 
   /**
    * Forget all information about the current submission (and the fact that we
@@ -105,7 +80,7 @@ class HTMLFormElement final : public nsGenericHTMLElement,
    */
   void ForgetCurrentSubmission();
 
-  virtual nsresult Clone(dom::NodeInfo*, nsINode** aResult) const override;
+  nsresult Clone(dom::NodeInfo*, nsINode** aResult) const override;
 
   NS_DECL_CYCLE_COLLECTION_CLASS_INHERITED(HTMLFormElement,
                                            nsGenericHTMLElement)
@@ -204,19 +179,10 @@ class HTMLFormElement final : public nsGenericHTMLElement,
   bool ImplicitSubmissionIsDisabled() const;
 
   /**
-   * Check whether a given nsIFormControl is the last single line input control
-   * that is not disabled. aControl is expected to not be null.
+   * Check whether a given nsGenericHTMLFormElement is the last single line
+   * input control that is not disabled. aElement is expected to not be null.
    */
-  bool IsLastActiveElement(const nsIFormControl* aControl) const;
-
-  /**
-   * Check whether a given nsIFormControl is the default submit
-   * element.  This is different from just comparing to
-   * GetDefaultSubmitElement() in certain situations inside an update
-   * when GetDefaultSubmitElement() might not be up to date.  aControl
-   * is expected to not be null.
-   */
-  bool IsDefaultSubmitElement(const nsIFormControl* aControl) const;
+  bool IsLastActiveElement(const nsGenericHTMLFormElement* aElement) const;
 
   /**
    * Flag the form to know that a button or image triggered scripted form
@@ -227,8 +193,8 @@ class HTMLFormElement final : public nsGenericHTMLElement,
   void OnSubmitClickEnd();
 
   /**
-   * This method will update the form validity so the submit controls states
-   * will be updated (for -moz-submit-invalid pseudo-class).
+   * This method will update the form validity.
+   *
    * This method has to be called by form elements whenever their validity state
    * or status regarding constraint validation changes.
    *
@@ -240,27 +206,17 @@ class HTMLFormElement final : public nsGenericHTMLElement,
    */
   void UpdateValidity(bool aElementValidityState);
 
-  /**
-   * Returns the form validity based on the last UpdateValidity() call.
-   *
-   * @return Whether the form was valid the last time UpdateValidity() was
-   * called.
-   *
-   * @note This method may not return the *current* validity state!
-   */
-  bool GetValidity() const { return !mInvalidElementsCount; }
-
   /**
    * This method check the form validity and make invalid form elements send
    * invalid event if needed.
    *
    * @return Whether the form is valid.
    *
-   * @note Do not call this method if novalidate/formnovalidate is used.
    * @note This method might disappear with bug 592124, hopefuly.
    * @see
    * https://html.spec.whatwg.org/multipage/form-control-infrastructure.html#interactively-validate-the-constraints
    */
+  MOZ_CAN_RUN_SCRIPT
   bool CheckValidFormSubmission();
 
   /**
@@ -270,16 +226,8 @@ class HTMLFormElement final : public nsGenericHTMLElement,
    *
    * @param aFormData the form data object
    */
-  nsresult ConstructEntryList(FormData* aFormData);
-
-  /**
-   * Whether the submission of this form has been ever prevented because of
-   * being invalid.
-   *
-   * @return Whether the submission of this form has been prevented because of
-   * being invalid.
-   */
-  bool HasEverTriedInvalidSubmit() const { return mEverTriedInvalidSubmit; }
+  // TODO: Convert this to MOZ_CAN_RUN_SCRIPT (bug 1415230)
+  MOZ_CAN_RUN_SCRIPT_BOUNDARY nsresult ConstructEntryList(FormData*);
 
   /**
    * Implements form[name]. Returns form controls in this form with the correct
@@ -341,9 +289,15 @@ class HTMLFormElement final : public nsGenericHTMLElement,
     SetHTMLAttr(nsGkAtoms::target, aValue, aRv);
   }
 
+  void GetRel(DOMString& aValue) { GetHTMLAttr(nsGkAtoms::rel, aValue); }
+  void SetRel(const nsAString& aRel, ErrorResult& aError) {
+    SetHTMLAttr(nsGkAtoms::rel, aRel, aError);
+  }
+  nsDOMTokenList* RelList();
+
   // it's only out-of-line because the class definition is not available in the
   // header
-  nsIHTMLCollection* Elements();
+  HTMLFormControlsCollection* Elements();
 
   int32_t Length();
 
@@ -372,10 +326,11 @@ class HTMLFormElement final : public nsGenericHTMLElement,
   MOZ_CAN_RUN_SCRIPT void RequestSubmit(nsGenericHTMLElement* aSubmitter,
                                         ErrorResult& aRv);
 
-  void Reset();
+  MOZ_CAN_RUN_SCRIPT void Reset();
 
   bool CheckValidity() { return CheckFormValidity(nullptr); }
 
+  MOZ_CAN_RUN_SCRIPT
   bool ReportValidity() { return CheckValidFormSubmission(); }
 
   Element* IndexedGetter(uint32_t aIndex, bool& aFound);
@@ -385,9 +340,6 @@ class HTMLFormElement final : public nsGenericHTMLElement,
 
   void GetSupportedNames(nsTArray<nsString>& aRetval);
 
-  static int32_t CompareFormControlPosition(Element* aElement1,
-                                            Element* aElement2,
-                                            const nsIContent* aForm);
 #ifdef DEBUG
   static void AssertDocumentOrder(
       const nsTArray<nsGenericHTMLFormElement*>& aControls, nsIContent* aForm);
@@ -399,14 +351,7 @@ class HTMLFormElement final : public nsGenericHTMLElement,
   JS::ExpandoAndGeneration mExpandoAndGeneration;
 
  protected:
-  virtual JSObject* WrapNode(JSContext* aCx,
-                             JS::Handle<JSObject*> aGivenProto) override;
-
-  void PostPasswordEvent();
-  void PostPossibleUsernameEvent();
-
-  RefPtr<AsyncEventDispatcher> mFormPasswordEventDispatcher;
-  RefPtr<AsyncEventDispatcher> mFormPossibleUsernameEventDispatcher;
+  JSObject* WrapNode(JSContext*, JS::Handle<JSObject*> aGivenProto) override;
 
   class RemoveElementRunnable;
   friend class RemoveElementRunnable;
@@ -536,6 +481,11 @@ class HTMLFormElement final : public nsGenericHTMLElement,
    */
   nsresult GetActionURL(nsIURI** aActionURL, Element* aOriginatingElement);
 
+  // Get the target to submit to. This is either the submitter's |formtarget| or
+  // the form's |target| (Including <base>).
+  void GetSubmissionTarget(nsGenericHTMLElement* aSubmitter,
+                           nsAString& aTarget);
+
   // Returns a number for this form that is unique within its owner document.
   // This is used by nsContentUtils::GenerateStateKey to identify form controls
   // that are inserted into the document by the parser.
@@ -556,10 +506,6 @@ class HTMLFormElement final : public nsGenericHTMLElement,
 
   /** The pending submission object */
   UniquePtr<HTMLFormSubmission> mPendingSubmission;
-  /** The request currently being submitted */
-  nsCOMPtr<nsIRequest> mSubmittingRequest;
-  /** The web progress object we are currently listening to */
-  nsWeakPtr mWebProgress;
 
   /** The target browsing context, if any. */
   RefPtr<BrowsingContext> mTargetContext;
@@ -580,7 +526,8 @@ class HTMLFormElement final : public nsGenericHTMLElement,
   // This is needed to properly clean up the bi-directional references
   // (both weak and strong) between the form and its HTMLImageElements.
 
-  nsTArray<HTMLImageElement*> mImageElements;  // Holds WEAK references
+  // Holds WEAK references
+  TreeOrderedArray<HTMLImageElement*> mImageElements;
 
   // A map from an ID or NAME attribute to the HTMLImageElement(s), this
   // hash holds strong references either to the named HTMLImageElement, or
@@ -598,10 +545,11 @@ class HTMLFormElement final : public nsGenericHTMLElement,
   /** Keep track of what the popup state was when the submit was initiated */
   PopupBlocker::PopupControlState mSubmitPopupState;
 
+  RefPtr<nsDOMTokenList> mRelList;
+
   /**
    * Number of invalid and candidate for constraint validation elements in the
    * form the last time UpdateValidity has been called.
-   * @note Should only be used by UpdateValidity() and GetValidity()!
    */
   int32_t mInvalidElementsCount;
 
@@ -631,14 +579,20 @@ class HTMLFormElement final : public nsGenericHTMLElement,
  private:
   bool IsSubmitting() const;
 
+  void SetDefaultSubmitElement(nsGenericHTMLFormElement*);
+
   NotNull<const Encoding*> GetSubmitEncoding();
 
   /**
    * Fire an event when the form is removed from the DOM tree. This is now only
-   * used by the password manager.
+   * used by the password manager and formautofill.
    */
   void MaybeFireFormRemoved();
 
+  MOZ_CAN_RUN_SCRIPT
+  void ReportInvalidUnfocusableElements(
+      const nsTArray<RefPtr<Element>>&& aInvalidElements);
+
   ~HTMLFormElement();
 };
 
