@@ -5,11 +5,9 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "HTMLFormSubmission.h"
-
 #include "HTMLFormElement.h"
+#include "HTMLFormSubmissionConstants.h"
 #include "nsCOMPtr.h"
-#include "nsIForm.h"
-#include "mozilla/dom/Document.h"
 #include "nsComponentManagerUtils.h"
 #include "nsGkAtoms.h"
 #include "nsIFormControl.h"
@@ -31,12 +29,15 @@
 #include "nsCExternalHandlerService.h"
 #include "nsContentUtils.h"
 
+#include "mozilla/dom/Document.h"
 #include "mozilla/dom/AncestorIterator.h"
 #include "mozilla/dom/Directory.h"
 #include "mozilla/dom/File.h"
 #include "mozilla/StaticPrefs_dom.h"
 #include "mozilla/RandomNum.h"
 
+#include <tuple>
+
 namespace mozilla::dom {
 
 namespace {
@@ -687,11 +688,10 @@ nsresult FSTextPlain::GetEncodedSubmission(nsIURI* aURI,
 
 HTMLFormSubmission::HTMLFormSubmission(
     nsIURI* aActionURL, const nsAString& aTarget,
-    mozilla::NotNull<const mozilla::Encoding*> aEncoding, Element* aSubmitter)
+    mozilla::NotNull<const mozilla::Encoding*> aEncoding)
     : mActionURL(aActionURL),
       mTarget(aTarget),
       mEncoding(aEncoding),
-      mSubmitter(aSubmitter),
       mInitiatedFromUserInput(UserActivation::IsHandlingUserInput()) {
   MOZ_COUNT_CTOR(HTMLFormSubmission);
 }
@@ -699,7 +699,7 @@ HTMLFormSubmission::HTMLFormSubmission(
 EncodingFormSubmission::EncodingFormSubmission(
     nsIURI* aActionURL, const nsAString& aTarget,
     NotNull<const Encoding*> aEncoding, Element* aSubmitter)
-    : HTMLFormSubmission(aActionURL, aTarget, aEncoding, aSubmitter) {
+    : HTMLFormSubmission(aActionURL, aTarget, aEncoding) {
   if (!aEncoding->CanEncodeEverything()) {
     nsAutoCString name;
     aEncoding->Name(name);
@@ -717,8 +717,7 @@ nsresult EncodingFormSubmission::EncodeVal(const nsAString& aStr,
                                            nsCString& aOut,
                                            EncodeType aEncodeType) {
   nsresult rv;
-  const Encoding* ignored;
-  Tie(rv, ignored) = mEncoding->Encode(aStr, aOut);
+  std::tie(rv, std::ignore) = mEncoding->Encode(aStr, aOut);
   if (NS_FAILED(rv)) {
     return rv;
   }
@@ -779,6 +778,32 @@ nsresult HTMLFormSubmission::GetFromForm(HTMLFormElement* aForm,
 
   nsresult rv;
 
+  // Get method (default: GET)
+  int32_t method = NS_FORM_METHOD_GET;
+  if (aSubmitter && aSubmitter->HasAttr(nsGkAtoms::formmethod)) {
+    GetEnumAttr(aSubmitter, nsGkAtoms::formmethod, &method);
+  } else {
+    GetEnumAttr(aForm, nsGkAtoms::method, &method);
+  }
+
+  if (method == NS_FORM_METHOD_DIALOG) {
+    HTMLDialogElement* dialog = aForm->FirstAncestorOfType<HTMLDialogElement>();
+
+    // If there isn't one, do nothing.
+    if (!dialog) {
+      return NS_ERROR_FAILURE;
+    }
+
+    nsAutoString result;
+    if (aSubmitter) {
+      aSubmitter->ResultForDialogSubmit(result);
+    }
+    *aFormSubmission = new DialogFormSubmission(result, aEncoding, dialog);
+    return NS_OK;
+  }
+
+  MOZ_ASSERT(method != NS_FORM_METHOD_DIALOG);
+
   // Get action
   nsCOMPtr<nsIURI> actionURL;
   rv = aForm->GetActionURL(getter_AddRefs(actionURL), aSubmitter);
@@ -793,7 +818,8 @@ nsresult HTMLFormSubmission::GetFromForm(HTMLFormElement* aForm,
     // policy - do *not* consult default-src, see:
     // http://www.w3.org/TR/CSP2/#directive-default-src
     rv = csp->Permits(aForm, nullptr /* nsICSPEventListener */, actionURL,
-                      nsIContentSecurityPolicy::FORM_ACTION_DIRECTIVE, true,
+                      nsIContentSecurityPolicy::FORM_ACTION_DIRECTIVE,
+                      true /* aSpecific */, true /* aSendViolationReports */,
                       &permitsFormAction);
     NS_ENSURE_SUCCESS(rv, rv);
     if (!permitsFormAction) {
@@ -802,58 +828,17 @@ nsresult HTMLFormSubmission::GetFromForm(HTMLFormElement* aForm,
   }
 
   // Get target
-  // The target is the submitter element formtarget attribute if the element
-  // is a submit control and has such an attribute.
-  // Otherwise, the target is the form owner's target attribute,
-  // if it has such an attribute.
-  // Finally, if one of the child nodes of the head element is a base element
-  // with a target attribute, then the value of the target attribute of the
-  // first such base element; or, if there is no such element, the empty string.
   nsAutoString target;
-  if (!(aSubmitter && aSubmitter->GetAttr(kNameSpaceID_None,
-                                          nsGkAtoms::formtarget, target)) &&
-      !aForm->GetAttr(kNameSpaceID_None, nsGkAtoms::target, target)) {
-    aForm->GetBaseTarget(target);
-  }
+  aForm->GetSubmissionTarget(aSubmitter, target);
 
   // Get encoding type (default: urlencoded)
   int32_t enctype = NS_FORM_ENCTYPE_URLENCODED;
-  if (aSubmitter &&
-      aSubmitter->HasAttr(kNameSpaceID_None, nsGkAtoms::formenctype)) {
+  if (aSubmitter && aSubmitter->HasAttr(nsGkAtoms::formenctype)) {
     GetEnumAttr(aSubmitter, nsGkAtoms::formenctype, &enctype);
   } else {
     GetEnumAttr(aForm, nsGkAtoms::enctype, &enctype);
   }
 
-  // Get method (default: GET)
-  int32_t method = NS_FORM_METHOD_GET;
-  if (aSubmitter &&
-      aSubmitter->HasAttr(kNameSpaceID_None, nsGkAtoms::formmethod)) {
-    GetEnumAttr(aSubmitter, nsGkAtoms::formmethod, &method);
-  } else {
-    GetEnumAttr(aForm, nsGkAtoms::method, &method);
-  }
-
-  if (method == NS_FORM_METHOD_DIALOG) {
-    HTMLDialogElement* dialog = aForm->FirstAncestorOfType<HTMLDialogElement>();
-
-    // If there isn't one, or if it does not have an open attribute, do
-    // nothing.
-    if (!dialog || !dialog->Open()) {
-      return NS_ERROR_FAILURE;
-    }
-
-    nsAutoString result;
-    if (aSubmitter) {
-      aSubmitter->ResultForDialogSubmit(result);
-    }
-    *aFormSubmission = new DialogFormSubmission(result, actionURL, target,
-                                                aEncoding, aSubmitter, dialog);
-    return NS_OK;
-  }
-
-  MOZ_ASSERT(method != NS_FORM_METHOD_DIALOG);
-
   // Choose encoder
   if (method == NS_FORM_METHOD_POST && enctype == NS_FORM_ENCTYPE_MULTIPART) {
     *aFormSubmission =
@@ -868,12 +853,10 @@ nsresult HTMLFormSubmission::GetFromForm(HTMLFormElement* aForm,
         enctype == NS_FORM_ENCTYPE_TEXTPLAIN) {
       AutoTArray<nsString, 1> args;
       nsString& enctypeStr = *args.AppendElement();
-      if (aSubmitter &&
-          aSubmitter->HasAttr(kNameSpaceID_None, nsGkAtoms::formenctype)) {
-        aSubmitter->GetAttr(kNameSpaceID_None, nsGkAtoms::formenctype,
-                            enctypeStr);
+      if (aSubmitter && aSubmitter->HasAttr(nsGkAtoms::formenctype)) {
+        aSubmitter->GetAttr(nsGkAtoms::formenctype, enctypeStr);
       } else {
-        aForm->GetAttr(kNameSpaceID_None, nsGkAtoms::enctype, enctypeStr);
+        aForm->GetAttr(nsGkAtoms::enctype, enctypeStr);
       }
 
       SendJSWarning(doc, "ForgotPostWarning", args);
