@@ -12,6 +12,7 @@
 #include "IndexedDatabase.h"  // for StructuredCloneFile...
 #include "IndexedDatabaseInlines.h"
 #include "IndexedDatabaseManager.h"
+#include "IndexedDBCipherKeyManager.h"
 #include "IndexedDBCommon.h"
 #include "ReportInternalError.h"
 
@@ -36,11 +37,13 @@
 #include "mozilla/ProfilerLabels.h"
 #include "mozilla/RefPtr.h"
 #include "mozilla/ResultExtensions.h"
+#include "mozilla/StaticPrefs_dom.h"
 #include "mozilla/StaticPtr.h"
 #include "mozilla/Telemetry.h"
 #include "mozilla/TelemetryScalarEnums.h"
 #include "mozilla/dom/quota/DecryptingInputStream_impl.h"
 #include "mozilla/dom/quota/QuotaCommon.h"
+#include "mozilla/dom/quota/ResultExtensions.h"
 #include "mozilla/dom/quota/ScopedLogExtraInfo.h"
 #include "mozilla/fallible.h"
 #include "mozilla/ipc/BackgroundParent.h"
@@ -63,10 +66,12 @@ class nsIFile;
 
 namespace mozilla::dom::indexedDB {
 
-static_assert(SNAPPY_VERSION == 0x010108);
+static_assert(SNAPPY_VERSION == 0x010200);
 
 using mozilla::ipc::IsOnBackgroundThread;
 
+const nsLiteralString kJournalDirectoryName = u"journals"_ns;
+
 namespace {
 
 constexpr StructuredCloneFileBase::FileType ToStructuredCloneFileType(
@@ -101,10 +106,10 @@ Result<StructuredCloneFileParent, nsresult> DeserializeStructuredCloneFile(
   const StructuredCloneFileBase::FileType type =
       ToStructuredCloneFileType(aText.First());
 
-  QM_TRY_INSPECT(
-      const auto& id,
-      ToResultGet<int32_t>(
-          ToInteger, type == StructuredCloneFileBase::eBlob
+  QM_TRY_INSPECT(const auto& id,
+                 MOZ_TO_RESULT_GET_TYPED(
+                     int32_t, ToInteger,
+                     type == StructuredCloneFileBase::eBlob
                          ? aText
                          : static_cast<const nsAString&>(Substring(aText, 1))));
 
@@ -113,14 +118,7 @@ Result<StructuredCloneFileParent, nsresult> DeserializeStructuredCloneFile(
   // XXX In bug 1432133, for some reasons DatabaseFileInfo object cannot be
   // got. This is just a short-term fix, and we are working on finding the real
   // cause in bug 1519859.
-  if (!fileInfo) {
-    IDB_WARNING(
-        "Corrupt structured clone data detected in IndexedDB. Failing the "
-        "database request. Bug 1519859 will address this problem.");
-    Telemetry::ScalarAdd(Telemetry::ScalarID::IDB_FAILURE_FILEINFO_ERROR, 1);
-
-    return Err(NS_ERROR_DOM_INDEXEDDB_UNKNOWN_ERR);
-  }
+  QM_TRY(OkIf((bool)fileInfo), Err(NS_ERROR_DOM_INDEXEDDB_UNKNOWN_ERR));
 
   return StructuredCloneFileParent{type, std::move(fileInfo)};
 }
@@ -157,7 +155,9 @@ class SandboxHolder final {
           NullPrincipal::CreateWithoutOriginAttributes();
 
       JS::Rooted<JSObject*> sandbox(aCx);
-      QM_TRY(xpc->CreateSandbox(aCx, principal, sandbox.address()), nullptr);
+      QM_TRY(
+          MOZ_TO_RESULT(xpc->CreateSandbox(aCx, principal, sandbox.address())),
+          nullptr);
 
       mSandbox = new JSObjectHolder(aCx, sandbox);
     }
@@ -309,8 +309,9 @@ nsresult ReadCompressedIndexDataValuesFromSource(
   MOZ_ASSERT(aOutIndexValues);
   MOZ_ASSERT(aOutIndexValues->IsEmpty());
 
-  QM_TRY_INSPECT(const int32_t& columnType,
-                 MOZ_TO_RESULT_INVOKE(aSource, GetTypeOfIndex, aColumnIndex));
+  QM_TRY_INSPECT(
+      const int32_t& columnType,
+      MOZ_TO_RESULT_INVOKE_MEMBER(aSource, GetTypeOfIndex, aColumnIndex));
 
   switch (columnType) {
     case mozIStorageStatement::VALUE_TYPE_NULL:
@@ -321,13 +322,14 @@ nsresult ReadCompressedIndexDataValuesFromSource(
       // we also can't use QM_TRY_UNWRAP/QM_TRY_INSPECT here.
       const uint8_t* blobData;
       uint32_t blobDataLength;
-      QM_TRY(aSource.GetSharedBlob(aColumnIndex, &blobDataLength, &blobData));
+      QM_TRY(MOZ_TO_RESULT(
+          aSource.GetSharedBlob(aColumnIndex, &blobDataLength, &blobData)));
 
       QM_TRY(OkIf(blobDataLength), NS_ERROR_FILE_CORRUPTED,
              IDB_REPORT_INTERNAL_ERR_LAMBDA);
 
-      QM_TRY(ReadCompressedIndexDataValuesFromBlob(
-          Span(blobData, blobDataLength), aOutIndexValues));
+      QM_TRY(MOZ_TO_RESULT(ReadCompressedIndexDataValuesFromBlob(
+          Span(blobData, blobDataLength), aOutIndexValues)));
 
       return NS_OK;
     }
@@ -341,8 +343,7 @@ Result<StructuredCloneReadInfoParent, nsresult>
 GetStructuredCloneReadInfoFromBlob(const uint8_t* aBlobData,
                                    uint32_t aBlobDataLength,
                                    const DatabaseFileManager& aFileManager,
-                                   const nsAString& aFileIds,
-                                   const Maybe<CipherKey>& aMaybeKey) {
+                                   const nsAString& aFileIds) {
   MOZ_ASSERT(!IsOnBackgroundThread());
 
   AUTO_PROFILER_LABEL("GetStructuredCloneReadInfoFromBlob", DOM);
@@ -355,7 +356,9 @@ GetStructuredCloneReadInfoFromBlob(const uint8_t* aBlobData,
                                             &uncompressedLength)),
          Err(NS_ERROR_FILE_CORRUPTED));
 
-  AutoTArray<uint8_t, 512> uncompressed;
+  // `data` (JSStructuredCloneData) currently uses 4k buffer internally.
+  // For performance reasons, it's better to align `uncompressed` with that.
+  AutoTArray<uint8_t, 4096> uncompressed;
   QM_TRY(OkIf(uncompressed.SetLength(uncompressedLength, fallible)),
          Err(NS_ERROR_OUT_OF_MEMORY));
 
@@ -383,7 +386,7 @@ GetStructuredCloneReadInfoFromBlob(const uint8_t* aBlobData,
 Result<StructuredCloneReadInfoParent, nsresult>
 GetStructuredCloneReadInfoFromExternalBlob(
     uint64_t aIntData, const DatabaseFileManager& aFileManager,
-    const nsAString& aFileIds, const Maybe<CipherKey>& aMaybeKey) {
+    const nsAString& aFileIds) {
   MOZ_ASSERT(!IsOnBackgroundThread());
 
   AUTO_PROFILER_LABEL("GetStructuredCloneReadInfoFromExternalBlob", DOM);
@@ -401,7 +404,7 @@ GetStructuredCloneReadInfoFromExternalBlob(
   QM_TRY(OkIf(index < files.Length()), Err(NS_ERROR_UNEXPECTED),
          [](const auto&) { MOZ_ASSERT(false, "Bad index value!"); });
 
-  if (IndexedDatabaseManager::PreprocessingEnabled()) {
+  if (StaticPrefs::dom_indexedDB_preprocessing()) {
     return StructuredCloneReadInfoParent{
         JSStructuredCloneData{JS::StructuredCloneScope::DifferentProcess},
         std::move(files), true};
@@ -411,6 +414,15 @@ GetStructuredCloneReadInfoFromExternalBlob(
   const StructuredCloneFileParent& file = files[index];
   MOZ_ASSERT(file.Type() == StructuredCloneFileBase::eStructuredClone);
 
+  Maybe<CipherKey> maybeKey;
+
+  if (aFileManager.IsInPrivateBrowsingMode()) {
+    nsCString fileKeyId;
+    fileKeyId.AppendInt(file.FileInfo().Id());
+
+    maybeKey = aFileManager.MutableCipherKeyManagerRef().Get(fileKeyId);
+  }
+
   auto data = JSStructuredCloneData{JS::StructuredCloneScope::DifferentProcess};
 
   {
@@ -420,19 +432,20 @@ GetStructuredCloneReadInfoFromExternalBlob(
     QM_TRY_INSPECT(
         const auto& fileInputStream,
         NS_NewLocalFileInputStream(nativeFile)
-            .andThen([aMaybeKey](auto fileInputStream)
+            .andThen([maybeKey](auto fileInputStream)
                          -> Result<nsCOMPtr<nsIInputStream>, nsresult> {
-              if (aMaybeKey) {
+              if (maybeKey) {
                 return nsCOMPtr<nsIInputStream>{MakeRefPtr<
                     quota::DecryptingInputStream<IndexedDBCipherStrategy>>(
                     WrapNotNull(std::move(fileInputStream)),
-                    kEncryptedStreamBlockSize, *aMaybeKey)};
+                    kEncryptedStreamBlockSize, *maybeKey)};
               }
 
               return fileInputStream;
             }));
 
-    QM_TRY(SnappyUncompressStructuredCloneData(*fileInputStream, data));
+    QM_TRY(MOZ_TO_RESULT(
+        SnappyUncompressStructuredCloneData(*fileInputStream, data)));
   }
 
   return StructuredCloneReadInfoParent{std::move(data), std::move(files),
@@ -443,43 +456,45 @@ template <typename T>
 Result<StructuredCloneReadInfoParent, nsresult>
 GetStructuredCloneReadInfoFromSource(T* aSource, uint32_t aDataIndex,
                                      uint32_t aFileIdsIndex,
-                                     const DatabaseFileManager& aFileManager,
-                                     const Maybe<CipherKey>& aMaybeKey) {
+                                     const DatabaseFileManager& aFileManager) {
   MOZ_ASSERT(!IsOnBackgroundThread());
   MOZ_ASSERT(aSource);
 
-  QM_TRY_INSPECT(const int32_t& columnType,
-                 MOZ_TO_RESULT_INVOKE(aSource, GetTypeOfIndex, aDataIndex));
+  QM_TRY_INSPECT(
+      const int32_t& columnType,
+      MOZ_TO_RESULT_INVOKE_MEMBER(aSource, GetTypeOfIndex, aDataIndex));
 
-  QM_TRY_INSPECT(const bool& isNull,
-                 MOZ_TO_RESULT_INVOKE(aSource, GetIsNull, aFileIdsIndex));
+  QM_TRY_INSPECT(const bool& isNull, MOZ_TO_RESULT_INVOKE_MEMBER(
+                                         aSource, GetIsNull, aFileIdsIndex));
 
   QM_TRY_INSPECT(const nsString& fileIds, ([aSource, aFileIdsIndex, isNull] {
                    return isNull ? Result<nsString, nsresult>{VoidString()}
-                                 : MOZ_TO_RESULT_INVOKE_TYPED(nsString, aSource,
-                                                              GetString,
-                                                              aFileIdsIndex);
+                                 : MOZ_TO_RESULT_INVOKE_MEMBER_TYPED(
+                                       nsString, aSource, GetString,
+                                       aFileIdsIndex);
                  }()));
 
   switch (columnType) {
     case mozIStorageStatement::VALUE_TYPE_INTEGER: {
-      QM_TRY_INSPECT(const int64_t& intData,
-                     MOZ_TO_RESULT_INVOKE(aSource, GetInt64, aDataIndex));
+      QM_TRY_INSPECT(
+          const int64_t& intData,
+          MOZ_TO_RESULT_INVOKE_MEMBER(aSource, GetInt64, aDataIndex));
 
       uint64_t uintData;
       memcpy(&uintData, &intData, sizeof(uint64_t));
 
       return GetStructuredCloneReadInfoFromExternalBlob(uintData, aFileManager,
-                                                        fileIds, aMaybeKey);
+                                                        fileIds);
     }
 
     case mozIStorageStatement::VALUE_TYPE_BLOB: {
       const uint8_t* blobData;
       uint32_t blobDataLength;
-      QM_TRY(aSource->GetSharedBlob(aDataIndex, &blobDataLength, &blobData));
+      QM_TRY(MOZ_TO_RESULT(
+          aSource->GetSharedBlob(aDataIndex, &blobDataLength, &blobData)));
 
-      return GetStructuredCloneReadInfoFromBlob(
-          blobData, blobDataLength, aFileManager, fileIds, aMaybeKey);
+      return GetStructuredCloneReadInfoFromBlob(blobData, blobDataLength,
+                                                aFileManager, fileIds);
     }
 
     default:
@@ -493,8 +508,8 @@ IndexDataValue::IndexDataValue() : mIndexId(0), mUnique(false) {
   MOZ_COUNT_CTOR(IndexDataValue);
 }
 
-#ifdef NS_BUILD_REFCNT_LOGGING
-IndexDataValue::IndexDataValue(IndexDataValue&& aOther)
+#if defined(DEBUG) || defined(NS_BUILD_REFCNT_LOGGING)
+IndexDataValue::IndexDataValue(IndexDataValue&& aOther) noexcept
     : mIndexId(aOther.mIndexId),
       mPosition(std::move(aOther.mPosition)),
       mLocaleAwarePosition(std::move(aOther.mLocaleAwarePosition)),
@@ -624,8 +639,9 @@ nsresult ReadCompressedIndexDataValues(
 template <typename T>
 Result<IndexDataValuesAutoArray, nsresult> ReadCompressedIndexDataValues(
     T& aValues, uint32_t aColumnIndex) {
-  return ToResultInvoke<IndexDataValuesAutoArray>(
-      &ReadCompressedIndexDataValuesFromSource<T>, aValues, aColumnIndex);
+  return MOZ_TO_RESULT_INVOKE_TYPED(IndexDataValuesAutoArray,
+                                    &ReadCompressedIndexDataValuesFromSource<T>,
+                                    aValues, aColumnIndex);
 }
 
 template Result<IndexDataValuesAutoArray, nsresult>
@@ -675,20 +691,17 @@ ReadCompressedNumber(const Span<const uint8_t> aSpan) {
 Result<StructuredCloneReadInfoParent, nsresult>
 GetStructuredCloneReadInfoFromValueArray(
     mozIStorageValueArray* aValues, uint32_t aDataIndex, uint32_t aFileIdsIndex,
-    const DatabaseFileManager& aFileManager,
-    const Maybe<CipherKey>& aMaybeKey) {
-  return GetStructuredCloneReadInfoFromSource(
-      aValues, aDataIndex, aFileIdsIndex, aFileManager, aMaybeKey);
+    const DatabaseFileManager& aFileManager) {
+  return GetStructuredCloneReadInfoFromSource(aValues, aDataIndex,
+                                              aFileIdsIndex, aFileManager);
 }
 
 Result<StructuredCloneReadInfoParent, nsresult>
-GetStructuredCloneReadInfoFromStatement(mozIStorageStatement* aStatement,
-                                        uint32_t aDataIndex,
-                                        uint32_t aFileIdsIndex,
-                                        const DatabaseFileManager& aFileManager,
-                                        const Maybe<CipherKey>& aMaybeKey) {
-  return GetStructuredCloneReadInfoFromSource(
-      aStatement, aDataIndex, aFileIdsIndex, aFileManager, aMaybeKey);
+GetStructuredCloneReadInfoFromStatement(
+    mozIStorageStatement* aStatement, uint32_t aDataIndex,
+    uint32_t aFileIdsIndex, const DatabaseFileManager& aFileManager) {
+  return GetStructuredCloneReadInfoFromSource(aStatement, aDataIndex,
+                                              aFileIdsIndex, aFileManager);
 }
 
 Result<nsTArray<StructuredCloneFileParent>, nsresult>
@@ -715,9 +728,9 @@ nsresult ExecuteSimpleSQLSequence(mozIStorageConnection& aConnection,
                                   Span<const nsLiteralCString> aSQLCommands) {
   for (const auto& aSQLCommand : aSQLCommands) {
     const auto extraInfo = quota::ScopedLogExtraInfo{
-        quota::ScopedLogExtraInfo::kTagQuery, aSQLCommand};
+        quota::ScopedLogExtraInfo::kTagQueryTainted, aSQLCommand};
 
-    QM_TRY(aConnection.ExecuteSimpleSQL(aSQLCommand));
+    QM_TRY(MOZ_TO_RESULT(aConnection.ExecuteSimpleSQL(aSQLCommand)));
   }
 
   return NS_OK;
