@@ -35,15 +35,14 @@ using media::TimeUnit;
 
 AudioSink::AudioSink(AbstractThread* aThread,
                      MediaQueue<AudioData>& aAudioQueue, const AudioInfo& aInfo,
-                     AudioDeviceInfo* aAudioDevice)
-    : mInfo(aInfo),
-      mAudioDevice(aAudioDevice),
-      mPlaying(true),
+                     bool aShouldResistFingerprinting)
+    : mPlaying(true),
       mWritten(0),
       mErrored(false),
       mOwnerThread(aThread),
       mFramesParsed(0),
-      mOutputRate(DecideAudioPlaybackSampleRate(aInfo)),
+      mOutputRate(
+          DecideAudioPlaybackSampleRate(aInfo, aShouldResistFingerprinting)),
       mOutputChannels(DecideAudioPlaybackChannels(aInfo)),
       mAudibilityMonitor(
           mOutputRate,
@@ -88,18 +87,18 @@ AudioSink::AudioSink(AbstractThread* aThread,
 }
 
 AudioSink::~AudioSink() {
-  // Generally instances of AudioSink should be properly Shutdown manually.
-  // The only way deleting an AudioSink without shutdown an happen is if the
+  // Generally instances of AudioSink should be properly shut down manually.
+  // The only way deleting an AudioSink without shutdown can happen is if the
   // dispatch back to the MDSM thread after initializing it asynchronously
   // fails. When that's the case, the stream has been initialized but not
   // started. Manually shutdown the AudioStream in this case.
   if (mAudioStream) {
-    mAudioStream->Shutdown();
+    mAudioStream->ShutDown();
   }
 }
 
 nsresult AudioSink::InitializeAudioStream(
-    const PlaybackParams& aParams,
+    const PlaybackParams& aParams, const RefPtr<AudioDeviceInfo>& aAudioDevice,
     AudioSink::InitializationType aInitializationType) {
   if (aInitializationType == AudioSink::InitializationType::UNMUTING) {
     // Consider the stream to be audible immediately, before initialization
@@ -127,9 +126,9 @@ nsresult AudioSink::InitializeAudioStream(
   MOZ_ASSERT(!mAudioStream);
   mAudioStream =
       new AudioStream(*this, mOutputRate, mOutputChannels, channelMap);
-  nsresult rv = mAudioStream->Init(mAudioDevice);
+  nsresult rv = mAudioStream->Init(aAudioDevice);
   if (NS_FAILED(rv)) {
-    mAudioStream->Shutdown();
+    mAudioStream->ShutDown();
     mAudioStream = nullptr;
     return rv;
   }
@@ -143,9 +142,8 @@ nsresult AudioSink::InitializeAudioStream(
   return NS_OK;
 }
 
-nsresult AudioSink::Start(
-    const media::TimeUnit& aStartTime,
-    MozPromiseHolder<MediaSink::EndedPromise>& aEndedPromise) {
+RefPtr<MediaSink::EndedPromise> AudioSink::Start(
+    const media::TimeUnit& aStartTime) {
   MOZ_ASSERT(mOwnerThread->IsCurrentThreadIn());
 
   mAudioQueueListener = mAudioQueue.PushEvent().Connect(
@@ -161,7 +159,7 @@ nsresult AudioSink::Start(
   // ready to be played.
   NotifyAudioNeeded();
 
-  return mAudioStream->Start(aEndedPromise);
+  return mAudioStream->Start();
 }
 
 TimeUnit AudioSink::GetPosition() {
@@ -196,14 +194,14 @@ TimeUnit AudioSink::UnplayedDuration() const {
 }
 
 void AudioSink::ReenqueueUnplayedAudioDataIfNeeded() {
-  // This is OK: the AudioStream has been shut down. Shutdown guarantees that
+  // This is OK: the AudioStream has been shut down. ShutDown guarantees that
   // the audio callback thread won't call back again.
-  mProcessedSPSCQueue->ResetThreadIds();
+  mProcessedSPSCQueue->ResetConsumerThreadId();
 
   // construct an AudioData
-  int sampleCount = mProcessedSPSCQueue->AvailableRead();
+  int sampleInRingbuffer = mProcessedSPSCQueue->AvailableRead();
 
-  if (!sampleCount) {
+  if (!sampleInRingbuffer) {
     return;
   }
 
@@ -217,64 +215,85 @@ void AudioSink::ReenqueueUnplayedAudioDataIfNeeded() {
     rate = mOutputRate;
   }
 
-  uint32_t frameCount = sampleCount / channelCount;
-  auto duration = FramesToTimeUnit(frameCount, rate);
-  if (!duration.IsValid()) {
-    NS_WARNING("Int overflow in AudioSink");
-    mErrored = true;
-    return;
-  }
-
-  AlignedAudioBuffer queuedAudio(sampleCount);
-  DebugOnly<int> samplesRead =
-      mProcessedSPSCQueue->Dequeue(queuedAudio.Data(), sampleCount);
-  MOZ_ASSERT(samplesRead == sampleCount);
+  uint32_t framesRemaining = sampleInRingbuffer / channelCount;
 
+  nsTArray<AlignedAudioBuffer> packetsToReenqueue;
+  RefPtr<AudioData> frontPacket = mAudioQueue.PeekFront();
+  uint32_t offset;
+  TimeUnit time;
+  uint32_t typicalPacketFrameCount;
   // Extrapolate mOffset, mTime from the front of the queue
   // We can't really find a good value for `mOffset`, so we take what we have
   // at the front of the queue.
   // For `mTime`, assume there hasn't been a discontinuity recently.
-  RefPtr<AudioData> frontPacket = mAudioQueue.PeekFront();
-  uint32_t offset;
-  TimeUnit time;
   if (!frontPacket) {
     // We do our best here, but it's not going to be perfect.
+    typicalPacketFrameCount = 1024;  // typical for e.g. AAC
     offset = 0;
-    time = std::max(GetPosition() - duration, TimeUnit::Zero());
+    time = GetPosition();
   } else {
+    typicalPacketFrameCount = frontPacket->Frames();
     offset = frontPacket->mOffset;
-    time = frontPacket->mTime - duration;
+    time = frontPacket->mTime;
   }
-  RefPtr<AudioData> data =
-      new AudioData(offset, time, std::move(queuedAudio), channelCount, rate);
-  MOZ_DIAGNOSTIC_ASSERT(duration == data->mDuration, "must be equal");
 
-  SINK_LOG(
-      "Muting: Pushing back %u frames (%lfms) from the ring buffer back into "
-      "the audio queue",
-      frameCount, static_cast<float>(frameCount) / rate);
+  // Extract all audio data from the ring buffer, we can only read the data from
+  // the most recent, so we reenqueue the data, packetized, in a temporary
+  // array.
+  while (framesRemaining) {
+    uint32_t packetFrameCount =
+        std::min(framesRemaining, typicalPacketFrameCount);
+    framesRemaining -= packetFrameCount;
 
-  mAudioQueue.PushFront(data);
+    int packetSampleCount = packetFrameCount * channelCount;
+    AlignedAudioBuffer packetData(packetSampleCount);
+    DebugOnly<int> samplesRead =
+        mProcessedSPSCQueue->Dequeue(packetData.Data(), packetSampleCount);
+    MOZ_ASSERT(samplesRead == packetSampleCount);
+
+    packetsToReenqueue.AppendElement(packetData);
+  }
+  // Reenqueue in the audio queue in correct order in the audio queue, starting
+  // with the end of the temporary array.
+  while (!packetsToReenqueue.IsEmpty()) {
+    auto packetData = packetsToReenqueue.PopLastElement();
+    uint32_t packetFrameCount = packetData.Length() / channelCount;
+    auto duration = TimeUnit(packetFrameCount, rate);
+    if (!duration.IsValid()) {
+      NS_WARNING("Int overflow in AudioSink");
+      mErrored = true;
+      return;
+    }
+    time -= duration;
+    RefPtr<AudioData> packet =
+        new AudioData(offset, time, std::move(packetData), channelCount, rate);
+    MOZ_DIAGNOSTIC_ASSERT(duration == packet->mDuration, "must be equal");
+
+    SINK_LOG(
+        "Muting: Pushing back %u frames (%lfms) from the ring buffer back into "
+        "the audio queue at pts %lf",
+        packetFrameCount, 1000 * static_cast<float>(packetFrameCount) / rate,
+        time.ToSeconds());
+    // The audio data's timestamp would be adjusted already if we're in looping,
+    // so we don't want to adjust them again.
+    mAudioQueue.PushFront(packet,
+                          MediaQueue<AudioData>::TimestampAdjustment::Disable);
+  }
 }
 
-Maybe<MozPromiseHolder<MediaSink::EndedPromise>> AudioSink::Shutdown(
-    ShutdownCause aShutdownCause) {
+void AudioSink::ShutDown() {
   MOZ_ASSERT(mOwnerThread->IsCurrentThreadIn());
 
   mAudioQueueListener.DisconnectIfExists();
   mAudioQueueFinishListener.DisconnectIfExists();
   mProcessedQueueListener.DisconnectIfExists();
 
-  Maybe<MozPromiseHolder<MediaSink::EndedPromise>> rv;
-
   if (mAudioStream) {
-    rv = mAudioStream->Shutdown(aShutdownCause);
+    mAudioStream->ShutDown();
     mAudioStream = nullptr;
     ReenqueueUnplayedAudioDataIfNeeded();
   }
   mProcessedQueueFinished = true;
-
-  return rv;
 }
 
 void AudioSink::SetVolume(double aVolume) {
@@ -318,7 +337,8 @@ void AudioSink::SetPlaying(bool aPlaying) {
 }
 
 TimeUnit AudioSink::GetEndTime() const {
-  TimeUnit played = FramesToTimeUnit(mWritten, mOutputRate) + mStartTime;
+  uint64_t written = mWritten;
+  TimeUnit played = media::TimeUnit(written, mOutputRate) + mStartTime;
   if (!played.IsValid()) {
     NS_WARNING("Int overflow calculating audio end time");
     return TimeUnit::Zero();
@@ -335,7 +355,7 @@ uint32_t AudioSink::PopFrames(AudioDataValue* aBuffer, uint32_t aFrames,
   // happen when not using cubeb remoting, and often when changing audio device
   // at the system level.
   if (aAudioThreadChanged) {
-    mProcessedSPSCQueue->ResetThreadIds();
+    mProcessedSPSCQueue->ResetConsumerThreadId();
   }
 
   TRACE_COMMENT("AudioSink::PopFrames", "%u frames (ringbuffer: %u/%u)",
@@ -363,8 +383,6 @@ uint32_t AudioSink::PopFrames(AudioDataValue* aBuffer, uint32_t aFrames,
   SINK_LOG_V("Popping %u frames. Remaining in ringbuffer %u / %u\n", aFrames,
              SampleToFrame(mProcessedSPSCQueue->AvailableRead()),
              SampleToFrame(mProcessedSPSCQueue->Capacity()));
-
-  // Don't consider the silence added because of an underrun.
   CheckIsAudible(Span(aBuffer, samplesRead), mOutputChannels);
 
   return SampleToFrame(samplesRead);
@@ -500,6 +518,9 @@ void AudioSink::NotifyAudioNeeded() {
       // we pushed to the audio hardware. We must push silence into the audio
       // hardware so that the next audio packet begins playback at the correct
       // time. But don't push more than the ring buffer can receive.
+      SINK_LOG("Sample time %" PRId64 " > frames parsed %" PRId64,
+               sampleTime.value(), mFramesParsed);
+
       missingFrames = std::min<int64_t>(
           std::min<int64_t>(INT32_MAX, missingFrames.value()),
           SampleToFrame(mProcessedSPSCQueue->AvailableWrite()));
@@ -571,7 +592,7 @@ already_AddRefed<AudioData> AudioSink::CreateAudioFromBuffer(
   if (!frames) {
     return nullptr;
   }
-  auto duration = FramesToTimeUnit(frames, mOutputRate);
+  auto duration = media::TimeUnit(frames, mOutputRate);
   if (!duration.IsValid()) {
     NS_WARNING("Int overflow in AudioSink");
     mErrored = true;
