@@ -13,16 +13,18 @@
 #include "AnnexB.h"
 #include "BufferStream.h"
 #include "H264.h"
-#include "Index.h"
+#include "H265.h"
 #include "MP4Decoder.h"
 #include "MP4Metadata.h"
 #include "MoofParser.h"
 #include "ResourceStream.h"
+#include "TimeUnits.h"
 #include "VPXDecoder.h"
 #include "mozilla/Span.h"
 #include "mozilla/StaticPrefs_media.h"
 #include "mozilla/Telemetry.h"
 #include "nsPrintfCString.h"
+#include "SampleIterator.h"
 
 extern mozilla::LazyLogModule gMediaDemuxerLog;
 mozilla::LogModule* GetDemuxerLog() { return gMediaDemuxerLog; }
@@ -33,28 +35,32 @@ mozilla::LogModule* GetDemuxerLog() { return gMediaDemuxerLog; }
 
 namespace mozilla {
 
+using TimeUnit = media::TimeUnit;
+using TimeInterval = media::TimeInterval;
+using TimeIntervals = media::TimeIntervals;
+
 DDLoggedTypeDeclNameAndBase(MP4TrackDemuxer, MediaTrackDemuxer);
 
 class MP4TrackDemuxer : public MediaTrackDemuxer,
                         public DecoderDoctorLifeLogger<MP4TrackDemuxer> {
  public:
   MP4TrackDemuxer(MediaResource* aResource, UniquePtr<TrackInfo>&& aInfo,
-                  const IndiceWrapper& aIndices);
+                  const IndiceWrapper& aIndices, uint32_t aTimeScale);
 
   UniquePtr<TrackInfo> GetInfo() const override;
 
-  RefPtr<SeekPromise> Seek(const media::TimeUnit& aTime) override;
+  RefPtr<SeekPromise> Seek(const TimeUnit& aTime) override;
 
   RefPtr<SamplesPromise> GetSamples(int32_t aNumSamples = 1) override;
 
   void Reset() override;
 
-  nsresult GetNextRandomAccessPoint(media::TimeUnit* aTime) override;
+  nsresult GetNextRandomAccessPoint(TimeUnit* aTime) override;
 
   RefPtr<SkipAccessPointPromise> SkipToNextRandomAccessPoint(
-      const media::TimeUnit& aTimeThreshold) override;
+      const TimeUnit& aTimeThreshold) override;
 
-  media::TimeIntervals GetBuffered() override;
+  TimeIntervals GetBuffered() override;
 
   void NotifyDataRemoved();
   void NotifyDataArrived();
@@ -66,13 +72,13 @@ class MP4TrackDemuxer : public MediaTrackDemuxer,
   RefPtr<MediaResource> mResource;
   RefPtr<ResourceStream> mStream;
   UniquePtr<TrackInfo> mInfo;
-  RefPtr<Index> mIndex;
+  RefPtr<MP4SampleIndex> mIndex;
   UniquePtr<SampleIterator> mIterator;
-  Maybe<media::TimeUnit> mNextKeyframeTime;
+  Maybe<TimeUnit> mNextKeyframeTime;
   // Queued samples extracted by the demuxer, but not yet returned.
   RefPtr<MediaRawData> mQueuedSample;
   bool mNeedReIndex;
-  enum CodecType { kH264, kVP9, kOther } mType = kOther;
+  enum CodecType { kH264, kVP9, kAAC, kHEVC, kOther } mType = kOther;
 };
 
 MP4Demuxer::MP4Demuxer(MediaResource* aResource)
@@ -184,8 +190,11 @@ RefPtr<MP4Demuxer::InitPromise> MP4Demuxer::Init() {
         }
         continue;
       }
-      RefPtr<MP4TrackDemuxer> demuxer = new MP4TrackDemuxer(
-          mResource, std::move(info.Ref()), *indices.Ref().get());
+      LOG("Created audio track demuxer for info (%s)",
+          info.Ref()->ToString().get());
+      RefPtr<MP4TrackDemuxer> demuxer =
+          new MP4TrackDemuxer(mResource, std::move(info.Ref()),
+                              *indices.Ref().get(), info.Ref()->mTimeScale);
       DDLINKCHILD("audio demuxer", demuxer.get());
       mAudioDemuxers.AppendElement(std::move(demuxer));
     }
@@ -221,8 +230,11 @@ RefPtr<MP4Demuxer::InitPromise> MP4Demuxer::Init() {
         }
         continue;
       }
-      RefPtr<MP4TrackDemuxer> demuxer = new MP4TrackDemuxer(
-          mResource, std::move(info.Ref()), *indices.Ref().get());
+      LOG("Created video track demuxer for info (%s)",
+          info.Ref()->ToString().get());
+      RefPtr<MP4TrackDemuxer> demuxer =
+          new MP4TrackDemuxer(mResource, std::move(info.Ref()),
+                              *indices.Ref().get(), info.Ref()->mTimeScale);
       DDLINKCHILD("video demuxer", demuxer.get());
       mVideoDemuxers.AppendElement(std::move(demuxer));
     }
@@ -298,23 +310,26 @@ UniquePtr<EncryptionInfo> MP4Demuxer::GetCrypto() {
   UniquePtr<EncryptionInfo> crypto;
   if (!mCryptoInitData.IsEmpty()) {
     crypto.reset(new EncryptionInfo{});
-    crypto->AddInitData(NS_LITERAL_STRING("cenc"), mCryptoInitData);
+    crypto->AddInitData(u"cenc"_ns, mCryptoInitData);
   }
   return crypto;
 }
 
 MP4TrackDemuxer::MP4TrackDemuxer(MediaResource* aResource,
                                  UniquePtr<TrackInfo>&& aInfo,
-                                 const IndiceWrapper& aIndices)
+                                 const IndiceWrapper& aIndices,
+                                 uint32_t aTimeScale)
     : mResource(aResource),
       mStream(new ResourceStream(aResource)),
       mInfo(std::move(aInfo)),
-      mIndex(new Index(aIndices, mStream, mInfo->mTrackId, mInfo->IsAudio())),
+      mIndex(new MP4SampleIndex(aIndices, mStream, mInfo->mTrackId,
+                                mInfo->IsAudio(), aTimeScale)),
       mIterator(MakeUnique<SampleIterator>(mIndex)),
       mNeedReIndex(true) {
   EnsureUpToDateIndex();  // Force update of index
 
   VideoInfo* videoInfo = mInfo->GetAsVideoInfo();
+  AudioInfo* audioInfo = mInfo->GetAsAudioInfo();
   if (videoInfo && MP4Decoder::IsH264(mInfo->mMimeType)) {
     mType = kH264;
     RefPtr<MediaByteBuffer> extraData = videoInfo->mExtraData;
@@ -327,9 +342,19 @@ MP4TrackDemuxer::MP4TrackDemuxer(MediaResource* aResource,
       videoInfo->mDisplay.width = spsdata.display_width;
       videoInfo->mDisplay.height = spsdata.display_height;
     }
-  } else {
-    if (videoInfo && VPXDecoder::IsVP9(mInfo->mMimeType)) {
-      mType = kVP9;
+  } else if (videoInfo && VPXDecoder::IsVP9(mInfo->mMimeType)) {
+    mType = kVP9;
+  } else if (audioInfo && MP4Decoder::IsAAC(mInfo->mMimeType)) {
+    mType = kAAC;
+  } else if (videoInfo && MP4Decoder::IsHEVC(mInfo->mMimeType)) {
+    mType = kHEVC;
+    if (auto rv = H265::DecodeSPSFromHVCCExtraData(videoInfo->mExtraData);
+        rv.isOk()) {
+      const auto sps = rv.unwrap();
+      videoInfo->mImage.width = sps.GetImageSize().Width();
+      videoInfo->mImage.height = sps.GetImageSize().Height();
+      videoInfo->mDisplay.width = sps.GetDisplaySize().Width();
+      videoInfo->mDisplay.height = sps.GetDisplaySize().Height();
     }
   }
 }
@@ -351,11 +376,11 @@ void MP4TrackDemuxer::EnsureUpToDateIndex() {
 }
 
 RefPtr<MP4TrackDemuxer::SeekPromise> MP4TrackDemuxer::Seek(
-    const media::TimeUnit& aTime) {
+    const TimeUnit& aTime) {
   auto seekTime = aTime;
   mQueuedSample = nullptr;
 
-  mIterator->Seek(seekTime.ToMicroseconds());
+  mIterator->Seek(seekTime);
 
   // Check what time we actually seeked to.
   do {
@@ -422,7 +447,7 @@ already_AddRefed<MediaRawData> MP4TrackDemuxer::GetNextSample() {
       }
     } else if (mType == kVP9 && !sample->mCrypto.IsEncrypted()) {
       bool keyframe = VPXDecoder::IsKeyframe(
-          MakeSpan<const uint8_t>(sample->Data(), sample->Size()),
+          Span<const uint8_t>(sample->Data(), sample->Size()),
           VPXDecoder::Codec::VP9);
       if (sample->mKeyframe != keyframe) {
         NS_WARNING(nsPrintfCString(
@@ -437,6 +462,60 @@ already_AddRefed<MediaRawData> MP4TrackDemuxer::GetNextSample() {
     }
   }
 
+  // Adjust trimming information if needed.
+  if (mInfo->GetAsAudioInfo()) {
+    AudioInfo* info = mInfo->GetAsAudioInfo();
+    TimeUnit originalPts = sample->mTime;
+    TimeUnit originalEnd = sample->GetEndTime();
+    if (sample->mTime.IsNegative()) {
+      sample->mTime = TimeUnit::Zero(originalPts);
+      sample->mDuration = std::max(TimeUnit::Zero(sample->mTime),
+                                   originalPts + sample->mDuration);
+      sample->mOriginalPresentationWindow =
+          Some(TimeInterval{originalPts, originalEnd});
+    }
+    // The demuxer only knows the presentation time of the packet, not the
+    // actual number of samples that will be decoded from this packet.
+    // However we need to trim the last packet to the correct duration.
+    // Find the actual size of the decoded packet to know how many samples to
+    // trim. This only works because the packet size are constant.
+    TimeUnit totalMediaDurationIncludingTrimming =
+        info->mDuration - info->mMediaTime;
+    if (mType == kAAC &&
+        sample->GetEndTime() >= totalMediaDurationIncludingTrimming &&
+        totalMediaDurationIncludingTrimming.IsPositive()) {
+      // Seek backward a bit.
+      mIterator->Seek(sample->mTime - sample->mDuration);
+      RefPtr<MediaRawData> previousSample = mIterator->GetNext();
+      if (previousSample) {
+        TimeInterval fullPacketDuration{previousSample->mTime,
+                                        previousSample->GetEndTime()};
+        sample->mOriginalPresentationWindow = Some(TimeInterval{
+            originalPts, originalPts + fullPacketDuration.Length()});
+      }
+      // Seek back so we're back at the original location -- there's no packet
+      // left anyway.
+      mIterator->Seek(sample->mTime);
+      RefPtr<MediaRawData> dummy = mIterator->GetNext();
+    }
+  }
+
+  if (MOZ_LOG_TEST(GetDemuxerLog(), LogLevel::Verbose)) {
+    bool isAudio = mInfo->GetAsAudioInfo();
+    TimeUnit originalStart = TimeUnit::Invalid();
+    TimeUnit originalEnd = TimeUnit::Invalid();
+    if (sample->mOriginalPresentationWindow) {
+      originalStart = sample->mOriginalPresentationWindow->mStart;
+      originalEnd = sample->mOriginalPresentationWindow->mEnd;
+    }
+    LOG("%s packet demuxed (track id: %d): [%s,%s], duration: %s (original "
+        "time: [%s,%s])",
+        isAudio ? "Audio" : "Video", mInfo->mTrackId,
+        sample->mTime.ToString().get(), sample->GetEndTime().ToString().get(),
+        sample->mDuration.ToString().get(), originalStart.ToString().get(),
+        originalEnd.ToString().get());
+  }
+
   return sample.forget();
 }
 
@@ -479,23 +558,23 @@ RefPtr<MP4TrackDemuxer::SamplesPromise> MP4TrackDemuxer::GetSamples(
 
 void MP4TrackDemuxer::SetNextKeyFrameTime() {
   mNextKeyframeTime.reset();
-  Microseconds frameTime = mIterator->GetNextKeyframeTime();
-  if (frameTime != -1) {
-    mNextKeyframeTime.emplace(media::TimeUnit::FromMicroseconds(frameTime));
+  TimeUnit frameTime = mIterator->GetNextKeyframeTime();
+  if (frameTime.IsValid()) {
+    mNextKeyframeTime.emplace(frameTime);
   }
 }
 
 void MP4TrackDemuxer::Reset() {
   mQueuedSample = nullptr;
-  // TODO, Seek to first frame available, which isn't always 0.
-  mIterator->Seek(0);
+  // TODO: verify this
+  mIterator->Seek(TimeUnit::FromNegativeInfinity());
   SetNextKeyFrameTime();
 }
 
-nsresult MP4TrackDemuxer::GetNextRandomAccessPoint(media::TimeUnit* aTime) {
+nsresult MP4TrackDemuxer::GetNextRandomAccessPoint(TimeUnit* aTime) {
   if (mNextKeyframeTime.isNothing()) {
     // There's no next key frame.
-    *aTime = media::TimeUnit::FromInfinity();
+    *aTime = TimeUnit::FromInfinity();
   } else {
     *aTime = mNextKeyframeTime.value();
   }
@@ -503,8 +582,7 @@ nsresult MP4TrackDemuxer::GetNextRandomAccessPoint(media::TimeUnit* aTime) {
 }
 
 RefPtr<MP4TrackDemuxer::SkipAccessPointPromise>
-MP4TrackDemuxer::SkipToNextRandomAccessPoint(
-    const media::TimeUnit& aTimeThreshold) {
+MP4TrackDemuxer::SkipToNextRandomAccessPoint(const TimeUnit& aTimeThreshold) {
   mQueuedSample = nullptr;
   // Loop until we reach the next keyframe after the threshold.
   uint32_t parsed = 0;
@@ -526,17 +604,32 @@ MP4TrackDemuxer::SkipToNextRandomAccessPoint(
   return SkipAccessPointPromise::CreateAndReject(std::move(failure), __func__);
 }
 
-media::TimeIntervals MP4TrackDemuxer::GetBuffered() {
+TimeIntervals MP4TrackDemuxer::GetBuffered() {
   EnsureUpToDateIndex();
   AutoPinned<MediaResource> resource(mResource);
   MediaByteRangeSet byteRanges;
   nsresult rv = resource->GetCachedRanges(byteRanges);
 
   if (NS_FAILED(rv)) {
-    return media::TimeIntervals();
+    return TimeIntervals();
   }
 
-  return mIndex->ConvertByteRangesToTimeRanges(byteRanges);
+  TimeIntervals timeRanges = mIndex->ConvertByteRangesToTimeRanges(byteRanges);
+  if (AudioInfo* info = mInfo->GetAsAudioInfo(); info) {
+    // Trim as in GetNextSample().
+    TimeUnit totalMediaDurationIncludingTrimming =
+        info->mDuration - info->mMediaTime;
+    auto end = TimeUnit::FromInfinity();
+    if (mType == kAAC && totalMediaDurationIncludingTrimming.IsPositive()) {
+      end = info->mDuration;
+    }
+    if (timeRanges.GetStart().IsNegative() || timeRanges.GetEnd() > end) {
+      TimeInterval trimming(TimeUnit::Zero(timeRanges.GetStart()), end);
+      timeRanges = timeRanges.Intersection(trimming);
+    }
+  }
+
+  return timeRanges;
 }
 
 void MP4TrackDemuxer::NotifyDataArrived() { mNeedReIndex = true; }
