@@ -7,8 +7,10 @@
 #ifndef AudioContext_h_
 #define AudioContext_h_
 
+#include "X11UndefineNone.h"
 #include "AudioParamDescriptorMap.h"
 #include "mozilla/dom/OfflineAudioContextBinding.h"
+#include "mozilla/dom/AudioContextBinding.h"
 #include "MediaBufferDecoder.h"
 #include "mozilla/Attributes.h"
 #include "mozilla/DOMEventTargetHelper.h"
@@ -19,17 +21,12 @@
 #include "mozilla/UniquePtr.h"
 #include "nsCOMPtr.h"
 #include "nsCycleCollectionParticipant.h"
+#include "nsTHashMap.h"
 #include "nsHashKeys.h"
-#include "nsTHashtable.h"
+#include "nsTHashSet.h"
 #include "js/TypeDecls.h"
 #include "nsIMemoryReporter.h"
 
-// X11 has a #define for CurrentTime. Unbelievable :-(.
-// See dom/media/DOMMediaStream.h for more fun!
-#ifdef CurrentTime
-#  undef CurrentTime
-#endif
-
 namespace WebCore {
 class PeriodicWave;
 }  // namespace WebCore
@@ -40,9 +37,9 @@ namespace mozilla {
 
 class DOMMediaStream;
 class ErrorResult;
-class MediaStream;
-class MediaStreamGraph;
-class AudioNodeStream;
+class MediaTrack;
+class MediaTrackGraph;
+class AudioNodeTrack;
 
 namespace dom {
 
@@ -54,6 +51,7 @@ class AudioDestinationNode;
 class AudioListener;
 class AudioNode;
 class BiquadFilterNode;
+class BrowsingContext;
 class ChannelMergerNode;
 class ChannelSplitterNode;
 class ConstantSourceNode;
@@ -81,7 +79,7 @@ class Promise;
 enum class OscillatorType : uint8_t;
 
 // This is addrefed by the OscillatorNodeEngine on the main thread
-// and then used from the MSG thread.
+// and then used from the MTG thread.
 // It can be released either from the graph thread or the main thread.
 class BasicWaveFormCache {
  public:
@@ -97,7 +95,7 @@ class BasicWaveFormCache {
   uint32_t mSampleRate;
 };
 
-/* This runnable allows the MSG to notify the main thread when audio is actually
+/* This runnable allows the MTG to notify the main thread when audio is actually
  * flowing */
 class StateChangeTask final : public Runnable {
  public:
@@ -108,7 +106,7 @@ class StateChangeTask final : public Runnable {
 
   /* This constructor should be used when this event is sent from the audio
    * thread. */
-  StateChangeTask(AudioNodeStream* aStream, void* aPromise,
+  StateChangeTask(AudioNodeTrack* aTrack, void* aPromise,
                   AudioContextState aNewState);
 
   NS_IMETHOD Run() override;
@@ -116,11 +114,13 @@ class StateChangeTask final : public Runnable {
  private:
   RefPtr<AudioContext> mAudioContext;
   void* mPromise;
-  RefPtr<AudioNodeStream> mAudioNodeStream;
+  RefPtr<AudioNodeTrack> mAudioNodeTrack;
   AudioContextState mNewState;
 };
 
-enum class AudioContextOperation { Suspend, Resume, Close };
+enum class AudioContextOperation : uint8_t { Suspend, Resume, Close };
+static const char* const kAudioContextOptionsStrings[] = {"Suspend", "Resume",
+                                                          "Close"};
 // When suspending or resuming an AudioContext, some operations have to notify
 // the main thread, so that the Promise is resolved, the state is modified, and
 // the statechanged event is sent. Some other operations don't go back to the
@@ -140,8 +140,6 @@ class AudioContext final : public DOMEventTargetHelper,
                float aSampleRate = 0.0f);
   ~AudioContext();
 
-  nsresult Init();
-
  public:
   typedef uint64_t AudioContextId;
 
@@ -151,10 +149,11 @@ class AudioContext final : public DOMEventTargetHelper,
 
   nsPIDOMWindowInner* GetParentObject() const { return GetOwner(); }
 
-  virtual void DisconnectFromOwner() override;
-  virtual void BindToOwner(nsIGlobalObject* aNew) override;
+  nsISerialEventTarget* GetMainThread() const;
 
-  void Shutdown();  // idempotent
+  virtual void DisconnectFromOwner() override;
+
+  void OnWindowDestroy();  // idempotent
 
   JSObject* WrapObject(JSContext* aCx,
                        JS::Handle<JSObject*> aGivenProto) override;
@@ -184,14 +183,25 @@ class AudioContext final : public DOMEventTargetHelper,
 
   float SampleRate() const { return mSampleRate; }
 
-  bool ShouldSuspendNewStream() const { return mSuspendCalled; }
-
+  bool ShouldSuspendNewTrack() const {
+    return mTracksAreSuspended || mCloseCalled;
+  }
   double CurrentTime();
 
   AudioListener* Listener();
 
   AudioContextState State() const { return mAudioContextState; }
 
+  double BaseLatency() const {
+    // Gecko does not do any buffering between rendering the audio and sending
+    // it to the audio subsystem.
+    return 0.0;
+  }
+
+  double OutputLatency();
+
+  void GetOutputTimestamp(AudioTimestamp& aTimeStamp);
+
   Worklet* GetAudioWorklet(ErrorResult& aRv);
 
   bool IsRunning() const;
@@ -201,10 +211,10 @@ class AudioContext final : public DOMEventTargetHelper,
   void StartBlockedAudioContextIfAllowed();
 
   // Those three methods return a promise to content, that is resolved when an
-  // (possibly long) operation is completed on the MSG (and possibly other)
+  // (possibly long) operation is completed on the MTG (and possibly other)
   // thread(s). To avoid having to match the calls and asychronous result when
   // the operation is completed, we keep a reference to the promises on the main
-  // thread, and then send the promises pointers down the MSG thread, as a void*
+  // thread, and then send the promises pointers down the MTG thread, as a void*
   // (to make it very clear that the pointer is to merely be treated as an ID).
   // When back on the main thread, we can resolve or reject the promise, by
   // casting it back to a `Promise*` while asserting we're back on the main
@@ -219,10 +229,12 @@ class AudioContext final : public DOMEventTargetHelper,
   // calling from inner window, so we won't need to return promise for caller.
   void SuspendFromChrome();
   void ResumeFromChrome();
+  // Called on completion of offline rendering:
+  void OfflineClose();
 
-  already_AddRefed<AudioBufferSourceNode> CreateBufferSource(ErrorResult& aRv);
+  already_AddRefed<AudioBufferSourceNode> CreateBufferSource();
 
-  already_AddRefed<ConstantSourceNode> CreateConstantSource(ErrorResult& aRv);
+  already_AddRefed<ConstantSourceNode> CreateConstantSource();
 
   already_AddRefed<AudioBuffer> CreateBuffer(uint32_t aNumberOfChannels,
                                              uint32_t aLength,
@@ -277,7 +289,7 @@ class AudioContext final : public DOMEventTargetHelper,
   already_AddRefed<OscillatorNode> CreateOscillator(ErrorResult& aRv);
 
   already_AddRefed<PeriodicWave> CreatePeriodicWave(
-      const Float32Array& aRealData, const Float32Array& aImagData,
+      const Sequence<float>& aRealData, const Sequence<float>& aImagData,
       const PeriodicWaveConstraints& aConstraints, ErrorResult& aRv);
 
   already_AddRefed<Promise> DecodeAudioData(
@@ -293,8 +305,12 @@ class AudioContext final : public DOMEventTargetHelper,
 
   bool IsOffline() const { return mIsOffline; }
 
-  MediaStreamGraph* Graph() const;
-  AudioNodeStream* DestinationStream() const;
+  bool ShouldResistFingerprinting() const {
+    return mShouldResistFingerprinting;
+  }
+
+  MediaTrackGraph* Graph() const;
+  AudioNodeTrack* DestinationTrack() const;
 
   // Nodes register here if they will produce sound even if they have silent
   // or no input connections.  The AudioContext will keep registered nodes
@@ -329,13 +345,14 @@ class AudioContext final : public DOMEventTargetHelper,
                                  AudioParamDescriptorMap* aParamMap);
   const AudioParamDescriptorMap* GetParamMapForWorkletName(
       const nsAString& aName) {
-    return mWorkletParamDescriptors.GetValue(aName);
+    return mWorkletParamDescriptors.Lookup(aName).DataPtrOrNull();
   }
 
   void Dispatch(already_AddRefed<nsIRunnable>&& aRunnable);
 
  private:
   void DisconnectFromWindow();
+  already_AddRefed<Promise> CreatePromise(ErrorResult& aRv);
   void RemoveFromDecodeQueue(WebAudioDecodeJob* aDecodeJob);
   void ShutdownDecoder();
 
@@ -344,9 +361,9 @@ class AudioContext final : public DOMEventTargetHelper,
 
   friend struct ::mozilla::WebAudioDecodeJob;
 
-  nsTArray<MediaStream*> GetAllStreams() const;
+  nsTArray<RefPtr<mozilla::MediaTrack>> GetAllTracks() const;
 
-  void ResumeInternal(AudioContextOperationFlags aFlags);
+  void ResumeInternal();
   void SuspendInternal(void* aPromise, AudioContextOperationFlags aFlags);
   void CloseInternal(void* aPromise, AudioContextOperationFlags aFlags);
 
@@ -367,10 +384,19 @@ class AudioContext final : public DOMEventTargetHelper,
   void MaybeUpdateAutoplayTelemetry();
   void MaybeUpdateAutoplayTelemetryWhenShutdown();
 
+  // If the pref `dom.suspend_inactive.enabled` is enabled, the dom window will
+  // be suspended when the window becomes inactive. In order to keep audio
+  // context running still, we will ask pages to keep awake in that situation.
+  void MaybeUpdatePageAwakeRequest();
+  void MaybeClearPageAwakeRequest();
+  void SetPageAwakeRequest(bool aShouldSet);
+
+  BrowsingContext* GetTopLevelBrowsingContext();
+
  private:
-  // Each AudioContext has an id, that is passed down the MediaStreams that
+  // Each AudioContext has an id, that is passed down the MediaTracks that
   // back the AudioNodes, so we can easily compute the set of all the
-  // MediaStreams for a given context, on the MediasStreamGraph side.
+  // MediaTracks for a given context, on the MediasTrackGraph side.
   const AudioContextId mId;
   // Note that it's important for mSampleRate to be initialized before
   // mDestination, as mDestination's constructor needs to access it!
@@ -390,28 +416,36 @@ class AudioContext final : public DOMEventTargetHelper,
   nsTArray<RefPtr<Promise>> mPendingResumePromises;
   // See RegisterActiveNode.  These will keep the AudioContext alive while it
   // is rendering and the window remains alive.
-  nsTHashtable<nsRefPtrHashKey<AudioNode>> mActiveNodes;
+  nsTHashSet<RefPtr<AudioNode>> mActiveNodes;
   // Raw (non-owning) references to all AudioNodes for this AudioContext.
-  nsTHashtable<nsPtrHashKey<AudioNode>> mAllNodes;
-  nsDataHashtable<nsStringHashKey, AudioParamDescriptorMap>
-      mWorkletParamDescriptors;
+  nsTHashSet<AudioNode*> mAllNodes;
+  nsTHashMap<nsStringHashKey, AudioParamDescriptorMap> mWorkletParamDescriptors;
   // Cache to avoid recomputing basic waveforms all the time.
   RefPtr<BasicWaveFormCache> mBasicWaveFormCache;
   // Number of channels passed in the OfflineAudioContext ctor.
   uint32_t mNumberOfChannels;
-  bool mIsOffline;
+  const RTPCallerType mRTPCallerType;
+  const bool mShouldResistFingerprinting;
+  const bool mIsOffline;
+  // true iff realtime or startRendering() has been called.
   bool mIsStarted;
   bool mIsShutDown;
-  // Close has been called, reject suspend and resume call.
-  bool mCloseCalled;
-  // Suspend has been called with no following resume.
-  bool mSuspendCalled;
   bool mIsDisconnecting;
+  // Close has been called; reject suspend and resume calls.
+  bool mCloseCalled;
+  // Whether the MediaTracks are suspended, due to one or more of
+  // !mWasAllowedToStart, mSuspendedByContent, or mSuspendedByChrome.
+  // false if offline.
+  bool mTracksAreSuspended;
   // This flag stores the value of previous status of `allowed-to-start`.
+  // true if offline.
   bool mWasAllowedToStart;
-
-  // True if this AudioContext has been suspended by the page.
+  // Whether this AudioContext is suspended because the page called suspend().
+  // Unused if offline.
   bool mSuspendedByContent;
+  // Whether this AudioContext is suspended because the Window is suspended.
+  // Unused if offline.
+  bool mSuspendedByChrome;
 
   // These variables are used for telemetry, they're not reflect the actual
   // status of AudioContext, they are based on the "assumption" of enabling
@@ -426,6 +460,11 @@ class AudioContext final : public DOMEventTargetHelper,
   bool mWasEverAllowedToStart;
   bool mWasEverBlockedToStart;
   bool mWouldBeAllowedToStart;
+
+  // Whether we have set the page awake reqeust when non-offline audio context
+  // is running. That will keep the audio context being able to continue running
+  // even if the window is inactive.
+  bool mSetPageAwakeRequest = false;
 };
 
 static const dom::AudioContext::AudioContextId NO_AUDIO_CONTEXT = 0;
