@@ -109,20 +109,9 @@ nsReturnRef<HRTFKernel> HRTFElevation::calculateKernelForAzimuthElevation(
   const int16_t(&impulse_response_data)[ResponseFrameSize] =
       irc_composite_c_r0195[elevationIndex].azimuths[azimuthIndex];
 
-  // When libspeex_resampler is compiled with FIXED_POINT, samples in
-  // speex_resampler_process_float are rounded directly to int16_t, which
-  // only works well if the floats are in the range +/-32767.  On such
-  // platforms it's better to resample before converting to float anyway.
-#ifdef MOZ_SAMPLE_TYPE_S16
-#  define RESAMPLER_PROCESS speex_resampler_process_int
-  const int16_t* response = impulse_response_data;
-  const int16_t* resampledResponse;
-#else
-#  define RESAMPLER_PROCESS speex_resampler_process_float
   float response[ResponseFrameSize];
   ConvertAudioSamples(impulse_response_data, response, ResponseFrameSize);
   float* resampledResponse;
-#endif
 
   // Note that depending on the fftSize returned by the panner, we may be
   // truncating the impulse response.
@@ -140,8 +129,8 @@ nsReturnRef<HRTFKernel> HRTFElevation::calculateKernelForAzimuthElevation(
     // Feed the input buffer into the resampler.
     spx_uint32_t in_len = ResponseFrameSize;
     spx_uint32_t out_len = resampled.Length();
-    RESAMPLER_PROCESS(resampler, 0, response, &in_len, resampled.Elements(),
-                      &out_len);
+    speex_resampler_process_float(resampler, 0, response, &in_len,
+                                  resampled.Elements(), &out_len);
 
     if (out_len < resampled.Length()) {
       // The input should have all been processed.
@@ -150,8 +139,8 @@ nsReturnRef<HRTFKernel> HRTFElevation::calculateKernelForAzimuthElevation(
       spx_uint32_t out_index = out_len;
       in_len = speex_resampler_get_input_latency(resampler);
       out_len = resampled.Length() - out_index;
-      RESAMPLER_PROCESS(resampler, 0, nullptr, &in_len,
-                        resampled.Elements() + out_index, &out_len);
+      speex_resampler_process_float(resampler, 0, nullptr, &in_len,
+                                    resampled.Elements() + out_index, &out_len);
       out_index += out_len;
       // There may be some uninitialized samples remaining for very low
       // sample rates.
@@ -161,18 +150,8 @@ nsReturnRef<HRTFKernel> HRTFElevation::calculateKernelForAzimuthElevation(
     speex_resampler_reset_mem(resampler);
   }
 
-#ifdef MOZ_SAMPLE_TYPE_S16
-  AutoTArray<float, 2 * ResponseFrameSize> floatArray;
-  floatArray.SetLength(resampledResponseLength);
-  float* floatResponse = floatArray.Elements();
-  ConvertAudioSamples(resampledResponse, floatResponse,
-                      resampledResponseLength);
-#else
-  float* floatResponse = resampledResponse;
-#endif
-#undef RESAMPLER_PROCESS
-
-  return HRTFKernel::create(floatResponse, resampledResponseLength, sampleRate);
+  return HRTFKernel::create(resampledResponse, resampledResponseLength,
+                            sampleRate);
 }
 
 // The range of elevations for the IRCAM impulse responses varies depending on
@@ -266,7 +245,7 @@ nsReturnRef<HRTFElevation> HRTFElevation::createBuiltin(int elevation,
   }
 
   return nsReturnRef<HRTFElevation>(
-      new HRTFElevation(&kernelListL, elevation, sampleRate));
+      new HRTFElevation(std::move(kernelListL), elevation, sampleRate));
 }
 
 nsReturnRef<HRTFElevation> HRTFElevation::createByInterpolatingSlices(
@@ -293,8 +272,8 @@ nsReturnRef<HRTFElevation> HRTFElevation::createByInterpolatingSlices(
   double angle = (1.0 - x) * hrtfElevation1->elevationAngle() +
                  x * hrtfElevation2->elevationAngle();
 
-  return nsReturnRef<HRTFElevation>(
-      new HRTFElevation(&kernelListL, static_cast<int>(angle), sampleRate));
+  return nsReturnRef<HRTFElevation>(new HRTFElevation(
+      std::move(kernelListL), static_cast<int>(angle), sampleRate));
 }
 
 void HRTFElevation::getKernelsFromAzimuth(
