@@ -14,6 +14,7 @@
 // Config stuff
 #include "mozilla/dom/RTCConfigurationBinding.h"
 #include "mozilla/Preferences.h"
+#include "mozilla/StaticPrefs_network.h"
 
 // Parsing STUN/TURN URIs
 #include "nsIURI.h"
@@ -42,12 +43,22 @@
 
 #include "nsISocketTransportService.h"
 #include "nsDNSService2.h"
-#include "nsNetUtil.h"  // NS_CheckPortSafety
 
 #include <string>
 #include <vector>
 #include <map>
 
+#ifdef MOZ_GECKO_PROFILER
+#  include "mozilla/ProfilerMarkers.h"
+
+#  define MEDIA_TRANSPORT_HANDLER_PACKET_RECEIVED(aPacket)              \
+    PROFILER_MARKER_TEXT("WebRTC Packet Received", MEDIA_RT, {},        \
+                         ProfilerString8View::WrapNullTerminatedString( \
+                             PacketTypeToString((aPacket).type())));
+#else
+#  define MEDIA_TRANSPORT_HANDLER_PACKET_RECEIVED(aPacket)
+#endif
+
 namespace mozilla {
 
 static const char* mthLogTag = "MediaTransportHandler";
@@ -66,8 +77,9 @@ class MediaTransportHandlerSTS : public MediaTransportHandler,
   void EnterPrivateMode() override;
   void ExitPrivateMode() override;
 
-  nsresult CreateIceCtx(const std::string& aName,
-                        const nsTArray<dom::RTCIceServer>& aIceServers,
+  void CreateIceCtx(const std::string& aName) override;
+
+  nsresult SetIceConfig(const nsTArray<dom::RTCIceServer>& aIceServers,
                         dom::RTCIceTransportPolicy aIcePolicy) override;
 
   // We will probably be able to move the proxy lookup stuff into
@@ -77,7 +89,7 @@ class MediaTransportHandlerSTS : public MediaTransportHandler,
   void EnsureProvisionalTransport(const std::string& aTransportId,
                                   const std::string& aUfrag,
                                   const std::string& aPwd,
-                                  size_t aComponentCount) override;
+                                  int aComponentCount) override;
 
   void SetTargetForDefaultLocalAddressLookup(const std::string& aTargetIp,
                                              uint16_t aTargetPort) override;
@@ -125,12 +137,10 @@ class MediaTransportHandlerSTS : public MediaTransportHandler,
   void Destroy_s();
   void DestroyFinal();
   void Shutdown_s();
-  RefPtr<TransportFlow> CreateTransportFlow(const std::string& aTransportId,
-                                            bool aIsRtcp,
-                                            RefPtr<DtlsIdentity> aDtlsIdentity,
-                                            bool aDtlsClient,
-                                            const DtlsDigestList& aDigests,
-                                            bool aPrivacyRequested);
+  RefPtr<TransportFlow> CreateTransportFlow(
+      const std::string& aTransportId, bool aIsRtcp,
+      const RefPtr<DtlsIdentity>& aDtlsIdentity, bool aDtlsClient,
+      const DtlsDigestList& aDigests, bool aPrivacyRequested);
 
   struct Transport {
     RefPtr<TransportFlow> mFlow;
@@ -146,9 +156,9 @@ class MediaTransportHandlerSTS : public MediaTransportHandler,
   using MediaTransportHandler::OnRtcpStateChange;
   using MediaTransportHandler::OnStateChange;
 
-  void OnGatheringStateChange(NrIceCtx* aIceCtx,
-                              NrIceCtx::GatheringState aState);
-  void OnConnectionStateChange(NrIceCtx* aIceCtx,
+  void OnGatheringStateChange(const std::string& aTransportId,
+                              NrIceMediaStream::GatheringState aState);
+  void OnConnectionStateChange(NrIceMediaStream* aIceStream,
                                NrIceCtx::ConnectionState aState);
   void OnCandidateFound(NrIceMediaStream* aStream,
                         const std::string& aCandidate,
@@ -169,13 +179,16 @@ class MediaTransportHandlerSTS : public MediaTransportHandler,
   RefPtr<NrIceResolver> mDNSResolver;
   std::map<std::string, Transport> mTransports;
   bool mObfuscateHostAddresses = false;
+  bool mTurnDisabled = false;
   uint32_t mMinDtlsVersion = 0;
   uint32_t mMaxDtlsVersion = 0;
+  bool mForceNoHost = false;
+  Maybe<NrIceCtx::NatSimulatorConfig> mNatConfig;
 
   std::set<std::string> mSignaledAddresses;
 
   // Init can only be done on main, but we want this to be usable on any thread
-  typedef MozPromise<bool, std::string, false> InitPromise;
+  using InitPromise = MozPromise<bool, std::string, false>;
   RefPtr<InitPromise> mInitPromise;
 };
 
@@ -185,11 +198,12 @@ already_AddRefed<MediaTransportHandler> MediaTransportHandler::Create(
   RefPtr<MediaTransportHandler> result;
   if (XRE_IsContentProcess() &&
       Preferences::GetBool("media.peerconnection.mtransport_process") &&
-      Preferences::GetBool("network.process.enabled")) {
+      StaticPrefs::network_process_enabled()) {
     result = new MediaTransportHandlerIPC(aCallbackThread);
   } else {
     result = new MediaTransportHandlerSTS(aCallbackThread);
   }
+  result->Initialize();
   return result.forget();
 }
 
@@ -246,7 +260,7 @@ class STSShutdownHandler : public nsISTSShutdownObserver {
   }
 
  private:
-  virtual ~STSShutdownHandler() {}
+  virtual ~STSShutdownHandler() = default;
 
   // Raw ptrs, registered on init, deregistered on destruction, all on main
   std::set<MediaTransportHandlerSTS*> mHandlers;
@@ -276,11 +290,7 @@ static NrIceCtx::Policy toNrIcePolicy(dom::RTCIceTransportPolicy aPolicy) {
     case dom::RTCIceTransportPolicy::Relay:
       return NrIceCtx::ICE_POLICY_RELAY;
     case dom::RTCIceTransportPolicy::All:
-      if (Preferences::GetBool("media.peerconnection.ice.no_host", false)) {
-        return NrIceCtx::ICE_POLICY_NO_HOST;
-      } else {
-        return NrIceCtx::ICE_POLICY_ALL;
-      }
+      return NrIceCtx::ICE_POLICY_ALL;
     default:
       MOZ_CRASH();
   }
@@ -351,14 +361,16 @@ static nsresult addNrIceServer(const nsString& aIceUrl,
     }
 
     rv = net_GetAuthURLParser()->ParseAuthority(
-        path.get(), path.Length(), nullptr, nullptr, nullptr, nullptr, &hostPos,
-        &hostLen, &port);
+        path.get(), static_cast<int>(path.Length()), nullptr, nullptr, nullptr,
+        nullptr, &hostPos, &hostLen, &port);
     NS_ENSURE_SUCCESS(rv, rv);
     if (!hostLen) {
       return NS_ERROR_FAILURE;
     }
-    if (hostPos > 1) /* The username was removed */
+    if (hostPos > 1) {
+      /* The username was removed */
       return NS_ERROR_FAILURE;
+    }
     path.Mid(host, hostPos, hostLen);
     // Strip off brackets around IPv6 literals
     host.Trim("[]");
@@ -475,6 +487,8 @@ static Maybe<NrIceCtx::NatSimulatorConfig> GetNatConfig() {
       "media.peerconnection.nat_simulator.block_tcp", false);
   bool block_udp = Preferences::GetBool(
       "media.peerconnection.nat_simulator.block_udp", false);
+  bool block_tls = Preferences::GetBool(
+      "media.peerconnection.nat_simulator.block_tls", false);
   int error_code_for_drop = Preferences::GetInt(
       "media.peerconnection.nat_simulator.error_code_for_drop", 0);
   nsAutoCString mapping_type;
@@ -490,13 +504,14 @@ static Maybe<NrIceCtx::NatSimulatorConfig> GetNatConfig() {
   (void)Preferences::GetCString(
       "media.peerconnection.nat_simulator.redirect_targets", redirect_targets);
 
-  if (block_udp || block_tcp || !mapping_type.IsEmpty() ||
+  if (block_udp || block_tcp || block_tls || !mapping_type.IsEmpty() ||
       !filtering_type.IsEmpty() || !redirect_address.IsEmpty()) {
     CSFLogDebug(LOGTAG, "NAT filtering type: %s", filtering_type.get());
     CSFLogDebug(LOGTAG, "NAT mapping type: %s", mapping_type.get());
     NrIceCtx::NatSimulatorConfig natConfig;
     natConfig.mBlockUdp = block_udp;
     natConfig.mBlockTcp = block_tcp;
+    natConfig.mBlockTls = block_tls;
     natConfig.mErrorCodeForDrop = error_code_for_drop;
     natConfig.mFilteringType = filtering_type;
     natConfig.mMappingType = mapping_type;
@@ -516,17 +531,7 @@ static Maybe<NrIceCtx::NatSimulatorConfig> GetNatConfig() {
   return Nothing();
 }
 
-nsresult MediaTransportHandlerSTS::CreateIceCtx(
-    const std::string& aName, const nsTArray<dom::RTCIceServer>& aIceServers,
-    dom::RTCIceTransportPolicy aIcePolicy) {
-  // We rely on getting an error when this happens, so do it up front.
-  std::vector<NrIceStunServer> stunServers;
-  std::vector<NrIceTurnServer> turnServers;
-  nsresult rv = ConvertIceServers(aIceServers, &stunServers, &turnServers);
-  if (NS_FAILED(rv)) {
-    return rv;
-  }
-
+void MediaTransportHandlerSTS::CreateIceCtx(const std::string& aName) {
   mInitPromise = InvokeAsync(
       GetMainThreadSerialEventTarget(), __func__,
       [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
@@ -560,7 +565,7 @@ nsresult MediaTransportHandlerSTS::CreateIceCtx(
         }
 
         // Give us a way to globally turn off TURN support
-        bool turnDisabled =
+        mTurnDisabled =
             Preferences::GetBool("media.peerconnection.turn.disable", false);
         // We are reading these here, because when we setup the DTLS transport
         // we are on the wrong thread to read prefs
@@ -568,10 +573,9 @@ nsresult MediaTransportHandlerSTS::CreateIceCtx(
             Preferences::GetUint("media.peerconnection.dtls.version.min");
         mMaxDtlsVersion =
             Preferences::GetUint("media.peerconnection.dtls.version.max");
-
-        NrIceCtx::Config config;
-        config.mPolicy = toNrIcePolicy(aIcePolicy);
-        config.mNatSimulatorConfig = GetNatConfig();
+        mForceNoHost =
+            Preferences::GetBool("media.peerconnection.ice.no_host", false);
+        mNatConfig = GetNatConfig();
 
         MOZ_RELEASE_ASSERT(STSShutdownHandler::Instance());
         STSShutdownHandler::Instance()->Register(this);
@@ -579,38 +583,17 @@ nsresult MediaTransportHandlerSTS::CreateIceCtx(
         return InvokeAsync(
             mStsThread, __func__,
             [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
-              mIceCtx = NrIceCtx::Create(aName, config);
+              mIceCtx = NrIceCtx::Create(aName);
               if (!mIceCtx) {
                 return InitPromise::CreateAndReject("NrIceCtx::Create failed",
                                                     __func__);
               }
 
-              mIceCtx->SignalGatheringStateChange.connect(
-                  this, &MediaTransportHandlerSTS::OnGatheringStateChange);
               mIceCtx->SignalConnectionStateChange.connect(
                   this, &MediaTransportHandlerSTS::OnConnectionStateChange);
 
-              nsresult rv;
-
-              if (NS_FAILED(rv = mIceCtx->SetStunServers(stunServers))) {
-                CSFLogError(LOGTAG, "%s: Failed to set stun servers",
-                            __FUNCTION__);
-                return InitPromise::CreateAndReject(
-                    "Failed to set stun servers", __func__);
-              }
-              if (!turnDisabled) {
-                if (NS_FAILED(rv = mIceCtx->SetTurnServers(turnServers))) {
-                  CSFLogError(LOGTAG, "%s: Failed to set turn servers",
-                              __FUNCTION__);
-                  return InitPromise::CreateAndReject(
-                      "Failed to set turn servers", __func__);
-                }
-              } else if (!turnServers.empty()) {
-                CSFLogError(LOGTAG, "%s: Setting turn servers disabled",
-                            __FUNCTION__);
-              }
-
               mDNSResolver = new NrIceResolver;
+              nsresult rv;
               if (NS_FAILED(rv = mDNSResolver->Init())) {
                 CSFLogError(LOGTAG, "%s: Failed to initialize dns resolver",
                             __FUNCTION__);
@@ -629,6 +612,55 @@ nsresult MediaTransportHandlerSTS::CreateIceCtx(
               return InitPromise::CreateAndResolve(true, __func__);
             });
       });
+}
+
+nsresult MediaTransportHandlerSTS::SetIceConfig(
+    const nsTArray<dom::RTCIceServer>& aIceServers,
+    dom::RTCIceTransportPolicy aIcePolicy) {
+  // We rely on getting an error when this happens, so do it up front.
+  std::vector<NrIceStunServer> stunServers;
+  std::vector<NrIceTurnServer> turnServers;
+  nsresult rv = ConvertIceServers(aIceServers, &stunServers, &turnServers);
+  if (NS_FAILED(rv)) {
+    return rv;
+  }
+
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
+  mInitPromise->Then(
+      mStsThread, __func__,
+      [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
+        if (!mIceCtx) {
+          CSFLogError(LOGTAG, "%s: mIceCtx is null", __FUNCTION__);
+          return;
+        }
+        NrIceCtx::Config config;
+        config.mPolicy = toNrIcePolicy(aIcePolicy);
+        if (config.mPolicy == NrIceCtx::ICE_POLICY_ALL && mForceNoHost) {
+          config.mPolicy = NrIceCtx::ICE_POLICY_NO_HOST;
+        }
+        config.mNatSimulatorConfig = mNatConfig;
+
+        nsresult rv;
+
+        if (NS_FAILED(rv = mIceCtx->SetStunServers(stunServers))) {
+          CSFLogError(LOGTAG, "%s: Failed to set stun servers", __FUNCTION__);
+          return;
+        }
+        if (!mTurnDisabled) {
+          if (NS_FAILED(rv = mIceCtx->SetTurnServers(turnServers))) {
+            CSFLogError(LOGTAG, "%s: Failed to set turn servers", __FUNCTION__);
+            return;
+          }
+        } else if (!turnServers.empty()) {
+          CSFLogError(LOGTAG, "%s: Setting turn servers disabled",
+                      __FUNCTION__);
+        }
+        if (NS_FAILED(rv = mIceCtx->SetIceConfig(config))) {
+          CSFLogError(LOGTAG, "%s: Failed to set config", __FUNCTION__);
+        }
+      });
+
   return NS_OK;
 }
 
@@ -662,8 +694,9 @@ void MediaTransportHandlerSTS::Destroy() {
   CSFLogDebug(LOGTAG, "%s %p", __func__, this);
   // Our "destruction tour" starts on main, because we need to deregister.
   if (!NS_IsMainThread()) {
-    GetMainThreadEventTarget()->Dispatch(NewNonOwningRunnableMethod(
-        __func__, this, &MediaTransportHandlerSTS::Destroy));
+    GetMainThreadSerialEventTarget()->Dispatch(
+        NewNonOwningRunnableMethod("MediaTransportHandlerSTS::Destroy", this,
+                                   &MediaTransportHandlerSTS::Destroy));
     return;
   }
 
@@ -677,8 +710,9 @@ void MediaTransportHandlerSTS::Destroy() {
   // and clean up there. However, by the time _that_ happens, we may have
   // dispatched a signal callback to mCallbackThread, so we have to dispatch
   // the final destruction to mCallbackThread.
-  nsresult rv = mStsThread->Dispatch(NewNonOwningRunnableMethod(
-      __func__, this, &MediaTransportHandlerSTS::Destroy_s));
+  nsresult rv = mStsThread->Dispatch(
+      NewNonOwningRunnableMethod("MediaTransportHandlerSTS::Destroy_s", this,
+                                 &MediaTransportHandlerSTS::Destroy_s));
   if (NS_WARN_IF(NS_FAILED(rv))) {
     CSFLogError(LOGTAG,
                 "Unable to dispatch to STS: why has the XPCOM shutdown handler "
@@ -703,6 +737,8 @@ void MediaTransportHandlerSTS::DestroyFinal() { delete this; }
 
 void MediaTransportHandlerSTS::SetProxyConfig(
     NrSocketProxyConfig&& aProxyConfig) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [this, self = RefPtr<MediaTransportHandlerSTS>(this),
@@ -718,7 +754,9 @@ void MediaTransportHandlerSTS::SetProxyConfig(
 
 void MediaTransportHandlerSTS::EnsureProvisionalTransport(
     const std::string& aTransportId, const std::string& aUfrag,
-    const std::string& aPwd, size_t aComponentCount) {
+    const std::string& aPwd, int aComponentCount) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
@@ -728,9 +766,9 @@ void MediaTransportHandlerSTS::EnsureProvisionalTransport(
 
         RefPtr<NrIceMediaStream> stream(mIceCtx->GetStream(aTransportId));
         if (!stream) {
-          CSFLogDebug(LOGTAG, "%s: Creating ICE media stream=%s components=%u",
+          CSFLogDebug(LOGTAG, "%s: Creating ICE media stream=%s components=%d",
                       mIceCtx->name().c_str(), aTransportId.c_str(),
-                      static_cast<unsigned>(aComponentCount));
+                      aComponentCount);
 
           std::ostringstream os;
           os << mIceCtx->name() << " transport-id=" << aTransportId;
@@ -744,6 +782,8 @@ void MediaTransportHandlerSTS::EnsureProvisionalTransport(
 
           stream->SignalCandidate.connect(
               this, &MediaTransportHandlerSTS::OnCandidateFound);
+          stream->SignalGatheringStateChange.connect(
+              this, &MediaTransportHandlerSTS::OnGatheringStateChange);
         }
 
         // Begins an ICE restart if this stream has a different ufrag/pwd
@@ -762,6 +802,8 @@ void MediaTransportHandlerSTS::ActivateTransport(
     const nsTArray<uint8_t>& aKeyDer, const nsTArray<uint8_t>& aCertDer,
     SSLKEAType aAuthType, bool aDtlsClient, const DtlsDigestList& aDigests,
     bool aPrivacyRequested) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [=, keyDer = aKeyDer.Clone(), certDer = aCertDer.Clone(),
@@ -847,6 +889,8 @@ void MediaTransportHandlerSTS::ActivateTransport(
 
 void MediaTransportHandlerSTS::SetTargetForDefaultLocalAddressLookup(
     const std::string& aTargetIp, uint16_t aTargetPort) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
@@ -862,6 +906,8 @@ void MediaTransportHandlerSTS::SetTargetForDefaultLocalAddressLookup(
 void MediaTransportHandlerSTS::StartIceGathering(
     bool aDefaultRouteOnly, bool aObfuscateHostAddresses,
     const nsTArray<NrIceStunAddr>& aStunAddrs) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [=, stunAddrs = aStunAddrs.Clone(),
@@ -892,6 +938,8 @@ void MediaTransportHandlerSTS::StartIceGathering(
 
 void MediaTransportHandlerSTS::StartIceChecks(
     bool aIsControlling, const std::vector<std::string>& aIceOptions) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
@@ -937,6 +985,8 @@ void TokenizeCandidate(const std::string& aCandidate,
 void MediaTransportHandlerSTS::AddIceCandidate(
     const std::string& aTransportId, const std::string& aCandidate,
     const std::string& aUfrag, const std::string& aObfuscatedAddress) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
@@ -976,6 +1026,8 @@ void MediaTransportHandlerSTS::AddIceCandidate(
 }
 
 void MediaTransportHandlerSTS::UpdateNetworkState(bool aOnline) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
@@ -990,6 +1042,8 @@ void MediaTransportHandlerSTS::UpdateNetworkState(bool aOnline) {
 
 void MediaTransportHandlerSTS::RemoveTransportsExcept(
     const std::set<std::string>& aTransportIds) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
@@ -1030,6 +1084,8 @@ void MediaTransportHandlerSTS::RemoveTransportsExcept(
 
 void MediaTransportHandlerSTS::SendPacket(const std::string& aTransportId,
                                           MediaPacket&& aPacket) {
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
   mInitPromise->Then(
       mStsThread, __func__,
       [this, self = RefPtr<MediaTransportHandlerSTS>(this), aTransportId,
@@ -1125,31 +1181,31 @@ void MediaTransportHandler::OnAlpnNegotiated(const std::string& aAlpn) {
 }
 
 void MediaTransportHandler::OnGatheringStateChange(
-    dom::RTCIceGatheringState aState) {
+    const std::string& aTransportId, dom::RTCIceGathererState aState) {
   if (mCallbackThread && !mCallbackThread->IsOnCurrentThread()) {
     mCallbackThread->Dispatch(
         // This is being called from sigslot, which does not hold a strong ref.
         WrapRunnable(this, &MediaTransportHandler::OnGatheringStateChange,
-                     aState),
+                     aTransportId, aState),
         NS_DISPATCH_NORMAL);
     return;
   }
 
-  SignalGatheringStateChange(aState);
+  SignalGatheringStateChange(aTransportId, aState);
 }
 
 void MediaTransportHandler::OnConnectionStateChange(
-    dom::RTCIceConnectionState aState) {
+    const std::string& aTransportId, dom::RTCIceTransportState aState) {
   if (mCallbackThread && !mCallbackThread->IsOnCurrentThread()) {
     mCallbackThread->Dispatch(
         // This is being called from sigslot, which does not hold a strong ref.
         WrapRunnable(this, &MediaTransportHandler::OnConnectionStateChange,
-                     aState),
+                     aTransportId, aState),
         NS_DISPATCH_NORMAL);
     return;
   }
 
-  SignalConnectionStateChange(aState);
+  SignalConnectionStateChange(aTransportId, aState);
 }
 
 void MediaTransportHandler::OnPacketReceived(const std::string& aTransportId,
@@ -1158,7 +1214,7 @@ void MediaTransportHandler::OnPacketReceived(const std::string& aTransportId,
     mCallbackThread->Dispatch(
         // This is being called from sigslot, which does not hold a strong ref.
         WrapRunnable(this, &MediaTransportHandler::OnPacketReceived,
-                     aTransportId, const_cast<MediaPacket&>(aPacket)),
+                     aTransportId, aPacket.Clone()),
         NS_DISPATCH_NORMAL);
     return;
   }
@@ -1172,7 +1228,7 @@ void MediaTransportHandler::OnEncryptedSending(const std::string& aTransportId,
     mCallbackThread->Dispatch(
         // This is being called from sigslot, which does not hold a strong ref.
         WrapRunnable(this, &MediaTransportHandler::OnEncryptedSending,
-                     aTransportId, const_cast<MediaPacket&>(aPacket)),
+                     aTransportId, aPacket.Clone()),
         NS_DISPATCH_NORMAL);
     return;
   }
@@ -1220,20 +1276,19 @@ void MediaTransportHandler::OnRtcpStateChange(const std::string& aTransportId,
 
 RefPtr<dom::RTCStatsPromise> MediaTransportHandlerSTS::GetIceStats(
     const std::string& aTransportId, DOMHighResTimeStamp aNow) {
-  return mInitPromise->Then(
-      mStsThread, __func__,
-      [=, self = RefPtr<MediaTransportHandlerSTS>(this)]() {
-        UniquePtr<dom::RTCStatsCollection> stats(new dom::RTCStatsCollection);
-        if (mIceCtx) {
-          for (const auto& stream : mIceCtx->GetStreams()) {
-            if (aTransportId.empty() || aTransportId == stream->GetId()) {
-              GetIceStats(*stream, aNow, stats.get());
-            }
-          }
+  MOZ_RELEASE_ASSERT(mInitPromise);
+
+  return mInitPromise->Then(mStsThread, __func__, [=, self = RefPtr(this)]() {
+    UniquePtr<dom::RTCStatsCollection> stats(new dom::RTCStatsCollection);
+    if (mIceCtx) {
+      for (const auto& stream : mIceCtx->GetStreams()) {
+        if (aTransportId.empty() || aTransportId == stream->GetId()) {
+          GetIceStats(*stream, aNow, stats.get());
         }
-        return dom::RTCStatsPromise::CreateAndResolve(std::move(stats),
-                                                      __func__);
-      });
+      }
+    }
+    return dom::RTCStatsPromise::CreateAndResolve(std::move(stats), __func__);
+  });
 }
 
 RefPtr<MediaTransportHandler::IceLogPromise>
@@ -1459,7 +1514,7 @@ RefPtr<TransportFlow> MediaTransportHandlerSTS::GetTransportFlow(
 
 RefPtr<TransportFlow> MediaTransportHandlerSTS::CreateTransportFlow(
     const std::string& aTransportId, bool aIsRtcp,
-    RefPtr<DtlsIdentity> aDtlsIdentity, bool aDtlsClient,
+    const RefPtr<DtlsIdentity>& aDtlsIdentity, bool aDtlsClient,
     const DtlsDigestList& aDigests, bool aPrivacyRequested) {
   nsresult rv;
   RefPtr<TransportFlow> flow = new TransportFlow(aTransportId);
@@ -1528,48 +1583,48 @@ RefPtr<TransportFlow> MediaTransportHandlerSTS::CreateTransportFlow(
   return flow;
 }
 
-static mozilla::dom::RTCIceGatheringState toDomIceGatheringState(
-    NrIceCtx::GatheringState aState) {
+static mozilla::dom::RTCIceGathererState toDomIceGathererState(
+    NrIceMediaStream::GatheringState aState) {
   switch (aState) {
-    case NrIceCtx::ICE_CTX_GATHER_INIT:
-      return dom::RTCIceGatheringState::New;
-    case NrIceCtx::ICE_CTX_GATHER_STARTED:
-      return dom::RTCIceGatheringState::Gathering;
-    case NrIceCtx::ICE_CTX_GATHER_COMPLETE:
-      return dom::RTCIceGatheringState::Complete;
+    case NrIceMediaStream::ICE_STREAM_GATHER_INIT:
+      return dom::RTCIceGathererState::New;
+    case NrIceMediaStream::ICE_STREAM_GATHER_STARTED:
+      return dom::RTCIceGathererState::Gathering;
+    case NrIceMediaStream::ICE_STREAM_GATHER_COMPLETE:
+      return dom::RTCIceGathererState::Complete;
   }
   MOZ_CRASH();
 }
 
 void MediaTransportHandlerSTS::OnGatheringStateChange(
-    NrIceCtx* aIceCtx, NrIceCtx::GatheringState aState) {
-  OnGatheringStateChange(toDomIceGatheringState(aState));
+    const std::string& aTransportId, NrIceMediaStream::GatheringState aState) {
+  OnGatheringStateChange(aTransportId, toDomIceGathererState(aState));
 }
 
-static mozilla::dom::RTCIceConnectionState toDomIceConnectionState(
+static mozilla::dom::RTCIceTransportState toDomIceTransportState(
     NrIceCtx::ConnectionState aState) {
   switch (aState) {
     case NrIceCtx::ICE_CTX_INIT:
-      return dom::RTCIceConnectionState::New;
+      return dom::RTCIceTransportState::New;
     case NrIceCtx::ICE_CTX_CHECKING:
-      return dom::RTCIceConnectionState::Checking;
+      return dom::RTCIceTransportState::Checking;
     case NrIceCtx::ICE_CTX_CONNECTED:
-      return dom::RTCIceConnectionState::Connected;
+      return dom::RTCIceTransportState::Connected;
     case NrIceCtx::ICE_CTX_COMPLETED:
-      return dom::RTCIceConnectionState::Completed;
+      return dom::RTCIceTransportState::Completed;
     case NrIceCtx::ICE_CTX_FAILED:
-      return dom::RTCIceConnectionState::Failed;
+      return dom::RTCIceTransportState::Failed;
     case NrIceCtx::ICE_CTX_DISCONNECTED:
-      return dom::RTCIceConnectionState::Disconnected;
+      return dom::RTCIceTransportState::Disconnected;
     case NrIceCtx::ICE_CTX_CLOSED:
-      return dom::RTCIceConnectionState::Closed;
+      return dom::RTCIceTransportState::Closed;
   }
   MOZ_CRASH();
 }
 
 void MediaTransportHandlerSTS::OnConnectionStateChange(
-    NrIceCtx* aIceCtx, NrIceCtx::ConnectionState aState) {
-  OnConnectionStateChange(toDomIceConnectionState(aState));
+    NrIceMediaStream* aIceStream, NrIceCtx::ConnectionState aState) {
+  OnConnectionStateChange(aIceStream->GetId(), toDomIceTransportState(aState));
 }
 
 // The stuff below here will eventually go into the MediaTransportChild class
@@ -1634,8 +1689,31 @@ void MediaTransportHandlerSTS::OnRtcpStateChange(TransportLayer* aLayer,
   MediaTransportHandler::OnRtcpStateChange(aLayer->flow_id(), aState);
 }
 
+constexpr static const char* PacketTypeToString(MediaPacket::Type type) {
+  switch (type) {
+    case MediaPacket::Type::UNCLASSIFIED:
+      return "UNCLASSIFIED";
+    case MediaPacket::Type::SRTP:
+      return "SRTP";
+    case MediaPacket::Type::SRTCP:
+      return "SRTCP";
+    case MediaPacket::Type::DTLS:
+      return "DTLS";
+    case MediaPacket::Type::RTP:
+      return "RTP";
+    case MediaPacket::Type::RTCP:
+      return "RTCP";
+    case MediaPacket::Type::SCTP:
+      return "SCTP";
+    default:
+      MOZ_ASSERT(false, "unreached");
+      return "";
+  }
+}
+
 void MediaTransportHandlerSTS::PacketReceived(TransportLayer* aLayer,
                                               MediaPacket& aPacket) {
+  MEDIA_TRANSPORT_HANDLER_PACKET_RECEIVED(aPacket);
   OnPacketReceived(aLayer->flow_id(), aPacket);
 }
 
@@ -1645,3 +1723,5 @@ void MediaTransportHandlerSTS::EncryptedPacketSending(TransportLayer* aLayer,
 }
 
 }  // namespace mozilla
+
+#undef MEDIA_TRANSPORT_HANDLER_PACKET_RECEIVED
