@@ -5,11 +5,14 @@
 #ifndef _JSEPCODECDESCRIPTION_H_
 #define _JSEPCODECDESCRIPTION_H_
 
+#include <cmath>
 #include <string>
 #include "sdp/SdpMediaSection.h"
 #include "sdp/SdpHelper.h"
 #include "nsCRT.h"
+#include "nsString.h"
 #include "mozilla/net/DataChannelProtocol.h"
+#include "mozilla/Preferences.h"
 
 namespace mozilla {
 
@@ -19,11 +22,9 @@ namespace mozilla {
 // A single entry in our list of known codecs.
 class JsepCodecDescription {
  public:
-  JsepCodecDescription(mozilla::SdpMediaSection::MediaType type,
-                       const std::string& defaultPt, const std::string& name,
+  JsepCodecDescription(const std::string& defaultPt, const std::string& name,
                        uint32_t clock, uint32_t channels, bool enabled)
-      : mType(type),
-        mDefaultPt(defaultPt),
+      : mDefaultPt(defaultPt),
         mName(name),
         mClock(clock),
         mChannels(channels),
@@ -32,16 +33,31 @@ class JsepCodecDescription {
         mDirection(sdp::kSend) {}
   virtual ~JsepCodecDescription() {}
 
+  virtual SdpMediaSection::MediaType Type() const = 0;
+
   virtual JsepCodecDescription* Clone() const = 0;
 
   bool GetPtAsInt(uint16_t* ptOutparam) const {
     return SdpHelper::GetPtAsInt(mDefaultPt, ptOutparam);
   }
 
+  // The id used for codec stats, to uniquely identify this codec configuration
+  // within a transport.
+  const nsString& StatsId() const {
+    if (!mStatsId) {
+      mStatsId.emplace();
+      mStatsId->AppendPrintf(
+          "_%s_%s/%s_%u_%u_%s", mDefaultPt.c_str(),
+          Type() == SdpMediaSection::kVideo ? "video" : "audio", mName.c_str(),
+          mClock, mChannels, mSdpFmtpLine ? mSdpFmtpLine->c_str() : "nothing");
+    }
+    return *mStatsId;
+  }
+
   virtual bool Matches(const std::string& fmt,
                        const SdpMediaSection& remoteMsection) const {
     // note: fmt here is remote fmt (to go with remoteMsection)
-    if (mType != remoteMsection.GetMediaType()) {
+    if (Type() != remoteMsection.GetMediaType()) {
       return false;
     }
 
@@ -67,16 +83,40 @@ class JsepCodecDescription {
     return true;
   }
 
+  Maybe<std::string> GetMatchingFormat(
+      const SdpMediaSection& remoteMsection) const {
+    for (const auto& fmt : remoteMsection.GetFormats()) {
+      if (Matches(fmt, remoteMsection)) {
+        return Some(fmt);
+      }
+    }
+    return Nothing();
+  }
+
   virtual bool Negotiate(const std::string& pt,
-                         const SdpMediaSection& remoteMsection, bool isOffer) {
-    if (mDirection == sdp::kSend || isOffer) {
+                         const SdpMediaSection& remoteMsection,
+                         bool remoteIsOffer,
+                         Maybe<const SdpMediaSection&> localMsection) {
+    // Configuration might change. Invalidate the stats id.
+    mStatsId = Nothing();
+    if (mDirection == sdp::kSend || remoteIsOffer) {
       mDefaultPt = pt;
     }
+    if (localMsection) {
+      // Offer/answer is concluding. Update the sdpFmtpLine.
+      MOZ_ASSERT(mDirection == sdp::kSend || mDirection == sdp::kRecv);
+      const SdpMediaSection& msection =
+          mDirection == sdp::kSend ? remoteMsection : *localMsection;
+      UpdateSdpFmtpLine(ToMaybeRef(msection.FindFmtp(mDefaultPt)));
+    }
     return true;
   }
 
+  virtual void ApplyConfigToFmtp(
+      UniquePtr<SdpFmtpAttributeList::Parameters>& aFmtp) const = 0;
+
   virtual void AddToMediaSection(SdpMediaSection& msection) const {
-    if (mEnabled && msection.GetMediaType() == mType) {
+    if (mEnabled && msection.GetMediaType() == Type()) {
       if (mDirection == sdp::kRecv) {
         msection.AddCodec(mDefaultPt, mName, mClock, mChannels);
       }
@@ -119,9 +159,21 @@ class JsepCodecDescription {
     return false;
   }
 
-  mozilla::SdpMediaSection::MediaType mType;
+  // TODO Bug 1751671: Take a verbatim fmtp line (std::string or eq.) instead
+  // of fmtp parameters that have to be (re-)serialized.
+  void UpdateSdpFmtpLine(
+      const Maybe<const SdpFmtpAttributeList::Parameters&> aParams) {
+    mSdpFmtpLine = aParams.map([](const auto& aFmtp) {
+      std::stringstream ss;
+      aFmtp.Serialize(ss);
+      return ss.str();
+    });
+  }
+
   std::string mDefaultPt;
   std::string mName;
+  Maybe<std::string> mSdpFmtpLine;
+  mutable Maybe<nsString> mStatsId;
   uint32_t mClock;
   uint32_t mChannels;
   bool mEnabled;
@@ -136,8 +188,7 @@ class JsepAudioCodecDescription : public JsepCodecDescription {
   JsepAudioCodecDescription(const std::string& defaultPt,
                             const std::string& name, uint32_t clock,
                             uint32_t channels, bool enabled = true)
-      : JsepCodecDescription(mozilla::SdpMediaSection::kAudio, defaultPt, name,
-                             clock, channels, enabled),
+      : JsepCodecDescription(defaultPt, name, clock, channels, enabled),
         mMaxPlaybackRate(0),
         mForceMono(false),
         mFECEnabled(false),
@@ -149,8 +200,41 @@ class JsepAudioCodecDescription : public JsepCodecDescription {
         mMaxFrameSizeMs(0),
         mCbrEnabled(false) {}
 
+  static constexpr SdpMediaSection::MediaType type = SdpMediaSection::kAudio;
+
+  SdpMediaSection::MediaType Type() const override { return type; }
+
   JSEP_CODEC_CLONE(JsepAudioCodecDescription)
 
+  static UniquePtr<JsepAudioCodecDescription> CreateDefaultOpus() {
+    // Per jmspeex on IRC:
+    // For 32KHz sampling, 28 is ok, 32 is good, 40 should be really good
+    // quality.  Note that 1-2Kbps will be wasted on a stereo Opus channel
+    // with mono input compared to configuring it for mono.
+    // If we reduce bitrate enough Opus will low-pass us; 16000 will kill a
+    // 9KHz tone.  This should be adaptive when we're at the low-end of video
+    // bandwidth (say <100Kbps), and if we're audio-only, down to 8 or
+    // 12Kbps.
+    return MakeUnique<JsepAudioCodecDescription>("109", "opus", 48000, 2);
+  }
+
+  static UniquePtr<JsepAudioCodecDescription> CreateDefaultG722() {
+    return MakeUnique<JsepAudioCodecDescription>("9", "G722", 8000, 1);
+  }
+
+  static UniquePtr<JsepAudioCodecDescription> CreateDefaultPCMU() {
+    return MakeUnique<JsepAudioCodecDescription>("0", "PCMU", 8000, 1);
+  }
+
+  static UniquePtr<JsepAudioCodecDescription> CreateDefaultPCMA() {
+    return MakeUnique<JsepAudioCodecDescription>("8", "PCMA", 8000, 1);
+  }
+
+  static UniquePtr<JsepAudioCodecDescription> CreateDefaultTelephoneEvent() {
+    return MakeUnique<JsepAudioCodecDescription>("101", "telephone-event", 8000,
+                                                 1);
+  }
+
   SdpFmtpAttributeList::OpusParameters GetOpusParameters(
       const std::string& pt, const SdpMediaSection& msection) const {
     // Will contain defaults if nothing else
@@ -187,23 +271,13 @@ class JsepAudioCodecDescription : public JsepCodecDescription {
     }
 
     if (mName == "opus") {
-      SdpFmtpAttributeList::OpusParameters opusParams(
-          GetOpusParameters(mDefaultPt, msection));
-      if (mMaxPlaybackRate) {
-        opusParams.maxplaybackrate = mMaxPlaybackRate;
-      }
-      opusParams.maxAverageBitrate = mMaxAverageBitrate;
-      if (mChannels == 2 && !mForceMono) {
-        // We prefer to receive stereo, if available.
-        opusParams.stereo = 1;
-      }
-      opusParams.useInBandFec = mFECEnabled ? 1 : 0;
-      opusParams.useDTX = mDTXEnabled;
-      opusParams.frameSizeMs = mFrameSizeMs;
-      opusParams.minFrameSizeMs = mMinFrameSizeMs;
-      opusParams.maxFrameSizeMs = mMaxFrameSizeMs;
-      opusParams.useCbr = mCbrEnabled;
-      msection.SetFmtp(SdpFmtpAttributeList::Fmtp(mDefaultPt, opusParams));
+      UniquePtr<SdpFmtpAttributeList::Parameters> opusParams =
+          MakeUnique<SdpFmtpAttributeList::OpusParameters>(
+              GetOpusParameters(mDefaultPt, msection));
+
+      ApplyConfigToFmtp(opusParams);
+
+      msection.SetFmtp(SdpFmtpAttributeList::Fmtp(mDefaultPt, *opusParams));
     } else if (mName == "telephone-event") {
       // add the default dtmf tones
       SdpFmtpAttributeList::TelephoneEventParameters teParams(
@@ -213,8 +287,10 @@ class JsepAudioCodecDescription : public JsepCodecDescription {
   }
 
   bool Negotiate(const std::string& pt, const SdpMediaSection& remoteMsection,
-                 bool isOffer) override {
-    JsepCodecDescription::Negotiate(pt, remoteMsection, isOffer);
+                 bool remoteIsOffer,
+                 Maybe<const SdpMediaSection&> localMsection) override {
+    JsepCodecDescription::Negotiate(pt, remoteMsection, remoteIsOffer,
+                                    localMsection);
     if (mName == "opus" && mDirection == sdp::kSend) {
       SdpFmtpAttributeList::OpusParameters opusParams(
           GetOpusParameters(mDefaultPt, remoteMsection));
@@ -239,7 +315,7 @@ class JsepAudioCodecDescription : public JsepCodecDescription {
       mMinFrameSizeMs = opusParams.minFrameSizeMs;
       if (remoteMsection.GetAttributeList().HasAttribute(
               SdpAttribute::kMaxptimeAttribute)) {
-        mFrameSizeMs = remoteMsection.GetAttributeList().GetMaxptime();
+        mMaxFrameSizeMs = remoteMsection.GetAttributeList().GetMaxptime();
       } else {
         mMaxFrameSizeMs = opusParams.maxFrameSizeMs;
       }
@@ -249,6 +325,46 @@ class JsepAudioCodecDescription : public JsepCodecDescription {
     return true;
   }
 
+  void ApplyConfigToFmtp(
+      UniquePtr<SdpFmtpAttributeList::Parameters>& aFmtp) const override {
+    if (mName == "opus") {
+      SdpFmtpAttributeList::OpusParameters opusParams;
+      if (aFmtp) {
+        MOZ_RELEASE_ASSERT(aFmtp->codec_type == SdpRtpmapAttributeList::kOpus);
+        opusParams =
+            static_cast<const SdpFmtpAttributeList::OpusParameters&>(*aFmtp);
+        opusParams.useInBandFec = mFECEnabled ? 1 : 0;
+      } else {
+        // If we weren't passed a fmtp to use then show we can do in band FEC
+        // for getCapabilities queries.
+        opusParams.useInBandFec = 1;
+      }
+      if (mMaxPlaybackRate) {
+        opusParams.maxplaybackrate = mMaxPlaybackRate;
+      }
+      opusParams.maxAverageBitrate = mMaxAverageBitrate;
+
+      if (mChannels == 2 &&
+          !Preferences::GetBool("media.peerconnection.sdp.disable_stereo_fmtp",
+                                false) &&
+          !mForceMono) {
+        // We prefer to receive stereo, if available.
+        opusParams.stereo = 1;
+      }
+      opusParams.useDTX = mDTXEnabled;
+      opusParams.frameSizeMs = mFrameSizeMs;
+      opusParams.minFrameSizeMs = mMinFrameSizeMs;
+      opusParams.maxFrameSizeMs = mMaxFrameSizeMs;
+      opusParams.useCbr = mCbrEnabled;
+      aFmtp.reset(opusParams.Clone());
+    } else if (mName == "telephone-event") {
+      if (!aFmtp) {
+        // We only use the default right now
+        aFmtp.reset(new SdpFmtpAttributeList::TelephoneEventParameters);
+      }
+    }
+  };
+
   uint32_t mMaxPlaybackRate;
   bool mForceMono;
   bool mFECEnabled;
@@ -266,8 +382,7 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
   JsepVideoCodecDescription(const std::string& defaultPt,
                             const std::string& name, uint32_t clock,
                             bool enabled = true)
-      : JsepCodecDescription(mozilla::SdpMediaSection::kVideo, defaultPt, name,
-                             clock, 0, enabled),
+      : JsepCodecDescription(defaultPt, name, clock, 0, enabled),
         mTmmbrEnabled(false),
         mRembEnabled(false),
         mFECEnabled(false),
@@ -281,67 +396,107 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
     mCcmFbTypes.push_back(SdpRtcpFbAttributeList::fir);
   }
 
-  virtual void EnableTmmbr() {
-    // EnableTmmbr can be called multiple times due to multiple calls to
-    // PeerConnectionImpl::ConfigureJsepSessionCodecs
-    if (!mTmmbrEnabled) {
-      mTmmbrEnabled = true;
-      mCcmFbTypes.push_back(SdpRtcpFbAttributeList::tmmbr);
+  static constexpr SdpMediaSection::MediaType type = SdpMediaSection::kVideo;
+
+  SdpMediaSection::MediaType Type() const override { return type; }
+
+  static UniquePtr<JsepVideoCodecDescription> CreateDefaultVP8(bool aUseRtx) {
+    auto codec = MakeUnique<JsepVideoCodecDescription>("120", "VP8", 90000);
+    // Defaults for mandatory params
+    codec->mConstraints.maxFs = 12288;  // Enough for 2048x1536
+    codec->mConstraints.maxFps = Some(60);
+    if (aUseRtx) {
+      codec->EnableRtx("124");
     }
+    return codec;
   }
 
-  virtual void EnableRemb() {
-    // EnableRemb can be called multiple times due to multiple calls to
-    // PeerConnectionImpl::ConfigureJsepSessionCodecs
-    if (!mRembEnabled) {
-      mRembEnabled = true;
-      mOtherFbTypes.push_back({"", SdpRtcpFbAttributeList::kRemb, "", ""});
+  static UniquePtr<JsepVideoCodecDescription> CreateDefaultVP9(bool aUseRtx) {
+    auto codec = MakeUnique<JsepVideoCodecDescription>("121", "VP9", 90000);
+    // Defaults for mandatory params
+    codec->mConstraints.maxFs = 12288;  // Enough for 2048x1536
+    codec->mConstraints.maxFps = Some(60);
+    if (aUseRtx) {
+      codec->EnableRtx("125");
     }
+    return codec;
   }
 
-  virtual void EnableFec(std::string redPayloadType,
-                         std::string ulpfecPayloadType) {
-    // Enabling FEC for video works a little differently than enabling
-    // REMB or TMMBR.  Support for FEC is indicated by the presence of
-    // particular codes (red and ulpfec) instead of using rtcpfb
-    // attributes on a given codec.  There is no rtcpfb to push for FEC
-    // as can be seen above when REMB or TMMBR are enabled.
-
-    // Ensure we have valid payload types. This returns zero on failure, which
-    // is a valid payload type.
-    uint16_t redPt, ulpfecPt;
-    if (!SdpHelper::GetPtAsInt(redPayloadType, &redPt) ||
-        !SdpHelper::GetPtAsInt(ulpfecPayloadType, &ulpfecPt)) {
-      return;
+  static UniquePtr<JsepVideoCodecDescription> CreateDefaultH264_0(
+      bool aUseRtx) {
+    auto codec = MakeUnique<JsepVideoCodecDescription>("97", "H264", 90000);
+    codec->mPacketizationMode = 0;
+    // Defaults for mandatory params
+    codec->mProfileLevelId = 0x42E01F;
+    if (aUseRtx) {
+      codec->EnableRtx("98");
     }
-
-    mFECEnabled = true;
-    mREDPayloadType = redPayloadType;
-    mULPFECPayloadType = ulpfecPayloadType;
+    return codec;
   }
 
-  virtual void EnableTransportCC() {
-    if (!mTransportCCEnabled) {
-      mTransportCCEnabled = true;
-      mOtherFbTypes.push_back(
-          {"", SdpRtcpFbAttributeList::kTransportCC, "", ""});
+  static UniquePtr<JsepVideoCodecDescription> CreateDefaultH264_1(
+      bool aUseRtx) {
+    auto codec = MakeUnique<JsepVideoCodecDescription>("126", "H264", 90000);
+    codec->mPacketizationMode = 1;
+    // Defaults for mandatory params
+    codec->mProfileLevelId = 0x42E01F;
+    if (aUseRtx) {
+      codec->EnableRtx("127");
     }
+    return codec;
   }
 
-  void EnableRtx(const std::string& rtxPayloadType) {
-    mRtxEnabled = true;
-    mRtxPayloadType = rtxPayloadType;
+  static UniquePtr<JsepVideoCodecDescription> CreateDefaultH264Baseline_0(
+      bool aUseRtx) {
+    auto codec = MakeUnique<JsepVideoCodecDescription>("103", "H264", 90000);
+    codec->mPacketizationMode = 0;
+    // Defaults for mandatory params
+    codec->mProfileLevelId = 0x42001F;
+    if (aUseRtx) {
+      codec->EnableRtx("104");
+    }
+    return codec;
   }
 
-  void AddParametersToMSection(SdpMediaSection& msection) const override {
-    AddFmtpsToMSection(msection);
-    AddRtcpFbsToMSection(msection);
+  static UniquePtr<JsepVideoCodecDescription> CreateDefaultH264Baseline_1(
+      bool aUseRtx) {
+    auto codec = MakeUnique<JsepVideoCodecDescription>("105", "H264", 90000);
+    codec->mPacketizationMode = 1;
+    // Defaults for mandatory params
+    codec->mProfileLevelId = 0x42001F;
+    if (aUseRtx) {
+      codec->EnableRtx("106");
+    }
+    return codec;
   }
 
-  void AddFmtpsToMSection(SdpMediaSection& msection) const {
+  static UniquePtr<JsepVideoCodecDescription> CreateDefaultUlpFec() {
+    return MakeUnique<JsepVideoCodecDescription>(
+        "123",     // payload type
+        "ulpfec",  // codec name
+        90000      // clock rate (match other video codecs)
+    );
+  }
+
+  static UniquePtr<JsepVideoCodecDescription> CreateDefaultRed() {
+    auto codec = MakeUnique<JsepVideoCodecDescription>(
+        "122",  // payload type
+        "red",  // codec name
+        90000   // clock rate (match other video codecs)
+    );
+    codec->EnableRtx("119");
+    return codec;
+  }
+
+  void ApplyConfigToFmtp(
+      UniquePtr<SdpFmtpAttributeList::Parameters>& aFmtp) const override {
     if (mName == "H264") {
-      SdpFmtpAttributeList::H264Parameters h264Params(
-          GetH264Parameters(mDefaultPt, msection));
+      SdpFmtpAttributeList::H264Parameters h264Params;
+      if (aFmtp) {
+        MOZ_RELEASE_ASSERT(aFmtp->codec_type == SdpRtpmapAttributeList::kH264);
+        h264Params =
+            static_cast<const SdpFmtpAttributeList::H264Parameters&>(*aFmtp);
+      }
 
       if (mDirection == sdp::kSend) {
         if (!h264Params.level_asymmetry_allowed) {
@@ -367,21 +522,111 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
       // Hard-coded, may need to change someday?
       h264Params.level_asymmetry_allowed = true;
 
-      msection.SetFmtp(SdpFmtpAttributeList::Fmtp(mDefaultPt, h264Params));
-    } else if (mName == "red" && !mRedundantEncodings.empty()) {
-      SdpFmtpAttributeList::RedParameters redParams(
-          GetRedParameters(mDefaultPt, msection));
-      redParams.encodings = mRedundantEncodings;
-      msection.SetFmtp(SdpFmtpAttributeList::Fmtp(mDefaultPt, redParams));
+      // Parameters that apply to both the send and recv directions
+      h264Params.packetization_mode = mPacketizationMode;
+      // Hard-coded, may need to change someday?
+      h264Params.level_asymmetry_allowed = true;
+      aFmtp.reset(h264Params.Clone());
+    } else if (mName == "VP8" || mName == "VP9") {
+      SdpRtpmapAttributeList::CodecType type =
+          mName == "VP8" ? SdpRtpmapAttributeList::CodecType::kVP8
+                         : SdpRtpmapAttributeList::CodecType::kVP9;
+      auto vp8Params = SdpFmtpAttributeList::VP8Parameters(type);
+
+      if (aFmtp) {
+        MOZ_RELEASE_ASSERT(aFmtp->codec_type == type);
+        vp8Params =
+            static_cast<const SdpFmtpAttributeList::VP8Parameters&>(*aFmtp);
+      }
+      // VP8 and VP9 share the same SDP parameters thus far
+      vp8Params.max_fs = mConstraints.maxFs;
+      if (mConstraints.maxFps.isSome()) {
+        vp8Params.max_fr =
+            static_cast<unsigned int>(std::round(*mConstraints.maxFps));
+      } else {
+        vp8Params.max_fr = 60;
+      }
+      aFmtp.reset(vp8Params.Clone());
+    }
+  }
+
+  virtual void EnableTmmbr() {
+    // EnableTmmbr can be called multiple times due to multiple calls to
+    // PeerConnectionImpl::ConfigureJsepSessionCodecs
+    if (!mTmmbrEnabled) {
+      mTmmbrEnabled = true;
+      mCcmFbTypes.push_back(SdpRtcpFbAttributeList::tmmbr);
+    }
+  }
+
+  virtual void EnableRemb() {
+    // EnableRemb can be called multiple times due to multiple calls to
+    // PeerConnectionImpl::ConfigureJsepSessionCodecs
+    if (!mRembEnabled) {
+      mRembEnabled = true;
+      mOtherFbTypes.push_back({"", SdpRtcpFbAttributeList::kRemb, "", ""});
+    }
+  }
+
+  virtual void EnableFec(std::string redPayloadType,
+                         std::string ulpfecPayloadType,
+                         std::string redRtxPayloadType) {
+    // Enabling FEC for video works a little differently than enabling
+    // REMB or TMMBR.  Support for FEC is indicated by the presence of
+    // particular codes (red and ulpfec) instead of using rtcpfb
+    // attributes on a given codec.  There is no rtcpfb to push for FEC
+    // as can be seen above when REMB or TMMBR are enabled.
+
+    // Ensure we have valid payload types. This returns zero on failure, which
+    // is a valid payload type.
+    uint16_t redPt, ulpfecPt, redRtxPt;
+    if (!SdpHelper::GetPtAsInt(redPayloadType, &redPt) ||
+        !SdpHelper::GetPtAsInt(ulpfecPayloadType, &ulpfecPt) ||
+        !SdpHelper::GetPtAsInt(redRtxPayloadType, &redRtxPt)) {
+      return;
+    }
+
+    mFECEnabled = true;
+    mREDPayloadType = redPayloadType;
+    mULPFECPayloadType = ulpfecPayloadType;
+    mREDRTXPayloadType = redRtxPayloadType;
+  }
+
+  virtual void EnableTransportCC() {
+    if (!mTransportCCEnabled) {
+      mTransportCCEnabled = true;
+      mOtherFbTypes.push_back(
+          {"", SdpRtcpFbAttributeList::kTransportCC, "", ""});
+    }
+  }
+
+  void EnableRtx(const std::string& rtxPayloadType) {
+    mRtxEnabled = true;
+    mRtxPayloadType = rtxPayloadType;
+  }
+
+  void AddParametersToMSection(SdpMediaSection& msection) const override {
+    AddFmtpsToMSection(msection);
+    AddRtcpFbsToMSection(msection);
+  }
+
+  void AddFmtpsToMSection(SdpMediaSection& msection) const {
+    if (mName == "H264") {
+      UniquePtr<SdpFmtpAttributeList::Parameters> h264Params =
+          MakeUnique<SdpFmtpAttributeList::H264Parameters>(
+              GetH264Parameters(mDefaultPt, msection));
+
+      ApplyConfigToFmtp(h264Params);
+
+      msection.SetFmtp(SdpFmtpAttributeList::Fmtp(mDefaultPt, *h264Params));
     } else if (mName == "VP8" || mName == "VP9") {
       if (mDirection == sdp::kRecv) {
         // VP8 and VP9 share the same SDP parameters thus far
-        SdpFmtpAttributeList::VP8Parameters vp8Params(
-            GetVP8Parameters(mDefaultPt, msection));
-
-        vp8Params.max_fs = mConstraints.maxFs;
-        vp8Params.max_fr = mConstraints.maxFps;
-        msection.SetFmtp(SdpFmtpAttributeList::Fmtp(mDefaultPt, vp8Params));
+        UniquePtr<SdpFmtpAttributeList::Parameters> vp8Params =
+            MakeUnique<SdpFmtpAttributeList::VP8Parameters>(
+                GetVP8Parameters(mDefaultPt, msection));
+        ApplyConfigToFmtp(vp8Params);
+        msection.SetFmtp(SdpFmtpAttributeList::Fmtp(mDefaultPt, *vp8Params));
       }
     }
 
@@ -510,9 +755,13 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
   void NegotiateRtcpFb(const SdpMediaSection& remoteMsection,
                        SdpRtcpFbAttributeList::Type type,
                        std::vector<std::string>* supportedTypes) {
+    Maybe<std::string> remoteFmt = GetMatchingFormat(remoteMsection);
+    if (!remoteFmt) {
+      return;
+    }
     std::vector<std::string> temp;
     for (auto& subType : *supportedTypes) {
-      if (remoteMsection.HasRtcpFb(mDefaultPt, type, subType)) {
+      if (remoteMsection.HasRtcpFb(*remoteFmt, type, subType)) {
         temp.push_back(subType);
       }
     }
@@ -522,9 +771,13 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
   void NegotiateRtcpFb(
       const SdpMediaSection& remoteMsection,
       std::vector<SdpRtcpFbAttributeList::Feedback>* supportedFbs) {
+    Maybe<std::string> remoteFmt = GetMatchingFormat(remoteMsection);
+    if (!remoteFmt) {
+      return;
+    }
     std::vector<SdpRtcpFbAttributeList::Feedback> temp;
     for (auto& fb : *supportedFbs) {
-      if (remoteMsection.HasRtcpFb(mDefaultPt, fb.type, fb.parameter)) {
+      if (remoteMsection.HasRtcpFb(*remoteFmt, fb.type, fb.parameter)) {
         temp.push_back(fb);
       }
     }
@@ -541,8 +794,10 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
 
   virtual bool Negotiate(const std::string& pt,
                          const SdpMediaSection& remoteMsection,
-                         bool isOffer) override {
-    JsepCodecDescription::Negotiate(pt, remoteMsection, isOffer);
+                         bool remoteIsOffer,
+                         Maybe<const SdpMediaSection&> localMsection) override {
+    JsepCodecDescription::Negotiate(pt, remoteMsection, remoteIsOffer,
+                                    localMsection);
     if (mName == "H264") {
       SdpFmtpAttributeList::H264Parameters h264Params(
           GetH264Parameters(mDefaultPt, remoteMsection));
@@ -570,21 +825,21 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
       } else {
         // TODO(bug 1143709): max-recv-level support
       }
-    } else if (mName == "red") {
-      SdpFmtpAttributeList::RedParameters redParams(
-          GetRedParameters(mDefaultPt, remoteMsection));
-      mRedundantEncodings = redParams.encodings;
     } else if (mName == "VP8" || mName == "VP9") {
       if (mDirection == sdp::kSend) {
         SdpFmtpAttributeList::VP8Parameters vp8Params(
             GetVP8Parameters(mDefaultPt, remoteMsection));
 
         mConstraints.maxFs = vp8Params.max_fs;
-        mConstraints.maxFps = vp8Params.max_fr;
+        // Right now, we treat max-fr=0 (or the absence of max-fr) as no limit.
+        // We will eventually want to stop doing this (bug 1762600).
+        if (vp8Params.max_fr) {
+          mConstraints.maxFps = Some(vp8Params.max_fr);
+        }
       }
     }
 
-    if (mRtxEnabled && (mDirection == sdp::kSend || isOffer)) {
+    if (mRtxEnabled && (mDirection == sdp::kSend || remoteIsOffer)) {
       Maybe<std::string> rtxPt = GetRtxPtByApt(mDefaultPt, remoteMsection);
       if (rtxPt.isSome()) {
         EnableRtx(*rtxPt);
@@ -805,20 +1060,6 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
     return false;
   }
 
-  virtual void UpdateRedundantEncodings(
-      const std::vector<UniquePtr<JsepCodecDescription>>& codecs) {
-    for (const auto& codec : codecs) {
-      if (codec->mType == SdpMediaSection::kVideo && codec->mEnabled &&
-          codec->mName != "red") {
-        uint16_t pt;
-        if (!SdpHelper::GetPtAsInt(codec->mDefaultPt, &pt)) {
-          continue;
-        }
-        mRedundantEncodings.push_back(pt);
-      }
-    }
-  }
-
   void EnsureNoDuplicatePayloadTypes(std::set<std::string>& aUsedPts) override {
     JsepCodecDescription::EnsureNoDuplicatePayloadTypes(aUsedPts);
     if (mFECEnabled) {
@@ -842,9 +1083,9 @@ class JsepVideoCodecDescription : public JsepCodecDescription {
   bool mTransportCCEnabled;
   bool mRtxEnabled;
   std::string mREDPayloadType;
+  std::string mREDRTXPayloadType;
   std::string mULPFECPayloadType;
   std::string mRtxPayloadType;
-  std::vector<uint8_t> mRedundantEncodings;
 
   // H264-specific stuff
   uint32_t mProfileLevelId;
@@ -859,20 +1100,31 @@ class JsepApplicationCodecDescription : public JsepCodecDescription {
                                   uint16_t localPort,
                                   uint32_t localMaxMessageSize,
                                   bool enabled = true)
-      : JsepCodecDescription(mozilla::SdpMediaSection::kApplication, "", name,
-                             0, channels, enabled),
+      : JsepCodecDescription("", name, 0, channels, enabled),
         mLocalPort(localPort),
         mLocalMaxMessageSize(localMaxMessageSize),
         mRemotePort(0),
         mRemoteMaxMessageSize(0),
         mRemoteMMSSet(false) {}
 
+  static constexpr SdpMediaSection::MediaType type =
+      SdpMediaSection::kApplication;
+
+  SdpMediaSection::MediaType Type() const override { return type; }
+
   JSEP_CODEC_CLONE(JsepApplicationCodecDescription)
 
+  static UniquePtr<JsepApplicationCodecDescription> CreateDefault() {
+    return MakeUnique<JsepApplicationCodecDescription>(
+        "webrtc-datachannel", WEBRTC_DATACHANNEL_STREAMS_DEFAULT,
+        WEBRTC_DATACHANNEL_PORT_DEFAULT,
+        WEBRTC_DATACHANNEL_MAX_MESSAGE_SIZE_LOCAL);
+  }
+
   // Override, uses sctpport or sctpmap instead of rtpmap
   virtual bool Matches(const std::string& fmt,
                        const SdpMediaSection& remoteMsection) const override {
-    if (mType != remoteMsection.GetMediaType()) {
+    if (type != remoteMsection.GetMediaType()) {
       return false;
     }
 
@@ -896,7 +1148,7 @@ class JsepApplicationCodecDescription : public JsepCodecDescription {
   }
 
   virtual void AddToMediaSection(SdpMediaSection& msection) const override {
-    if (mEnabled && msection.GetMediaType() == mType) {
+    if (mEnabled && msection.GetMediaType() == type) {
       if (mDirection == sdp::kRecv) {
         msection.AddDataChannel(mName, mLocalPort, mChannels,
                                 mLocalMaxMessageSize);
@@ -907,8 +1159,10 @@ class JsepApplicationCodecDescription : public JsepCodecDescription {
   }
 
   bool Negotiate(const std::string& pt, const SdpMediaSection& remoteMsection,
-                 bool isOffer) override {
-    JsepCodecDescription::Negotiate(pt, remoteMsection, isOffer);
+                 bool remoteIsOffer,
+                 Maybe<const SdpMediaSection&> localMsection) override {
+    JsepCodecDescription::Negotiate(pt, remoteMsection, remoteIsOffer,
+                                    localMsection);
 
     uint32_t message_size;
     mRemoteMMSSet = remoteMsection.GetMaxMessageSize(&message_size);
@@ -939,6 +1193,9 @@ class JsepApplicationCodecDescription : public JsepCodecDescription {
   void EnsureNoDuplicatePayloadTypes(std::set<std::string>& aUsedPts) override {
   }
 
+  void ApplyConfigToFmtp(
+      UniquePtr<SdpFmtpAttributeList::Parameters>& aFmtp) const override {};
+
   uint16_t mLocalPort;
   uint32_t mLocalMaxMessageSize;
   uint16_t mRemotePort;
