@@ -5,6 +5,10 @@
 #ifndef GMPVIDEOCODEC_H_
 #define GMPVIDEOCODEC_H_
 
+#include <string>
+
+#include "PerformanceRecorder.h"
+
 namespace mozilla {
 
 class WebrtcVideoDecoder;
@@ -12,8 +16,10 @@ class WebrtcVideoEncoder;
 
 class GmpVideoCodec {
  public:
-  static WebrtcVideoEncoder* CreateEncoder(std::string aPCHandle);
-  static WebrtcVideoDecoder* CreateDecoder(std::string aPCHandle);
+  static WebrtcVideoEncoder* CreateEncoder(
+      const webrtc::SdpVideoFormat& aFormat, std::string aPCHandle);
+  static WebrtcVideoDecoder* CreateDecoder(std::string aPCHandle,
+                                           TrackingId aTrackingId);
 };
 
 }  // namespace mozilla
