@@ -16,6 +16,7 @@
 #include "mozilla/dom/MessagePortBinding.h"
 #include "mozilla/dom/MessagePortChild.h"
 #include "mozilla/dom/PMessagePort.h"
+#include "mozilla/dom/RootedDictionary.h"
 #include "mozilla/dom/ScriptSettings.h"
 #include "mozilla/dom/StructuredCloneTags.h"
 #include "mozilla/dom/WorkerCommon.h"
@@ -25,19 +26,11 @@
 #include "mozilla/dom/RefMessageBodyService.h"
 #include "mozilla/dom/SharedMessageBody.h"
 #include "mozilla/ipc/PBackgroundChild.h"
-#include "mozilla/MessagePortTimelineMarker.h"
 #include "mozilla/ScopeExit.h"
-#include "mozilla/TimelineConsumers.h"
-#include "mozilla/TimelineMarker.h"
 #include "mozilla/Unused.h"
 #include "nsContentUtils.h"
-#include "nsGlobalWindow.h"
 #include "nsPresContext.h"
 
-#include "nsIBFCacheEntry.h"
-#include "mozilla/dom/Document.h"
-#include "nsServiceManagerUtils.h"
-
 #ifdef XP_WIN
 #  undef PostMessage
 #endif
@@ -116,28 +109,9 @@ class PostMessageRunnable final : public CancelableRunnable {
     IgnoredErrorResult rv;
     JS::Rooted<JS::Value> value(cx);
 
-    UniquePtr<AbstractTimelineMarker> start;
-    UniquePtr<AbstractTimelineMarker> end;
-    RefPtr<TimelineConsumers> timelines = TimelineConsumers::Get();
-    bool isTimelineRecording = timelines && !timelines->IsEmpty();
-
-    if (isTimelineRecording) {
-      start = MakeUnique<MessagePortTimelineMarker>(
-          ProfileTimelineMessagePortOperationType::DeserializeData,
-          MarkerTracingType::START);
-    }
-
     mData->Read(cx, &value, mPort->mRefMessageBodyService,
                 SharedMessageBody::ReadMethod::StealRefMessageBody, rv);
 
-    if (isTimelineRecording) {
-      end = MakeUnique<MessagePortTimelineMarker>(
-          ProfileTimelineMessagePortOperationType::DeserializeData,
-          MarkerTracingType::END);
-      timelines->AddMarkerForAllObservedDocShells(start);
-      timelines->AddMarkerForAllObservedDocShells(end);
-    }
-
     if (NS_WARN_IF(rv.Failed())) {
       JS_ClearPendingException(cx);
       mPort->DispatchError();
@@ -215,6 +189,11 @@ MessagePort::MessagePort(nsIGlobalObject* aGlobal, State aState)
 
 MessagePort::~MessagePort() {
   CloseForced();
+  MOZ_ASSERT(!mActor);
+  if (mActor) {
+    mActor->SetPort(nullptr);
+    mActor = nullptr;
+  }
   MOZ_ASSERT(!mWorkerRef);
 }
 
@@ -225,7 +204,8 @@ already_AddRefed<MessagePort> MessagePort::Create(nsIGlobalObject* aGlobal,
                                                   ErrorResult& aRv) {
   MOZ_ASSERT(aGlobal);
 
-  RefPtr<MessagePort> mp = new MessagePort(aGlobal, eStateUnshippedEntangled);
+  RefPtr<MessagePort> mp =
+      new MessagePort(aGlobal, eStateInitializingUnshippedEntangled);
   mp->Initialize(aUUID, aDestinationUUID, 1 /* 0 is an invalid sequence ID */,
                  false /* Neutered */, aRv);
   return mp.forget();
@@ -244,11 +224,16 @@ already_AddRefed<MessagePort> MessagePort::Create(
   return mp.forget();
 }
 
-void MessagePort::UnshippedEntangle(MessagePort* aEntangledPort) {
+void MessagePort::UnshippedEntangle(RefPtr<MessagePort>& aEntangledPort) {
   MOZ_DIAGNOSTIC_ASSERT(aEntangledPort);
   MOZ_DIAGNOSTIC_ASSERT(!mUnshippedEntangledPort);
 
-  mUnshippedEntangledPort = aEntangledPort;
+  if (mState == eStateInitializingUnshippedEntangled) {
+    mUnshippedEntangledPort = aEntangledPort;
+    mState = eStateUnshippedEntangled;
+  } else {
+    MOZ_ASSERT_UNREACHABLE("Should not have been called.");
+  }
 }
 
 void MessagePort::Initialize(const nsID& aUUID, const nsID& aDestinationUUID,
@@ -272,7 +257,7 @@ void MessagePort::Initialize(const nsID& aUUID, const nsID& aDestinationUUID,
       return;
     }
   } else {
-    MOZ_ASSERT(mState == eStateUnshippedEntangled);
+    MOZ_ASSERT(mState == eStateInitializingUnshippedEntangled);
   }
 
   // The port has to keep itself alive until it's entangled.
@@ -287,8 +272,7 @@ void MessagePort::Initialize(const nsID& aUUID, const nsID& aDestinationUUID,
         workerPrivate, "MessagePort", [self]() { self->CloseForced(); });
     if (NS_WARN_IF(!strongWorkerRef)) {
       // The worker is shutting down.
-      mState = eStateDisentangled;
-      UpdateMustKeepAlive();
+      CloseForced();
       aRv.Throw(NS_ERROR_FAILURE);
       return;
     }
@@ -342,28 +326,9 @@ void MessagePort::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
   RefPtr<SharedMessageBody> data = new SharedMessageBody(
       StructuredCloneHolder::TransferringSupported, agentClusterId);
 
-  UniquePtr<AbstractTimelineMarker> start;
-  UniquePtr<AbstractTimelineMarker> end;
-  RefPtr<TimelineConsumers> timelines = TimelineConsumers::Get();
-  bool isTimelineRecording = timelines && !timelines->IsEmpty();
-
-  if (isTimelineRecording) {
-    start = MakeUnique<MessagePortTimelineMarker>(
-        ProfileTimelineMessagePortOperationType::SerializeData,
-        MarkerTracingType::START);
-  }
-
   data->Write(aCx, aMessage, transferable, mIdentifier->uuid(),
               mRefMessageBodyService, aRv);
 
-  if (isTimelineRecording) {
-    end = MakeUnique<MessagePortTimelineMarker>(
-        ProfileTimelineMessagePortOperationType::SerializeData,
-        MarkerTracingType::END);
-    timelines->AddMarkerForAllObservedDocShells(start);
-    timelines->AddMarkerForAllObservedDocShells(end);
-  }
-
   if (NS_WARN_IF(aRv.Failed())) {
     return;
   }
@@ -373,6 +338,12 @@ void MessagePort::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
     return;
   }
 
+  if (mState == eStateInitializingUnshippedEntangled) {
+    MOZ_ASSERT_UNREACHABLE(
+        "Should be eStateUnshippedEntangled or eStateDisentangled by now.");
+    return;
+  }
+
   // If we are unshipped we are connected to the other port on the same thread.
   if (mState == eStateUnshippedEntangled) {
     MOZ_DIAGNOSTIC_ASSERT(mUnshippedEntangledPort);
@@ -411,7 +382,7 @@ void MessagePort::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
 }
 
 void MessagePort::PostMessage(JSContext* aCx, JS::Handle<JS::Value> aMessage,
-                              const PostMessageOptions& aOptions,
+                              const StructuredSerializeOptions& aOptions,
                               ErrorResult& aRv) {
   PostMessage(aCx, aMessage, aOptions.mTransfer, aRv);
 }
@@ -431,6 +402,11 @@ void MessagePort::Dispatch() {
   }
 
   switch (mState) {
+    case eStateInitializingUnshippedEntangled:
+      MOZ_ASSERT_UNREACHABLE(
+          "Should be eStateUnshippedEntangled or eStateDisentangled by now.");
+      break;
+
     case eStateUnshippedEntangled:
       // Everything is fine here. We have messages because the other
       // port populates our queue directly.
@@ -479,14 +455,6 @@ void MessagePort::Dispatch() {
   mMessages.RemoveElementAt(0);
 
   mPostMessageRunnable = new PostMessageRunnable(this, data);
-
-  nsCOMPtr<nsIGlobalObject> global = GetOwnerGlobal();
-  if (NS_IsMainThread() && global) {
-    MOZ_ALWAYS_SUCCEEDS(
-        global->Dispatch(TaskCategory::Other, do_AddRef(mPostMessageRunnable)));
-    return;
-  }
-
   MOZ_ALWAYS_SUCCEEDS(NS_DispatchToCurrentThread(mPostMessageRunnable));
 }
 
@@ -508,6 +476,17 @@ void MessagePort::CloseInternal(bool aSoftly) {
   // now invalid.
   mRefMessageBodyService->ForgetPort(mIdentifier->uuid());
 
+  if (mState == eStateInitializingUnshippedEntangled) {
+    // We can end up here if we failed to create our WorkerRef.  Our Create
+    // method will end up returning an error and MessageChannel will not bother
+    // creating our counterpart port or calling UnshippedEntangle (and we
+    // assert on this).
+    mState = eStateDisentangledForClose;
+
+    UpdateMustKeepAlive();
+    return;
+  }
+
   if (mState == eStateUnshippedEntangled) {
     MOZ_DIAGNOSTIC_ASSERT(mUnshippedEntangledPort);
 
