@@ -9,23 +9,28 @@
 #include "mozilla/ArrayUtils.h"
 #include "mozilla/ErrorResult.h"
 #include "mozilla/ScopeExit.h"
+#include "mozilla/StaticPrefs_svg.h"
 #include "mozilla/SVGObserverUtils.h"
 #include "mozilla/SVGUseFrame.h"
 #include "mozilla/URLExtraData.h"
 #include "mozilla/dom/Document.h"
-#include "mozilla/dom/Element.h"
+#include "mozilla/dom/ReferrerInfo.h"
+#include "mozilla/dom/ShadowIncludingTreeIterator.h"
+#include "mozilla/dom/SVGGraphicsElement.h"
 #include "mozilla/dom/SVGLengthBinding.h"
 #include "mozilla/dom/SVGSVGElement.h"
+#include "mozilla/dom/SVGSwitchElement.h"
+#include "mozilla/dom/SVGSymbolElement.h"
 #include "mozilla/dom/SVGUseElementBinding.h"
 #include "nsGkAtoms.h"
 #include "nsContentUtils.h"
+#include "nsIReferrerInfo.h"
 #include "nsIURI.h"
 #include "SVGGeometryProperty.h"
 
 NS_IMPL_NS_NEW_SVG_ELEMENT(Use)
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 JSObject* SVGUseElement::WrapNode(JSContext* aCx,
                                   JS::Handle<JSObject*> aGivenProto) {
@@ -57,7 +62,6 @@ NS_IMPL_CYCLE_COLLECTION_CLASS(SVGUseElement)
 
 NS_IMPL_CYCLE_COLLECTION_UNLINK_BEGIN_INHERITED(SVGUseElement,
                                                 SVGUseElementBase)
-  nsAutoScriptBlocker scriptBlocker;
   NS_IMPL_CYCLE_COLLECTION_UNLINK(mOriginal)
   tmp->UnlinkSource();
 NS_IMPL_CYCLE_COLLECTION_UNLINK_END
@@ -75,8 +79,11 @@ NS_IMPL_ISUPPORTS_CYCLE_COLLECTION_INHERITED(SVGUseElement, SVGUseElementBase,
 
 SVGUseElement::SVGUseElement(
     already_AddRefed<mozilla::dom::NodeInfo>&& aNodeInfo)
-    : SVGUseElementBase(std::move(aNodeInfo)),
-      mReferencedElementTracker(this) {}
+    : SVGUseElementBase(std::move(aNodeInfo)), mReferencedElementTracker(this) {
+  SetEnabledCallbacks(kCharacterDataChanged | kAttributeChanged |
+                      kContentAppended | kContentInserted | kContentRemoved |
+                      kNodeWillBeDestroyed);
+}
 
 SVGUseElement::~SVGUseElement() {
   UnlinkSource();
@@ -89,12 +96,11 @@ namespace SVGT = SVGGeometryProperty::Tags;
 //----------------------------------------------------------------------
 // nsINode methods
 
-bool SVGUseElement::IsNodeOfType(uint32_t aFlags) const {
-  return !(aFlags & ~eUSE_TARGET);
-}
-
 void SVGUseElement::ProcessAttributeChange(int32_t aNamespaceID,
                                            nsAtom* aAttribute) {
+  if (OwnerDoc()->CloningForSVGUse()) {
+    return;
+  }
   if (aNamespaceID == kNameSpaceID_None) {
     if (aAttribute == nsGkAtoms::width || aAttribute == nsGkAtoms::height) {
       const bool hadValidDimensions = HasValidDimensions();
@@ -116,17 +122,21 @@ void SVGUseElement::ProcessAttributeChange(int32_t aNamespaceID,
     if (auto* frame = GetFrame()) {
       frame->HrefChanged();
     }
-    mOriginal = nullptr;
     UnlinkSource();
     TriggerReclone();
   }
 }
 
-nsresult SVGUseElement::AfterSetAttr(int32_t aNamespaceID, nsAtom* aAttribute,
-                                     const nsAttrValue* aValue,
-                                     const nsAttrValue* aOldValue,
-                                     nsIPrincipal* aSubjectPrincipal,
-                                     bool aNotify) {
+void SVGUseElement::DidAnimateAttribute(int32_t aNameSpaceID,
+                                        nsAtom* aAttribute) {
+  ProcessAttributeChange(aNameSpaceID, aAttribute);
+}
+
+void SVGUseElement::AfterSetAttr(int32_t aNamespaceID, nsAtom* aAttribute,
+                                 const nsAttrValue* aValue,
+                                 const nsAttrValue* aOldValue,
+                                 nsIPrincipal* aSubjectPrincipal,
+                                 bool aNotify) {
   ProcessAttributeChange(aNamespaceID, aAttribute);
   return SVGUseElementBase::AfterSetAttr(aNamespaceID, aAttribute, aValue,
                                          aOldValue, aSubjectPrincipal, aNotify);
@@ -142,8 +152,10 @@ nsresult SVGUseElement::Clone(dom::NodeInfo* aNodeInfo,
   nsresult rv1 = it->Init();
   nsresult rv2 = const_cast<SVGUseElement*>(this)->CopyInnerTo(it);
 
-  // SVGUseElement specific portion - record who we cloned from
-  it->mOriginal = const_cast<SVGUseElement*>(this);
+  if (aNodeInfo->GetDocument()->CloningForSVGUse()) {
+    // SVGUseElement specific portion - record who we cloned from
+    it->mOriginal = const_cast<SVGUseElement*>(this);
+  }
 
   if (NS_SUCCEEDED(rv1) && NS_SUCCEEDED(rv2)) {
     kungFuDeathGrip.swap(*aResult);
@@ -160,8 +172,8 @@ nsresult SVGUseElement::BindToTree(BindContext& aContext, nsINode& aParent) {
   return NS_OK;
 }
 
-void SVGUseElement::UnbindFromTree(bool aNullParent) {
-  SVGUseElementBase::UnbindFromTree(aNullParent);
+void SVGUseElement::UnbindFromTree(UnbindContext& aContext) {
+  SVGUseElementBase::UnbindFromTree(aContext);
   OwnerDoc()->UnscheduleSVGUseElementShadowTreeUpdate(*this);
 }
 
@@ -235,33 +247,76 @@ void SVGUseElement::ContentRemoved(nsIContent* aChild,
   }
 }
 
-void SVGUseElement::NodeWillBeDestroyed(const nsINode* aNode) {
+void SVGUseElement::NodeWillBeDestroyed(nsINode* aNode) {
   nsCOMPtr<nsIMutationObserver> kungFuDeathGrip(this);
   UnlinkSource();
 }
 
 // Returns whether this node could ever be displayed.
 static bool NodeCouldBeRendered(const nsINode& aNode) {
-  if (aNode.IsSVGElement(nsGkAtoms::symbol)) {
-    // Only <symbol> elements in the root of a <svg:use> shadow tree are
-    // displayed.
-    auto* shadowRoot = ShadowRoot::FromNodeOrNull(aNode.GetParentNode());
-    return shadowRoot && shadowRoot->Host()->IsSVGElement(nsGkAtoms::use);
+  if (const auto* symbol = SVGSymbolElement::FromNode(aNode)) {
+    return symbol->CouldBeRendered();
+  }
+  if (const auto* svgGraphics = SVGGraphicsElement::FromNode(aNode)) {
+    if (!svgGraphics->PassesConditionalProcessingTests()) {
+      return false;
+    }
+  }
+  if (auto* svgSwitch =
+          SVGSwitchElement::FromNodeOrNull(aNode.GetParentNode())) {
+    if (&aNode != svgSwitch->GetActiveChild()) {
+      return false;
+    }
   }
-  // TODO: Do we have other cases we can optimize out easily?
   return true;
 }
 
+// <svg:use> can be used (no pun intended) to trivially cause an explosion of
+// clones that could potentially DoS the browser. We have a configurable limit
+// to control this.
+static bool IsTooMuchRecursion(uint32_t aCount) {
+  switch (StaticPrefs::svg_use_element_recursive_clone_limit_enabled()) {
+    case 0:
+      return false;
+    case 1:
+      break;
+    default:
+      if (!XRE_IsParentProcess()) {
+        return false;
+      }
+      break;
+  }
+  return aCount >= StaticPrefs::svg_use_element_recursive_clone_limit();
+}
+
 // Circular loop detection, plus detection of whether this shadow tree is
 // rendered at all.
 auto SVGUseElement::ScanAncestors(const Element& aTarget) const -> ScanResult {
+  uint32_t count = 0;
+  return ScanAncestorsInternal(aTarget, count);
+}
+
+auto SVGUseElement::ScanAncestorsInternal(const Element& aTarget,
+                                          uint32_t& aCount) const
+    -> ScanResult {
   if (&aTarget == this) {
     return ScanResult::CyclicReference;
   }
-  if (mOriginal &&
-      mOriginal->ScanAncestors(aTarget) == ScanResult::CyclicReference) {
-    return ScanResult::CyclicReference;
+  if (mOriginal) {
+    if (IsTooMuchRecursion(++aCount)) {
+      return ScanResult::TooDeep;
+    }
+    auto result = mOriginal->ScanAncestorsInternal(aTarget, aCount);
+    switch (result) {
+      case ScanResult::TooDeep:
+      case ScanResult::CyclicReference:
+        return result;
+      case ScanResult::Ok:
+      case ScanResult::Invisible:
+        break;
+    }
   }
+
   auto result = ScanResult::Ok;
   for (nsINode* parent = GetParentOrShadowHostNode(); parent;
        parent = parent->GetParentOrShadowHostNode()) {
@@ -269,6 +324,9 @@ auto SVGUseElement::ScanAncestors(const Element& aTarget) const -> ScanResult {
       return ScanResult::CyclicReference;
     }
     if (auto* use = SVGUseElement::FromNode(*parent)) {
+      if (IsTooMuchRecursion(++aCount)) {
+        return ScanResult::TooDeep;
+      }
       if (mOriginal && use->mOriginal == mOriginal) {
         return ScanResult::CyclicReference;
       }
@@ -285,6 +343,70 @@ auto SVGUseElement::ScanAncestors(const Element& aTarget) const -> ScanResult {
 
 //----------------------------------------------------------------------
 
+static bool IsForbiddenUseNode(const nsINode& aNode) {
+  if (!aNode.IsElement()) {
+    return false;
+  }
+  const auto* svg = SVGElement::FromNode(aNode);
+  return !svg || !svg->IsSVGGraphicsElement();
+}
+
+static void CollectForbiddenNodes(Element& aRoot,
+                                  nsTArray<RefPtr<nsINode>>& aNodes) {
+  auto iter = dom::ShadowIncludingTreeIterator(aRoot);
+  while (iter) {
+    nsINode* node = *iter;
+    if (IsForbiddenUseNode(*node)) {
+      aNodes.AppendElement(node);
+      iter.SkipChildren();
+      continue;
+    }
+    ++iter;
+  }
+}
+
+// SVG1 restricted <use> trees to SVGGraphicsElements.
+// https://www.w3.org/TR/SVG11/struct.html#UseElement:
+//
+//    Any ‘svg’, ‘symbol’, ‘g’, graphics element or other ‘use’ is potentially a
+//    template object that can be re-used (i.e., "instanced") in the SVG
+//    document via a ‘use’ element. The ‘use’ element references another element
+//    and indicates that the graphical contents of that element is
+//    included/drawn at that given point in the document.
+//
+// SVG2 doesn't have that same restriction.
+// https://www.w3.org/TR/SVG2/struct.html#UseShadowTree:
+//
+//    Previous versions of SVG restricted the contents of the shadow tree to SVG
+//    graphics elements. This specification allows any valid SVG document
+//    subtree to be cloned. Cloning non-graphical content, however, will not
+//    usually have any visible effect.
+//
+// But it's pretty ambiguous as to what the behavior should be for some
+// elements, because <script> is inert, but <iframe> is not, see:
+// https://github.com/w3c/svgwg/issues/876
+//
+// So, fairly confusing, all-in-all.
+static void RemoveForbiddenNodes(Element& aRoot, bool aIsCrossDocument) {
+  switch (StaticPrefs::svg_use_element_graphics_element_restrictions()) {
+    case 0:
+      return;
+    case 1:
+      if (!aIsCrossDocument) {
+        return;
+      }
+      break;
+    default:
+      break;
+  }
+
+  AutoTArray<RefPtr<nsINode>, 10> unsafeNodes;
+  CollectForbiddenNodes(aRoot, unsafeNodes);
+  for (auto& unsafeNode : unsafeNodes) {
+    unsafeNode->Remove();
+  }
+}
+
 void SVGUseElement::UpdateShadowTree() {
   MOZ_ASSERT(IsInComposedDoc());
 
@@ -300,12 +422,12 @@ void SVGUseElement::UpdateShadowTree() {
   }
   MOZ_ASSERT(shadow);
 
-  Element* targetElement = mReferencedElementTracker.get();
+  auto* targetElement =
+      SVGGraphicsElement::FromNodeOrNull(mReferencedElementTracker.get());
   RefPtr<Element> newElement;
 
   auto UpdateShadowTree = mozilla::MakeScopeExit([&]() {
-    nsIContent* firstChild = shadow->GetFirstChild();
-    if (firstChild) {
+    if (nsIContent* firstChild = shadow->GetFirstChild()) {
       MOZ_ASSERT(!firstChild->GetNextSibling());
       shadow->RemoveChildNode(firstChild, /* aNotify = */ true);
     }
@@ -316,7 +438,7 @@ void SVGUseElement::UpdateShadowTree() {
   });
 
   // make sure target is valid type for <use>
-  if (!targetElement || !targetElement->IsNodeOfType(nsINode::eUSE_TARGET)) {
+  if (!targetElement) {
     return;
   }
 
@@ -330,9 +452,10 @@ void SVGUseElement::UpdateShadowTree() {
   }
 
   {
-    nsNodeInfoManager* nodeInfoManager = targetElement->OwnerDoc() == OwnerDoc()
-                                             ? nullptr
-                                             : OwnerDoc()->NodeInfoManager();
+    const bool isCrossDocument = targetElement->OwnerDoc() != OwnerDoc();
+
+    nsNodeInfoManager* nodeInfoManager =
+        isCrossDocument ? OwnerDoc()->NodeInfoManager() : nullptr;
 
     nsCOMPtr<nsINode> newNode =
         targetElement->Clone(true, nodeInfoManager, IgnoreErrors());
@@ -342,6 +465,7 @@ void SVGUseElement::UpdateShadowTree() {
 
     MOZ_ASSERT(newNode->IsElement());
     newElement = newNode.forget().downcast<Element>();
+    RemoveForbiddenNodes(*newElement, isCrossDocument);
   }
 
   if (newElement->IsAnyOfSVGElements(nsGkAtoms::svg, nsGkAtoms::symbol)) {
@@ -438,20 +562,31 @@ void SVGUseElement::LookupHref() {
     return;
   }
 
-  nsCOMPtr<nsIURI> originURI =
-      mOriginal ? mOriginal->GetBaseURI() : GetBaseURI();
-  nsCOMPtr<nsIURI> baseURI =
-      nsContentUtils::IsLocalRefURL(href)
-          ? SVGObserverUtils::GetBaseURLForLocalRef(this, originURI)
-          : originURI;
+  Element* treeToWatch = mOriginal ? mOriginal.get() : this;
+  if (nsContentUtils::IsLocalRefURL(href)) {
+    mReferencedElementTracker.ResetWithLocalRef(*treeToWatch, href);
+    return;
+  }
 
+  nsCOMPtr<nsIURI> baseURI = treeToWatch->GetBaseURI();
   nsCOMPtr<nsIURI> targetURI;
   nsContentUtils::NewURIWithDocumentCharset(getter_AddRefs(targetURI), href,
                                             GetComposedDoc(), baseURI);
-  nsCOMPtr<nsIReferrerInfo> referrerInfo =
-      ReferrerInfo::CreateForSVGResources(OwnerDoc());
+  if (!targetURI) {
+    return;
+  }
 
-  mReferencedElementTracker.ResetToURIFragmentID(this, targetURI, referrerInfo);
+  // Don't allow <use href="data:...">. Using "#ref" inside a data: document is
+  // handled above.
+  if (targetURI->SchemeIs("data") &&
+      !StaticPrefs::svg_use_element_data_url_href_allowed()) {
+    return;
+  }
+
+  nsIReferrerInfo* referrer =
+      OwnerDoc()->ReferrerInfoForInternalCSSAndSVGResources();
+  mReferencedElementTracker.ResetToURIFragmentID(treeToWatch, targetURI,
+                                                 referrer);
 }
 
 void SVGUseElement::TriggerReclone() {
@@ -510,6 +645,10 @@ gfxMatrix SVGUseElement::PrependLocalTransformsTo(
 
 /* virtual */
 bool SVGUseElement::HasValidDimensions() const {
+  if (!OurWidthAndHeightAreUsed()) {
+    return true;
+  }
+
   return (!mLengthAttributes[ATTR_WIDTH].IsExplicitlySet() ||
           mLengthAttributes[ATTR_WIDTH].GetAnimValInSpecifiedUnits() > 0) &&
          (!mLengthAttributes[ATTR_HEIGHT].IsExplicitlySet() ||
@@ -542,17 +681,7 @@ SVGUseFrame* SVGUseElement::GetFrame() const {
 
 NS_IMETHODIMP_(bool)
 SVGUseElement::IsAttributeMapped(const nsAtom* name) const {
-  static const MappedAttributeEntry* const map[] = {sFEFloodMap,
-                                                    sFiltersMap,
-                                                    sFontSpecificationMap,
-                                                    sGradientStopMap,
-                                                    sLightingEffectsMap,
-                                                    sMarkersMap,
-                                                    sTextContentElementsMap,
-                                                    sViewportsMap};
-
   return name == nsGkAtoms::x || name == nsGkAtoms::y ||
-         FindAttributeDependence(name, map) ||
          SVGUseElementBase::IsAttributeMapped(name);
 }
 
@@ -568,5 +697,4 @@ nsCSSPropertyID SVGUseElement::GetCSSPropertyIdForAttrEnum(uint8_t aAttrEnum) {
   }
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
