@@ -14,14 +14,13 @@
 #include "mozilla/dom/Promise.h"
 #include "mozilla/dom/FeaturePolicyUtils.h"
 #include "mozilla/dom/PermissionMessageUtils.h"
-#include "nsGlobalWindow.h"
+#include "nsGlobalWindowInner.h"
 #include "nsThreadUtils.h"
 #include "gfxVR.h"
 #include "VRDisplayClient.h"
 #include "VRManagerChild.h"
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 using namespace gfx;
 
@@ -123,8 +122,8 @@ already_AddRefed<Promise> XRSystem::IsSessionSupported(XRSessionMode aMode,
 }
 
 already_AddRefed<Promise> XRSystem::RequestSession(
-    JSContext* aCx, XRSessionMode aMode, const XRSessionInit& aOptions,
-    CallerType aCallerType, ErrorResult& aRv) {
+    XRSessionMode aMode, const XRSessionInit& aOptions, CallerType aCallerType,
+    ErrorResult& aRv) {
   nsCOMPtr<nsIGlobalObject> global = GetParentObject();
   NS_ENSURE_TRUE(global, nullptr);
 
@@ -167,49 +166,25 @@ already_AddRefed<Promise> XRSystem::RequestSession(
     requiredReferenceSpaceTypes.AppendElement(XRReferenceSpaceType::Local);
   }
 
-  BindingCallContext callCx(aCx, "XRSystem.requestSession");
-
   if (aOptions.mRequiredFeatures.WasPassed()) {
-    const Sequence<JS::Value>& arr = (aOptions.mRequiredFeatures.Value());
-    for (const JS::Value& val : arr) {
-      if (!val.isNull() && !val.isUndefined()) {
-        bool bFound = false;
-        JS::RootedValue v(aCx, val);
-        int index = 0;
-        if (FindEnumStringIndex<false>(
-                callCx, v, XRReferenceSpaceTypeValues::strings,
-                "XRReferenceSpaceType", "Argument 2 of XR.requestSession",
-                &index)) {
-          if (index >= 0) {
-            requiredReferenceSpaceTypes.AppendElement(
-                static_cast<XRReferenceSpaceType>(index));
-            bFound = true;
-          }
-        }
-        if (!bFound) {
-          promise->MaybeRejectWithNotSupportedError(
-              "A required feature for the XRSession is not available.");
-          return promise.forget();
-        }
+    for (const nsString& val : aOptions.mRequiredFeatures.Value()) {
+      Maybe<XRReferenceSpaceType> type =
+          StringToEnum<XRReferenceSpaceType>(val);
+      if (type.isNothing()) {
+        promise->MaybeRejectWithNotSupportedError(
+            "A required feature for the XRSession is not available.");
+        return promise.forget();
       }
+      requiredReferenceSpaceTypes.AppendElement(type.value());
     }
   }
 
   if (aOptions.mOptionalFeatures.WasPassed()) {
-    const Sequence<JS::Value>& arr = (aOptions.mOptionalFeatures.Value());
-    for (const JS::Value& val : arr) {
-      if (!val.isNull() && !val.isUndefined()) {
-        JS::RootedValue v(aCx, val);
-        int index = 0;
-        if (FindEnumStringIndex<false>(
-                callCx, v, XRReferenceSpaceTypeValues::strings,
-                "XRReferenceSpaceType", "Argument 2 of XR.requestSession",
-                &index)) {
-          if (index >= 0) {
-            optionalReferenceSpaceTypes.AppendElement(
-                static_cast<XRReferenceSpaceType>(index));
-          }
-        }
+    for (const nsString& val : aOptions.mOptionalFeatures.Value()) {
+      Maybe<XRReferenceSpaceType> type =
+          StringToEnum<XRReferenceSpaceType>(val);
+      if (type.isSome()) {
+        optionalReferenceSpaceTypes.AppendElement(type.value());
       }
     }
   }
@@ -614,9 +589,6 @@ uint32_t RequestSessionRequest::GetPresentationGroup() const {
 // IsSessionSupportedRequest cycle collection
 NS_IMPL_CYCLE_COLLECTION(IsSessionSupportedRequest, mPromise)
 
-NS_IMPL_CYCLE_COLLECTION_ROOT_NATIVE(IsSessionSupportedRequest, AddRef)
-NS_IMPL_CYCLE_COLLECTION_UNROOT_NATIVE(IsSessionSupportedRequest, Release)
-
 XRSessionMode IsSessionSupportedRequest::GetSessionMode() const {
   return mSessionMode;
 }
@@ -658,7 +630,7 @@ XRRequestSessionPermissionRequest::Cancel() {
 }
 
 NS_IMETHODIMP
-XRRequestSessionPermissionRequest::Allow(JS::HandleValue aChoices) {
+XRRequestSessionPermissionRequest::Allow(JS::Handle<JS::Value> aChoices) {
   nsTArray<PermissionChoice> choices;
   nsresult rv = TranslateChoices(aChoices, mPermissionRequests, choices);
   if (NS_FAILED(rv)) {
@@ -704,8 +676,4 @@ XRRequestSessionPermissionRequest::Create(
 // RequestSessionRequest cycle collection
 NS_IMPL_CYCLE_COLLECTION(RequestSessionRequest, mPromise)
 
-NS_IMPL_CYCLE_COLLECTION_ROOT_NATIVE(RequestSessionRequest, AddRef)
-NS_IMPL_CYCLE_COLLECTION_UNROOT_NATIVE(RequestSessionRequest, Release)
-
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
