@@ -15,7 +15,6 @@
 #include "nsNetUtil.h"
 #include "nsIClassOfService.h"
 #include "nsIInterfaceRequestorUtils.h"
-#include "nsILoadContext.h"
 #include "nsIPrivateBrowsingChannel.h"
 #include "nsComponentManagerUtils.h"
 #include "nsIStorageStream.h"
@@ -26,9 +25,7 @@
 #include "nsICacheInfoChannel.h"
 #include "nsIFileChannel.h"
 #include "nsEscape.h"
-#include "nsUnicharUtils.h"
 #include "nsIStringEnumerator.h"
-#include "nsContentCID.h"
 #include "nsStreamUtils.h"
 
 #include "nsCExternalHandlerService.h"
@@ -38,7 +35,6 @@
 #include "nsIWebProgressListener.h"
 #include "nsIAuthPrompt.h"
 #include "nsIPrompt.h"
-#include "nsIFormControl.h"
 #include "nsIThreadRetargetableRequest.h"
 #include "nsContentUtils.h"
 
@@ -50,8 +46,7 @@
 
 #include "nsIContent.h"
 #include "nsIMIMEInfo.h"
-#include "mozilla/dom/HTMLInputElement.h"
-#include "mozilla/dom/HTMLSharedElement.h"
+#include "mozilla/dom/Document.h"
 #include "mozilla/net/CookieJarSettings.h"
 #include "mozilla/Mutex.h"
 #include "mozilla/Printf.h"
@@ -121,7 +116,7 @@ struct nsWebBrowserPersist::OutputData {
   nsCOMPtr<nsIURI> mFile;
   nsCOMPtr<nsIURI> mOriginalLocation;
   nsCOMPtr<nsIOutputStream> mStream;
-  Mutex mStreamMutex;
+  Mutex mStreamMutex MOZ_UNANNOTATED;
   int64_t mSelfProgress;
   int64_t mSelfProgressMax;
   bool mCalcFileExt;
@@ -412,17 +407,6 @@ NS_IMETHODIMP nsWebBrowserPersist::SetProgressListener(
 }
 
 NS_IMETHODIMP nsWebBrowserPersist::SaveURI(
-    nsIURI* aURI, nsIPrincipal* aPrincipal, uint32_t aCacheKey,
-    nsIReferrerInfo* aReferrerInfo, nsICookieJarSettings* aCookieJarSettings,
-    nsIInputStream* aPostData, const char* aExtraHeaders, nsISupports* aFile,
-    nsContentPolicyType aContentPolicyType, nsILoadContext* aPrivacyContext) {
-  bool isPrivate = aPrivacyContext && aPrivacyContext->UsePrivateBrowsing();
-  return SavePrivacyAwareURI(aURI, aPrincipal, aCacheKey, aReferrerInfo,
-                             aCookieJarSettings, aPostData, aExtraHeaders,
-                             aFile, aContentPolicyType, isPrivate);
-}
-
-NS_IMETHODIMP nsWebBrowserPersist::SavePrivacyAwareURI(
     nsIURI* aURI, nsIPrincipal* aPrincipal, uint32_t aCacheKey,
     nsIReferrerInfo* aReferrerInfo, nsICookieJarSettings* aCookieJarSettings,
     nsIInputStream* aPostData, const char* aExtraHeaders, nsISupports* aFile,
@@ -435,7 +419,7 @@ NS_IMETHODIMP nsWebBrowserPersist::SavePrivacyAwareURI(
   rv = GetValidURIFromObject(aFile, getter_AddRefs(fileAsURI));
   NS_ENSURE_SUCCESS(rv, NS_ERROR_INVALID_ARG);
 
-  // SaveURI doesn't like broken uris.
+  // SaveURIInternal doesn't like broken uris.
   mPersistFlags |= PERSIST_FLAGS_FAIL_ON_BROKEN_LINKS;
   rv = SaveURIInternal(aURI, aPrincipal, aContentPolicy, aCacheKey,
                        aReferrerInfo, aCookieJarSettings, aPostData,
@@ -456,7 +440,7 @@ NS_IMETHODIMP nsWebBrowserPersist::SaveChannel(nsIChannel* aChannel,
   rv = aChannel->GetURI(getter_AddRefs(mURI));
   NS_ENSURE_SUCCESS(rv, rv);
 
-  // SaveURI doesn't like broken uris.
+  // SaveChannelInternal doesn't like broken uris.
   mPersistFlags |= PERSIST_FLAGS_FAIL_ON_BROKEN_LINKS;
   rv = SaveChannelInternal(aChannel, fileAsURI, false);
   return NS_FAILED(rv) ? rv : NS_OK;
@@ -1093,6 +1077,9 @@ nsWebBrowserPersist::OnDataAvailable(nsIRequest* request,
 
 NS_IMETHODIMP nsWebBrowserPersist::CheckListenerChain() { return NS_OK; }
 
+NS_IMETHODIMP
+nsWebBrowserPersist::OnDataFinished(nsresult) { return NS_OK; }
+
 //*****************************************************************************
 // nsWebBrowserPersist::nsIProgressEventSink
 //*****************************************************************************
@@ -1343,10 +1330,23 @@ nsresult nsWebBrowserPersist::SaveURIInternal(
   // current state of the prefs/permissions.
   nsCOMPtr<nsICookieJarSettings> cookieJarSettings = aCookieJarSettings;
   if (!cookieJarSettings) {
+    // Although the variable is called 'triggering principal', it is used as the
+    // loading principal in the download channel, so we treat it as a loading
+    // principal also.
+    bool shouldResistFingerprinting =
+        nsContentUtils::ShouldResistFingerprinting_dangerous(
+            aTriggeringPrincipal,
+            "We are creating a new CookieJar Settings, so none exists "
+            "currently. Although the variable is called 'triggering principal',"
+            "it is used as the loading principal in the download channel, so we"
+            "treat it as a loading principal also.",
+            RFPTarget::IsAlwaysEnabledForPrecompute);
     cookieJarSettings =
         aIsPrivate
-            ? net::CookieJarSettings::Create(net::CookieJarSettings::ePrivate)
-            : net::CookieJarSettings::Create(net::CookieJarSettings::eRegular);
+            ? net::CookieJarSettings::Create(net::CookieJarSettings::ePrivate,
+                                             shouldResistFingerprinting)
+            : net::CookieJarSettings::Create(net::CookieJarSettings::eRegular,
+                                             shouldResistFingerprinting);
   }
 
   // Open a channel to the URI
@@ -1377,6 +1377,9 @@ nsresult nsWebBrowserPersist::SaveURIInternal(
     }
   }
 
+  nsCOMPtr<nsILoadInfo> loadInfo = inputChannel->LoadInfo();
+  loadInfo->SetIsUserTriggeredSave(true);
+
   // Set the referrer, post data and headers if any
   nsCOMPtr<nsIHttpChannel> httpChannel(do_QueryInterface(inputChannel));
   if (httpChannel) {
@@ -1415,7 +1418,7 @@ nsresult nsWebBrowserPersist::SaveURIInternal(
       const char* kWhitespace = "\b\t\r\n ";
       nsAutoCString extraHeaders(aExtraHeaders);
       while (true) {
-        crlf = extraHeaders.Find("\r\n", true);
+        crlf = extraHeaders.Find("\r\n");
         if (crlf == -1) break;
         extraHeaders.Mid(oneHeader, 0, crlf);
         extraHeaders.Cut(0, crlf + 2);
@@ -1454,8 +1457,7 @@ nsresult nsWebBrowserPersist::SaveChannelInternal(nsIChannel* aChannel,
 
   if (fc && !fu) {
     nsCOMPtr<nsIInputStream> fileInputStream, bufferedInputStream;
-    nsresult rv =
-        NS_MaybeOpenChannelUsingOpen(aChannel, getter_AddRefs(fileInputStream));
+    nsresult rv = aChannel->Open(getter_AddRefs(fileInputStream));
     NS_ENSURE_SUCCESS(rv, rv);
     rv = NS_NewBufferedInputStream(getter_AddRefs(bufferedInputStream),
                                    fileInputStream.forget(),
@@ -1473,7 +1475,7 @@ nsresult nsWebBrowserPersist::SaveChannelInternal(nsIChannel* aChannel,
   }
 
   // Read from the input channel
-  nsresult rv = NS_MaybeOpenChannelUsingAsyncOpen(aChannel, this);
+  nsresult rv = aChannel->AsyncOpen(this);
   if (rv == NS_ERROR_NO_CONTENT) {
     // Assume this is a protocol such as mailto: which does not feed out
     // data and just ignore it.
@@ -1705,7 +1707,7 @@ nsWebBrowserPersist::OnWalk::VisitBrowsingContext(
     return NS_ERROR_FAILURE;
   }
 
-  UniquePtr<WebBrowserPersistDocumentParent> actor(
+  RefPtr<WebBrowserPersistDocumentParent> actor(
       new WebBrowserPersistDocumentParent());
 
   nsCOMPtr<nsIWebBrowserPersistDocumentReceiver> receiver =
@@ -1717,7 +1719,7 @@ nsWebBrowserPersist::OnWalk::VisitBrowsingContext(
 
   bool ok =
       context->GetContentParent()->SendPWebBrowserPersistDocumentConstructor(
-          actor.release(), browserParent, context);
+          actor, browserParent, context);
 
   if (NS_WARN_IF(!ok)) {
     // (The actor will be destroyed on constructor failure.)
@@ -1804,7 +1806,7 @@ void nsWebBrowserPersist::FinishSaveDocumentInternal(nsIURI* aFile,
   if (mWalkStack.Length() > 0) {
     mozilla::UniquePtr<WalkData> toWalk = mWalkStack.PopLastElement();
     // Bounce this off the event loop to avoid stack overflow.
-    typedef StoreCopyPassByRRef<decltype(toWalk)> WalkStorage;
+    using WalkStorage = StoreCopyPassByRRef<decltype(toWalk)>;
     auto saveMethod = &nsWebBrowserPersist::SaveDocumentDeferred;
     nsCOMPtr<nsIRunnable> saveLater = NewRunnableMethod<WalkStorage>(
         "nsWebBrowserPersist::FinishSaveDocumentInternal", this, saveMethod,
@@ -2064,11 +2066,11 @@ nsresult nsWebBrowserPersist::CalculateUniqueFilename(
 
       // Resync the URI with the file after the extension has been appended
       return NS_MutateURI(aURI)
-          .Apply(NS_MutatorMethod(&nsIFileURLMutator::SetFile, localFile))
+          .Apply(&nsIFileURLMutator::SetFile, localFile)
           .Finalize(aOutURI);
     }
     return NS_MutateURI(url)
-        .Apply(NS_MutatorMethod(&nsIURLMutator::SetFileName, filename, nullptr))
+        .Apply(&nsIURLMutator::SetFileName, filename, nullptr)
         .Finalize(aOutURI);
   }
 
@@ -2152,79 +2154,26 @@ nsresult nsWebBrowserPersist::CalculateAndAppendFileExt(
     mMIMEService->GetTypeFromURI(uri, contentType);
   }
 
-  // Append the extension onto the file
+  // Validate the filename
   if (!contentType.IsEmpty()) {
-    nsCOMPtr<nsIMIMEInfo> mimeInfo;
-    mMIMEService->GetFromTypeAndExtension(contentType, ""_ns,
-                                          getter_AddRefs(mimeInfo));
+    nsAutoString newFileName;
+    if (NS_SUCCEEDED(mMIMEService->GetValidFileName(
+            aChannel, contentType, aOriginalURIWithExtension,
+            nsIMIMEService::VALIDATE_DEFAULT, newFileName))) {
+      nsCOMPtr<nsIFile> localFile;
+      GetLocalFileFromURI(aURI, getter_AddRefs(localFile));
+      if (localFile) {
+        localFile->SetLeafName(newFileName);
 
-    nsCOMPtr<nsIFile> localFile;
-    GetLocalFileFromURI(aURI, getter_AddRefs(localFile));
-
-    if (mimeInfo) {
-      nsCOMPtr<nsIURL> url(do_QueryInterface(aURI));
-      NS_ENSURE_TRUE(url, NS_ERROR_FAILURE);
-
-      nsAutoCString newFileName;
-      url->GetFileName(newFileName);
-
-      // Test if the current extension is current for the mime type
-      bool hasExtension = false;
-      int32_t ext = newFileName.RFind(".");
-      if (ext != -1) {
-        mimeInfo->ExtensionExists(Substring(newFileName, ext + 1),
-                                  &hasExtension);
-      }
-
-      // Append the mime file extension
-      nsAutoCString fileExt;
-      if (!hasExtension) {
-        // Test if previous extension is acceptable
-        nsCOMPtr<nsIURL> oldurl(do_QueryInterface(aOriginalURIWithExtension));
-        NS_ENSURE_TRUE(oldurl, NS_ERROR_FAILURE);
-        oldurl->GetFileExtension(fileExt);
-        bool useOldExt = false;
-        if (!fileExt.IsEmpty()) {
-          mimeInfo->ExtensionExists(fileExt, &useOldExt);
-        }
-
-        // If the url doesn't have an extension, or we don't know the extension,
-        // try to use the primary extension for the type. If we don't know the
-        // primary extension for the type, just continue with the url extension.
-        if (!useOldExt) {
-          nsAutoCString primaryExt;
-          mimeInfo->GetPrimaryExtension(primaryExt);
-          if (!primaryExt.IsEmpty()) {
-            fileExt = primaryExt;
-          }
-        }
-
-        if (!fileExt.IsEmpty()) {
-          uint32_t newLength = newFileName.Length() + fileExt.Length() + 1;
-          if (newLength > kDefaultMaxFilenameLength) {
-            if (fileExt.Length() > kDefaultMaxFilenameLength / 2)
-              fileExt.Truncate(kDefaultMaxFilenameLength / 2);
-
-            uint32_t diff = kDefaultMaxFilenameLength - 1 - fileExt.Length();
-            if (newFileName.Length() > diff) newFileName.Truncate(diff);
-          }
-          newFileName.Append('.');
-          newFileName.Append(fileExt);
-        }
-
-        if (localFile) {
-          localFile->SetLeafName(NS_ConvertUTF8toUTF16(newFileName));
-
-          // Resync the URI with the file after the extension has been appended
-          return NS_MutateURI(url)
-              .Apply(NS_MutatorMethod(&nsIFileURLMutator::SetFile, localFile))
-              .Finalize(aOutURI);
-        }
-        return NS_MutateURI(url)
-            .Apply(NS_MutatorMethod(&nsIURLMutator::SetFileName, newFileName,
-                                    nullptr))
+        // Resync the URI with the file after the extension has been appended
+        return NS_MutateURI(aURI)
+            .Apply(&nsIFileURLMutator::SetFile, localFile)
             .Finalize(aOutURI);
       }
+      return NS_MutateURI(aURI)
+          .Apply(&nsIURLMutator::SetFileName,
+                 NS_ConvertUTF16toUTF8(newFileName), nullptr)
+          .Finalize(aOutURI);
     }
   }
 
