@@ -17,7 +17,6 @@
 #include "jsfriendapi.h"
 
 #include "nsCOMPtr.h"
-#include "nsHTMLStyleSheet.h"
 #include "nsIContentSink.h"
 #include "mozilla/dom/Document.h"
 #include "nsIFormControl.h"
@@ -28,7 +27,6 @@
 #include "nsParserBase.h"
 #include "nsViewManager.h"
 #include "nsIScriptSecurityManager.h"
-#include "nsLayoutCID.h"
 #include "nsNetUtil.h"
 #include "nsString.h"
 #include "nsReadableUtils.h"
@@ -169,22 +167,9 @@ NS_IMPL_CYCLE_COLLECTING_RELEASE(XULContentSinkImpl)
 //----------------------------------------------------------------------
 // nsIContentSink interface
 
-NS_IMETHODIMP
-XULContentSinkImpl::WillBuildModel(nsDTDMode aDTDMode) {
-#if FIXME
-  if (!mParentContentSink) {
-    // If we're _not_ an overlay, then notify the document that
-    // the load is beginning.
-    mDocument->BeginLoad();
-  }
-#endif
-
-  return NS_OK;
-}
-
 NS_IMETHODIMP
 XULContentSinkImpl::DidBuildModel(bool aTerminated) {
-  nsCOMPtr<Document> doc = do_QueryReferent(mDocument);
+  nsCOMPtr<Document> doc(mDocument);
   if (doc) {
     mPrototype->NotifyLoadDone();
     mDocument = nullptr;
@@ -202,10 +187,8 @@ XULContentSinkImpl::WillInterrupt(void) {
   return NS_OK;
 }
 
-NS_IMETHODIMP
-XULContentSinkImpl::WillResume(void) {
+void XULContentSinkImpl::WillResume() {
   // XXX Notify the docshell, if necessary
-  return NS_OK;
 }
 
 NS_IMETHODIMP
@@ -216,16 +199,13 @@ XULContentSinkImpl::SetParser(nsParserBase* aParser) {
 
 void XULContentSinkImpl::SetDocumentCharset(
     NotNull<const Encoding*> aEncoding) {
-  nsCOMPtr<Document> doc = do_QueryReferent(mDocument);
+  nsCOMPtr<Document> doc(mDocument);
   if (doc) {
     doc->SetDocumentCharacterSet(aEncoding);
   }
 }
 
-nsISupports* XULContentSinkImpl::GetTarget() {
-  nsCOMPtr<Document> doc = do_QueryReferent(mDocument);
-  return ToSupports(doc);
-}
+nsISupports* XULContentSinkImpl::GetTarget() { return ToSupports(mDocument); }
 
 //----------------------------------------------------------------------
 
@@ -234,7 +214,7 @@ nsresult XULContentSinkImpl::Init(Document* aDocument,
   MOZ_ASSERT(aDocument != nullptr, "null ptr");
   if (!aDocument) return NS_ERROR_NULL_POINTER;
 
-  mDocument = do_GetWeakReference(aDocument);
+  mDocument = aDocument;
   mPrototype = aPrototype;
 
   mDocumentURL = mPrototype->GetURI();
@@ -427,13 +407,13 @@ XULContentSinkImpl::HandleEndElement(const char16_t* aName) {
           static_cast<nsXULPrototypeScript*>(node.get());
 
       // If given a src= attribute, we must ignore script tag content.
-      if (!script->mSrcURI && !script->HasScriptObject()) {
-        nsCOMPtr<Document> doc = do_QueryReferent(mDocument);
+      if (!script->mSrcURI && !script->HasStencil()) {
+        nsCOMPtr<Document> doc(mDocument);
 
         script->mOutOfLine = false;
         if (doc) {
-          script->Compile(mText, mTextLength, JS::SourceOwnership::Borrowed,
-                          mDocumentURL, script->mLineNo, doc);
+          script->Compile(mText, mTextLength, mDocumentURL, script->mLineNo,
+                          doc);
         }
       }
 
@@ -564,7 +544,7 @@ XULContentSinkImpl::ReportError(const char16_t* aErrorText,
 
   // return leaving the document empty if we're asked to not add a <parsererror>
   // root node
-  nsCOMPtr<Document> idoc = do_QueryReferent(mDocument);
+  nsCOMPtr<Document> idoc(mDocument);
   if (idoc && idoc->SuppressParserErrorElement()) {
     return NS_OK;
   };
@@ -741,7 +721,7 @@ nsresult XULContentSinkImpl::OpenScript(const char16_t** aAttributes,
     return NS_OK;
   }
 
-  nsCOMPtr<Document> doc(do_QueryReferent(mDocument));
+  nsCOMPtr<Document> doc(mDocument);
   nsCOMPtr<nsIScriptGlobalObject> globalObject;
   if (doc) globalObject = do_QueryInterface(doc->GetWindow());
   RefPtr<nsXULPrototypeScript> script = new nsXULPrototypeScript(aLineNumber);
@@ -757,14 +737,10 @@ nsresult XULContentSinkImpl::OpenScript(const char16_t** aAttributes,
     if (NS_SUCCEEDED(rv)) {
       if (!mSecMan)
         mSecMan = do_GetService(NS_SCRIPTSECURITYMANAGER_CONTRACTID, &rv);
-      if (NS_SUCCEEDED(rv)) {
-        nsCOMPtr<Document> doc = do_QueryReferent(mDocument, &rv);
-
-        if (NS_SUCCEEDED(rv)) {
-          rv = mSecMan->CheckLoadURIWithPrincipal(
-              doc->NodePrincipal(), script->mSrcURI,
-              nsIScriptSecurityManager::ALLOW_CHROME, doc->InnerWindowID());
-        }
+      if (NS_SUCCEEDED(rv) && doc) {
+        rv = mSecMan->CheckLoadURIWithPrincipal(
+            doc->NodePrincipal(), script->mSrcURI,
+            nsIScriptSecurityManager::ALLOW_CHROME, doc->InnerWindowID());
       }
     }
 
