@@ -19,6 +19,7 @@
 #include "mozilla/layers/APZPublicUtils.h"
 #include "mozilla/layers/KeyboardScrollAction.h"
 #include "mozilla/TextEvents.h"
+#include "mozilla/ipc/IPCForwards.h"
 
 template <class E>
 struct already_AddRefed;
@@ -76,14 +77,8 @@ class InputData {
   // Warning, this class is serialized and sent over IPC. Any change to its
   // fields must be reflected in its ParamTraits<>, in nsGUIEventIPC.h
   InputType mInputType;
-  // Time in milliseconds that this data is relevant to. This only really
-  // matters when this data is used as an event. We use uint32_t instead of
-  // TimeStamp because it is easier to convert from WidgetInputEvent. The time
-  // is platform-specific but it in the case of B2G and Fennec it is since
-  // startup.
-  uint32_t mTime;
-  // Set in parallel to mTime until we determine it is safe to drop
-  // platform-specific event times (see bug 77992).
+  // Time that this data is relevant to. This only really matters when this data
+  // is used as an event.
   TimeStamp mTimeStamp;
   // The sequence number of the last potentially focus changing event handled
   // by APZ. This is used to track when that event has been processed by
@@ -108,8 +103,7 @@ class InputData {
   explicit InputData(InputType aInputType);
 
  protected:
-  InputData(InputType aInputType, uint32_t aTime, TimeStamp aTimeStamp,
-            Modifiers aModifiers);
+  InputData(InputType aInputType, TimeStamp aTimeStamp, Modifiers aModifiers);
 };
 
 /**
@@ -196,9 +190,9 @@ class SingleTouchData {
   // How hard the screen is being pressed.
   float mForce;
 
-  uint32_t mTiltX = 0;
-  uint32_t mTiltY = 0;
-  uint32_t mTwist = 0;
+  int32_t mTiltX = 0;
+  int32_t mTiltY = 0;
+  int32_t mTwist = 0;
 };
 
 /**
@@ -227,7 +221,7 @@ class MultiTouchInput : public InputData {
                   Modifiers aModifiers);
   MultiTouchInput();
   MultiTouchInput(MultiTouchInput&&) = default;
-  MultiTouchInput(const MultiTouchInput& aOther);
+  MultiTouchInput(const MultiTouchInput&) = default;
   explicit MultiTouchInput(const WidgetTouchEvent& aTouchEvent);
 
   MultiTouchInput& operator=(MultiTouchInput&&) = default;
@@ -264,6 +258,7 @@ class MouseInput : public InputData {
  protected:
   friend mozilla::layers::APZInputBridgeChild;
   friend mozilla::layers::PAPZInputBridgeParent;
+  ALLOW_DEPRECATED_READPARAM
 
   MouseInput();
 
@@ -279,7 +274,9 @@ class MouseInput : public InputData {
       MOUSE_DRAG_END,
       MOUSE_WIDGET_ENTER,
       MOUSE_WIDGET_EXIT,
-      MOUSE_HITTEST
+      MOUSE_HITTEST,
+      MOUSE_EXPLORE_BY_TOUCH,
+      MOUSE_CONTEXTMENU
   ));
 
   MOZ_DEFINE_ENUM_AT_CLASS_SCOPE(
@@ -292,8 +289,8 @@ class MouseInput : public InputData {
   // clang-format on
 
   MouseInput(MouseType aType, ButtonType aButtonType, uint16_t aInputSource,
-             int16_t aButtons, const ScreenPoint& aPoint, uint32_t aTime,
-             TimeStamp aTimeStamp, Modifiers aModifiers);
+             int16_t aButtons, const ScreenPoint& aPoint, TimeStamp aTimeStamp,
+             Modifiers aModifiers);
   explicit MouseInput(const WidgetMouseEventBase& aMouseEvent);
 
   bool IsLeftButton() const;
@@ -322,9 +319,12 @@ class MouseInput : public InputData {
  * These events are currently only used for scrolling on desktop.
  */
 class PanGestureInput : public InputData {
+  friend struct IPC::ParamTraits<PanGestureInput>;
+
  protected:
   friend mozilla::layers::APZInputBridgeChild;
   friend mozilla::layers::PAPZInputBridgeParent;
+  ALLOW_DEPRECATED_READPARAM
 
   PanGestureInput();
 
@@ -393,10 +393,19 @@ class PanGestureInput : public InputData {
   ));
   // clang-format on
 
-  PanGestureInput(PanGestureType aType, uint32_t aTime, TimeStamp aTimeStamp,
+  PanGestureInput(PanGestureType aType, TimeStamp aTimeStamp,
                   const ScreenPoint& aPanStartPoint,
                   const ScreenPoint& aPanDisplacement, Modifiers aModifiers);
 
+  enum class IsEligibleForSwipe : bool { No, Yes };
+  PanGestureInput(PanGestureType aType, TimeStamp aTimeStamp,
+                  const ScreenPoint& aPanStartPoint,
+                  const ScreenPoint& aPanDisplacement, Modifiers aModifiers,
+                  IsEligibleForSwipe aIsEligibleForSwipe);
+
+  void SetLineOrPageDeltas(int32_t aLineOrPageDeltaX,
+                           int32_t aLineOrPageDeltaY);
+
   bool IsMomentum() const;
 
   WidgetWheelEvent ToWidgetEvent(nsIWidget* aWidget) const;
@@ -406,6 +415,29 @@ class PanGestureInput : public InputData {
   ScreenPoint UserMultipliedPanDisplacement() const;
   ParentLayerPoint UserMultipliedLocalPanDisplacement() const;
 
+  void SetHandledByAPZ(bool aHandled) { mHandledByAPZ = aHandled; }
+  void SetOverscrollBehaviorAllowsSwipe(bool aAllows) {
+    mOverscrollBehaviorAllowsSwipe = aAllows;
+  }
+  void SetSimulateMomentum(bool aSimulate) { mSimulateMomentum = aSimulate; }
+  void SetIsNoLineOrPageDelta(bool aIsNoLineOrPageDelta) {
+    mIsNoLineOrPageDelta = aIsNoLineOrPageDelta;
+  }
+
+  // Returns true if this pan gesture event is elligible for browser swipe
+  // gesture considering the overscroll-behavior property of the target
+  // scroll container.
+  bool AllowsSwipe() const {
+    MOZ_ASSERT(mHandledByAPZ);
+    return mMayTriggerSwipe && mOverscrollBehaviorAllowsSwipe;
+  }
+
+  // Similar to above AllowsSwipe() but this doesn't care the
+  // overscroll-behavior property, this function should be only used for cases
+  // where APZ isn't involved.
+  bool MayTriggerSwipe() const { return mMayTriggerSwipe; }
+  bool RequiresContentResponseIfCannotScrollHorizontallyInStartDirection();
+
   static gfx::IntPoint GetIntegerDeltaForEvent(bool aIsStart, float x, float y);
 
   // Warning, this class is serialized and sent over IPC. Any change to its
@@ -433,20 +465,8 @@ class PanGestureInput : public InputData {
 
   bool mHandledByAPZ : 1;
 
-  // true if this is a PANGESTURE_END event that will be followed by a
-  // PANGESTURE_MOMENTUMSTART event.
-  bool mFollowedByMomentum : 1;
-
-  // If this is true, and this event started a new input block that couldn't
-  // find a scrollable target which is scrollable in the horizontal component
-  // of the scroll start direction, then this input block needs to be put on
-  // hold until a content response has arrived, even if the block has a
-  // confirmed target.
-  // This is used by events that can result in a swipe instead of a scroll.
-  bool mRequiresContentResponseIfCannotScrollHorizontallyInStartDirection : 1;
-
-  // This is used by APZ to communicate to the macOS widget code whether
-  // the overscroll-behavior of the scroll frame handling this swipe allows
+  // This is used by APZ to communicate to widget code whether the
+  // overscroll-behavior of the scroll frame handling this swipe allows
   // non-local overscroll behaviors in the horizontal direction (such as
   // swipe navigation).
   bool mOverscrollBehaviorAllowsSwipe : 1;
@@ -456,19 +476,22 @@ class PanGestureInput : public InputData {
   // events.)
   bool mSimulateMomentum : 1;
 
-  void SetHandledByAPZ(bool aHandled) { mHandledByAPZ = aHandled; }
-  void SetFollowedByMomentum(bool aFollowed) {
-    mFollowedByMomentum = aFollowed;
-  }
-  void SetRequiresContentResponseIfCannotScrollHorizontallyInStartDirection(
-      bool aRequires) {
-    mRequiresContentResponseIfCannotScrollHorizontallyInStartDirection =
-        aRequires;
-  }
-  void SetOverscrollBehaviorAllowsSwipe(bool aAllows) {
-    mOverscrollBehaviorAllowsSwipe = aAllows;
-  }
-  void SetSimulateMomentum(bool aSimulate) { mSimulateMomentum = aSimulate; }
+  // true if the creator of this object does not set the mLineOrPageDeltaX/Y
+  // fields and when/if WidgetWheelEvent's are generated from this object wants
+  // the corresponding mLineOrPageDeltaX/Y fields in the WidgetWheelEvent to be
+  // automatically calculated (upon event dispatch by the EventStateManager
+  // code).
+  bool mIsNoLineOrPageDelta : 1;
+
+ private:
+  // If this is true, and this event started a new input block that couldn't
+  // find a scrollable target which is scrollable in the horizontal component
+  // of the scroll start direction, then this input block needs to be put on
+  // hold until a content response has arrived, even if the block has a
+  // confirmed target.
+  // This is used by events that can result in a swipe instead of a scroll.
+  bool mMayTriggerSwipe : 1;
+  void SetMayTriggerSwipe(bool aValue) { mMayTriggerSwipe = aValue; }
 };
 
 /**
@@ -480,6 +503,7 @@ class PinchGestureInput : public InputData {
  protected:
   friend mozilla::layers::APZInputBridgeChild;
   friend mozilla::layers::PAPZInputBridgeParent;
+  ALLOW_DEPRECATED_READPARAM
 
   PinchGestureInput();
 
@@ -514,8 +538,7 @@ class PinchGestureInput : public InputData {
 
   // Construct a pinch gesture from a Screen point.
   PinchGestureInput(PinchGestureType aType, PinchGestureSource aSource,
-                    uint32_t aTime, TimeStamp aTimeStamp,
-                    const ExternalPoint& aScreenOffset,
+                    TimeStamp aTimeStamp, const ExternalPoint& aScreenOffset,
                     const ScreenPoint& aFocusPoint, ScreenCoord aCurrentSpan,
                     ScreenCoord aPreviousSpan, Modifiers aModifiers);
 
@@ -592,6 +615,7 @@ class TapGestureInput : public InputData {
  protected:
   friend mozilla::layers::APZInputBridgeChild;
   friend mozilla::layers::PAPZInputBridgeParent;
+  ALLOW_DEPRECATED_READPARAM
 
   TapGestureInput();
 
@@ -611,12 +635,12 @@ class TapGestureInput : public InputData {
 
   // Construct a tap gesture from a Screen point.
   // mLocalPoint remains (0,0) unless it's set later.
-  TapGestureInput(TapGestureType aType, uint32_t aTime, TimeStamp aTimeStamp,
+  TapGestureInput(TapGestureType aType, TimeStamp aTimeStamp,
                   const ScreenIntPoint& aPoint, Modifiers aModifiers);
 
   // Construct a tap gesture from a ParentLayer point.
   // mPoint remains (0,0) unless it's set later.
-  TapGestureInput(TapGestureType aType, uint32_t aTime, TimeStamp aTimeStamp,
+  TapGestureInput(TapGestureType aType, TimeStamp aTimeStamp,
                   const ParentLayerPoint& aLocalPoint, Modifiers aModifiers);
 
   bool TransformToLocal(const ScreenToParentLayerMatrix4x4& aTransform);
@@ -642,6 +666,7 @@ class ScrollWheelInput : public InputData {
  protected:
   friend mozilla::layers::APZInputBridgeChild;
   friend mozilla::layers::PAPZInputBridgeParent;
+  ALLOW_DEPRECATED_READPARAM
 
   typedef mozilla::layers::APZWheelAction APZWheelAction;
 
@@ -666,7 +691,7 @@ class ScrollWheelInput : public InputData {
   );
   // clang-format on
 
-  ScrollWheelInput(uint32_t aTime, TimeStamp aTimeStamp, Modifiers aModifiers,
+  ScrollWheelInput(TimeStamp aTimeStamp, Modifiers aModifiers,
                    ScrollMode aScrollMode, ScrollDeltaType aDeltaType,
                    const ScreenPoint& aOrigin, double aDeltaX, double aDeltaY,
                    bool aAllowToOverrideSystemScrollSpeed,
@@ -684,7 +709,11 @@ class ScrollWheelInput : public InputData {
 
   // The following two functions are for auto-dir scrolling. For detailed
   // information on auto-dir, @see mozilla::WheelDeltaAdjustmentStrategy
-  bool IsAutoDir() const {
+  bool IsAutoDir(bool aForce = false) const {
+    if (aForce) {
+      return true;
+    }
+
     switch (mWheelDeltaAdjustmentStrategy) {
       case WheelDeltaAdjustmentStrategy::eAutoDir:
       case WheelDeltaAdjustmentStrategy::eAutoDirWithRootHonour:
@@ -703,9 +732,10 @@ class ScrollWheelInput : public InputData {
   // not an auto-dir scroll.
   // For detailed information on auto-dir,
   // @see mozilla::WheelDeltaAdjustmentStrategy
-  bool HonoursRoot() const {
+  bool HonoursRoot(bool aForce = false) const {
     return WheelDeltaAdjustmentStrategy::eAutoDirWithRootHonour ==
-           mWheelDeltaAdjustmentStrategy;
+               mWheelDeltaAdjustmentStrategy ||
+           aForce;
   }
 
   // Warning, this class is serialized and sent over IPC. Any change to its
@@ -770,7 +800,7 @@ class KeyboardInput : public InputData {
     KEY_DOWN,
     KEY_PRESS,
     KEY_UP,
-    // Any other key event such as eKeyDownOnPlugin
+    // Any other key event such as eAccessKeyNotFound
     KEY_OTHER,
 
     // Used as an upper bound for ContiguousEnumSerializer
@@ -796,6 +826,7 @@ class KeyboardInput : public InputData {
  protected:
   friend mozilla::layers::APZInputBridgeChild;
   friend mozilla::layers::PAPZInputBridgeParent;
+  ALLOW_DEPRECATED_READPARAM
 
   KeyboardInput();
 };
