@@ -44,7 +44,7 @@ struct ScrollbarParams {
   bool isSmall = false;
   bool isHorizontal = false;
   bool isRtl = false;
-  bool isOnDarkBackground = false;
+  bool isDark = false;
   bool isCustom = false;
   // Two colors only used when custom is true.
   nscolor trackColor = NS_RGBA(0, 0, 0, 0);
@@ -53,6 +53,7 @@ struct ScrollbarParams {
 
 static ScrollbarParams ComputeScrollbarParams(nsIFrame* aFrame,
                                               const ComputedStyle& aStyle,
+                                              const ThemeColors& aColors,
                                               ScrollbarKind aScrollbarKind) {
   ScrollbarParams params;
   params.isOverlay =
@@ -62,8 +63,7 @@ static ScrollbarParams ComputeScrollbarParams(nsIFrame* aFrame,
       aStyle.StyleUIReset()->ScrollbarWidth() == StyleScrollbarWidth::Thin;
   params.isRtl = aScrollbarKind == ScrollbarKind::VerticalLeft;
   params.isHorizontal = aScrollbarKind == ScrollbarKind::Horizontal;
-  params.isOnDarkBackground = !StaticPrefs::widget_disable_dark_scrollbar() &&
-                              nsNativeTheme::IsDarkBackground(aFrame);
+  params.isDark = aColors.IsDark();
 
   const nsStyleUI* ui = aStyle.StyleUI();
   if (ui->HasCustomScrollbars()) {
@@ -81,75 +81,39 @@ LayoutDeviceIntSize ScrollbarDrawingCocoa::GetMinimumWidgetSize(
     nsIFrame* aFrame) {
   MOZ_ASSERT(nsNativeTheme::IsWidgetScrollbarPart(aAppearance));
 
-  auto minSize = [&] {
+  auto minSize = [&]() -> CSSIntSize {
     switch (aAppearance) {
       case StyleAppearance::ScrollbarthumbHorizontal:
-        return IntSize{26, 0};
+        return {26, 0};
       case StyleAppearance::ScrollbarthumbVertical:
-        return IntSize{0, 26};
+        return {0, 26};
       case StyleAppearance::ScrollbarVertical:
       case StyleAppearance::ScrollbarHorizontal:
       case StyleAppearance::ScrollbartrackVertical:
       case StyleAppearance::ScrollbartrackHorizontal: {
         ComputedStyle* style = nsLayoutUtils::StyleForScrollbar(aFrame);
         auto scrollbarWidth = style->StyleUIReset()->ScrollbarWidth();
-        auto size = GetScrollbarSize(
-            scrollbarWidth,
-            LookAndFeel::GetInt(LookAndFeel::IntID::UseOverlayScrollbars));
-        return IntSize{size, size};
+        auto size = GetCSSScrollbarSize(
+            scrollbarWidth, Overlay(aPresContext->UseOverlayScrollbars()));
+        return {size, size};
       }
       case StyleAppearance::ScrollbarbuttonUp:
       case StyleAppearance::ScrollbarbuttonDown:
-        return IntSize{15, 16};
+        return {15, 16};
       case StyleAppearance::ScrollbarbuttonLeft:
       case StyleAppearance::ScrollbarbuttonRight:
-        return IntSize{16, 15};
+        return {16, 15};
       default:
-        return IntSize{};
+        return {};
     }
   }();
 
-  auto dpi = GetDPIRatioForScrollbarPart(aPresContext).scale;
-  if (dpi >= 2.0f) {
-    return LayoutDeviceIntSize{minSize.width * 2, minSize.height * 2};
-  }
-  return LayoutDeviceIntSize{minSize.width, minSize.height};
-}
-
-/*static*/
-CSSIntCoord ScrollbarDrawingCocoa::GetScrollbarSize(StyleScrollbarWidth aWidth,
-                                                    bool aOverlay) {
-  bool isSmall = aWidth == StyleScrollbarWidth::Thin;
-  if (aOverlay) {
-    return isSmall ? 14 : 16;
-  }
-  return isSmall ? 11 : 15;
-}
-
-/*static*/
-LayoutDeviceIntCoord ScrollbarDrawingCocoa::GetScrollbarSize(
-    StyleScrollbarWidth aWidth, bool aOverlay, DPIRatio aDpiRatio) {
-  CSSIntCoord size = GetScrollbarSize(aWidth, aOverlay);
-  if (aDpiRatio.scale >= 2.0f) {
-    return int32_t(size) * 2;
-  }
-  return int32_t(size);
-}
-
-auto ScrollbarDrawingCocoa::GetScrollbarSizes(nsPresContext* aPresContext,
-                                              StyleScrollbarWidth aWidth,
-                                              Overlay aOverlay)
-    -> ScrollbarSizes {
-  auto size = GetScrollbarSize(aWidth, aOverlay == Overlay::Yes,
-                               GetDPIRatioForScrollbarPart(aPresContext));
-  return {size, size};
+  auto dpi = GetDPIRatioForScrollbarPart(aPresContext);
+  return LayoutDeviceIntSize::Round(CSSSize(minSize) * dpi);
 }
 
 static ThumbRect GetThumbRect(const LayoutDeviceRect& aRect,
                               const ScrollbarParams& aParams, float aScale) {
-  // This matches the sizing checks in GetMinimumWidgetSize etc.
-  aScale = aScale >= 2.0f ? 2.0f : 1.0f;
-
   // Compute the thumb thickness. This varies based on aParams.small,
   // aParams.overlay and aParams.rolledOver. non-overlay: 6 / 8, overlay
   // non-hovered: 5 / 7, overlay hovered: 9 / 11
@@ -186,9 +150,9 @@ static ThumbRect GetThumbRect(const LayoutDeviceRect& aRect,
     faceColor = aParams.faceColor;
   } else {
     if (aParams.isOverlay) {
-      faceColor = aParams.isOnDarkBackground ? NS_RGBA(255, 255, 255, 128)
-                                             : NS_RGBA(0, 0, 0, 128);
-    } else if (aParams.isOnDarkBackground) {
+      faceColor =
+          aParams.isDark ? NS_RGBA(255, 255, 255, 128) : NS_RGBA(0, 0, 0, 128);
+    } else if (aParams.isDark) {
       faceColor = aParams.isRolledOver ? NS_RGBA(158, 158, 158, 255)
                                        : NS_RGBA(117, 117, 117, 255);
     } else {
@@ -208,12 +172,11 @@ static ThumbRect GetThumbRect(const LayoutDeviceRect& aRect,
     const uint8_t strokeAlpha =
         uint8_t(clamped(NS_GET_A(faceColor) * kAlphaScaling, 0.0f, 48.0f));
     if (strokeAlpha) {
-      strokeOutset = (aParams.isOnDarkBackground ? 0.3f : 0.5f) * aScale;
-      strokeWidth = (aParams.isOnDarkBackground ? 0.6f : 0.8f) * aScale;
+      strokeOutset = (aParams.isDark ? 0.3f : 0.5f) * aScale;
+      strokeWidth = (aParams.isDark ? 0.6f : 0.8f) * aScale;
 
-      strokeColor = aParams.isOnDarkBackground
-                        ? NS_RGBA(0, 0, 0, strokeAlpha)
-                        : NS_RGBA(255, 255, 255, strokeAlpha);
+      strokeColor = aParams.isDark ? NS_RGBA(0, 0, 0, strokeAlpha)
+                                   : NS_RGBA(255, 255, 255, strokeAlpha);
     }
   }
 
@@ -256,19 +219,16 @@ static bool GetScrollbarTrackRects(const LayoutDeviceRect& aRect,
     return false;
   }
 
-  // This matches the sizing checks in GetMinimumWidgetSize etc.
-  aScale = aScale >= 2.0f ? 2.0f : 1.0f;
-
   nscolor trackColor;
   if (aParams.isCustom) {
     trackColor = aParams.trackColor;
   } else {
     if (aParams.isOverlay) {
-      trackColor = aParams.isOnDarkBackground ? NS_RGBA(201, 201, 201, 38)
-                                              : NS_RGBA(250, 250, 250, 191);
+      trackColor = aParams.isDark ? NS_RGBA(201, 201, 201, 38)
+                                  : NS_RGBA(250, 250, 250, 191);
     } else {
-      trackColor = aParams.isOnDarkBackground ? NS_RGBA(46, 46, 46, 255)
-                                              : NS_RGBA(250, 250, 250, 255);
+      trackColor = aParams.isDark ? NS_RGBA(46, 46, 46, 255)
+                                  : NS_RGBA(250, 250, 250, 255);
     }
   }
 
@@ -324,9 +284,6 @@ static bool GetScrollCornerRects(const LayoutDeviceRect& aRect,
     return false;
   }
 
-  // This matches the sizing checks in GetMinimumWidgetSize etc.
-  aScale = aScale >= 2.0f ? 2.0f : 1.0f;
-
   // Draw the following scroll corner.
   //
   //        Output:                      Rectangles:
@@ -348,8 +305,8 @@ static bool GetScrollCornerRects(const LayoutDeviceRect& aRect,
   if (aParams.isCustom) {
     trackColor = aParams.trackColor;
   } else {
-    trackColor = aParams.isOnDarkBackground ? NS_RGBA(46, 46, 46, 255)
-                                            : NS_RGBA(250, 250, 250, 255);
+    trackColor =
+        aParams.isDark ? NS_RGBA(46, 46, 46, 255) : NS_RGBA(250, 250, 250, 255);
   }
   ScrollbarTrackDecorationColors colors =
       ComputeScrollbarTrackDecorationColors(trackColor);
@@ -386,10 +343,10 @@ template <typename PaintBackendData>
 void ScrollbarDrawingCocoa::DoPaintScrollbarThumb(
     PaintBackendData& aPaintData, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aElementState, const EventStates& aDocumentState,
-    const DPIRatio& aDpiRatio) {
+    const ElementState& aElementState, const DocumentState& aDocumentState,
+    const Colors& aColors, const DPIRatio& aDpiRatio) {
   ScrollbarParams params =
-      ComputeScrollbarParams(aFrame, aStyle, aScrollbarKind);
+      ComputeScrollbarParams(aFrame, aStyle, aColors, aScrollbarKind);
   auto thumb = GetThumbRect(aRect, params, aDpiRatio.scale);
   LayoutDeviceCoord radius =
       (params.isHorizontal ? thumb.mRect.Height() : thumb.mRect.Width()) / 2.0f;
@@ -415,22 +372,20 @@ void ScrollbarDrawingCocoa::DoPaintScrollbarThumb(
 bool ScrollbarDrawingCocoa::PaintScrollbarThumb(
     DrawTarget& aDt, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aElementState, const EventStates& aDocumentState,
-    const Colors&, const DPIRatio& aDpiRatio) {
-  // TODO: Maybe respect the UseSystemColors setting?
+    const ElementState& aElementState, const DocumentState& aDocumentState,
+    const Colors& aColors, const DPIRatio& aDpiRatio) {
   DoPaintScrollbarThumb(aDt, aRect, aScrollbarKind, aFrame, aStyle,
-                        aElementState, aDocumentState, aDpiRatio);
+                        aElementState, aDocumentState, aColors, aDpiRatio);
   return true;
 }
 
 bool ScrollbarDrawingCocoa::PaintScrollbarThumb(
     WebRenderBackendData& aWrData, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aElementState, const EventStates& aDocumentState,
-    const Colors&, const DPIRatio& aDpiRatio) {
-  // TODO: Maybe respect the UseSystemColors setting?
+    const ElementState& aElementState, const DocumentState& aDocumentState,
+    const Colors& aColors, const DPIRatio& aDpiRatio) {
   DoPaintScrollbarThumb(aWrData, aRect, aScrollbarKind, aFrame, aStyle,
-                        aElementState, aDocumentState, aDpiRatio);
+                        aElementState, aDocumentState, aColors, aDpiRatio);
   return true;
 }
 
@@ -438,9 +393,10 @@ template <typename PaintBackendData>
 void ScrollbarDrawingCocoa::DoPaintScrollbarTrack(
     PaintBackendData& aPaintData, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aDocumentState, const DPIRatio& aDpiRatio) {
+    const DocumentState& aDocumentState, const Colors& aColors,
+    const DPIRatio& aDpiRatio) {
   ScrollbarParams params =
-      ComputeScrollbarParams(aFrame, aStyle, aScrollbarKind);
+      ComputeScrollbarParams(aFrame, aStyle, aColors, aScrollbarKind);
   ScrollbarTrackRects rects;
   if (GetScrollbarTrackRects(aRect, params, aDpiRatio.scale, rects)) {
     for (const auto& rect : rects) {
@@ -453,22 +409,20 @@ void ScrollbarDrawingCocoa::DoPaintScrollbarTrack(
 bool ScrollbarDrawingCocoa::PaintScrollbarTrack(
     DrawTarget& aDt, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aDocumentState, const Colors&,
+    const DocumentState& aDocumentState, const Colors& aColors,
     const DPIRatio& aDpiRatio) {
-  // TODO: Maybe respect the UseSystemColors setting?
   DoPaintScrollbarTrack(aDt, aRect, aScrollbarKind, aFrame, aStyle,
-                        aDocumentState, aDpiRatio);
+                        aDocumentState, aColors, aDpiRatio);
   return true;
 }
 
 bool ScrollbarDrawingCocoa::PaintScrollbarTrack(
     WebRenderBackendData& aWrData, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aDocumentState, const Colors&,
+    const DocumentState& aDocumentState, const Colors& aColors,
     const DPIRatio& aDpiRatio) {
-  // TODO: Maybe respect the UseSystemColors setting?
   DoPaintScrollbarTrack(aWrData, aRect, aScrollbarKind, aFrame, aStyle,
-                        aDocumentState, aDpiRatio);
+                        aDocumentState, aColors, aDpiRatio);
   return true;
 }
 
@@ -476,9 +430,10 @@ template <typename PaintBackendData>
 void ScrollbarDrawingCocoa::DoPaintScrollCorner(
     PaintBackendData& aPaintData, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aDocumentState, const DPIRatio& aDpiRatio) {
+    const DocumentState& aDocumentState, const Colors& aColors,
+    const DPIRatio& aDpiRatio) {
   ScrollbarParams params =
-      ComputeScrollbarParams(aFrame, aStyle, aScrollbarKind);
+      ComputeScrollbarParams(aFrame, aStyle, aColors, aScrollbarKind);
   ScrollCornerRects rects;
   if (GetScrollCornerRects(aRect, params, aDpiRatio.scale, rects)) {
     for (const auto& rect : rects) {
@@ -491,33 +446,31 @@ void ScrollbarDrawingCocoa::DoPaintScrollCorner(
 bool ScrollbarDrawingCocoa::PaintScrollCorner(
     DrawTarget& aDt, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aDocumentState, const Colors&,
+    const DocumentState& aDocumentState, const Colors& aColors,
     const DPIRatio& aDpiRatio) {
-  // TODO: Maybe respect the UseSystemColors setting?
   DoPaintScrollCorner(aDt, aRect, aScrollbarKind, aFrame, aStyle,
-                      aDocumentState, aDpiRatio);
+                      aDocumentState, aColors, aDpiRatio);
   return true;
 }
 
 bool ScrollbarDrawingCocoa::PaintScrollCorner(
     WebRenderBackendData& aWrData, const LayoutDeviceRect& aRect,
     ScrollbarKind aScrollbarKind, nsIFrame* aFrame, const ComputedStyle& aStyle,
-    const EventStates& aDocumentState, const Colors&,
+    const DocumentState& aDocumentState, const Colors& aColors,
     const DPIRatio& aDpiRatio) {
-  // TODO: Maybe respect the UseSystemColors setting?
   DoPaintScrollCorner(aWrData, aRect, aScrollbarKind, aFrame, aStyle,
-                      aDocumentState, aDpiRatio);
+                      aDocumentState, aColors, aDpiRatio);
   return true;
 }
 
 void ScrollbarDrawingCocoa::RecomputeScrollbarParams() {
-  uint32_t defaultSize = 17;
-  uint32_t overrideSize =
-      StaticPrefs::widget_non_native_theme_scrollbar_size_override();
-  if (overrideSize > 0) {
-    defaultSize = overrideSize;
-  }
-  mHorizontalScrollbarHeight = mVerticalScrollbarWidth = defaultSize;
+  // FIXME(emilio): This doesn't respect the
+  // StaticPrefs::widget_non_native_theme_scrollbar_size_override() pref;
+  ConfigureScrollbarSize(15);  // Just in case, for future-proofing
+  ConfigureScrollbarSize(StyleScrollbarWidth::Auto, Overlay::No, 15);
+  ConfigureScrollbarSize(StyleScrollbarWidth::Thin, Overlay::No, 11);
+  ConfigureScrollbarSize(StyleScrollbarWidth::Auto, Overlay::Yes, 16);
+  ConfigureScrollbarSize(StyleScrollbarWidth::Thin, Overlay::Yes, 14);
 }
 
 }  // namespace mozilla::widget
