@@ -3,13 +3,11 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-#include "AppearanceOverride.h"
 #include "mozilla/widget/ThemeChangeKind.h"
 #include "nsLookAndFeel.h"
 #include "nsCocoaFeatures.h"
 #include "nsNativeThemeColors.h"
 #include "nsStyleConsts.h"
-#include "nsCocoaFeatures.h"
 #include "nsIContent.h"
 #include "gfxFont.h"
 #include "gfxFontConstants.h"
@@ -18,8 +16,8 @@
 #include "mozilla/FontPropertyTypes.h"
 #include "mozilla/gfx/2D.h"
 #include "mozilla/StaticPrefs_widget.h"
+#include "mozilla/Telemetry.h"
 #include "mozilla/widget/WidgetMessageUtils.h"
-#include "SDKDeclarations.h"
 
 #import <Cocoa/Cocoa.h>
 #import <AppKit/NSColor.h>
@@ -33,32 +31,58 @@ using namespace mozilla;
 + (void)startObserving;
 @end
 
-nsLookAndFeel::nsLookAndFeel() = default;
+nsLookAndFeel::nsLookAndFeel() {
+  [MOZLookAndFeelDynamicChangeObserver startObserving];
+}
 
 nsLookAndFeel::~nsLookAndFeel() = default;
 
-void nsLookAndFeel::NativeInit() {
+void nsLookAndFeel::EnsureInit() {
+  if (mInitialized) {
+    return;
+  }
+
   NS_OBJC_BEGIN_TRY_ABORT_BLOCK
 
-  [MOZLookAndFeelDynamicChangeObserver startObserving];
+  mInitialized = true;
+  NSWindow* window =
+      [[NSWindow alloc] initWithContentRect:NSZeroRect
+                                  styleMask:NSWindowStyleMaskTitled
+                                    backing:NSBackingStoreBuffered
+                                      defer:NO];
+  auto release = MakeScopeExit([&] { [window release]; });
+
+  mRtl = window.windowTitlebarLayoutDirection ==
+         NSUserInterfaceLayoutDirectionRightToLeft;
+  mTitlebarHeight = std::ceil(window.frame.size.height);
+
   RecordTelemetry();
 
   NS_OBJC_END_TRY_ABORT_BLOCK
 }
 
+void nsLookAndFeel::RefreshImpl() {
+  mInitialized = false;
+  nsXPLookAndFeel::RefreshImpl();
+}
+
 static nscolor GetColorFromNSColor(NSColor* aColor) {
-  NSColor* deviceColor = [aColor colorUsingColorSpaceName:NSDeviceRGBColorSpace];
+  NSColor* deviceColor =
+      [aColor colorUsingColorSpace:NSColorSpace.deviceRGBColorSpace];
   return NS_RGBA((unsigned int)(deviceColor.redComponent * 255.0),
                  (unsigned int)(deviceColor.greenComponent * 255.0),
                  (unsigned int)(deviceColor.blueComponent * 255.0),
                  (unsigned int)(deviceColor.alphaComponent * 255.0));
 }
 
-static nscolor GetColorFromNSColorWithCustomAlpha(NSColor* aColor, float alpha) {
-  NSColor* deviceColor = [aColor colorUsingColorSpaceName:NSDeviceRGBColorSpace];
+static nscolor GetColorFromNSColorWithCustomAlpha(NSColor* aColor,
+                                                  float alpha) {
+  NSColor* deviceColor =
+      [aColor colorUsingColorSpace:[NSColorSpace deviceRGBColorSpace]];
   return NS_RGBA((unsigned int)(deviceColor.redComponent * 255.0),
                  (unsigned int)(deviceColor.greenComponent * 255.0),
-                 (unsigned int)(deviceColor.blueComponent * 255.0), (unsigned int)(alpha * 255.0));
+                 (unsigned int)(deviceColor.blueComponent * 255.0),
+                 (unsigned int)(alpha * 255.0));
 }
 
 // Turns an opaque selection color into a partially transparent selection color,
@@ -74,11 +98,11 @@ static nscolor GetColorFromNSColorWithCustomAlpha(NSColor* aColor, float alpha)
 // whereas white text on dark blue (which what you get if you mix
 // partially-transparent light blue with the black textbox background) has much
 // better contrast.
-nscolor nsLookAndFeel::ProcessSelectionBackground(nscolor aColor, ColorScheme aScheme) {
+static nscolor ProcessSelectionBackground(nscolor aColor, ColorScheme aScheme) {
   if (aScheme == ColorScheme::Dark) {
     // When we use a dark selection color, we do not change alpha because we do
-    // not use dark selection in content. The dark system color is appropriate for
-    // Firefox UI without needing to adjust its alpha.
+    // not use dark selection in content. The dark system color is appropriate
+    // for Firefox UI without needing to adjust its alpha.
     return aColor;
   }
   uint16_t hue, sat, value;
@@ -100,42 +124,51 @@ nscolor nsLookAndFeel::ProcessSelectionBackground(nscolor aColor, ColorScheme aS
   return resultColor;
 }
 
-nsresult nsLookAndFeel::NativeGetColor(ColorID aID, ColorScheme aScheme, nscolor& aColor) {
+nsresult nsLookAndFeel::NativeGetColor(ColorID aID, ColorScheme aScheme,
+                                       nscolor& aColor) {
   NS_OBJC_BEGIN_TRY_ABORT_BLOCK
 
-  if (@available(macOS 10.14, *)) {
-    // No-op. macOS 10.14+ supports dark mode, so currentAppearance can be set
-    // to either Light or Dark.
-  } else {
-    // System colors before 10.14 are always Light.
-    aScheme = ColorScheme::Light;
-  }
-
   NSAppearance.currentAppearance = NSAppearanceForColorScheme(aScheme);
 
   nscolor color = 0;
   switch (aID) {
     case ColorID::Infobackground:
-      color = NS_RGB(0xdd, 0xdd, 0xdd);
+      color = aScheme == ColorScheme::Light
+                  ? NS_RGB(0xdd, 0xdd, 0xdd)
+                  : GetColorFromNSColor(NSColor.windowBackgroundColor);
       break;
     case ColorID::Highlight:
-      color = ProcessSelectionBackground(GetColorFromNSColor(NSColor.selectedTextBackgroundColor),
-                                         aScheme);
+      color = ProcessSelectionBackground(
+          GetColorFromNSColor(NSColor.selectedTextBackgroundColor), aScheme);
       break;
     // This is used to gray out the selection when it's not focused. Used with
     // nsISelectionController::SELECTION_DISABLED.
     case ColorID::TextSelectDisabledBackground:
-      color = ProcessSelectionBackground(GetColorFromNSColor(NSColor.secondarySelectedControlColor),
-                                         aScheme);
+      color = ProcessSelectionBackground(
+          GetColorFromNSColor(NSColor.secondarySelectedControlColor), aScheme);
+      break;
+    case ColorID::MozMenuhoverdisabled:
+      aColor = NS_TRANSPARENT;
+      break;
+    case ColorID::Accentcolor:
+      color = GetColorFromNSColor(NSColor.controlAccentColor);
       break;
     case ColorID::MozMenuhover:
     case ColorID::Selecteditem:
-      color = GetColorFromNSColor(NSColor.alternateSelectedControlColor);
+      color = GetColorFromNSColor(NSColor.selectedContentBackgroundColor);
+      if (aID == ColorID::MozMenuhover &&
+          !LookAndFeel::GetInt(IntID::PrefersReducedTransparency)) {
+        // Wash the color a little bit with semi-transparent white to match a
+        // bit closer the native NSVisualEffectSelection on menus.
+        color = NS_ComposeColors(
+            color,
+            NS_RGBA(255, 255, 255, aScheme == ColorScheme::Light ? 51 : 25));
+      }
       break;
-    case ColorID::MozAccentColorForeground:
+    case ColorID::Accentcolortext:
     case ColorID::MozMenuhovertext:
     case ColorID::Selecteditemtext:
-      color = GetColorFromNSColor(NSColor.alternateSelectedControlTextColor);
+      color = GetColorFromNSColor(NSColor.selectedMenuItemTextColor);
       break;
     case ColorID::IMESelectedRawTextBackground:
     case ColorID::IMESelectedConvertedTextBackground:
@@ -158,46 +191,43 @@ nsresult nsLookAndFeel::NativeGetColor(ColorID aID, ColorScheme aScheme, nscolor
     case ColorID::IMESelectedConvertedTextUnderline:
       color = NS_SAME_AS_FOREGROUND_COLOR;
       break;
-    case ColorID::SpellCheckerUnderline:
-      color = NS_RGB(0xff, 0, 0);
-      break;
 
       //
       // css2 system colors http://www.w3.org/TR/REC-CSS2/ui.html#system-colors
       //
-      // It's really hard to effectively map these to the Appearance Manager properly,
-      // since they are modeled word for word after the win32 system colors and don't have any
-      // real counterparts in the Mac world. I'm sure we'll be tweaking these for
-      // years to come.
+      // It's really hard to effectively map these to the Appearance Manager
+      // properly, since they are modeled word for word after the win32 system
+      // colors and don't have any real counterparts in the Mac world. I'm sure
+      // we'll be tweaking these for years to come.
       //
-      // Thanks to mpt26@student.canterbury.ac.nz for the hardcoded values that form the defaults
+      // Thanks to mpt26@student.canterbury.ac.nz for the hardcoded values that
+      // form the defaults
       //  if querying the Appearance Manager fails ;)
       //
     case ColorID::MozMacDefaultbuttontext:
       color = NS_RGB(0xFF, 0xFF, 0xFF);
       break;
-    case ColorID::MozButtonactivetext:
-      // Pre-macOS 12, pressed buttons were filled with the highlight color and the text was white.
-      // Starting with macOS 12, pressed (non-default) buttons are filled with medium gray and the
-      // text color is the same as in the non-pressed state.
-      color = nsCocoaFeatures::OnMontereyOrLater() ? GetColorFromNSColor(NSColor.controlTextColor)
-                                                   : NS_RGB(0xFF, 0xFF, 0xFF);
+    case ColorID::MozSidebar:
+      color = aScheme == ColorScheme::Light ? NS_RGB(0xf6, 0xf6, 0xf6)
+                                            : NS_RGB(0x2d, 0x2d, 0x2d);
       break;
-    case ColorID::Captiontext:
-    case ColorID::Menutext:
-    case ColorID::Infotext:
-    case ColorID::MozMenubartext:
-      color = GetColorFromNSColor(NSColor.textColor);
+    case ColorID::MozSidebarborder:
+      // hsla(240, 5%, 5%, .1)
+      color = NS_RGBA(12, 12, 13, 26);
+      break;
+    case ColorID::MozButtonactivetext:
+      // Pre-macOS 12, pressed buttons were filled with the highlight color and
+      // the text was white. Starting with macOS 12, pressed (non-default)
+      // buttons are filled with medium gray and the text color is the same as
+      // in the non-pressed state.
+      color = nsCocoaFeatures::OnMontereyOrLater()
+                  ? GetColorFromNSColor(NSColor.controlTextColor)
+                  : NS_RGB(0xFF, 0xFF, 0xFF);
       break;
     case ColorID::Windowtext:
+    case ColorID::MozDialogtext:
       color = GetColorFromNSColor(NSColor.windowFrameTextColor);
       break;
-    case ColorID::Activecaption:
-      color = GetColorFromNSColor(NSColor.gridColor);
-      break;
-    case ColorID::Activeborder:
-      color = GetColorFromNSColor(NSColor.keyboardFocusIndicatorColor);
-      break;
     case ColorID::Appworkspace:
       color = NS_RGB(0xFF, 0xFF, 0xFF);
       break;
@@ -216,9 +246,6 @@ nsresult nsLookAndFeel::NativeGetColor(ColorID aID, ColorScheme aScheme, nscolor
     case ColorID::Buttonhighlight:
       color = GetColorFromNSColor(NSColor.selectedControlColor);
       break;
-    case ColorID::Inactivecaptiontext:
-      color = NS_RGB(0x45, 0x45, 0x45);
-      break;
     case ColorID::Scrollbar:
       color = GetColorFromNSColor(NSColor.scrollBarColor);
       break;
@@ -227,98 +254,84 @@ nsresult nsLookAndFeel::NativeGetColor(ColorID aID, ColorScheme aScheme, nscolor
       break;
     case ColorID::Buttonshadow:
     case ColorID::Threeddarkshadow:
-      color = aScheme == ColorScheme::Dark ? *GenericDarkColor(aID) : NS_RGB(0xDC, 0xDC, 0xDC);
+      color = aScheme == ColorScheme::Dark ? *GenericDarkColor(aID)
+                                           : NS_RGB(0xDC, 0xDC, 0xDC);
       break;
     case ColorID::Threedshadow:
-      color = aScheme == ColorScheme::Dark ? *GenericDarkColor(aID) : NS_RGB(0xE0, 0xE0, 0xE0);
+      color = aScheme == ColorScheme::Dark ? *GenericDarkColor(aID)
+                                           : NS_RGB(0xE0, 0xE0, 0xE0);
       break;
     case ColorID::Threedface:
-      color = aScheme == ColorScheme::Dark ? *GenericDarkColor(aID) : NS_RGB(0xF0, 0xF0, 0xF0);
+      color = aScheme == ColorScheme::Dark ? *GenericDarkColor(aID)
+                                           : NS_RGB(0xF0, 0xF0, 0xF0);
       break;
     case ColorID::Threedlightshadow:
+    case ColorID::Buttonborder:
     case ColorID::MozDisabledfield:
-      color = aScheme == ColorScheme::Dark ? *GenericDarkColor(aID) : NS_RGB(0xDA, 0xDA, 0xDA);
+      color = aScheme == ColorScheme::Dark ? *GenericDarkColor(aID)
+                                           : NS_RGB(0xDA, 0xDA, 0xDA);
       break;
     case ColorID::Menu:
-      color = GetColorFromNSColor(NSColor.textBackgroundColor);
+      // Hand-picked from Sonoma because there doesn't seem to be any
+      // appropriate menu system color.
+      color = aScheme == ColorScheme::Dark ? NS_RGB(0x36, 0x36, 0x39)
+                                           : NS_RGB(0xeb, 0xeb, 0xeb);
       break;
     case ColorID::Windowframe:
       color = GetColorFromNSColor(NSColor.windowFrameColor);
       break;
-    case ColorID::Window: {
-      if (@available(macOS 10.14, *)) {
-        color = GetColorFromNSColor(NSColor.windowBackgroundColor);
-      } else {
-        // On 10.13 and below, NSColor.windowBackgroundColor is transparent black.
-        // Use a light grey instead (taken from macOS 11.5).
-        color = NS_RGB(0xF6, 0xF6, 0xF6);
-      }
+    case ColorID::MozDialog:
+    case ColorID::Window:
+      color = GetColorFromNSColor(aScheme == ColorScheme::Light
+                                      ? NSColor.windowBackgroundColor
+                                      : NSColor.underPageBackgroundColor);
       break;
-    }
     case ColorID::Field:
     case ColorID::MozCombobox:
-    case ColorID::Inactiveborder:
-    case ColorID::Inactivecaption:
-    case ColorID::MozDialog:
       color = GetColorFromNSColor(NSColor.controlBackgroundColor);
       break;
     case ColorID::Fieldtext:
     case ColorID::MozComboboxtext:
     case ColorID::Buttontext:
     case ColorID::MozButtonhovertext:
-    case ColorID::MozDialogtext:
+    case ColorID::Menutext:
+    case ColorID::Infotext:
     case ColorID::MozCellhighlighttext:
-    case ColorID::MozColheadertext:
-    case ColorID::MozColheaderhovertext:
+    case ColorID::MozSidebartext:
       color = GetColorFromNSColor(NSColor.controlTextColor);
       break;
-    case ColorID::MozDragtargetzone:
-      color = GetColorFromNSColor(NSColor.selectedControlColor);
-      break;
-    case ColorID::MozMacChromeActive: {
-      int grey = NativeGreyColorAsInt(toolbarFillGrey, true);
-      color = NS_RGB(grey, grey, grey);
-      break;
-    }
-    case ColorID::MozMacChromeInactive: {
-      int grey = NativeGreyColorAsInt(toolbarFillGrey, false);
-      color = NS_RGB(grey, grey, grey);
-      break;
-    }
     case ColorID::MozMacFocusring:
-      color = GetColorFromNSColorWithCustomAlpha(NSColor.keyboardFocusIndicatorColor, 0.48);
-      break;
-    case ColorID::MozMacMenushadow:
-      color = NS_RGB(0xA3, 0xA3, 0xA3);
-      break;
-    case ColorID::MozMacMenutextdisable:
-      color = NS_RGB(0x98, 0x98, 0x98);
-      break;
-    case ColorID::MozMacMenutextselect:
-      color = GetColorFromNSColor(NSColor.selectedMenuItemTextColor);
+      color = GetColorFromNSColorWithCustomAlpha(
+          NSColor.keyboardFocusIndicatorColor, 0.48);
       break;
     case ColorID::MozMacDisabledtoolbartext:
     case ColorID::Graytext:
       color = GetColorFromNSColor(NSColor.disabledControlTextColor);
       break;
-    case ColorID::MozMacMenuselect:
-      color = GetColorFromNSColor(NSColor.alternateSelectedControlColor);
-      break;
-    case ColorID::MozButtondefault:
-      color = NS_RGB(0xDC, 0xDC, 0xDC);
-      break;
     case ColorID::MozCellhighlight:
-    case ColorID::MozMacSecondaryhighlight:
       // For inactive list selection
       color = GetColorFromNSColor(NSColor.secondarySelectedControlColor);
       break;
+    case ColorID::MozColheadertext:
+    case ColorID::MozColheaderhovertext:
+    case ColorID::MozColheaderactivetext:
+      color = GetColorFromNSColor(NSColor.headerTextColor);
+      break;
+    case ColorID::MozColheaderactive:
+      color = GetColorFromNSColor(
+          NSColor.unemphasizedSelectedContentBackgroundColor);
+      break;
+    case ColorID::MozColheader:
+    case ColorID::MozColheaderhover:
     case ColorID::MozEventreerow:
       // Background color of even list rows.
-      color = GetColorFromNSColor(NSColor.controlAlternatingRowBackgroundColors[0]);
+      color =
+          GetColorFromNSColor(NSColor.controlAlternatingRowBackgroundColors[0]);
       break;
     case ColorID::MozOddtreerow:
       // Background color of odd list rows.
-      color = GetColorFromNSColor(NSColor.controlAlternatingRowBackgroundColors[1]);
+      color =
+          GetColorFromNSColor(NSColor.controlAlternatingRowBackgroundColors[1]);
       break;
     case ColorID::MozNativehyperlinktext:
       color = GetColorFromNSColor(NSColor.linkColor);
@@ -326,36 +339,29 @@ nsresult nsLookAndFeel::NativeGetColor(ColorID aID, ColorScheme aScheme, nscolor
     case ColorID::MozNativevisitedhyperlinktext:
       color = GetColorFromNSColor(NSColor.systemPurpleColor);
       break;
-    // The following colors are supposed to be used as font-smoothing background
-    // colors, in the chrome-only -moz-font-smoothing-background-color property.
-    // This property is used for text on "vibrant" -moz-appearances.
-    // The colors have been obtained from the system on 10.14 using the
-    // program at https://bugzilla.mozilla.org/attachment.cgi?id=9208594 .
-    // We could obtain them at runtime, but doing so may be expensive and
-    // requires the use of the private API
-    // -[NSVisualEffectView fontSmoothingBackgroundColor].
-    case ColorID::MozMacVibrantTitlebarLight:
-      color = NS_RGB(0xe6, 0xe6, 0xe6);
-      break;
-    case ColorID::MozMacVibrantTitlebarDark:
-      color = NS_RGB(0x28, 0x28, 0x28);
-      break;
-    case ColorID::MozMacTooltip:
-    case ColorID::MozMacMenupopup:
-    case ColorID::MozMacMenuitem:
-      color = aScheme == ColorScheme::Light ? NS_RGB(0xf6, 0xf6, 0xf6) : NS_RGB(0x28, 0x28, 0x28);
-      break;
-    case ColorID::MozMacSourceList:
-      color = aScheme == ColorScheme::Light ? NS_RGB(0xf6, 0xf6, 0xf6) : NS_RGB(0x2d, 0x2d, 0x2d);
-      break;
-    case ColorID::MozMacSourceListSelection:
-      color = aScheme == ColorScheme::Light ? NS_RGB(0xd3, 0xd3, 0xd3) : NS_RGB(0x2d, 0x2d, 0x2d);
-      break;
-    case ColorID::MozMacActiveMenuitem:
-    case ColorID::MozMacActiveSourceListSelection:
-    case ColorID::MozAccentColor:
-      color = GetColorFromNSColor(ControlAccentColor());
-      break;
+    case ColorID::MozHeaderbartext:
+    case ColorID::MozHeaderbarinactivetext:
+    case ColorID::Inactivecaptiontext:
+    case ColorID::Captiontext:
+      aColor = GetColorFromNSColor(NSColor.textColor);
+      return NS_OK;
+    case ColorID::MozHeaderbar:
+    case ColorID::MozHeaderbarinactive:
+    case ColorID::Inactivecaption:
+    case ColorID::Activecaption:
+      // This has better contrast than the stand-in colors.
+      aColor = GetColorFromNSColor(NSColor.windowBackgroundColor);
+      return NS_OK;
+    case ColorID::Marktext:
+    case ColorID::Mark:
+    case ColorID::SpellCheckerUnderline:
+    case ColorID::Activeborder:
+    case ColorID::Inactiveborder:
+    case ColorID::MozAutofillBackground:
+    case ColorID::TargetTextBackground:
+    case ColorID::TargetTextForeground:
+      aColor = GetStandinForNativeColor(aID, aScheme);
+      return NS_OK;
     default:
       aColor = NS_RGB(0xff, 0xff, 0xff);
       return NS_ERROR_FAILURE;
@@ -367,6 +373,16 @@ nsresult nsLookAndFeel::NativeGetColor(ColorID aID, ColorScheme aScheme, nscolor
   NS_OBJC_END_TRY_ABORT_BLOCK
 }
 
+static bool SystemWantsDarkTheme() {
+  // This returns true if the macOS system appearance is set to dark mode,
+  // false otherwise.
+  NSAppearanceName aquaOrDarkAqua =
+      [NSApp.effectiveAppearance bestMatchFromAppearancesWithNames:@[
+        NSAppearanceNameAqua, NSAppearanceNameDarkAqua
+      ]];
+  return [aquaOrDarkAqua isEqualToString:NSAppearanceNameDarkAqua];
+}
+
 nsresult nsLookAndFeel::NativeGetInt(IntID aID, int32_t& aResult) {
   NS_OBJC_BEGIN_TRY_BLOCK_RETURN;
 
@@ -386,9 +402,6 @@ nsresult nsLookAndFeel::NativeGetInt(IntID aID, int32_t& aResult) {
     case IntID::CaretWidth:
       aResult = 1;
       break;
-    case IntID::ShowCaretDuringSelection:
-      aResult = 0;
-      break;
     case IntID::SelectTextfieldsOnKeyFocus:
       // Select textfield content when focused by kbd
       // used by EventStateManager::sTextfieldSelectModel
@@ -397,9 +410,6 @@ nsresult nsLookAndFeel::NativeGetInt(IntID aID, int32_t& aResult) {
     case IntID::SubmenuDelay:
       aResult = 200;
       break;
-    case IntID::TooltipDelay:
-      aResult = 500;
-      break;
     case IntID::MenusCanOverlapOSBar:
       // xul popups are not allowed to overlap the menubar.
       aResult = 0;
@@ -414,9 +424,6 @@ nsresult nsLookAndFeel::NativeGetInt(IntID aID, int32_t& aResult) {
     case IntID::ScrollArrowStyle:
       aResult = eScrollArrow_None;
       break;
-    case IntID::ScrollSliderStyle:
-      aResult = eScrollThumbStyle_Proportional;
-      break;
     case IntID::UseOverlayScrollbars:
     case IntID::AllowOverlayScrollbarsOverlap:
       aResult = NSScroller.preferredScrollerStyle == NSScrollerStyleOverlay;
@@ -445,24 +452,23 @@ nsresult nsLookAndFeel::NativeGetInt(IntID aID, int32_t& aResult) {
     case IntID::TreeScrollLinesMax:
       aResult = 3;
       break;
-    case IntID::MacGraphiteTheme:
-      aResult = NSColor.currentControlTint == NSGraphiteControlTint;
-      break;
     case IntID::MacBigSurTheme:
       aResult = nsCocoaFeatures::OnBigSurOrLater();
       break;
     case IntID::MacRTL:
-      aResult = IsSystemOrientationRTL();
+      EnsureInit();
+      aResult = mRtl;
+      break;
+    case IntID::MacTitlebarHeight:
+      EnsureInit();
+      aResult = mTitlebarHeight;
       break;
     case IntID::AlertNotificationOrigin:
       aResult = NS_ALERT_TOP;
       break;
-    case IntID::TabFocusModel:
-      aResult = [NSApp isFullKeyboardAccessEnabled] ? nsIContent::eTabFocus_any
-                                                    : nsIContent::eTabFocus_textControlsMask;
-      break;
     case IntID::ScrollToClick: {
-      aResult = [[NSUserDefaults standardUserDefaults] boolForKey:@"AppleScrollerPagingBehavior"];
+      aResult = [[NSUserDefaults standardUserDefaults]
+          boolForKey:@"AppleScrollerPagingBehavior"];
     } break;
     case IntID::ChosenMenuItemsShouldBlink:
       aResult = 1;
@@ -471,10 +477,10 @@ nsresult nsLookAndFeel::NativeGetInt(IntID aID, int32_t& aResult) {
     case IntID::IMEConvertedTextUnderlineStyle:
     case IntID::IMESelectedRawTextUnderlineStyle:
     case IntID::IMESelectedConvertedTextUnderline:
-      aResult = NS_STYLE_TEXT_DECORATION_STYLE_SOLID;
+      aResult = static_cast<int32_t>(StyleTextDecorationStyle::Solid);
       break;
     case IntID::SpellCheckerUnderlineStyle:
-      aResult = NS_STYLE_TEXT_DECORATION_STYLE_DOTTED;
+      aResult = static_cast<int32_t>(StyleTextDecorationStyle::Dotted);
       break;
     case IntID::ScrollbarButtonAutoRepeatBehavior:
       aResult = 0;
@@ -492,13 +498,26 @@ nsresult nsLookAndFeel::NativeGetInt(IntID aID, int32_t& aResult) {
       aResult = SystemWantsDarkTheme();
       break;
     case IntID::PrefersReducedMotion:
-      aResult = NSWorkspace.sharedWorkspace.accessibilityDisplayShouldReduceMotion;
+      aResult =
+          NSWorkspace.sharedWorkspace.accessibilityDisplayShouldReduceMotion;
+      break;
+    case IntID::PrefersReducedTransparency:
+      aResult = NSWorkspace.sharedWorkspace
+                    .accessibilityDisplayShouldReduceTransparency;
+      break;
+    case IntID::InvertedColors:
+      aResult =
+          NSWorkspace.sharedWorkspace.accessibilityDisplayShouldInvertColors;
       break;
     case IntID::UseAccessibilityTheme:
-      aResult = NSWorkspace.sharedWorkspace.accessibilityDisplayShouldIncreaseContrast;
+      aResult = NSWorkspace.sharedWorkspace
+                    .accessibilityDisplayShouldIncreaseContrast;
       break;
-    case IntID::VideoDynamicRange:
-      aResult = nsCocoaFeatures::OnBigSurOrLater();
+    case IntID::PanelAnimations:
+      aResult = 1;
+      break;
+    case IntID::FullKeyboardAccess:
+      aResult = NSApp.isFullKeyboardAccessEnabled;
       break;
     default:
       aResult = 0;
@@ -522,7 +541,8 @@ nsresult nsLookAndFeel::NativeGetFloat(FloatID aID, float& aResult) {
       aResult = 2.0f;
       break;
     case FloatID::CursorScale: {
-      id uaDefaults = [[NSUserDefaults alloc] initWithSuiteName:@"com.apple.universalaccess"];
+      id uaDefaults = [[NSUserDefaults alloc]
+          initWithSuiteName:@"com.apple.universalaccess"];
       float f = [uaDefaults floatForKey:@"mouseDriverCursorSize"];
       [uaDefaults release];
       aResult = f > 0.0 ? f : 1.0;  // default to 1.0 if value not available
@@ -538,45 +558,10 @@ nsresult nsLookAndFeel::NativeGetFloat(FloatID aID, float& aResult) {
   NS_OBJC_END_TRY_BLOCK_RETURN(NS_ERROR_FAILURE);
 }
 
-bool nsLookAndFeel::SystemWantsDarkTheme() {
-  // This returns true if the macOS system appearance is set to dark mode on
-  // 10.14+, false otherwise.
-  if (@available(macOS 10.14, *)) {
-    NSAppearanceName aquaOrDarkAqua = [NSApp.effectiveAppearance
-        bestMatchFromAppearancesWithNames:@[ NSAppearanceNameAqua, NSAppearanceNameDarkAqua ]];
-    return [aquaOrDarkAqua isEqualToString:NSAppearanceNameDarkAqua];
-  }
-  return false;
-}
-
-/*static*/
-bool nsLookAndFeel::IsSystemOrientationRTL() {
-  NSWindow* window = [[NSWindow alloc] initWithContentRect:NSZeroRect
-                                                 styleMask:NSWindowStyleMaskBorderless
-                                                   backing:NSBackingStoreBuffered
-                                                     defer:NO];
-  auto direction = window.windowTitlebarLayoutDirection;
-  [window release];
-  return direction == NSUserInterfaceLayoutDirectionRightToLeft;
-}
-
-bool nsLookAndFeel::NativeGetFont(FontID aID, nsString& aFontName, gfxFontStyle& aFontStyle) {
+bool nsLookAndFeel::NativeGetFont(FontID aID, nsString& aFontName,
+                                  gfxFontStyle& aFontStyle) {
   NS_OBJC_BEGIN_TRY_BLOCK_RETURN;
 
-  // hack for now
-  if (aID == FontID::MozWindow || aID == FontID::MozDocument) {
-    aFontStyle.style = mozilla::FontSlantStyle::Normal();
-    aFontStyle.weight = mozilla::FontWeight::Normal();
-    aFontStyle.stretch = mozilla::FontStretch::Normal();
-    aFontStyle.size = 14;
-    aFontStyle.systemFont = true;
-
-    aFontName.AssignLiteral("sans-serif");
-    return true;
-  }
-
-  // TODO: Add caching? Note that it needs to be thread-safe for stylo use.
-
   nsAutoCString name;
   gfxPlatformMac::LookupSystemFont(aID, name, aFontStyle);
   aFontName.Append(NS_ConvertUTF8toUTF16(name));
@@ -586,6 +571,16 @@ bool nsLookAndFeel::NativeGetFont(FontID aID, nsString& aFontName, gfxFontStyle&
   NS_OBJC_END_TRY_BLOCK_RETURN(false);
 }
 
+void nsLookAndFeel::RecordAccessibilityTelemetry() {
+  if ([[NSWorkspace sharedWorkspace]
+          respondsToSelector:@selector
+          (accessibilityDisplayShouldInvertColors)]) {
+    bool val =
+        [[NSWorkspace sharedWorkspace] accessibilityDisplayShouldInvertColors];
+    Telemetry::ScalarSet(Telemetry::ScalarID::A11Y_INVERT_COLORS, val);
+  }
+}
+
 @implementation MOZLookAndFeelDynamicChangeObserver
 
 + (void)startObserving {
@@ -598,33 +593,28 @@ bool nsLookAndFeel::NativeGetFont(FontID aID, nsString& aFontName, gfxFontStyle&
 - (instancetype)init {
   self = [super init];
 
-  [NSNotificationCenter.defaultCenter addObserver:self
-                                         selector:@selector(colorsChanged)
-                                             name:NSControlTintDidChangeNotification
-                                           object:nil];
-  [NSNotificationCenter.defaultCenter addObserver:self
-                                         selector:@selector(colorsChanged)
-                                             name:NSSystemColorsDidChangeNotification
-                                           object:nil];
+  [NSNotificationCenter.defaultCenter
+      addObserver:self
+         selector:@selector(colorsChanged)
+             name:NSControlTintDidChangeNotification
+           object:nil];
+  [NSNotificationCenter.defaultCenter
+      addObserver:self
+         selector:@selector(colorsChanged)
+             name:NSSystemColorsDidChangeNotification
+           object:nil];
 
-  if (@available(macOS 10.14, *)) {
-    [NSWorkspace.sharedWorkspace.notificationCenter
-        addObserver:self
-           selector:@selector(mediaQueriesChanged)
-               name:NSWorkspaceAccessibilityDisplayOptionsDidChangeNotification
-             object:nil];
-  } else {
-    [NSNotificationCenter.defaultCenter
-        addObserver:self
-           selector:@selector(mediaQueriesChanged)
-               name:NSWorkspaceAccessibilityDisplayOptionsDidChangeNotification
-             object:nil];
-  }
+  [NSWorkspace.sharedWorkspace.notificationCenter
+      addObserver:self
+         selector:@selector(mediaQueriesChanged)
+             name:NSWorkspaceAccessibilityDisplayOptionsDidChangeNotification
+           object:nil];
 
-  [NSNotificationCenter.defaultCenter addObserver:self
-                                         selector:@selector(scrollbarsChanged)
-                                             name:NSPreferredScrollerStyleDidChangeNotification
-                                           object:nil];
+  [NSNotificationCenter.defaultCenter
+      addObserver:self
+         selector:@selector(scrollbarsChanged)
+             name:NSPreferredScrollerStyleDidChangeNotification
+           object:nil];
   [NSDistributedNotificationCenter.defaultCenter
              addObserver:self
                 selector:@selector(scrollbarsChanged)
@@ -644,11 +634,10 @@ bool nsLookAndFeel::NativeGetFont(FontID aID, nsString& aFontName, gfxFontStyle&
                   object:nil
       suspensionBehavior:NSNotificationSuspensionBehaviorDeliverImmediately];
 
-  [MOZGlobalAppearance.sharedInstance addObserver:self
-                                       forKeyPath:@"effectiveAppearance"
-                                          options:0
-                                          context:nil];
-  [NSApp addObserver:self forKeyPath:@"effectiveAppearance" options:0 context:nil];
+  [NSApp addObserver:self
+          forKeyPath:@"effectiveAppearance"
+             options:0
+             context:nil];
 
   return self;
 }
@@ -660,7 +649,10 @@ bool nsLookAndFeel::NativeGetFont(FontID aID, nsString& aFontName, gfxFontStyle&
   if ([keyPath isEqualToString:@"effectiveAppearance"]) {
     [self entireThemeChanged];
   } else {
-    [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
+    [super observeValueForKeyPath:keyPath
+                         ofObject:object
+                           change:change
+                          context:context];
   }
 }
 
@@ -673,7 +665,12 @@ bool nsLookAndFeel::NativeGetFont(FontID aID, nsString& aFontName, gfxFontStyle&
 }
 
 - (void)mediaQueriesChanged {
-  LookAndFeel::NotifyChangedAllWindows(widget::ThemeChangeKind::MediaQueriesOnly);
+  // Changing`Invert Colors` sends
+  // AccessibilityDisplayOptionsDidChangeNotifications. We monitor that setting
+  // via telemetry, so call into that recording method here.
+  nsLookAndFeel::RecordAccessibilityTelemetry();
+  LookAndFeel::NotifyChangedAllWindows(
+      widget::ThemeChangeKind::MediaQueriesOnly);
 }
 
 - (void)colorsChanged {
@@ -681,9 +678,12 @@ bool nsLookAndFeel::NativeGetFont(FontID aID, nsString& aFontName, gfxFontStyle&
 }
 
 - (void)cachedValuesChanged {
-  // We only need to re-cache (and broadcast) updated LookAndFeel values, so that they're up-to-date
-  // the next time they're queried. No further change handling is needed.
-  // TODO: Add a change hint for this which avoids the unnecessary media query invalidation.
-  LookAndFeel::NotifyChangedAllWindows(widget::ThemeChangeKind::MediaQueriesOnly);
+  // We only need to re-cache (and broadcast) updated LookAndFeel values, so
+  // that they're up-to-date the next time they're queried. No further change
+  // handling is needed.
+  // TODO: Add a change hint for this which avoids the unnecessary media query
+  // invalidation.
+  LookAndFeel::NotifyChangedAllWindows(
+      widget::ThemeChangeKind::MediaQueriesOnly);
 }
 @end
