@@ -68,32 +68,24 @@
 #  include "Units.h"
 #  include "nsWindow.h"
 extern mozilla::LazyLogModule gWidgetWaylandLog;
+extern mozilla::LazyLogModule gWidgetLog;
 #  define LOGWAYLAND(...) \
     MOZ_LOG(gWidgetWaylandLog, mozilla::LogLevel::Debug, (__VA_ARGS__))
+#  define LOGCONTAINER(...) \
+    MOZ_LOG(gWidgetLog, mozilla::LogLevel::Debug, (__VA_ARGS__))
 #else
 #  define LOGWAYLAND(...)
+#  define LOGCONTAINER(...)
 #endif /* MOZ_LOGGING */
 
 using namespace mozilla;
 using namespace mozilla::widget;
 
-/* init methods */
-static void moz_container_wayland_destroy(GtkWidget* widget);
-
-/* widget class methods */
-static void moz_container_wayland_map(GtkWidget* widget);
-static gboolean moz_container_wayland_map_event(GtkWidget* widget,
-                                                GdkEventAny* event);
-static void moz_container_wayland_unmap(GtkWidget* widget);
-static void moz_container_wayland_size_allocate(GtkWidget* widget,
-                                                GtkAllocation* allocation);
 static bool moz_container_wayland_surface_create_locked(
     const MutexAutoLock& aProofOfLock, MozContainer* container);
 static void moz_container_wayland_set_opaque_region_locked(
-    const MutexAutoLock& aProofOfLock, MozContainer* container);
-
-// Imlemented in MozContainer.cpp
-void moz_container_realize(GtkWidget* widget);
+    const MutexAutoLock& aProofOfLock, MozContainer* container,
+    const LayoutDeviceIntRegion&);
 
 // Invalidate gtk wl_surface to commit changes to wl_subsurface.
 // wl_subsurface changes are effective when parent surface is commited.
@@ -113,7 +105,7 @@ static void moz_container_wayland_invalidate(MozContainer* container) {
 // events from Gtk+.
 static void moz_container_clear_input_region(MozContainer* container) {
   struct wl_compositor* compositor = WaylandDisplayGet()->GetCompositor();
-  MozContainerWayland* wl_container = &container->wl_container;
+  MozContainerWayland* wl_container = &container->data.wl_container;
   wl_region* region = wl_compositor_create_region(compositor);
   wl_surface_set_input_region(wl_container->surface, region);
   wl_region_destroy(region);
@@ -122,10 +114,10 @@ static void moz_container_clear_input_region(MozContainer* container) {
 static void moz_container_wayland_move_locked(const MutexAutoLock& aProofOfLock,
                                               MozContainer* container, int dx,
                                               int dy) {
-  LOGWAYLAND("moz_container_wayland_move [%p] %d,%d\n",
-             (void*)moz_container_get_nsWindow(container), dx, dy);
+  LOGCONTAINER("moz_container_wayland_move [%p] %d,%d\n",
+               (void*)moz_container_get_nsWindow(container), dx, dy);
 
-  MozContainerWayland* wl_container = &container->wl_container;
+  MozContainerWayland* wl_container = &container->data.wl_container;
   if (!wl_container->subsurface || (wl_container->subsurface_dx == dx &&
                                     wl_container->subsurface_dy == dy)) {
     return;
@@ -139,69 +131,63 @@ static void moz_container_wayland_move_locked(const MutexAutoLock& aProofOfLock,
 }
 
 // This is called from layout/compositor code only with
-// size equal to GL rendering context. Otherwise there are
-// rendering artifacts as wl_egl_window size does not match
-// GL rendering pipeline setup.
-void moz_container_wayland_egl_window_set_size(MozContainer* container,
-                                               int width, int height) {
-  MozContainerWayland* wl_container = &container->wl_container;
-  MutexAutoLock lock(*wl_container->container_lock);
-  if (wl_container->eglwindow) {
-    wl_egl_window_resize(wl_container->eglwindow, width, height, 0, 0);
+// size equal to GL rendering context.
+
+// Return false if scale factor doesn't match buffer size.
+// We need to skip painting in such case do avoid Wayland compositor freaking.
+bool moz_container_wayland_egl_window_set_size(MozContainer* container,
+                                               nsIntSize aSize, int aScale) {
+  MozContainerWayland* wl_container = &container->data.wl_container;
+  MutexAutoLock lock(wl_container->container_lock);
+
+  // We may be called after unmap so we're missing egl window completelly.
+  // In such case don't return false which would block compositor.
+  // We return true here and don't block flush WebRender queue.
+  // We'll be repainted if our window become visible again anyway.
+  if (!wl_container->eglwindow) {
+    return true;
   }
-}
 
-void moz_container_wayland_class_init(MozContainerClass* klass) {
-  /*GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
-    GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass); */
-  GtkWidgetClass* widget_class = GTK_WIDGET_CLASS(klass);
-
-  widget_class->map = moz_container_wayland_map;
-  widget_class->map_event = moz_container_wayland_map_event;
-  widget_class->destroy = moz_container_wayland_destroy;
-  widget_class->unmap = moz_container_wayland_unmap;
-  widget_class->realize = moz_container_realize;
-  widget_class->size_allocate = moz_container_wayland_size_allocate;
-}
-
-void moz_container_wayland_init(MozContainerWayland* container) {
-  container->surface = nullptr;
-  container->subsurface = nullptr;
-  container->eglwindow = nullptr;
-  container->frame_callback_handler = nullptr;
-  container->viewport = nullptr;
-  container->ready_to_draw = false;
-  container->opaque_region_needs_updates = false;
-  container->opaque_region_corner_radius = 0;
-  container->opaque_region_used = false;
-  container->subsurface_dx = 0;
-  container->subsurface_dy = 0;
-  container->before_first_size_alloc = true;
-  container->buffer_scale = 1;
-  container->initial_draw_cbs.clear();
-  container->container_lock = new mozilla::Mutex("MozContainer lock");
-  container->commit_to_parent = false;
-}
-
-static void moz_container_wayland_destroy(GtkWidget* widget) {
-  MozContainerWayland* container = &MOZ_CONTAINER(widget)->wl_container;
-  if (!container->container_lock) {
-    // moz_container_wayland_init was not called - it's a hidden container.
-    return;
+  if (wl_container->buffer_scale != aScale) {
+    moz_container_wayland_set_scale_factor_locked(lock, container, aScale);
   }
-  moz_container_wayland_clear_initial_draw_callback(MOZ_CONTAINER(widget));
-  delete container->container_lock;
-  container->container_lock = nullptr;
+
+  /* Enable for size changes logging
+    LOGCONTAINER(
+        "moz_container_wayland_egl_window_set_size [%p] %d x %d scale %d "
+        "(unscaled %d x %d)",
+        (void*)moz_container_get_nsWindow(container), aSize.width, aSize.height,
+        aScale, aSize.width / aScale, aSize.height / aScale);
+  */
+  wl_egl_window_resize(wl_container->eglwindow, aSize.width, aSize.height, 0,
+                       0);
+
+  return moz_container_wayland_size_matches_scale_factor_locked(
+      lock, container, aSize.width, aSize.height);
 }
 
-void moz_container_wayland_add_initial_draw_callback(
+void moz_container_wayland_add_initial_draw_callback_locked(
     MozContainer* container, const std::function<void(void)>& initial_draw_cb) {
-  MozContainerWayland* wl_container = &MOZ_CONTAINER(container)->wl_container;
+  MozContainerWayland* wl_container = &container->data.wl_container;
+
+  if (wl_container->ready_to_draw && !wl_container->surface) {
+    NS_WARNING(
+        "moz_container_wayland_add_or_fire_initial_draw_callback:"
+        " ready to draw without wayland surface!");
+  }
+  MOZ_DIAGNOSTIC_ASSERT(!wl_container->ready_to_draw || !wl_container->surface);
+  wl_container->initial_draw_cbs.push_back(initial_draw_cb);
+}
+
+void moz_container_wayland_add_or_fire_initial_draw_callback(
+    MozContainer* container, const std::function<void(void)>& initial_draw_cb) {
+  MozContainerWayland* wl_container = &container->data.wl_container;
   {
-    MutexAutoLock lock(*container->wl_container.container_lock);
+    MutexAutoLock lock(wl_container->container_lock);
     if (wl_container->ready_to_draw && !wl_container->surface) {
       NS_WARNING(
-          "moz_container_wayland_add_initial_draw_callback: ready to draw "
+          "moz_container_wayland_add_or_fire_initial_draw_callback: ready to "
+          "draw "
           "without wayland surface!");
     }
     if (!wl_container->ready_to_draw || !wl_container->surface) {
@@ -218,20 +204,20 @@ void moz_container_wayland_add_initial_draw_callback(
 
 static void moz_container_wayland_clear_initial_draw_callback_locked(
     const MutexAutoLock& aProofOfLock, MozContainer* container) {
-  MozContainerWayland* wl_container = &MOZ_CONTAINER(container)->wl_container;
-  g_clear_pointer(&wl_container->frame_callback_handler, wl_callback_destroy);
+  MozContainerWayland* wl_container = &container->data.wl_container;
+  MozClearPointer(wl_container->frame_callback_handler, wl_callback_destroy);
   wl_container->initial_draw_cbs.clear();
 }
 
 void moz_container_wayland_clear_initial_draw_callback(
     MozContainer* container) {
-  MutexAutoLock lock(*container->wl_container.container_lock);
+  MutexAutoLock lock(container->data.wl_container.container_lock);
   moz_container_wayland_clear_initial_draw_callback_locked(lock, container);
 }
 
 static void moz_container_wayland_frame_callback_handler(
     void* data, struct wl_callback* callback, uint32_t time) {
-  MozContainerWayland* wl_container = &MOZ_CONTAINER(data)->wl_container;
+  MozContainerWayland* wl_container = MOZ_WL_CONTAINER(data);
 
   LOGWAYLAND(
       "%s [%p] frame_callback_handler %p ready_to_draw %d (set to true)"
@@ -243,8 +229,8 @@ static void moz_container_wayland_frame_callback_handler(
   std::vector<std::function<void(void)>> cbs;
   {
     // Protect mozcontainer internals changes by container_lock.
-    MutexAutoLock lock(*wl_container->container_lock);
-    g_clear_pointer(&wl_container->frame_callback_handler, wl_callback_destroy);
+    MutexAutoLock lock(wl_container->container_lock);
+    MozClearPointer(wl_container->frame_callback_handler, wl_callback_destroy);
     // It's possible that container is already unmapped so quit in such case.
     if (!wl_container->surface) {
       LOGWAYLAND("  container is unmapped, quit.");
@@ -262,7 +248,7 @@ static void moz_container_wayland_frame_callback_handler(
   }
 
   // Call the callbacks registered by
-  // moz_container_wayland_add_initial_draw_callback().
+  // moz_container_wayland_add_or_fire_initial_draw_callback().
   // and we can't do that under mozcontainer lock.
   for (auto const& cb : cbs) {
     cb();
@@ -274,10 +260,10 @@ static const struct wl_callback_listener moz_container_frame_listener = {
 
 static void after_frame_clock_after_paint(GdkFrameClock* clock,
                                           MozContainer* container) {
-  struct wl_surface* surface = moz_container_wayland_surface_lock(container);
+  MozContainerSurfaceLock lock(container);
+  struct wl_surface* surface = lock.GetSurface();
   if (surface) {
     wl_surface_commit(surface);
-    moz_container_wayland_surface_unlock(container, &surface);
   }
 }
 
@@ -293,7 +279,7 @@ static bool moz_gdk_wayland_window_add_frame_callback_surface_locked(
   }
 
   GdkWindow* window = gtk_widget_get_window(GTK_WIDGET(container));
-  MozContainerWayland* wl_container = &container->wl_container;
+  MozContainerWayland* wl_container = &container->data.wl_container;
 
   sGdkWaylandWindowAddCallbackSurface(window, wl_container->surface);
 
@@ -314,7 +300,7 @@ static void moz_gdk_wayland_window_remove_frame_callback_surface_locked(
   }
 
   GdkWindow* window = gtk_widget_get_window(GTK_WIDGET(container));
-  MozContainerWayland* wl_container = &container->wl_container;
+  MozContainerWayland* wl_container = &container->data.wl_container;
 
   if (wl_container->surface) {
     sGdkWaylandWindowRemoveCallbackSurface(window, wl_container->surface);
@@ -325,12 +311,18 @@ static void moz_gdk_wayland_window_remove_frame_callback_surface_locked(
       frame_clock, FuncToGpointer(after_frame_clock_after_paint), container);
 }
 
-static void moz_container_wayland_unmap_internal(MozContainer* container) {
-  MozContainerWayland* wl_container = &container->wl_container;
-  MutexAutoLock lock(*wl_container->container_lock);
+void moz_container_wayland_unmap(GtkWidget* widget) {
+  g_return_if_fail(IS_MOZ_CONTAINER(widget));
 
-  LOGWAYLAND("%s [%p]\n", __FUNCTION__,
-             (void*)moz_container_get_nsWindow(container));
+  // Unmap MozContainer first so we can remove our resources
+  moz_container_unmap(widget);
+
+  MozContainer* container = MOZ_CONTAINER(widget);
+  MozContainerWayland* wl_container = &container->data.wl_container;
+  MutexAutoLock lock(wl_container->container_lock);
+
+  LOGCONTAINER("%s [%p]\n", __FUNCTION__,
+               (void*)moz_container_get_nsWindow(container));
 
   moz_container_wayland_clear_initial_draw_callback_locked(lock, container);
 
@@ -342,26 +334,52 @@ static void moz_container_wayland_unmap_internal(MozContainer* container) {
     wl_container->surface = nullptr;
   }
 
-  g_clear_pointer(&wl_container->eglwindow, wl_egl_window_destroy);
-  g_clear_pointer(&wl_container->subsurface, wl_subsurface_destroy);
-  g_clear_pointer(&wl_container->surface, wl_surface_destroy);
-  g_clear_pointer(&wl_container->viewport, wp_viewport_destroy);
+  MozClearPointer(wl_container->eglwindow, wl_egl_window_destroy);
+  MozClearPointer(wl_container->subsurface, wl_subsurface_destroy);
+  MozClearPointer(wl_container->surface, wl_surface_destroy);
+  MozClearPointer(wl_container->viewport, wp_viewport_destroy);
+  MozClearPointer(wl_container->fractional_scale,
+                  wp_fractional_scale_v1_destroy);
 
   wl_container->ready_to_draw = false;
   wl_container->buffer_scale = 1;
+  wl_container->current_fractional_scale = 0.0;
 }
 
-static gboolean moz_container_wayland_map_event(GtkWidget* widget,
-                                                GdkEventAny* event) {
-  MozContainerWayland* wl_container = &MOZ_CONTAINER(widget)->wl_container;
-  MutexAutoLock lock(*wl_container->container_lock);
+gboolean moz_container_wayland_map_event(GtkWidget* widget,
+                                         GdkEventAny* event) {
+  MozContainerWayland* wl_container = &MOZ_CONTAINER(widget)->data.wl_container;
 
-  LOGWAYLAND("%s [%p]\n", __FUNCTION__,
-             (void*)moz_container_get_nsWindow(MOZ_CONTAINER(widget)));
+  LOGCONTAINER("%s [%p]\n", __FUNCTION__,
+               (void*)moz_container_get_nsWindow(MOZ_CONTAINER(widget)));
 
-  // We need to mark MozContainer as mapped to make sure
-  // moz_container_wayland_unmap() is called on hide/withdraw.
-  gtk_widget_set_mapped(widget, TRUE);
+  // Return early if we're not mapped. Gtk may send bogus map_event signal
+  // to unmapped widgets (see Bug 1875369).
+  if (!gtk_widget_get_mapped(widget)) {
+    return false;
+  }
+
+  // Make sure we're on main thread as we can't lock mozContainer here
+  // due to moz_container_wayland_add_or_fire_initial_draw_callback() call
+  // below.
+  MOZ_DIAGNOSTIC_ASSERT(NS_IsMainThread());
+
+  // Set waiting_to_show flag. It means the mozcontainer is cofigured/mapped
+  // and it's supposed to be visible. *But* it's really visible when we get
+  // moz_container_wayland_add_or_fire_initial_draw_callback() which means
+  // wayland compositor makes it live.
+  wl_container->waiting_to_show = true;
+  MozContainer* container = MOZ_CONTAINER(widget);
+  moz_container_wayland_add_or_fire_initial_draw_callback(
+      container, [container]() -> void {
+        LOGCONTAINER(
+            "[%p] moz_container_wayland_add_or_fire_initial_draw_callback set "
+            "visible",
+            moz_container_get_nsWindow(container));
+        moz_container_wayland_clear_waiting_to_show_flag(container);
+      });
+
+  MutexAutoLock lock(wl_container->container_lock);
 
   // Don't create wl_subsurface in map_event when it's already created or
   // if we create it for the first time.
@@ -376,55 +394,32 @@ static gboolean moz_container_wayland_map_event(GtkWidget* widget,
     }
   }
 
-  moz_container_wayland_set_scale_factor_locked(MOZ_CONTAINER(widget));
-  moz_container_wayland_set_opaque_region_locked(lock, MOZ_CONTAINER(widget));
+  nsWindow* window = moz_container_get_nsWindow(MOZ_CONTAINER(widget));
+  moz_container_wayland_set_scale_factor_locked(lock, MOZ_CONTAINER(widget),
+                                                window->GdkCeiledScaleFactor());
+  if (container->data.wl_container.opaque_region_needs_updates) {
+    moz_container_wayland_set_opaque_region_locked(lock, container,
+                                                   window->GetOpaqueRegion());
+  }
   moz_container_clear_input_region(MOZ_CONTAINER(widget));
   moz_container_wayland_invalidate(MOZ_CONTAINER(widget));
   return FALSE;
 }
 
-void moz_container_wayland_map(GtkWidget* widget) {
-  LOGWAYLAND("%s [%p]\n", __FUNCTION__,
-             (void*)moz_container_get_nsWindow(MOZ_CONTAINER(widget)));
-
-  g_return_if_fail(IS_MOZ_CONTAINER(widget));
-  gtk_widget_set_mapped(widget, TRUE);
-
-  if (gtk_widget_get_has_window(widget)) {
-    gdk_window_show(gtk_widget_get_window(widget));
-  }
-}
-
-void moz_container_wayland_unmap(GtkWidget* widget) {
-  LOGWAYLAND("%s [%p]\n", __FUNCTION__,
-             (void*)moz_container_get_nsWindow(MOZ_CONTAINER(widget)));
-
-  g_return_if_fail(IS_MOZ_CONTAINER(widget));
-  gtk_widget_set_mapped(widget, FALSE);
-
-  if (gtk_widget_get_has_window(widget)) {
-    gdk_window_hide(gtk_widget_get_window(widget));
-    moz_container_wayland_unmap_internal(MOZ_CONTAINER(widget));
-  }
-}
-
 void moz_container_wayland_size_allocate(GtkWidget* widget,
                                          GtkAllocation* allocation) {
-  MozContainer* container;
   GtkAllocation tmp_allocation;
 
   g_return_if_fail(IS_MOZ_CONTAINER(widget));
 
-  LOGWAYLAND("moz_container_wayland_size_allocate [%p] %d,%d -> %d x %d\n",
-             (void*)moz_container_get_nsWindow(MOZ_CONTAINER(widget)),
-             allocation->x, allocation->y, allocation->width,
-             allocation->height);
+  LOGCONTAINER("moz_container_wayland_size_allocate [%p] %d,%d -> %d x %d\n",
+               (void*)moz_container_get_nsWindow(MOZ_CONTAINER(widget)),
+               allocation->x, allocation->y, allocation->width,
+               allocation->height);
 
   /* short circuit if you can */
-  container = MOZ_CONTAINER(widget);
   gtk_widget_get_allocation(widget, &tmp_allocation);
-  if (!container->children && tmp_allocation.x == allocation->x &&
-      tmp_allocation.y == allocation->y &&
+  if (tmp_allocation.x == allocation->x && tmp_allocation.y == allocation->y &&
       tmp_allocation.width == allocation->width &&
       tmp_allocation.height == allocation->height) {
     return;
@@ -438,123 +433,152 @@ void moz_container_wayland_size_allocate(GtkWidget* widget,
     // We need to position our subsurface according to GdkWindow
     // when offset changes (GdkWindow is maximized for instance).
     // see gtk-clutter-embed.c for reference.
-    MutexAutoLock lock(*container->wl_container.container_lock);
-    if (!container->wl_container.surface) {
+    MozContainer* container = MOZ_CONTAINER(widget);
+    MutexAutoLock lock(container->data.wl_container.container_lock);
+    if (!container->data.wl_container.surface) {
       if (!moz_container_wayland_surface_create_locked(lock, container)) {
         return;
       }
     }
-    moz_container_wayland_set_scale_factor_locked(container);
-    moz_container_wayland_set_opaque_region_locked(lock, container);
+    nsWindow* window = moz_container_get_nsWindow(container);
+    moz_container_wayland_set_scale_factor_locked(
+        lock, container, window->GdkCeiledScaleFactor());
+    if (container->data.wl_container.opaque_region_needs_updates) {
+      moz_container_wayland_set_opaque_region_locked(lock, container,
+                                                     window->GetOpaqueRegion());
+    }
     moz_container_wayland_move_locked(lock, container, allocation->x,
                                       allocation->y);
     moz_container_clear_input_region(container);
     moz_container_wayland_invalidate(MOZ_CONTAINER(widget));
-    container->wl_container.before_first_size_alloc = false;
+    container->data.wl_container.before_first_size_alloc = false;
   }
 }
 
-static wl_region* moz_container_wayland_create_opaque_region(
-    int aX, int aY, int aWidth, int aHeight, int aCornerRadius) {
-  struct wl_compositor* compositor = WaylandDisplayGet()->GetCompositor();
-  wl_region* region = wl_compositor_create_region(compositor);
-  wl_region_add(region, aX, aY, aWidth, aHeight);
-  if (aCornerRadius) {
-    wl_region_subtract(region, aX, aY, aCornerRadius, aCornerRadius);
-    wl_region_subtract(region, aX + aWidth - aCornerRadius, aY, aCornerRadius,
-                       aCornerRadius);
-  }
-  return region;
-}
-
 static void moz_container_wayland_set_opaque_region_locked(
-    const MutexAutoLock& aProofOfLock, MozContainer* container) {
-  MozContainerWayland* wl_container = &container->wl_container;
-
-  if (!wl_container->opaque_region_needs_updates) {
+    const MutexAutoLock& aProofOfLock, MozContainer* container,
+    const LayoutDeviceIntRegion& aRegion) {
+  MozContainerWayland* wl_container = &container->data.wl_container;
+  MOZ_ASSERT(wl_container->opaque_region_needs_updates);
+  if (!wl_container->surface) {
     return;
   }
 
+  wl_container->opaque_region_needs_updates = false;
   if (!wl_container->opaque_region_used) {
-    wl_container->opaque_region_needs_updates = false;
     return;
   }
 
-  GtkAllocation allocation;
-  gtk_widget_get_allocation(GTK_WIDGET(container), &allocation);
-
-  wl_region* region = moz_container_wayland_create_opaque_region(
-      0, 0, allocation.width, allocation.height,
-      wl_container->opaque_region_corner_radius);
+  wl_region* region =
+      wl_compositor_create_region(WaylandDisplayGet()->GetCompositor());
+  for (auto iter = aRegion.RectIter(); !iter.Done(); iter.Next()) {
+    const auto& rect = iter.Get();
+    wl_region_add(region, rect.x, rect.y, rect.Width(), rect.Height());
+  }
   wl_surface_set_opaque_region(wl_container->surface, region);
   wl_region_destroy(region);
-  wl_container->opaque_region_needs_updates = false;
 }
 
-static void moz_container_wayland_set_opaque_region(MozContainer* container) {
-  MozContainerWayland* wl_container = &container->wl_container;
-  MutexAutoLock lock(*wl_container->container_lock);
-  if (wl_container->surface) {
-    moz_container_wayland_set_opaque_region_locked(lock, container);
+static void moz_container_wayland_surface_set_scale_locked(
+    const MutexAutoLock& aProofOfLock, MozContainerWayland* wl_container,
+    int scale) {
+  if (!wl_container->surface) {
+    return;
   }
+  if (wl_container->buffer_scale == scale) {
+    return;
+  }
+
+  LOGCONTAINER("%s scale %d\n", __FUNCTION__, scale);
+
+  // There is a chance that the attached wl_buffer has not yet been doubled
+  // on the main thread when scale factor changed to 2. This leads to
+  // crash with the following message:
+  // Buffer size (AxB) must be an integer multiple of the buffer_scale (2)
+  // Removing the possibly wrong wl_buffer to prevent that crash:
+  wl_surface_attach(wl_container->surface, nullptr, 0, 0);
+  wl_surface_set_buffer_scale(wl_container->surface, scale);
+  wl_container->buffer_scale = scale;
 }
 
-void moz_container_wayland_set_scale_factor_locked(MozContainer* container) {
+static void fractional_scale_handle_preferred_scale(
+    void* data, struct wp_fractional_scale_v1* info, uint32_t wire_scale) {
+  MozContainer* container = MOZ_CONTAINER(data);
+  MozContainerWayland* wl_container = &container->data.wl_container;
+  wl_container->current_fractional_scale = wire_scale / 120.0;
+
+  RefPtr<nsWindow> window = moz_container_get_nsWindow(container);
+  LOGWAYLAND("%s [%p] scale: %f\n", __func__, window.get(),
+             wl_container->current_fractional_scale);
+  MOZ_DIAGNOSTIC_ASSERT(window);
+  window->OnScaleChanged(/* aNotify = */ true);
+}
+
+static const struct wp_fractional_scale_v1_listener fractional_scale_listener =
+    {
+        .preferred_scale = fractional_scale_handle_preferred_scale,
+};
+
+void moz_container_wayland_set_scale_factor_locked(
+    const MutexAutoLock& aProofOfLock, MozContainer* container, int aScale) {
   if (gfx::gfxVars::UseWebRenderCompositor()) {
     // the compositor backend handles scaling itself
     return;
   }
 
-  MozContainerWayland* wl_container = &container->wl_container;
-  wl_container->container_lock->AssertCurrentThreadOwns();
+  MozContainerWayland* wl_container = &container->data.wl_container;
+  wl_container->container_lock.AssertCurrentThreadOwns();
 
-  nsWindow* window = moz_container_get_nsWindow(container);
-
-  if (window && window->UseFractionalScale()) {
-    if (!wl_container->viewport) {
-      wl_container->viewport = wp_viewporter_get_viewport(
-          WaylandDisplayGet()->GetViewporter(), wl_container->surface);
+  if (StaticPrefs::widget_wayland_fractional_scale_enabled_AtStartup()) {
+    if (!wl_container->fractional_scale) {
+      if (auto* manager = WaylandDisplayGet()->GetFractionalScaleManager()) {
+        wl_container->fractional_scale =
+            wp_fractional_scale_manager_v1_get_fractional_scale(
+                manager, wl_container->surface);
+        wp_fractional_scale_v1_add_listener(wl_container->fractional_scale,
+                                            &fractional_scale_listener,
+                                            container);
+      }
     }
 
-    GdkWindow* gdkWindow = gtk_widget_get_window(GTK_WIDGET(container));
-    wp_viewport_set_destination(wl_container->viewport,
-                                gdk_window_get_width(gdkWindow),
-                                gdk_window_get_height(gdkWindow));
-  } else {
-    int scale = window ? window->GdkCeiledScaleFactor() : 1;
-
-    if (scale == wl_container->buffer_scale) {
-      return;
+    if (wl_container->fractional_scale) {
+      if (!wl_container->viewport) {
+        if (auto* viewporter = WaylandDisplayGet()->GetViewporter()) {
+          wl_container->viewport =
+              wp_viewporter_get_viewport(viewporter, wl_container->surface);
+        }
+      }
+      if (wl_container->viewport) {
+        GdkWindow* gdkWindow = gtk_widget_get_window(GTK_WIDGET(container));
+        wp_viewport_set_destination(wl_container->viewport,
+                                    gdk_window_get_width(gdkWindow),
+                                    gdk_window_get_height(gdkWindow));
+        return;
+      }
     }
-
-    LOGWAYLAND("%s [%p] scale %d\n", __FUNCTION__,
-               (void*)moz_container_get_nsWindow(container), scale);
-    // There is a chance that the attached wl_buffer has not yet been doubled
-    // on the main thread when scale factor changed to 2. This leads to
-    // crash with the following message:
-    // Buffer size (AxB) must be an integer multiple of the buffer_scale (2)
-    // Removing the possibly wrong wl_buffer to prevent that crash:
-    wl_surface_attach(wl_container->surface, nullptr, 0, 0);
-    wl_surface_set_buffer_scale(wl_container->surface, scale);
-    wl_container->buffer_scale = scale;
   }
+
+  moz_container_wayland_surface_set_scale_locked(aProofOfLock, wl_container,
+                                                 aScale);
 }
 
-void moz_container_wayland_set_scale_factor(MozContainer* container) {
-  MutexAutoLock lock(*container->wl_container.container_lock);
-  if (container->wl_container.surface) {
-    moz_container_wayland_set_scale_factor_locked(container);
-  }
+bool moz_container_wayland_size_matches_scale_factor_locked(
+    const MutexAutoLock& aProofOfLock, MozContainer* container, int aWidth,
+    int aHeight) {
+  return aWidth % container->data.wl_container.buffer_scale == 0 &&
+         aHeight % container->data.wl_container.buffer_scale == 0;
 }
 
 static bool moz_container_wayland_surface_create_locked(
     const MutexAutoLock& aProofOfLock, MozContainer* container) {
-  MozContainerWayland* wl_container = &container->wl_container;
+  MozContainerWayland* wl_container = &container->data.wl_container;
 
   LOGWAYLAND("%s [%p]\n", __FUNCTION__,
              (void*)moz_container_get_nsWindow(container));
 
   GdkWindow* window = gtk_widget_get_window(GTK_WIDGET(container));
+  MOZ_DIAGNOSTIC_ASSERT(window);
+
   wl_surface* parent_surface = gdk_wayland_window_get_wl_surface(window);
   if (!parent_surface) {
     LOGWAYLAND("    Failed - missing parent surface!");
@@ -584,7 +608,7 @@ static bool moz_container_wayland_surface_create_locked(
       wl_subcompositor_get_subsurface(WaylandDisplayGet()->GetSubcompositor(),
                                       wl_container->surface, parent_surface);
   if (!wl_container->subsurface) {
-    g_clear_pointer(&wl_container->surface, wl_surface_destroy);
+    MozClearPointer(wl_container->surface, wl_surface_destroy);
     LOGWAYLAND("    Failed - can't create sub-surface!");
     return false;
   }
@@ -602,7 +626,7 @@ static bool moz_container_wayland_surface_create_locked(
   // If there's pending frame callback it's for wrong parent surface,
   // so delete it.
   if (wl_container->frame_callback_handler) {
-    g_clear_pointer(&wl_container->frame_callback_handler, wl_callback_destroy);
+    MozClearPointer(wl_container->frame_callback_handler, wl_callback_destroy);
   }
   wl_container->frame_callback_handler = wl_surface_frame(parent_surface);
   wl_callback_add_listener(wl_container->frame_callback_handler,
@@ -624,115 +648,125 @@ static bool moz_container_wayland_surface_create_locked(
 }
 
 struct wl_surface* moz_container_wayland_surface_lock(MozContainer* container)
-    NO_THREAD_SAFETY_ANALYSIS {
+    MOZ_NO_THREAD_SAFETY_ANALYSIS {
   // LOGWAYLAND("%s [%p] surface %p ready_to_draw %d\n", __FUNCTION__,
-  //           (void*)container, (void*)container->wl_container.surface,
-  //           container->wl_container.ready_to_draw);
-  if (!container->wl_container.surface ||
-      !container->wl_container.ready_to_draw) {
+  //           (void*)container, (void*)container->data.wl_container.surface,
+  //           container->data.wl_container.ready_to_draw);
+  container->data.wl_container.container_lock.Lock();
+  if (!container->data.wl_container.surface ||
+      !container->data.wl_container.ready_to_draw) {
     return nullptr;
   }
-  container->wl_container.container_lock->Lock();
-  return container->wl_container.surface;
+  return container->data.wl_container.surface;
 }
 
 void moz_container_wayland_surface_unlock(MozContainer* container,
                                           struct wl_surface** surface)
-    NO_THREAD_SAFETY_ANALYSIS {
+    MOZ_NO_THREAD_SAFETY_ANALYSIS {
   // Temporarily disabled to avoid log noise
   // LOGWAYLAND("%s [%p] surface %p\n", __FUNCTION__, (void*)container,
-  //            (void*)container->wl_container.surface);
+  //            (void*)container->data.wl_container.surface);
   if (*surface) {
-    container->wl_container.container_lock->Unlock();
     *surface = nullptr;
   }
-}
-
-struct wl_surface* moz_container_wayland_get_surface_locked(
-    const MutexAutoLock& aProofOfLock, MozContainer* container) {
-  LOGWAYLAND("%s [%p] surface %p ready_to_draw %d\n", __FUNCTION__,
-             (void*)moz_container_get_nsWindow(container),
-             (void*)container->wl_container.surface,
-             container->wl_container.ready_to_draw);
-  if (!container->wl_container.surface ||
-      !container->wl_container.ready_to_draw) {
-    return nullptr;
-  }
-  moz_container_wayland_set_scale_factor_locked(container);
-  return container->wl_container.surface;
-}
-
-void moz_container_wayland_lock(MozContainer* container)
-    NO_THREAD_SAFETY_ANALYSIS {
-  container->wl_container.container_lock->Lock();
-}
-
-void moz_container_wayland_unlock(MozContainer* container)
-    NO_THREAD_SAFETY_ANALYSIS {
-  container->wl_container.container_lock->Unlock();
+  container->data.wl_container.container_lock.Unlock();
 }
 
 struct wl_egl_window* moz_container_wayland_get_egl_window(
     MozContainer* container, double scale) {
-  MozContainerWayland* wl_container = &container->wl_container;
+  MozContainerWayland* wl_container = &container->data.wl_container;
 
-  LOGWAYLAND("%s [%p] eglwindow %p\n", __FUNCTION__,
-             (void*)moz_container_get_nsWindow(container),
-             (void*)wl_container->eglwindow);
+  LOGCONTAINER("%s [%p] eglwindow %p scale %d\n", __FUNCTION__,
+               (void*)moz_container_get_nsWindow(container),
+               (void*)wl_container->eglwindow, (int)scale);
 
-  MutexAutoLock lock(*wl_container->container_lock);
+  MutexAutoLock lock(wl_container->container_lock);
   if (!wl_container->surface || !wl_container->ready_to_draw) {
+    LOGCONTAINER(
+        "  quit, wl_container->surface %p wl_container->ready_to_draw %d\n",
+        wl_container->surface, wl_container->ready_to_draw);
     return nullptr;
   }
-  if (!wl_container->eglwindow) {
-    GdkWindow* window = gtk_widget_get_window(GTK_WIDGET(container));
-    wl_container->eglwindow = wl_egl_window_create(
-        wl_container->surface, (int)round(gdk_window_get_width(window) * scale),
-        (int)round(gdk_window_get_height(window) * scale));
 
-    LOGWAYLAND("%s [%p] created eglwindow %p\n", __FUNCTION__,
-               (void*)moz_container_get_nsWindow(container),
-               (void*)wl_container->eglwindow);
+  GdkWindow* window = gtk_widget_get_window(GTK_WIDGET(container));
+  nsIntSize requestedSize((int)round(gdk_window_get_width(window) * scale),
+                          (int)round(gdk_window_get_height(window) * scale));
+
+  if (!wl_container->eglwindow) {
+    wl_container->eglwindow = wl_egl_window_create(
+        wl_container->surface, requestedSize.width, requestedSize.height);
+
+    LOGCONTAINER("%s [%p] created eglwindow %p size %d x %d (with scale %f)\n",
+                 __FUNCTION__, (void*)moz_container_get_nsWindow(container),
+                 (void*)wl_container->eglwindow, requestedSize.width,
+                 requestedSize.height, scale);
+  } else {
+    nsIntSize recentSize;
+    wl_egl_window_get_attached_size(wl_container->eglwindow, &recentSize.width,
+                                    &recentSize.height);
+    if (requestedSize != recentSize || wl_container->buffer_scale != scale) {
+      LOGCONTAINER("%s [%p] resized to %d x %d (with scale %f)\n", __FUNCTION__,
+                   (void*)moz_container_get_nsWindow(container),
+                   requestedSize.width, requestedSize.height, scale);
+      wl_egl_window_resize(wl_container->eglwindow, requestedSize.width,
+                           requestedSize.height, 0, 0);
+    }
   }
+  moz_container_wayland_surface_set_scale_locked(lock, wl_container,
+                                                 static_cast<int>(scale));
   return wl_container->eglwindow;
 }
 
 gboolean moz_container_wayland_has_egl_window(MozContainer* container) {
-  return container->wl_container.eglwindow != nullptr;
+  return !!container->data.wl_container.eglwindow;
 }
 
-void moz_container_wayland_update_opaque_region(MozContainer* container,
-                                                int corner_radius) {
-  MozContainerWayland* wl_container = &container->wl_container;
+void moz_container_wayland_update_opaque_region(MozContainer* container) {
+  MozContainerWayland* wl_container = &container->data.wl_container;
+  MutexAutoLock lock(wl_container->container_lock);
   wl_container->opaque_region_needs_updates = true;
-  wl_container->opaque_region_corner_radius = corner_radius;
 
   // When GL compositor / WebRender is used,
   // moz_container_wayland_get_egl_window() is called only once when window
   // is created or resized so update opaque region now.
   if (moz_container_wayland_has_egl_window(container)) {
-    moz_container_wayland_set_opaque_region(container);
+    nsWindow* window = moz_container_get_nsWindow(container);
+    moz_container_wayland_set_opaque_region_locked(lock, container,
+                                                   window->GetOpaqueRegion());
   }
 }
 
 gboolean moz_container_wayland_can_draw(MozContainer* container) {
-  MozContainerWayland* wl_container = &container->wl_container;
-  MutexAutoLock lock(*wl_container->container_lock);
+  MozContainerWayland* wl_container = &container->data.wl_container;
+  MutexAutoLock lock(wl_container->container_lock);
   return wl_container->ready_to_draw;
 }
 
+double moz_container_wayland_get_fractional_scale(MozContainer* container) {
+  return container->data.wl_container.current_fractional_scale;
+}
+
 double moz_container_wayland_get_scale(MozContainer* container) {
   nsWindow* window = moz_container_get_nsWindow(container);
-  return window ? window->FractionalScaleFactor() : 1;
+  return window ? window->FractionalScaleFactor() : 1.0;
 }
 
 void moz_container_wayland_set_commit_to_parent(MozContainer* container) {
-  MozContainerWayland* wl_container = &container->wl_container;
+  MozContainerWayland* wl_container = &container->data.wl_container;
   MOZ_DIAGNOSTIC_ASSERT(!wl_container->surface);
   wl_container->commit_to_parent = true;
 }
 
 bool moz_container_wayland_is_commiting_to_parent(MozContainer* container) {
-  MozContainerWayland* wl_container = &container->wl_container;
-  return wl_container->commit_to_parent;
+  return container->data.wl_container.commit_to_parent;
+}
+
+bool moz_container_wayland_is_waiting_to_show(MozContainer* container) {
+  MOZ_DIAGNOSTIC_ASSERT(NS_IsMainThread());
+  return container->data.wl_container.waiting_to_show;
+}
+
+void moz_container_wayland_clear_waiting_to_show_flag(MozContainer* container) {
+  MOZ_DIAGNOSTIC_ASSERT(NS_IsMainThread());
+  container->data.wl_container.waiting_to_show = false;
 }
