@@ -5,20 +5,13 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-#include <unistd.h>
-
 #ifndef __WakeLockListener_h__
-#  define __WakeLockListener_h__
+#define __WakeLockListener_h__
 
-#  include "mozilla/StaticPtr.h"
-#  include "nsHashKeys.h"
-#  include "nsClassHashtable.h"
+#include "nsHashKeys.h"
+#include "nsRefPtrHashtable.h"
 
-#  include "nsIDOMWakeLockListener.h"
-
-#  ifdef MOZ_ENABLE_DBUS
-#    include "mozilla/DBusHelpers.h"
-#  endif
+#include "nsIDOMWakeLockListener.h"
 
 class WakeLockTopic;
 
@@ -30,26 +23,16 @@ class WakeLockListener final : public nsIDOMMozWakeLockListener {
  public:
   NS_DECL_ISUPPORTS;
 
-  static WakeLockListener* GetSingleton(bool aCreate = true);
-  static void Shutdown();
+  nsresult Callback(const nsAString& topic, const nsAString& state) override;
 
-  virtual nsresult Callback(const nsAString& topic,
-                            const nsAString& state) override;
+  WakeLockListener();
 
  private:
-  WakeLockListener();
-  ~WakeLockListener() = default;
+  ~WakeLockListener();
 
-  bool EnsureDBusConnection();
-
-  static mozilla::StaticRefPtr<WakeLockListener> sSingleton;
-
-#  ifdef MOZ_ENABLE_DBUS
-  RefPtr<DBusConnection> mConnection;
-#  endif
   // Map of topic names to |WakeLockTopic|s.
   // We assume a small, finite-sized set of topics.
-  nsClassHashtable<nsStringHashKey, WakeLockTopic> mTopics;
+  nsRefPtrHashtable<nsStringHashKey, WakeLockTopic> mTopics;
 };
 
 #endif  // __WakeLockListener_h__
