@@ -167,9 +167,6 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
   ElementState elementState = GetContentState(aFrame, aAppearance);
   if (aState) {
     memset(aState, 0, sizeof(GtkWidgetState));
-
-    // For XUL checkboxes and radio buttons, the state of the parent
-    // determines our state.
     if (aWidgetFlags) {
       if (elementState.HasState(ElementState::CHECKED)) {
         *aWidgetFlags |= MOZ_GTK_WIDGET_CHECKED;
@@ -191,18 +188,13 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
         aAppearance == StyleAppearance::Toolbarbutton ||
         aAppearance == StyleAppearance::Dualbutton ||
         aAppearance == StyleAppearance::ToolbarbuttonDropdown ||
+        aAppearance == StyleAppearance::MozWindowButtonMinimize ||
+        aAppearance == StyleAppearance::MozWindowButtonRestore ||
+        aAppearance == StyleAppearance::MozWindowButtonMaximize ||
+        aAppearance == StyleAppearance::MozWindowButtonClose ||
         aAppearance == StyleAppearance::Menulist ||
-        aAppearance == StyleAppearance::MenulistButton ||
-        aAppearance == StyleAppearance::MozMenulistArrowButton) {
+        aAppearance == StyleAppearance::MenulistButton) {
       aState->active &= aState->inHover;
-    } else if (aAppearance == StyleAppearance::Treetwisty ||
-               aAppearance == StyleAppearance::Treetwistyopen) {
-      if (nsTreeBodyFrame* treeBodyFrame = do_QueryFrame(aFrame)) {
-        const mozilla::AtomArray& atoms =
-            treeBodyFrame->GetPropertyArrayForCurrentDrawingItem();
-        aState->selected = atoms.Contains(nsGkAtoms::selected);
-        aState->inHover = atoms.Contains(nsGkAtoms::hover);
-      }
     }
 
     if (IsFrameContentNodeInNamespace(aFrame, kNameSpaceID_XUL)) {
@@ -217,16 +209,6 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
         aState->focused = FALSE;
       }
 
-      // menu item state is determined by the attribute "_moz-menuactive",
-      // and not by the mouse hovering (accessibility).  as a special case,
-      // menus which are children of a menu bar are only marked as prelight
-      // if they are open, not on normal hover.
-
-      if (aAppearance == StyleAppearance::Menuarrow) {
-        aState->inHover = CheckBooleanAttr(aFrame, nsGkAtoms::menuactive);
-        aState->active = FALSE;
-      }
-
       // A button with drop down menu open or an activated toggle button
       // should always appear depressed.
       if (aAppearance == StyleAppearance::Button ||
@@ -234,21 +216,12 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
           aAppearance == StyleAppearance::Dualbutton ||
           aAppearance == StyleAppearance::ToolbarbuttonDropdown ||
           aAppearance == StyleAppearance::Menulist ||
-          aAppearance == StyleAppearance::MenulistButton ||
-          aAppearance == StyleAppearance::MozMenulistArrowButton) {
+          aAppearance == StyleAppearance::MenulistButton) {
         bool menuOpen = IsOpenButton(aFrame);
         aState->depressed = IsCheckedButton(aFrame) || menuOpen;
         // we must not highlight buttons with open drop down menus on hover.
         aState->inHover = aState->inHover && !menuOpen;
       }
-
-      // When the input field of the drop down button has focus, some themes
-      // should draw focus for the drop down button as well.
-      if ((aAppearance == StyleAppearance::MenulistButton ||
-           aAppearance == StyleAppearance::MozMenulistArrowButton) &&
-          aWidgetFlags) {
-        *aWidgetFlags = CheckBooleanAttr(aFrame, nsGkAtoms::parentfocused);
-      }
     }
 
     if (aAppearance == StyleAppearance::MozWindowTitlebar ||
@@ -257,7 +230,8 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
         aAppearance == StyleAppearance::MozWindowButtonMinimize ||
         aAppearance == StyleAppearance::MozWindowButtonMaximize ||
         aAppearance == StyleAppearance::MozWindowButtonRestore) {
-      aState->backdrop = !nsWindow::GetTopLevelWindowActiveState(aFrame);
+      aState->backdrop = aFrame->PresContext()->Document()->State().HasState(
+          dom::DocumentState::WINDOW_INACTIVE);
     }
   }
 
@@ -272,10 +246,10 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
       aGtkWidgetType = MOZ_GTK_TOOLBAR_BUTTON;
       break;
     case StyleAppearance::Checkbox:
+      aGtkWidgetType = MOZ_GTK_CHECKBUTTON;
+      break;
     case StyleAppearance::Radio:
-      aGtkWidgetType = (aAppearance == StyleAppearance::Radio)
-                           ? MOZ_GTK_RADIOBUTTON
-                           : MOZ_GTK_CHECKBUTTON;
+      aGtkWidgetType = MOZ_GTK_RADIOBUTTON;
       break;
     case StyleAppearance::Spinner:
       aGtkWidgetType = MOZ_GTK_SPINBUTTON;
@@ -309,13 +283,8 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
       }
       break;
     }
-    case StyleAppearance::Separator:
-      aGtkWidgetType = MOZ_GTK_TOOLBAR_SEPARATOR;
-      break;
-    case StyleAppearance::Toolbargripper:
-      aGtkWidgetType = MOZ_GTK_GRIPPER;
-      break;
     case StyleAppearance::NumberInput:
+    case StyleAppearance::PasswordInput:
     case StyleAppearance::Textfield:
       aGtkWidgetType = MOZ_GTK_ENTRY;
       break;
@@ -323,48 +292,8 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
       aGtkWidgetType = MOZ_GTK_TEXT_VIEW;
       break;
     case StyleAppearance::Listbox:
-    case StyleAppearance::Treeview:
       aGtkWidgetType = MOZ_GTK_TREEVIEW;
       break;
-    case StyleAppearance::Treeheadercell:
-      if (aWidgetFlags) {
-        // In this case, the flag denotes whether the header is the sorted one
-        // or not
-        if (GetTreeSortDirection(aFrame) == eTreeSortDirection_Natural)
-          *aWidgetFlags = false;
-        else
-          *aWidgetFlags = true;
-      }
-      aGtkWidgetType = MOZ_GTK_TREE_HEADER_CELL;
-      break;
-    case StyleAppearance::Treeheadersortarrow:
-      if (aWidgetFlags) {
-        switch (GetTreeSortDirection(aFrame)) {
-          case eTreeSortDirection_Ascending:
-            *aWidgetFlags = GTK_ARROW_DOWN;
-            break;
-          case eTreeSortDirection_Descending:
-            *aWidgetFlags = GTK_ARROW_UP;
-            break;
-          case eTreeSortDirection_Natural:
-          default:
-            /* This prevents the treecolums from getting smaller
-             * and wider when switching sort direction off and on
-             * */
-            *aWidgetFlags = GTK_ARROW_NONE;
-            break;
-        }
-      }
-      aGtkWidgetType = MOZ_GTK_TREE_HEADER_SORTARROW;
-      break;
-    case StyleAppearance::Treetwisty:
-      aGtkWidgetType = MOZ_GTK_TREEVIEW_EXPANDER;
-      if (aWidgetFlags) *aWidgetFlags = GTK_EXPANDER_COLLAPSED;
-      break;
-    case StyleAppearance::Treetwistyopen:
-      aGtkWidgetType = MOZ_GTK_TREEVIEW_EXPANDER;
-      if (aWidgetFlags) *aWidgetFlags = GTK_EXPANDER_EXPANDED;
-      break;
     case StyleAppearance::MenulistButton:
     case StyleAppearance::Menulist:
       aGtkWidgetType = MOZ_GTK_DROPDOWN;
@@ -372,11 +301,6 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
         *aWidgetFlags =
             IsFrameContentNodeInNamespace(aFrame, kNameSpaceID_XHTML);
       break;
-    case StyleAppearance::MenulistText:
-      return false;  // nothing to do, but prevents the bg from being drawn
-    case StyleAppearance::MozMenulistArrowButton:
-      aGtkWidgetType = MOZ_GTK_DROPDOWN_ARROW;
-      break;
     case StyleAppearance::ToolbarbuttonDropdown:
     case StyleAppearance::ButtonArrowDown:
     case StyleAppearance::ButtonArrowUp:
@@ -394,21 +318,6 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
           *aWidgetFlags = GTK_ARROW_LEFT;
       }
       break;
-    case StyleAppearance::CheckboxContainer:
-      aGtkWidgetType = MOZ_GTK_CHECKBUTTON_CONTAINER;
-      break;
-    case StyleAppearance::RadioContainer:
-      aGtkWidgetType = MOZ_GTK_RADIOBUTTON_CONTAINER;
-      break;
-    case StyleAppearance::CheckboxLabel:
-      aGtkWidgetType = MOZ_GTK_CHECKBUTTON_LABEL;
-      break;
-    case StyleAppearance::RadioLabel:
-      aGtkWidgetType = MOZ_GTK_RADIOBUTTON_LABEL;
-      break;
-    case StyleAppearance::Toolbar:
-      aGtkWidgetType = MOZ_GTK_TOOLBAR;
-      break;
     case StyleAppearance::Tooltip:
       aGtkWidgetType = MOZ_GTK_TOOLTIP;
       break;
@@ -460,18 +369,15 @@ bool nsNativeThemeGTK::GetGtkWidgetAndState(StyleAppearance aAppearance,
       else
         aGtkWidgetType = MOZ_GTK_SPLITTER_HORIZONTAL;
       break;
-    case StyleAppearance::Menuarrow:
-      aGtkWidgetType = MOZ_GTK_MENUARROW;
-      break;
     case StyleAppearance::MozWindowTitlebar:
       aGtkWidgetType = MOZ_GTK_HEADER_BAR;
       break;
+    case StyleAppearance::MozWindowDecorations:
+      aGtkWidgetType = MOZ_GTK_WINDOW_DECORATION;
+      break;
     case StyleAppearance::MozWindowTitlebarMaximized:
       aGtkWidgetType = MOZ_GTK_HEADER_BAR_MAXIMIZED;
       break;
-    case StyleAppearance::MozWindowButtonBox:
-      aGtkWidgetType = MOZ_GTK_HEADER_BAR_BUTTON_BOX;
-      break;
     case StyleAppearance::MozWindowButtonClose:
       aGtkWidgetType = MOZ_GTK_HEADER_BAR_BUTTON_CLOSE;
       break;
@@ -740,8 +646,9 @@ CSSIntMargin nsNativeThemeGTK::GetExtraSizeForWidget(
       if (IsDefaultButton(aFrame)) {
         // Some themes draw a default indicator outside the widget,
         // include that in overflow
-        moz_gtk_button_get_default_overflow(&extra.top, &extra.left,
-                                            &extra.bottom, &extra.right);
+        moz_gtk_button_get_default_overflow(&extra.top.value, &extra.left.value,
+                                            &extra.bottom.value,
+                                            &extra.right.value);
         break;
       }
       return {};
@@ -752,16 +659,6 @@ CSSIntMargin nsNativeThemeGTK::GetExtraSizeForWidget(
   return extra;
 }
 
-bool nsNativeThemeGTK::IsWidgetVisible(StyleAppearance aAppearance) {
-  switch (aAppearance) {
-    case StyleAppearance::MozWindowButtonBox:
-      return false;
-    default:
-      break;
-  }
-  return true;
-}
-
 NS_IMETHODIMP
 nsNativeThemeGTK::DrawWidgetBackground(gfxContext* aContext, nsIFrame* aFrame,
                                        StyleAppearance aAppearance,
@@ -778,8 +675,7 @@ nsNativeThemeGTK::DrawWidgetBackground(gfxContext* aContext, nsIFrame* aFrame,
   GtkTextDirection direction = GetTextDirection(aFrame);
   gint flags;
 
-  if (!IsWidgetVisible(aAppearance) ||
-      !GetGtkWidgetAndState(aAppearance, aFrame, gtkWidgetType, &state,
+  if (!GetGtkWidgetAndState(aAppearance, aFrame, gtkWidgetType, &state,
                             &flags)) {
     return NS_OK;
   }
@@ -948,8 +844,9 @@ CSSIntMargin nsNativeThemeGTK::GetCachedWidgetBorder(
     if (mBorderCacheValid[cacheIndex] & cacheBit) {
       result = mBorderCache[gtkWidgetType];
     } else {
-      moz_gtk_get_widget_border(gtkWidgetType, &result.left, &result.top,
-                                &result.right, &result.bottom, aDirection);
+      moz_gtk_get_widget_border(gtkWidgetType, &result.left.value,
+                                &result.top.value, &result.right.value,
+                                &result.bottom.value, aDirection);
       if (gtkWidgetType != MOZ_GTK_DROPDOWN) {  // depends on aDirection
         mBorderCacheValid[cacheIndex] |= cacheBit;
         mBorderCache[gtkWidgetType] = result;
@@ -962,14 +859,13 @@ CSSIntMargin nsNativeThemeGTK::GetCachedWidgetBorder(
 
 LayoutDeviceIntMargin nsNativeThemeGTK::GetWidgetBorder(
     nsDeviceContext* aContext, nsIFrame* aFrame, StyleAppearance aAppearance) {
+  if (IsWidgetAlwaysNonNative(aFrame, aAppearance)) {
+    return Theme::GetWidgetBorder(aContext, aFrame, aAppearance);
+  }
+
   CSSIntMargin result;
   GtkTextDirection direction = GetTextDirection(aFrame);
   switch (aAppearance) {
-    case StyleAppearance::Toolbox:
-      // gtk has no toolbox equivalent.  So, although we map toolbox to
-      // gtk's 'toolbar' for purposes of painting the widget background,
-      // we don't use the toolbar border for toolbox.
-      break;
     case StyleAppearance::Dualbutton:
       // TOOLBAR_DUAL_BUTTON is an interesting case.  We want a border to draw
       // around the entire button + dropdown, and also an inner border if you're
@@ -986,9 +882,9 @@ LayoutDeviceIntMargin nsNativeThemeGTK::GetWidgetBorder(
                                 &flags)) {
         return {};
       }
-      moz_gtk_get_tab_border(&result.left, &result.top, &result.right,
-                             &result.bottom, direction, (GtkTabFlags)flags,
-                             gtkWidgetType);
+      moz_gtk_get_tab_border(&result.left.value, &result.top.value,
+                             &result.right.value, &result.bottom.value,
+                             direction, (GtkTabFlags)flags, gtkWidgetType);
     } break;
     default: {
       result = GetCachedWidgetBorder(aFrame, aAppearance, direction);
@@ -1002,10 +898,12 @@ bool nsNativeThemeGTK::GetWidgetPadding(nsDeviceContext* aContext,
                                         nsIFrame* aFrame,
                                         StyleAppearance aAppearance,
                                         LayoutDeviceIntMargin* aResult) {
+  if (IsWidgetAlwaysNonNative(aFrame, aAppearance)) {
+    return Theme::GetWidgetPadding(aContext, aFrame, aAppearance, aResult);
+  }
   switch (aAppearance) {
     case StyleAppearance::Toolbarbutton:
     case StyleAppearance::Tooltip:
-    case StyleAppearance::MozWindowButtonBox:
     case StyleAppearance::MozWindowButtonClose:
     case StyleAppearance::MozWindowButtonMinimize:
     case StyleAppearance::MozWindowButtonMaximize:
@@ -1013,7 +911,6 @@ bool nsNativeThemeGTK::GetWidgetPadding(nsDeviceContext* aContext,
     case StyleAppearance::Dualbutton:
     case StyleAppearance::TabScrollArrowBack:
     case StyleAppearance::TabScrollArrowForward:
-    case StyleAppearance::MozMenulistArrowButton:
     case StyleAppearance::ToolbarbuttonDropdown:
     case StyleAppearance::ButtonArrowUp:
     case StyleAppearance::ButtonArrowDown:
@@ -1053,15 +950,14 @@ bool nsNativeThemeGTK::GetWidgetOverflow(nsDeviceContext* aContext,
 auto nsNativeThemeGTK::IsWidgetNonNative(nsIFrame* aFrame,
                                          StyleAppearance aAppearance)
     -> NonNative {
-  if (IsWidgetScrollbarPart(aAppearance) ||
-      aAppearance == StyleAppearance::FocusOutline) {
+  if (IsWidgetAlwaysNonNative(aFrame, aAppearance)) {
     return NonNative::Always;
   }
 
   // If the current GTK theme color scheme matches our color-scheme, then we
   // can draw a native widget.
   if (LookAndFeel::ColorSchemeForFrame(aFrame) ==
-      LookAndFeel::ColorSchemeForChrome()) {
+      PreferenceSheet::ColorSchemeForChrome()) {
     return NonNative::No;
   }
 
@@ -1088,10 +984,16 @@ auto nsNativeThemeGTK::IsWidgetNonNative(nsIFrame* aFrame,
   return NonNative::BecauseColorMismatch;
 }
 
+bool nsNativeThemeGTK::IsWidgetAlwaysNonNative(nsIFrame* aFrame,
+                                               StyleAppearance aAppearance) {
+  return Theme::IsWidgetAlwaysNonNative(aFrame, aAppearance) ||
+         aAppearance == StyleAppearance::MozMenulistArrowButton;
+}
+
 LayoutDeviceIntSize nsNativeThemeGTK::GetMinimumWidgetSize(
     nsPresContext* aPresContext, nsIFrame* aFrame,
     StyleAppearance aAppearance) {
-  if (IsWidgetNonNative(aFrame, aAppearance) == NonNative::Always) {
+  if (IsWidgetAlwaysNonNative(aFrame, aAppearance)) {
     return Theme::GetMinimumWidgetSize(aPresContext, aFrame, aAppearance);
   }
 
@@ -1117,9 +1019,6 @@ LayoutDeviceIntSize nsNativeThemeGTK::GetMinimumWidgetSize(
     case StyleAppearance::TabScrollArrowForward: {
       moz_gtk_get_tab_scroll_arrow_size(&result.width, &result.height);
     } break;
-    case StyleAppearance::MozMenulistArrowButton: {
-      moz_gtk_get_combo_box_entry_button_size(&result.width, &result.height);
-    } break;
     case StyleAppearance::Checkbox:
     case StyleAppearance::Radio: {
       const ToggleGTKMetrics* metrics = GetToggleMetrics(
@@ -1139,34 +1038,28 @@ LayoutDeviceIntSize nsNativeThemeGTK::GetMinimumWidgetSize(
     case StyleAppearance::MozWindowButtonClose: {
       const ToolbarButtonGTKMetrics* metrics =
           GetToolbarButtonMetrics(MOZ_GTK_HEADER_BAR_BUTTON_CLOSE);
-      result.width = metrics->minSizeWithBorderMargin.width;
-      result.height = metrics->minSizeWithBorderMargin.height;
+      result.width = metrics->minSizeWithBorder.width;
+      result.height = metrics->minSizeWithBorder.height;
       break;
     }
     case StyleAppearance::MozWindowButtonMinimize: {
       const ToolbarButtonGTKMetrics* metrics =
           GetToolbarButtonMetrics(MOZ_GTK_HEADER_BAR_BUTTON_MINIMIZE);
-      result.width = metrics->minSizeWithBorderMargin.width;
-      result.height = metrics->minSizeWithBorderMargin.height;
+      result.width = metrics->minSizeWithBorder.width;
+      result.height = metrics->minSizeWithBorder.height;
       break;
     }
     case StyleAppearance::MozWindowButtonMaximize:
     case StyleAppearance::MozWindowButtonRestore: {
       const ToolbarButtonGTKMetrics* metrics =
           GetToolbarButtonMetrics(MOZ_GTK_HEADER_BAR_BUTTON_MAXIMIZE);
-      result.width = metrics->minSizeWithBorderMargin.width;
-      result.height = metrics->minSizeWithBorderMargin.height;
+      result.width = metrics->minSizeWithBorder.width;
+      result.height = metrics->minSizeWithBorder.height;
       break;
     }
-    case StyleAppearance::CheckboxContainer:
-    case StyleAppearance::RadioContainer:
-    case StyleAppearance::CheckboxLabel:
-    case StyleAppearance::RadioLabel:
     case StyleAppearance::Button:
     case StyleAppearance::Menulist:
-    case StyleAppearance::MenulistButton:
-    case StyleAppearance::Toolbarbutton:
-    case StyleAppearance::Treeheadercell: {
+    case StyleAppearance::MenulistButton: {
       if (aAppearance == StyleAppearance::Menulist ||
           aAppearance == StyleAppearance::MenulistButton) {
         // Include the arrow size.
@@ -1182,6 +1075,7 @@ LayoutDeviceIntSize nsNativeThemeGTK::GetMinimumWidgetSize(
       result.height += border.TopBottom();
     } break;
     case StyleAppearance::NumberInput:
+    case StyleAppearance::PasswordInput:
     case StyleAppearance::Textfield: {
       gint contentHeight = 0;
       gint borderPaddingHeight = 0;
@@ -1215,27 +1109,17 @@ LayoutDeviceIntSize nsNativeThemeGTK::GetMinimumWidgetSize(
         result.height = height;
       }
     } break;
-    case StyleAppearance::Separator: {
-      moz_gtk_get_toolbar_separator_width(&result.width);
-    } break;
     case StyleAppearance::Spinner:
       // hard code these sizes
       result.width = 14;
       result.height = 26;
       break;
-    case StyleAppearance::Treeheadersortarrow:
     case StyleAppearance::SpinnerUpbutton:
     case StyleAppearance::SpinnerDownbutton:
       // hard code these sizes
       result.width = 14;
       result.height = 13;
       break;
-    case StyleAppearance::Treetwisty:
-    case StyleAppearance::Treetwistyopen: {
-      gint expander_size;
-      moz_gtk_get_treeview_expander_size(&expander_size);
-      result.width = result.height = expander_size;
-    } break;
     default:
       break;
   }
@@ -1257,12 +1141,10 @@ nsNativeThemeGTK::WidgetStateChanged(nsIFrame* aFrame,
   }
 
   // Some widget types just never change state.
-  if (aAppearance == StyleAppearance::Toolbox ||
-      aAppearance == StyleAppearance::Toolbar ||
-      aAppearance == StyleAppearance::Progresschunk ||
+  if (aAppearance == StyleAppearance::Progresschunk ||
       aAppearance == StyleAppearance::ProgressBar ||
-      aAppearance == StyleAppearance::Menubar ||
-      aAppearance == StyleAppearance::Tooltip) {
+      aAppearance == StyleAppearance::Tooltip ||
+      aAppearance == StyleAppearance::MozWindowDecorations) {
     return NS_OK;
   }
 
@@ -1293,8 +1175,7 @@ nsNativeThemeGTK::WidgetStateChanged(nsIFrame* aFrame,
       aAttribute == nsGkAtoms::visuallyselected ||
       aAttribute == nsGkAtoms::focused || aAttribute == nsGkAtoms::readonly ||
       aAttribute == nsGkAtoms::_default ||
-      aAttribute == nsGkAtoms::menuactive || aAttribute == nsGkAtoms::open ||
-      aAttribute == nsGkAtoms::parentfocused) {
+      aAttribute == nsGkAtoms::menuactive || aAttribute == nsGkAtoms::open) {
     *aShouldRepaint = true;
     return NS_OK;
   }
@@ -1317,7 +1198,7 @@ nsNativeThemeGTK::ThemeSupportsWidget(nsPresContext* aPresContext,
     return false;
   }
 
-  if (IsWidgetNonNative(aFrame, aAppearance) == NonNative::Always) {
+  if (IsWidgetAlwaysNonNative(aFrame, aAppearance)) {
     return Theme::ThemeSupportsWidget(aPresContext, aFrame, aAppearance);
   }
 
@@ -1325,7 +1206,6 @@ nsNativeThemeGTK::ThemeSupportsWidget(nsPresContext* aPresContext,
     // Combobox dropdowns don't support native theming in vertical mode.
     case StyleAppearance::Menulist:
     case StyleAppearance::MenulistButton:
-    case StyleAppearance::MenulistText:
       if (aFrame && aFrame->GetWritingMode().IsVertical()) {
         return false;
       }
@@ -1334,8 +1214,6 @@ nsNativeThemeGTK::ThemeSupportsWidget(nsPresContext* aPresContext,
     case StyleAppearance::Button:
     case StyleAppearance::Radio:
     case StyleAppearance::Checkbox:
-    case StyleAppearance::Toolbox:  // N/A
-    case StyleAppearance::Toolbar:
     case StyleAppearance::Toolbarbutton:
     case StyleAppearance::Dualbutton:  // so we can override the border with 0
     case StyleAppearance::ToolbarbuttonDropdown:
@@ -1343,17 +1221,7 @@ nsNativeThemeGTK::ThemeSupportsWidget(nsPresContext* aPresContext,
     case StyleAppearance::ButtonArrowDown:
     case StyleAppearance::ButtonArrowNext:
     case StyleAppearance::ButtonArrowPrevious:
-    case StyleAppearance::Separator:
-    case StyleAppearance::Toolbargripper:
     case StyleAppearance::Listbox:
-    case StyleAppearance::Treeview:
-      // case StyleAppearance::Treeitem:
-    case StyleAppearance::Treetwisty:
-      // case StyleAppearance::Treeline:
-      // case StyleAppearance::Treeheader:
-    case StyleAppearance::Treeheadercell:
-    case StyleAppearance::Treeheadersortarrow:
-    case StyleAppearance::Treetwistyopen:
     case StyleAppearance::ProgressBar:
     case StyleAppearance::Progresschunk:
     case StyleAppearance::Tab:
@@ -1367,36 +1235,20 @@ nsNativeThemeGTK::ThemeSupportsWidget(nsPresContext* aPresContext,
     case StyleAppearance::SpinnerDownbutton:
     case StyleAppearance::SpinnerTextfield:
     case StyleAppearance::NumberInput:
+    case StyleAppearance::PasswordInput:
     case StyleAppearance::Textfield:
     case StyleAppearance::Textarea:
     case StyleAppearance::Range:
     case StyleAppearance::RangeThumb:
-    case StyleAppearance::CheckboxContainer:
-    case StyleAppearance::RadioContainer:
-    case StyleAppearance::CheckboxLabel:
-    case StyleAppearance::RadioLabel:
-    case StyleAppearance::Menuarrow:
-    case StyleAppearance::Radiomenuitem:
     case StyleAppearance::Splitter:
-    case StyleAppearance::MozWindowButtonBox:
     case StyleAppearance::MozWindowButtonClose:
     case StyleAppearance::MozWindowButtonMinimize:
     case StyleAppearance::MozWindowButtonMaximize:
     case StyleAppearance::MozWindowButtonRestore:
     case StyleAppearance::MozWindowTitlebar:
     case StyleAppearance::MozWindowTitlebarMaximized:
+    case StyleAppearance::MozWindowDecorations:
       return !IsWidgetStyled(aPresContext, aFrame, aAppearance);
-
-    case StyleAppearance::MozMenulistArrowButton:
-      if (aFrame && aFrame->GetWritingMode().IsVertical()) {
-        return false;
-      }
-      // "Native" dropdown buttons cause padding and margin problems, but only
-      // in HTML so allow them in XUL.
-      return (!aFrame ||
-              IsFrameContentNodeInNamespace(aFrame, kNameSpaceID_XUL)) &&
-             !IsWidgetStyled(aPresContext, aFrame, aAppearance);
-
     default:
       break;
   }
@@ -1407,8 +1259,7 @@ nsNativeThemeGTK::ThemeSupportsWidget(nsPresContext* aPresContext,
 NS_IMETHODIMP_(bool)
 nsNativeThemeGTK::WidgetIsContainer(StyleAppearance aAppearance) {
   // XXXdwh At some point flesh all of this out.
-  if (aAppearance == StyleAppearance::MozMenulistArrowButton ||
-      aAppearance == StyleAppearance::Radio ||
+  if (aAppearance == StyleAppearance::Radio ||
       aAppearance == StyleAppearance::RangeThumb ||
       aAppearance == StyleAppearance::Checkbox ||
       aAppearance == StyleAppearance::TabScrollArrowBack ||
@@ -1427,13 +1278,18 @@ bool nsNativeThemeGTK::ThemeDrawsFocusForWidget(nsIFrame* aFrame,
     return Theme::ThemeDrawsFocusForWidget(aFrame, aAppearance);
   }
   switch (aAppearance) {
+    case StyleAppearance::Checkbox:
+    case StyleAppearance::Radio:
+      // These are drawn only for non-XUL elements, but in XUL the label has
+      // the focus ring.
+      return true;
     case StyleAppearance::Button:
     case StyleAppearance::Menulist:
     case StyleAppearance::MenulistButton:
     case StyleAppearance::Textarea:
     case StyleAppearance::Textfield:
-    case StyleAppearance::Treeheadercell:
     case StyleAppearance::NumberInput:
+    case StyleAppearance::PasswordInput:
       return true;
     default:
       return false;
