@@ -3,6 +3,8 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
+#include "mozilla/Attributes.h"
+#include "mozilla/ScopeExit.h"
 #include "mozilla/ipc/MessageChannel.h"
 #include "mozilla/ipc/WindowsMessageLoop.h"
 #include "nsAppShell.h"
@@ -29,8 +31,11 @@
 #include "HeadlessScreenHelper.h"
 #include "mozilla/widget/ScreenManager.h"
 #include "mozilla/Atomics.h"
+#include "mozilla/NativeNt.h"
+#include "mozilla/WindowsDiagnostics.h"
 #include "mozilla/WindowsProcessMitigations.h"
-#include "mozilla/WindowsVersion.h"
+
+#include <winternl.h>
 
 #ifdef MOZ_BACKGROUNDTASKS
 #  include "mozilla/BackgroundTasks.h"
@@ -212,66 +217,13 @@ class WinWakeLockListener final : public nsIDOMMozWakeLockListener {
   HANDLE mNonDisplayHandle = nullptr;
 };
 NS_IMPL_ISUPPORTS(WinWakeLockListener, nsIDOMMozWakeLockListener)
-
-// This wakelock is used for the version older than Windows7.
-class LegacyWinWakeLockListener final : public nsIDOMMozWakeLockListener {
- public:
-  NS_DECL_ISUPPORTS
-  LegacyWinWakeLockListener() { MOZ_ASSERT(XRE_IsParentProcess()); }
-
- private:
-  ~LegacyWinWakeLockListener() {}
-
-  NS_IMETHOD Callback(const nsAString& aTopic,
-                      const nsAString& aState) override {
-    WAKE_LOCK_LOG("WinWakeLock: topic=%s, state=%s",
-                  NS_ConvertUTF16toUTF8(aTopic).get(),
-                  NS_ConvertUTF16toUTF8(aState).get());
-    if (!aTopic.EqualsASCII("screen") && !aTopic.EqualsASCII("audio-playing") &&
-        !aTopic.EqualsASCII("video-playing")) {
-      return NS_OK;
-    }
-
-    // Check what kind of lock we will require, if both display lock and non
-    // display lock are needed, we would require display lock because it has
-    // higher priority.
-    if (aTopic.EqualsASCII("audio-playing")) {
-      mRequireForNonDisplayLock = aState.EqualsASCII("locked-foreground") ||
-                                  aState.EqualsASCII("locked-background");
-    } else if (aTopic.EqualsASCII("screen") ||
-               aTopic.EqualsASCII("video-playing")) {
-      mRequireForDisplayLock = aState.EqualsASCII("locked-foreground");
-    }
-
-    if (mRequireForDisplayLock) {
-      WAKE_LOCK_LOG("WinWakeLock: Request display lock");
-      SetThreadExecutionState(ES_DISPLAY_REQUIRED | ES_CONTINUOUS);
-    } else if (mRequireForNonDisplayLock) {
-      WAKE_LOCK_LOG("WinWakeLock: Request non-display lock");
-      SetThreadExecutionState(ES_SYSTEM_REQUIRED | ES_CONTINUOUS);
-    } else {
-      WAKE_LOCK_LOG("WinWakeLock: reset lock");
-      SetThreadExecutionState(ES_CONTINUOUS);
-    }
-    return NS_OK;
-  }
-
-  bool mRequireForDisplayLock = false;
-  bool mRequireForNonDisplayLock = false;
-};
-
-NS_IMPL_ISUPPORTS(LegacyWinWakeLockListener, nsIDOMMozWakeLockListener)
 StaticRefPtr<nsIDOMMozWakeLockListener> sWakeLockListener;
 
 static void AddScreenWakeLockListener() {
   nsCOMPtr<nsIPowerManagerService> sPowerManagerService =
       do_GetService(POWERMANAGERSERVICE_CONTRACTID);
   if (sPowerManagerService) {
-    if (IsWin7SP1OrLater()) {
-      sWakeLockListener = new WinWakeLockListener();
-    } else {
-      sWakeLockListener = new LegacyWinWakeLockListener();
-    }
+    sWakeLockListener = new WinWakeLockListener();
     sPowerManagerService->AddWakeLockListener(sWakeLockListener);
   } else {
     NS_WARNING(
@@ -329,10 +281,10 @@ SingleNativeEventPump::AfterProcessNextEvent(nsIThreadInternal* aThread,
 // RegisterWindowMessage values
 // Native event callback message
 const wchar_t* kAppShellGeckoEventId = L"nsAppShell:EventID";
-UINT sAppShellGeckoMsgId;
+UINT sAppShellGeckoMsgId = 0x10001;  // initialize to invalid message ID
 // Taskbar button creation message
 const wchar_t* kTaskbarButtonEventId = L"TaskbarButtonCreated";
-UINT sTaskbarButtonCreatedMsg;
+UINT sTaskbarButtonCreatedMsg = 0x10002;  // initialize to invalid message ID
 
 /* static */
 UINT nsAppShell::GetTaskbarButtonCreatedMessage() {
@@ -356,6 +308,8 @@ static Atomic<size_t, ReleaseAcquire> sOutstandingNativeEventCallbacks;
 /*static*/ LRESULT CALLBACK nsAppShell::EventWindowProc(HWND hwnd, UINT uMsg,
                                                         WPARAM wParam,
                                                         LPARAM lParam) {
+  NativeEventLogger eventLogger("AppShell", hwnd, uMsg, wParam, lParam);
+
   if (uMsg == sAppShellGeckoMsgId) {
     // The app shell might have been destroyed between this message being
     // posted and being executed, so be extra careful.
@@ -368,7 +322,10 @@ static Atomic<size_t, ReleaseAcquire> sOutstandingNativeEventCallbacks;
     --sOutstandingNativeEventCallbacks;
     return TRUE;
   }
-  return DefWindowProc(hwnd, uMsg, wParam, lParam);
+
+  LRESULT ret = DefWindowProc(hwnd, uMsg, wParam, lParam);
+  eventLogger.SetResult(ret, false);
+  return ret;
 }
 
 nsAppShell::~nsAppShell() {
@@ -385,79 +342,6 @@ nsAppShell::~nsAppShell() {
   sOutstandingNativeEventCallbacks = 0;
 }
 
-#if defined(ACCESSIBILITY)
-
-static ULONG gUiaMsg;
-static HHOOK gUiaHook;
-static uint32_t gUiaAttempts;
-static const uint32_t kMaxUiaAttempts = 5;
-
-static void InitUIADetection();
-
-static LRESULT CALLBACK UiaHookProc(int aCode, WPARAM aWParam, LPARAM aLParam) {
-  if (aCode < 0) {
-    return ::CallNextHookEx(nullptr, aCode, aWParam, aLParam);
-  }
-
-  auto cwp = reinterpret_cast<CWPSTRUCT*>(aLParam);
-  if (gUiaMsg && cwp->message == gUiaMsg) {
-    if (gUiaAttempts < kMaxUiaAttempts) {
-      ++gUiaAttempts;
-
-      Maybe<bool> shouldCallNextHook =
-          a11y::Compatibility::OnUIAMessage(cwp->wParam, cwp->lParam);
-      if (shouldCallNextHook.isSome()) {
-        // We've got an instantiator.
-        if (!shouldCallNextHook.value()) {
-          // We're blocking this instantiation. We need to keep this hook set
-          // so that we can catch any future instantiation attempts.
-          return 0;
-        }
-
-        // We're allowing the instantiator to proceed, so this hook is no longer
-        // needed.
-        if (::UnhookWindowsHookEx(gUiaHook)) {
-          gUiaHook = nullptr;
-        }
-      } else {
-        // Our hook might be firing after UIA; let's try reinstalling ourselves.
-        InitUIADetection();
-      }
-    } else {
-      // We've maxed out our attempts. Let's unhook.
-      if (::UnhookWindowsHookEx(gUiaHook)) {
-        gUiaHook = nullptr;
-      }
-    }
-  }
-
-  return ::CallNextHookEx(nullptr, aCode, aWParam, aLParam);
-}
-
-static void InitUIADetection() {
-  if (gUiaHook) {
-    // In this case we want to re-hook so that the hook is always called ahead
-    // of UIA's hook.
-    if (::UnhookWindowsHookEx(gUiaHook)) {
-      gUiaHook = nullptr;
-    }
-  }
-
-  if (!gUiaMsg) {
-    // This is the message that UIA sends to trigger a command. UIA's
-    // CallWndProc looks for this message and then handles the request.
-    // Our hook gets in front of UIA's hook and examines the message first.
-    gUiaMsg = ::RegisterWindowMessageW(L"HOOKUTIL_MSG");
-  }
-
-  if (!gUiaHook) {
-    gUiaHook = ::SetWindowsHookEx(WH_CALLWNDPROC, &UiaHookProc, nullptr,
-                                  ::GetCurrentThreadId());
-  }
-}
-
-#endif  // defined(ACCESSIBILITY)
-
 NS_IMETHODIMP
 nsAppShell::Observe(nsISupports* aSubject, const char* aTopic,
                     const char16_t* aData) {
@@ -465,25 +349,6 @@ nsAppShell::Observe(nsISupports* aSubject, const char* aTopic,
     nsCOMPtr<nsIObserverService> obsServ(
         mozilla::services::GetObserverService());
 
-#if defined(ACCESSIBILITY)
-    if (!strcmp(aTopic, "dll-loaded-main-thread")) {
-      if (a11y::PlatformDisabledState() != a11y::ePlatformIsDisabled &&
-          !gUiaHook) {
-        nsDependentString dllName(aData);
-
-        if (StringEndsWith(dllName, u"uiautomationcore.dll"_ns,
-                           nsCaseInsensitiveStringComparator)) {
-          InitUIADetection();
-
-          // Now that we've handled the observer notification, we can remove it
-          obsServ->RemoveObserver(this, "dll-loaded-main-thread");
-        }
-      }
-
-      return NS_OK;
-    }
-#endif  // defined(ACCESSIBILITY)
-
     if (!strcmp(aTopic, "sessionstore-restoring-on-startup")) {
       nsWindow::SetIsRestoringSession(true);
       // Now that we've handled the observer notification, we can remove it
@@ -502,6 +367,215 @@ nsAppShell::Observe(nsISupports* aSubject, const char* aTopic,
   return nsBaseAppShell::Observe(aSubject, aTopic, aData);
 }
 
+namespace {
+
+// Struct containing information about the user atom table. (See
+// DiagnoseUserAtomTable(), below.)
+struct AtomTableInformation {
+  // Number of atoms in use. (Exactly 0x4000 == 16384, if all are.)
+  UINT in_use = 0;
+  // Number of atoms confirmed not in use.
+  UINT free = 0;
+  // Number of atoms which gave errors when checked.
+  UINT errors = 0;
+
+  // Last atom which gave an unexpected error...
+  UINT lastErrorAtom = ~0u;
+  // ... and the error it gave.
+  WinErrorState lastErrorState;
+};
+
+// Return a summary of the state of the atom table.
+MOZ_NEVER_INLINE static AtomTableInformation DiagnoseUserAtomTable() {
+  // Restore error state on exit, for the sake of automated minidump analyses.
+  auto const _restoreErrState =
+      mozilla::MakeScopeExit([oldErrState = WinErrorState::Get()]() {
+        WinErrorState::Apply(oldErrState);
+      });
+
+  AtomTableInformation retval;
+
+  // Expected error-state on failure-return when the atom is assigned, but not
+  // enough space was provided for the full string.
+  constexpr WinErrorState kBufferTooSmall = {
+      .error = ERROR_INSUFFICIENT_BUFFER,
+      .ntStatus = ((NTSTATUS)0xC0000023),  // == STATUS_BUFFER_TOO_SMALL
+  };
+  // Expected error-state on failure-return when the atom is not assigned.
+  constexpr WinErrorState kInvalidAtom = {
+      .error = ERROR_INVALID_HANDLE,
+      .ntStatus = ((NTSTATUS)STATUS_INVALID_HANDLE),
+  };
+
+  // Iterate over only the dynamic portion of the atom table.
+  for (UINT atom = 0xC000; atom <= 0xFFFF; ++atom) {
+    // The actual atom values are PII. Don't acquire them in their entirety, and
+    // don't keep more information about them than is needed.
+    WCHAR buf[2] = {};
+    // USE OF UNDOCUMENTED BEHAVIOR: The user atom table is shared by message
+    // names, window-class names, and clipboard-format names. Only the last has
+    // a documented getter-mechanism.
+    BOOL const ok = ::GetClipboardFormatNameW(atom, buf, 1);
+    WinErrorState const errState = WinErrorState::Get();
+    if (ok || errState == kBufferTooSmall) {
+      ++retval.in_use;
+    } else if (errState == kInvalidAtom) {
+      ++retval.free;
+    } else {
+      // Unexpected error-state.
+      ++retval.errors;
+      retval.lastErrorAtom = atom;
+      retval.lastErrorState = errState;
+    }
+  }
+
+  return retval;
+}
+
+#if defined(MOZ_DIAGNOSTIC_ASSERT_ENABLED) && defined(_M_X64)
+static constexpr int kMaxStepsUser32 = 0x1800;
+static constexpr int kMaxErrorStatesUser32 = 0x200;
+using User32SingleStepData =
+    ModuleSingleStepData<kMaxStepsUser32, kMaxErrorStatesUser32>;
+
+template <typename CallbackToRun, typename PostCollectionCallback>
+WindowsDiagnosticsError CollectUser32SingleStepData(
+    CallbackToRun aCallbackToRun,
+    PostCollectionCallback aPostCollectionCallback) {
+  return CollectModuleSingleStepData<kMaxStepsUser32, kMaxErrorStatesUser32>(
+      L"user32.dll", std::move(aCallbackToRun),
+      std::move(aPostCollectionCallback));
+}
+#endif  // MOZ_DIAGNOSTIC_ASSERT_ENABLED && _M_X64
+
+}  // namespace
+
+// Collect data for bug 1571516. We don't automatically send up `GetLastError`
+// or `GetLastNtStatus` data for beta/release builds, so extract the relevant
+// error values and store them on the stack, where they can be viewed in
+// minidumps -- in fact, do so after each individual API call. This takes the
+// form of various local variables whose initial character is an underscore,
+// most of which are also marked [[maybe_unused]].
+//
+// We tag this function `[[clang::optnone]]` to prevent the compiler from
+// eliding those values as _actually_ unused, as well as to generally simplify
+// the haruspex's task once the minidumps are in. (As this function should be
+// called at most once per process, the minor performance hit is not a concern.)
+//
+[[clang::optnone]] MOZ_NEVER_INLINE nsresult nsAppShell::InitHiddenWindow() {
+  // note the incoming error-state; this may be relevant to errors we get later
+  auto _initialErr [[maybe_unused]] = WinErrorState::Get();
+  // reset the error-state, to avoid ambiguity below
+  WinErrorState::Clear();
+
+  // Diagnostic variable. Only collected in the event of a failure in one of the
+  // functions that attempts to register an atom.
+  AtomTableInformation _atomTableInfo [[maybe_unused]];
+
+  // Attempt to register the window message. On failure, retain the initial
+  // value of `sAppShellGeckoMsgId`.
+  auto const _msgId = ::RegisterWindowMessageW(kAppShellGeckoEventId);
+  if (_msgId) {
+    sAppShellGeckoMsgId = _msgId;
+  }
+  auto const _sAppShellGeckoMsgId [[maybe_unused]] = sAppShellGeckoMsgId;
+  auto const _rwmErr [[maybe_unused]] = WinErrorState::Get();
+  if (!_msgId) _atomTableInfo = DiagnoseUserAtomTable();
+  NS_ASSERTION(sAppShellGeckoMsgId,
+               "Could not register hidden window event message!");
+
+  mLastNativeEventScheduled = TimeStamp::NowLoRes();
+
+  WNDCLASSW wc;
+  HINSTANCE const module = GetModuleHandle(nullptr);
+
+  constexpr const wchar_t* kWindowClass = L"nsAppShell:EventWindowClass";
+  // (Undocumented behavior note: on success, this will specifically be the
+  // window-class atom. We don't rely on this.)
+  BOOL const _gciwRet = ::GetClassInfoW(module, kWindowClass, &wc);
+  auto const _gciwErr [[maybe_unused]] = WinErrorState::Get();
+  WinErrorState::Clear();
+
+  WinErrorState _rcErr [[maybe_unused]];
+  if (!_gciwRet) {
+    wc.style = 0;
+    wc.lpfnWndProc = EventWindowProc;
+    wc.cbClsExtra = 0;
+    wc.cbWndExtra = 0;
+    wc.hInstance = module;
+    wc.hIcon = nullptr;
+    wc.hCursor = nullptr;
+    wc.hbrBackground = (HBRUSH) nullptr;
+    wc.lpszMenuName = (LPCWSTR) nullptr;
+    wc.lpszClassName = kWindowClass;
+
+    ATOM _windowClassAtom = ::RegisterClassW(&wc);
+    _rcErr = WinErrorState::Get();
+
+    if (!_windowClassAtom) _atomTableInfo = DiagnoseUserAtomTable();
+
+#if defined(MOZ_DIAGNOSTIC_ASSERT_ENABLED) && defined(_M_X64)
+    if (!_windowClassAtom) {
+      // Retry with single-step data collection
+      WindowsDiagnosticsError rv = CollectUser32SingleStepData(
+          [&wc, &_windowClassAtom]() {
+            _windowClassAtom = ::RegisterClassW(&wc);
+          },
+          [&_windowClassAtom](const User32SingleStepData& aData) {
+            // Crashing here gives access to the single step data on stack
+            MOZ_DIAGNOSTIC_ASSERT(
+                _windowClassAtom,
+                "RegisterClassW for EventWindowClass failed twice");
+          });
+      auto const _cssdErr [[maybe_unused]] = WinErrorState::Get();
+      MOZ_DIAGNOSTIC_ASSERT(
+          rv == WindowsDiagnosticsError::None,
+          "Failed to collect single step data for RegisterClassW");
+      // If we reach this point then somehow the single-stepped call succeeded
+      // and we can proceed
+    }
+#endif  // MOZ_DIAGNOSTIC_ASSERT_ENABLED && _M_X64
+
+    MOZ_DIAGNOSTIC_ASSERT(_windowClassAtom,
+                          "RegisterClassW for EventWindowClass failed");
+    WinErrorState::Clear();
+  }
+
+  mEventWnd = CreateWindowW(kWindowClass, L"nsAppShell:EventWindow", 0, 0, 0,
+                            10, 10, HWND_MESSAGE, nullptr, module, nullptr);
+  auto const _cwErr [[maybe_unused]] = WinErrorState::Get();
+
+#if defined(MOZ_DIAGNOSTIC_ASSERT_ENABLED) && defined(_M_X64)
+  if (!mEventWnd) {
+    // Retry with single-step data collection
+    HWND eventWnd{};
+    WindowsDiagnosticsError rv = CollectUser32SingleStepData(
+        [module, &eventWnd]() {
+          eventWnd =
+              CreateWindowW(kWindowClass, L"nsAppShell:EventWindow", 0, 0, 0,
+                            10, 10, HWND_MESSAGE, nullptr, module, nullptr);
+        },
+        [&eventWnd](const User32SingleStepData& aData) {
+          // Crashing here gives access to the single step data on stack
+          MOZ_DIAGNOSTIC_ASSERT(eventWnd,
+                                "CreateWindowW for EventWindow failed twice");
+        });
+    auto const _cssdErr [[maybe_unused]] = WinErrorState::Get();
+    MOZ_DIAGNOSTIC_ASSERT(
+        rv == WindowsDiagnosticsError::None,
+        "Failed to collect single step data for CreateWindowW");
+    // If we reach this point then somehow the single-stepped call succeeded and
+    // we can proceed
+    mEventWnd = eventWnd;
+  }
+#endif  // MOZ_DIAGNOSTIC_ASSERT_ENABLED && _M_X64
+
+  MOZ_DIAGNOSTIC_ASSERT(mEventWnd, "CreateWindowW for EventWindow failed");
+  NS_ENSURE_STATE(mEventWnd);
+
+  return NS_OK;
+}
+
 nsresult nsAppShell::Init() {
   LSPAnnotate();
 
@@ -518,35 +592,9 @@ nsresult nsAppShell::Init() {
   // we are processing native events. Disabling this is required for win32k
   // syscall lockdown.
   if (XRE_UseNativeEventProcessing()) {
-    sAppShellGeckoMsgId = ::RegisterWindowMessageW(kAppShellGeckoEventId);
-    NS_ASSERTION(sAppShellGeckoMsgId,
-                 "Could not register hidden window event message!");
-
-    mLastNativeEventScheduled = TimeStamp::NowLoRes();
-
-    WNDCLASSW wc;
-    HINSTANCE module = GetModuleHandle(nullptr);
-
-    const wchar_t* const kWindowClass = L"nsAppShell:EventWindowClass";
-    if (!GetClassInfoW(module, kWindowClass, &wc)) {
-      wc.style = 0;
-      wc.lpfnWndProc = EventWindowProc;
-      wc.cbClsExtra = 0;
-      wc.cbWndExtra = 0;
-      wc.hInstance = module;
-      wc.hIcon = nullptr;
-      wc.hCursor = nullptr;
-      wc.hbrBackground = (HBRUSH) nullptr;
-      wc.lpszMenuName = (LPCWSTR) nullptr;
-      wc.lpszClassName = kWindowClass;
-      [[maybe_unused]] ATOM wcA = RegisterClassW(&wc);
-      MOZ_DIAGNOSTIC_ASSERT(wcA, "RegisterClassW for EventWindowClass failed");
+    if (nsresult rv = this->InitHiddenWindow(); NS_FAILED(rv)) {
+      return rv;
     }
-
-    mEventWnd = CreateWindowW(kWindowClass, L"nsAppShell:EventWindow", 0, 0, 0,
-                              10, 10, HWND_MESSAGE, nullptr, module, nullptr);
-    MOZ_DIAGNOSTIC_ASSERT(mEventWnd, "CreateWindowW for EventWindow failed");
-    NS_ENSURE_STATE(mEventWnd);
   } else if (XRE_IsContentProcess() && !IsWin32kLockedDown()) {
     // We're not generally processing native events, but still using GDI and we
     // still have some internal windows, e.g. from calling CoInitializeEx.
@@ -575,14 +623,6 @@ nsresult nsAppShell::Init() {
 
     obsServ->AddObserver(this, "sessionstore-restoring-on-startup", false);
     obsServ->AddObserver(this, "sessionstore-windows-restored", false);
-
-#if defined(ACCESSIBILITY)
-    if (::GetModuleHandleW(L"uiautomationcore.dll")) {
-      InitUIADetection();
-    } else {
-      obsServ->AddObserver(this, "dll-loaded-main-thread", false);
-    }
-#endif  // defined(ACCESSIBILITY)
   }
 
   if (!WinUtils::GetTimezoneName(mTimezoneName)) {
@@ -625,23 +665,6 @@ nsAppShell::Run(void) {
   return rv;
 }
 
-NS_IMETHODIMP
-nsAppShell::Exit(void) {
-#if defined(ACCESSIBILITY)
-  if (XRE_IsParentProcess()) {
-    nsCOMPtr<nsIObserverService> obsServ(
-        mozilla::services::GetObserverService());
-    obsServ->RemoveObserver(this, "dll-loaded-main-thread");
-
-    if (gUiaHook && ::UnhookWindowsHookEx(gUiaHook)) {
-      gUiaHook = nullptr;
-    }
-  }
-#endif  // defined(ACCESSIBILITY)
-
-  return nsBaseAppShell::Exit();
-}
-
 void nsAppShell::DoProcessMoreGeckoEvents() {
   // Called by nsBaseAppShell's NativeEventCallback() after it has finished
   // processing pending gecko events and there are still gecko events pending
