@@ -7,33 +7,48 @@
 #ifndef nsNativeThemeWin_h
 #define nsNativeThemeWin_h
 
-#include "nsITheme.h"
-#include "nsCOMPtr.h"
-#include "nsAtom.h"
-#include "nsNativeTheme.h"
-#include "nsStyleConsts.h"
-#include "nsUXThemeConstants.h"
-#include "nsUXThemeData.h"
-#include "gfxTypes.h"
 #include <windows.h>
+
 #include "mozilla/Maybe.h"
 #include "mozilla/TimeStamp.h"
-#include "nsSize.h"
+#include "Theme.h"
+#include "nsUXThemeConstants.h"
+#include "nsUXThemeData.h"
 
-class nsNativeThemeWin : private nsNativeTheme, public nsITheme {
+namespace mozilla::widget {
+
+class nsNativeThemeWin : public Theme {
+ protected:
   virtual ~nsNativeThemeWin();
 
  public:
-  typedef mozilla::TimeStamp TimeStamp;
-  typedef mozilla::TimeDuration TimeDuration;
-
-  NS_DECL_ISUPPORTS_INHERITED
+  // Whether we draw a non-native widget.
+  //
+  // We always draw scrollbars as non-native so that all of Firefox has
+  // consistent scrollbar styles both in chrome and content (plus, the
+  // non-native scrollbars support scrollbar-width, auto-darkening...).
+  //
+  // We draw other widgets as non-native when their color-scheme is dark.  In
+  // that case (`BecauseColorMismatch`) we don't call into the non-native theme
+  // for sizing information (GetWidgetPadding/Border and GetMinimumWidgetSize),
+  // to avoid subtle sizing changes. The non-native theme can basically draw at
+  // any size, so we prefer to have consistent sizing information.
+  enum class NonNative { No, Always, BecauseColorMismatch };
+  static bool IsWidgetAlwaysNonNative(nsIFrame*, StyleAppearance);
+  NonNative IsWidgetNonNative(nsIFrame*, StyleAppearance);
 
   // The nsITheme interface.
   NS_IMETHOD DrawWidgetBackground(gfxContext* aContext, nsIFrame* aFrame,
                                   StyleAppearance aAppearance,
-                                  const nsRect& aRect,
-                                  const nsRect& aDirtyRect) override;
+                                  const nsRect& aRect, const nsRect& aDirtyRect,
+                                  DrawOverflow) override;
+
+  bool CreateWebRenderCommandsForWidget(wr::DisplayListBuilder&,
+                                        wr::IpcResourceUpdateQueue&,
+                                        const layers::StackingContextHelper&,
+                                        layers::RenderRootStateManager*,
+                                        nsIFrame*, StyleAppearance,
+                                        const nsRect&) override;
 
   [[nodiscard]] LayoutDeviceIntMargin GetWidgetBorder(
       nsDeviceContext* aContext, nsIFrame* aFrame,
@@ -47,10 +62,9 @@ class nsNativeThemeWin : private nsNativeTheme, public nsITheme {
                                  StyleAppearance aAppearance,
                                  nsRect* aOverflowRect) override;
 
-  NS_IMETHOD GetMinimumWidgetSize(nsPresContext* aPresContext, nsIFrame* aFrame,
-                                  StyleAppearance aAppearance,
-                                  mozilla::LayoutDeviceIntSize* aResult,
-                                  bool* aIsOverridable) override;
+  LayoutDeviceIntSize GetMinimumWidgetSize(
+      nsPresContext* aPresContext, nsIFrame* aFrame,
+      StyleAppearance aAppearance) override;
 
   virtual Transparency GetWidgetTransparency(
       nsIFrame* aFrame, StyleAppearance aAppearance) override;
@@ -64,25 +78,16 @@ class nsNativeThemeWin : private nsNativeTheme, public nsITheme {
   bool ThemeSupportsWidget(nsPresContext* aPresContext, nsIFrame* aFrame,
                            StyleAppearance aAppearance) override;
 
-  bool WidgetIsContainer(StyleAppearance aAppearance) override;
+  bool ThemeDrawsFocusForWidget(nsIFrame*, StyleAppearance) override;
 
-  bool ThemeDrawsFocusForWidget(StyleAppearance aAppearance) override;
-
-  bool ThemeWantsButtonInnerFocusRing(StyleAppearance) override { return true; }
+  bool ThemeWantsButtonInnerFocusRing() override { return true; }
 
   bool ThemeNeedsComboboxDropmarker() override;
 
-  virtual bool WidgetAppearanceDependsOnWindowFocus(
-      StyleAppearance aAppearance) override;
-
-  enum { eThemeGeometryTypeWindowButtons = eThemeGeometryTypeUnknown + 1 };
-  virtual ThemeGeometryType ThemeGeometryTypeForWidget(
-      nsIFrame* aFrame, StyleAppearance aAppearance) override;
-
   nsNativeThemeWin();
 
  protected:
-  mozilla::Maybe<nsUXThemeClass> GetThemeClass(StyleAppearance aAppearance);
+  Maybe<nsUXThemeClass> GetThemeClass(StyleAppearance aAppearance);
   HANDLE GetTheme(StyleAppearance aAppearance);
   nsresult GetThemePartAndState(nsIFrame* aFrame, StyleAppearance aAppearance,
                                 int32_t& aPart, int32_t& aState);
@@ -99,17 +104,12 @@ class nsNativeThemeWin : private nsNativeTheme, public nsITheme {
   bool ClassicGetWidgetPadding(nsDeviceContext* aContext, nsIFrame* aFrame,
                                StyleAppearance aAppearance,
                                LayoutDeviceIntMargin* aResult);
-  nsresult ClassicGetMinimumWidgetSize(nsIFrame* aFrame,
-                                       StyleAppearance aAppearance,
-                                       mozilla::LayoutDeviceIntSize* aResult,
-                                       bool* aIsOverridable);
+  LayoutDeviceIntSize ClassicGetMinimumWidgetSize(nsIFrame* aFrame,
+                                                  StyleAppearance aAppearance);
   bool ClassicThemeSupportsWidget(nsIFrame* aFrame,
                                   StyleAppearance aAppearance);
   void DrawCheckedRect(HDC hdc, const RECT& rc, int32_t fore, int32_t back,
                        HBRUSH defaultBack);
-  bool MayDrawCustomScrollbarPart(gfxContext* aContext, nsIFrame* aFrame,
-                                  StyleAppearance aAppearance,
-                                  const nsRect& aRect, const nsRect& aClipRect);
   uint32_t GetWidgetNativeDrawingFlags(StyleAppearance aAppearance);
   int32_t StandardGetState(nsIFrame* aFrame, StyleAppearance aAppearance,
                            bool wantFocused);
@@ -130,7 +130,7 @@ class nsNativeThemeWin : private nsNativeTheme, public nsITheme {
                                       StyleAppearance aAppearance,
                                       int32_t aPart, int32_t aState,
                                       THEMESIZE aSizeReq,
-                                      mozilla::LayoutDeviceIntSize* aResult);
+                                      LayoutDeviceIntSize* aResult);
 
   SIZE GetCachedGutterSize(HANDLE theme);
 
@@ -150,15 +150,17 @@ class nsNativeThemeWin : private nsNativeTheme, public nsITheme {
       mBorderCache[eUXNumClasses * THEME_PART_DISTINCT_VALUE_COUNT];
 
   // See the above not for mBorderCache and friends. However
-  // mozilla::LayoutDeviceIntSize is half the size of nsIntMargin, making the
+  // LayoutDeviceIntSize is half the size of nsIntMargin, making the
   // cache roughly half as large. In total the caches should come to about 18KB.
   uint8_t mMinimumWidgetSizeCacheValid
       [(eUXNumClasses * THEME_PART_DISTINCT_VALUE_COUNT + 7) / 8];
-  mozilla::LayoutDeviceIntSize
+  LayoutDeviceIntSize
       mMinimumWidgetSizeCache[eUXNumClasses * THEME_PART_DISTINCT_VALUE_COUNT];
 
   bool mGutterSizeCacheValid;
   SIZE mGutterSizeCache;
 };
 
+}  // namespace mozilla::widget
+
 #endif
